
import { Env } from "../../pi_sys/util/env";
import { Asset as NatvieAsset } from "../../webview/native_impl/core/asset";
import { Logger } from "pi_sys/util/logger";
import { readAssets } from "pi_sys/store/file_system";
import { utf8Decode } from "pi_sys/util/utf8";
import { Depend } from "pi_sys/util/depend";
import { get, post } from "pi_sys/net/http"
import { RESP_TYPE } from "pi_sys/net/http_base";
declare var module;
let logger = new Logger(module.name, "sys");

/**
 * 本地资源加载
 */

export namespace Asset {
    /**
     * 获取底层资源
     * @description 获取资源信息数组Json, [[file, time, size, sign], ...]
     * 读不到返回空数组
     * @note 底层真实资源都带有assets_前缀
     */
    export const get = (): Promise<string[][]> => {
        let dependPaths = [];
        // 这里的version对应 es6.version
        let version = Env.get("version");
        let paths = [];
        for (let key in version) {
            let path = version[key];
            if (typeof path === typeof "") {
                paths.push(path);
            } else {
                // 因为version有 一些对象，像 sys，imageDepend等，所以还要下一层取到所有的路径
                if (key === "imageDepend") {
                    const imgVersion = Env.get("imgVersion");
                    for (let key in path) {
                        if (key === imgVersion) {
                            paths.push(path[key]);
                        }
                    }
                }
            }
        }

        for (let path of paths) {
            // 路径得值是："path?sign"
            let len = path.indexOf("?");
            if (len >= 0) {
                path = path.slice(0, len);
            }
            // 只分析depend命名的变量；
            if (path.endsWith(".depend")) {
                dependPaths.push(path);
            }
        }

        let ps = [];
        let assetsCfg = Env.get("assetsCfg");
        // console.log("===== dependPaths ", dependPaths);
        for (let dependPath of dependPaths) {
            ps.push(new Promise((resolve) => {
                let json = assetsCfg[dependPath] || [];
                resolve(json);
            }));
        }

        return Promise.all(ps).then(arr => {
            // 取完之后，需要合并所有的 []，这数组就是所有底层的资源；
            return [].concat(...arr);
        });
    };

    /**
     * @description 读取资源
     * @example
     */
    export const read = (path: string, isString = false): Promise<ArrayBuffer | string> => {
        const platform_type = Env.get("platform_type2");
        if (platform_type === "windows") {
            return readAssets(path).then(buffer => {
                if (isString) {
                    return utf8Decode(buffer);
                } else {
                    return buffer;
                }
            })
        } else {
            return readAssetsByFile(path).then(buffer => {
                if (isString) {
                    return utf8Decode(buffer);
                } else {
                    return buffer;
                }
            })
        }


    };

    /**
     * 获取文件在底层的绝对路径
     * @param path 
     * @returns 
     */
    export const getSystemPath = (path: string): Promise<string> => {
        return new Promise((resolve, reject) => {
            NatvieAsset.init().then(() => {
                NatvieAsset.getSystemPath(path,
                    resolve,
                    reject
                );
            }).catch(reject);
        });
    }

    /**
     * 将assets目录下的文件拷贝到data目录下
     * @param assetsPath 
     * @param dataPath 
     * @returns 
     */
    export const copyAssetToData = (assetsPath: string, dataPath: string) => {
        return new Promise((resolve, reject) => {
            NatvieAsset.init().then(() => {
                NatvieAsset.copyAssetToData(assetsPath, dataPath,
                    resolve as any,
                    reject
                );
            }).catch(reject);
        });
    }
}


/**
 * 通过file协议获取本地资源
 */
const readAssetsByFile = (path: string) => {
    let cfg = Env.get("cfg");
    if (Array.isArray(cfg)) {
        cfg = cfg[0];
    }
    const gameName = cfg.first_app;
    const url = `file:///android_asset/app/${gameName}/${path}`;
    const TIMEOUT = 60 * 1000;
    return get(url, null, null, RESP_TYPE.BIN, TIMEOUT) as Promise<ArrayBuffer>;
}