import "./platform"; // 屏蔽各小游戏平台差异
/// 兼容window常用属性
export { default as HTMLElement } from './html_element';
export { default as Image } from './html_image_element';
export { default as HTMLCanvasElement } from './html_canvas_element';
export { default as HTMLImageElement } from './html_image_element';
export { default as document } from './document';
import { addListener, removeListener, dispatchEvent as dispatchEvent1 } from 'pi_sys/native/base/event';
import * as Mixin from "./minx";

import { Logger } from "pi_sys/util/logger";
declare var module;
let logger = new Logger(module.name, "sys");

const { platform } = wx.getSystemInfoSync()

export const addEventListener = addListener;
export const removeEventListener = removeListener;
export const dispatchEvent = dispatchEvent1;

export const alert = (msg: any) => {
    logger.debug(msg);
}

// logger.time, logger.timeEnd
if (platform !== 'devtools') {
    const wxPerf = wx.getPerformance ? wx.getPerformance() : Date;
    const consoleTimers = {};
    logger.time = function (name) {
        consoleTimers[name] = wxPerf.now();
    };

    logger.timeEnd = function (name) {
        const timeStart = consoleTimers[name];
        if (!timeStart) {
            return;
        }

        const timeElapsed = wxPerf.now() - timeStart;
        logger.debug(name + ": " + timeElapsed / 1000 + "ms");
        delete consoleTimers[name];
    };
}
// performance
let p: any;
if (wx.getPerformance && !window["qq"]) { // qq小游戏的performance有bug，存在死循环
    const { platform } = wx.getSystemInfoSync()
    const wxPerf = wx.getPerformance()
    const initTime = wxPerf.now()

    const clientPerfAdapter = Object.assign({}, wxPerf, {
        now: function () {
            return (wxPerf.now() - initTime) / 1000
        }
    })

    p = platform === 'devtools' ? wxPerf : clientPerfAdapter
} else {
    p = {
        timeOrigin: Date.now(),
        now: function () {
            return Date.now() - this.timeOrigin
        }
    }
}
export const performance = p;

// 如果使华为快游戏， 全局已经存在一个canvas
export let canvas = wx.createCanvas() as any;

canvas.focus = function () { };
canvas.blur = function () { };

canvas.addEventListener = addListener;

canvas.removeEventListener = removeListener;

canvas.dispatchEvent = dispatchEvent1;
Mixin.style(canvas);
canvas.setAttribute = (name, value) => {
    canvas[name] = value
}

canvas.getAttribute = (name) => {
    return canvas[name]
}

canvas.setAttributeNS = (name, value) => {
    canvas[name] = value
}

canvas.getAttributeNS = (name) => {
    return canvas[name]
}

//localStorage
export const localStorage = {
    get length() {
        const { keys } = wx.getStorageInfoSync()
        return keys.length
    },

    key(n) {
        const { keys } = wx.getStorageInfoSync()

        return keys[n]
    },

    getItem(key) {
        const value = wx.getStorageSync(key);
        return value === "" ? null : value;
    },

    setItem(key, value) {
        if (window.asyncStorage) {
            return wx.setStorage({
                key: key,
                data: value
            })
        }
        return wx.setStorageSync(key, value)
    },

    removeItem(key) {
        if (window.asyncStorage) {
            return wx.removeStorage({
                key: key
            })
        }
        return wx.removeStorageSync(key)
    },

    clear() {
        if (window.asyncStorage) {
            return wx.clearStorage()
        }
        return wx.clearStorageSync()
    }
}


const _events = new WeakMap()

class EventTarget {
    constructor() {
        _events.set(this, {})
    }

    addEventListener(type, listener, options = {}) {
        let events = _events.get(this)

        if (!events) {
            events = {}
            _events.set(this, events)
        }
        if (!events[type]) {
            events[type] = []
        }
        events[type].push(listener)

        if (options.capture) {
            // logger.warn('EventTarget.addEventListener: options.capture is not implemented.')
        }
        if (options.once) {
            // logger.warn('EventTarget.addEventListener: options.once is not implemented.')
        }
        if (options.passive) {
            // logger.warn('EventTarget.addEventListener: options.passive is not implemented.')
        }
    }

    removeEventListener(type, listener, options = {}) {
        const events = _events.get(this)

        if (events) {
            const listeners = events[type]

            if (listeners && listeners.length > 0) {
                for (let i = listeners.length; i--; i > 0) {
                    if (listeners[i] === listener) {
                        listeners.splice(i, 1)
                        break
                    }
                }
            }
        }
    }

    dispatchEvent(event = {}) {
        const listeners = _events.get(this)[event.type]

        if (listeners) {
            for (let i = 0; i < listeners.length; i++) {
                listeners[i](event)
            }
        }
    }
}

const _requestHeader = new WeakMap()
const _responseHeader = new WeakMap()
const _requestTask = new WeakMap()

function _triggerEvent(type, event = {}) {
    event.target = event.target || this

    if (typeof this[`on${type}`] === 'function') {
        this[`on${type}`].call(this, event)
    }
}

function _changeReadyState(readyState, event = {}) {
    this.readyState = readyState

    event.readyState = readyState;

    _triggerEvent.call(this, 'readystatechange', event)
}

function _isRelativePath(url) {
    return !(/^(http|https|ftp|wxfile):\/\/.*/i.test(url));
}

class XMLHttpRequest1 extends EventTarget {

    constructor() {
        super();

        /*
         * TODO 这一批事件应该是在 XMLHttpRequestEventTarget.prototype 上面的
         */
        this.onabort = null
        this.onerror = null
        this.onload = null
        this.onloadstart = null
        this.onprogress = null
        this.ontimeout = null
        this.onloadend = null

        this.onreadystatechange = null
        this.readyState = 0
        this.response = null
        this.responseText = null
        this.responseType = 'text'
        this.dataType = 'string'
        this.responseXML = null
        this.status = 0
        this.statusText = ''
        this.upload = {}
        this.withCredentials = false

        _requestHeader.set(this, {
            'content-type': 'application/x-www-form-urlencoded'
        })
        _responseHeader.set(this, {})
    }

    abort() {
        const myRequestTask = _requestTask.get(this)

        if (myRequestTask) {
            myRequestTask.abort()
        }
    }

    getAllResponseHeaders() {
        const responseHeader = _responseHeader.get(this)

        return Object.keys(responseHeader).map((header) => {
            return `${header}: ${responseHeader[header]}`
        }).join('\n')
    }

    getResponseHeader(header) {
        return _responseHeader.get(this)[header]
    }

    open(method, url /* async, user, password 这几个参数在小程序内不支持*/) {
        this._method = method
        this._url = url
        _changeReadyState.call(this, XMLHttpRequest.OPENED)
    }

    overrideMimeType() { }

    send(data = '') {
        if (this.readyState !== XMLHttpRequest.OPENED) {
            throw new Error("Failed to execute 'send' on 'XMLHttpRequest': The object's state must be OPENED.")
        } else {
            const url = this._url
            const header = _requestHeader.get(this)
            const responseType = this.responseType
            const dataType = this.dataType

            const relative = _isRelativePath(url)
            let encoding;

            if (responseType === 'arraybuffer') {
                // encoding = 'binary'
            } else {
                encoding = 'utf8'
            }

            delete this.response;
            this.response = null;

            const onSuccess = ({ data, statusCode, header }) => {
                statusCode = statusCode === undefined ? 200 : statusCode;
                if (typeof data !== 'string' && !(data instanceof ArrayBuffer)) {
                    try {
                        data = JSON.stringify(data)
                    } catch (e) {
                        data = data
                    }
                }

                this.status = statusCode
                if (header) {
                    _responseHeader.set(this, header)
                }
                _triggerEvent.call(this, 'loadstart')
                _changeReadyState.call(this, XMLHttpRequest.HEADERS_RECEIVED)
                _changeReadyState.call(this, XMLHttpRequest.LOADING)

                this.response = data

                if (data instanceof ArrayBuffer) {
                    Object.defineProperty(this, 'responseText', {
                        enumerable: true,
                        configurable: true,
                        get: function () {
                            throw "InvalidStateError : responseType is " + this.responseType;
                        }
                    });
                } else {
                    this.responseText = data
                }
                _changeReadyState.call(this, XMLHttpRequest.DONE)
                _triggerEvent.call(this, 'load')
                _triggerEvent.call(this, 'loadend')
            }

            const onFail = ({ errMsg }) => {
                // TODO 规范错误

                if (errMsg.indexOf('abort') !== -1) {
                    _triggerEvent.call(this, 'abort')
                } else {
                    _triggerEvent.call(this, 'error', {
                        message: errMsg
                    })
                }
                _triggerEvent.call(this, 'loadend')

                if (relative) {
                    // 用户即使没监听error事件, 也给出相应的警告
                    logger.warn(errMsg)
                }
            }

            if (relative) {
                const fs = wx.getFileSystemManager();

                var options = {
                    'filePath': url,
                    'success': onSuccess,
                    'fail': onFail
                }
                if (encoding) {
                    options['encoding'] = encoding;
                }
                fs.readFile(options)
                return
            }

            wx.request({
                data,
                url: url,
                method: this._method,
                header: header,
                dataType: dataType,
                responseType: responseType,
                success: onSuccess,
                fail: onFail
            })
        }
    }

    setRequestHeader(header, value) {
        const myHeader = _requestHeader.get(this)

        myHeader[header] = value
        _requestHeader.set(this, myHeader)
    }

    addEventListener(type, listener) {
        if (typeof listener !== 'function') {
            return;
        }

        this['on' + type] = (event = {}) => {
            event.target = event.target || this
            listener.call(this, event)
        }
    }

    removeEventListener(type, listener) {
        if (this['on' + type] === listener) {
            this['on' + type] = null;
        }
    }
}

// TODO 没法模拟 HEADERS_RECEIVED 和 LOADING 两个状态
XMLHttpRequest1.UNSEND = 0
XMLHttpRequest1.OPENED = 1
XMLHttpRequest1.HEADERS_RECEIVED = 2
XMLHttpRequest1.LOADING = 3
XMLHttpRequest1.DONE = 4

exports.XMLHttpRequest = XMLHttpRequest1;




const _socketTask = new WeakMap()

export class WebSocket {
    constructor(url, protocols = []) {
        this.binaryType = '' // TODO 更新 binaryType
        this.bufferedAmount = 0 // TODO 更新 bufferedAmount
        this.extensions = ''

        this.onclose = null
        this.onerror = null
        this.onmessage = null
        this.onopen = null

        this.protocol = '' // TODO 小程序内目前获取不到，实际上需要根据服务器选择的 sub-protocol 返回
        this.readyState = 3

        if (typeof url !== 'string' || !(/(^ws:\/\/)|(^wss:\/\/)/).test(url)) {
            throw new TypeError(`Failed to construct 'WebSocket': The URL '${url}' is invalid`)
        }

        this.url = url
        this.readyState = WebSocket.CONNECTING

        const socketTask = wx.connectSocket({
            url,
            protocols: Array.isArray(protocols) ? protocols : [protocols]
        })

        _socketTask.set(this, socketTask)

        socketTask.onClose((res) => {
            this.readyState = WebSocket.CLOSED
            if (typeof this.onclose === 'function') {
                this.onclose(res)
            }
        })

        socketTask.onMessage((res) => {
            if (typeof this.onmessage === 'function') {
                this.onmessage(res)
            }
        })

        socketTask.onOpen(() => {
            this.readyState = WebSocket.OPEN
            if (typeof this.onopen === 'function') {
                this.onopen()
            }
        })

        socketTask.onError((res) => {
            if (typeof this.onerror === 'function') {
                this.onerror(new Error(res.errMsg))
            }
        })

        return this
    }

    close(code, reason) {
        this.readyState = WebSocket.CLOSING
        const socketTask = _socketTask.get(this)

        socketTask.close({
            code,
            reason
        })
    }

    send(data) {
        if (typeof data !== 'string' && !(data instanceof ArrayBuffer)) {
            throw new TypeError(`Failed to send message: The data ${data} is invalid`)
        }

        const socketTask = _socketTask.get(this)

        socketTask.send({
            data
        })
    }
}

WebSocket.CONNECTING = 0 // The connection is not yet open.
WebSocket.OPEN = 1 // The connection is open and ready to communicate.
WebSocket.CLOSING = 2 // The connection is in the process of closing.
WebSocket.CLOSED = 3 // The connection is closed or couldn't be opened.


export class Worker {
    _onmessage: (e) => void;

    set onmessage(_onmessage) {
        if (!this._onmessage) {
            // 小游戏onMessage是add模式,并且没有找到取消监听的方法,只在第一次设置的时候监听
            this.worker.onMessage((data) => {
                this._onmessage({ data })
            });
        }
        this._onmessage = _onmessage;
    }
    constructor(url: string) {
        this.worker = wx.createWorker(url);
    }
    postMessage(data: any) {
        this.worker.postMessage(data)
    }


}