/**
 * 资源对象加载器， 图片、声音、视频, 字体
 * 小游戏加载该类资源时，先将其下载到本地临时目录中，
 * 后续通过拷贝将资源拷贝到本地用户目录下面
 */
// ============================== 导入
import { IFileDependInfo } from 'pi_sys/util/depend_base';
import { Depend } from "pi_sys/util/depend";
import { FileLoadParam } from '../../pi_sys/load/fileload_base';
import { IParamPiFont, PiFont } from 'pi_sys/render/font';
import type { TloadAudio, TloadAudioBuffer, TloadFile, TloadFont, TloadImage, TloadVideo } from 'pi_sys/load/interface';
import { PiImage } from 'pi_sys/render/image';
import { PiAudio } from 'pi_sys/media/audio';
import * as Sign from "pi_sys/load/sign";
import { Logger } from "pi_sys/util/logger";
import { PiAudioBuffer } from 'pi_sys/media/inner_audio';
import { loadRes } from 'pi_sys/load/res_load';
import { fileCache } from '../wx/file_cache';
import { Download } from 'pi_sys/load/down_load';
import { access } from 'pi_sys/store/file_system';
import { notifyResUsed } from 'pi_sys/util/collect';
import { IPiVideo, createVideo } from 'pi_sys/render/video';
import { Env } from 'pi_sys/util/env';
import { getUrl } from 'pi_sys/util/util';
import { BatchLoad } from 'pi_sys/load/batch_load';
declare var module;
let logger = new Logger(module.name, "sys");

// ============================== 导出



// ============================== 立即执行

/**
 * 资源加载工具
 */

export const loadImage: TloadImage = (filePath: string, args: any[]): Promise<PiImage> => {

    return new Promise((resolve, reject) => {
        let info = Depend.getFile(filePath);
        if (info) {
            loadLocalPath(info, filePath).then(localPath => {
                let img = wx.createImage();
                // 避免销毁时 置 src="", 响应 onerror
                let isOk = false;
                img.onload = () => {
                    isOk = true;
                    resolve(<any>img);
                }
                img.onerror = (err) => {
                    if (!isOk) {
                        reject(err);
                    }
                }
                img.src = localPath;
            }).catch(reject)
        } else {
            reject(`loadImage - Not Found File Depend Info ${filePath}`);
        }
    })
}

export const loadFont: TloadFont = (filePath: string, args: [IParamPiFont?]) => {
    return new Promise((resolve, reject) => {
        let info = Depend.getFile(filePath);

        if (info) {
            loadLocalPath(info, filePath).then(localPath => {
                let font = new PiFont();
                font.onload = () => {
                    resolve(font);
                }
                font.onerror = (err) => {
                    reject(err);
                }
                font.arg = args[0];
                font.src = localPath;
            }).catch(reject)
        } else {
            reject(`loadFont - Not Found File Depend Info ${filePath}`);
        }
    })

}

export const loadAudio: TloadAudio = (filePath: string, args: any[]) => {
    return new Promise((resolve, reject) => {
        let info = Depend.getFile(filePath);

        if (info) {
            loadLocalPath(info, filePath).then(localPath => {
                let audio = new PiAudio(localPath);
                resolve(audio);
            }).catch(reject)
        } else {
            reject(`loadAudio - Not Found File Depend Info ${filePath}`);
        }
    })
}

export const loadAudioBuffer: TloadAudioBuffer = (filePath: string, args: any[]): Promise<PiAudioBuffer> => {
    return new Promise((resolve, reject) => {
        let info = Depend.getFile(filePath);

        if (info) {
            loadLocalPath(info, filePath).then(localPath => {
                resolve(localPath);
            }).catch(reject)
        } else {
            reject(`loadAudioBuffer - Not Found File Depend Info ${filePath}`);
        }
    })
}

export const loadFile: TloadFile = (filePath: string, args: any[]): Promise<Uint8Array> => {
    return new Promise((resolve, reject) => {
        let info = Depend.getFile(filePath);
        if (info) {
            loadBuffer(info, filePath).then(r => {
                if (!ArrayBuffer.isView(r)) {
                    r = new Uint8Array(r);
                }
                resolve(r as Uint8Array);
            }).catch(reject)
        } else {
            reject(`loadFile - Not Found File Depend Info ${filePath}`);
        }
    })
}




const download = (file: IFileDependInfo, waitWrite: boolean = true) => {
    const down = new Download();
    (<any>down).waitWrite = waitWrite;
    down.add(file);
    return down.start();
}


export const loadLocalPath = (info: IFileDependInfo, path: string): Promise<string> => {
    notifyResUsed(info.path);
    const localPath = fileCache.pathFormat((<any>FileLoadParam.localStore).formatStorePath(path));
    let p;
    if (Sign.checkLocalSign(info, info.path)) {  // 本地存在
        p = access(localPath).then(r => {
            if (!r) {
                return download(info);
            }
        });
    } else {
        p = download(info);
    }
    return p.then(() => {
        return localPath;
    });
}



const getLocalPath = (info: IFileDependInfo, path: string): Promise<string> => {
    const localPath = fileCache.pathFormat((<any>FileLoadParam.localStore).formatStorePath(path));
    if (Sign.checkLocalSign(info, info.path)) {  // 本地存在
        return access(localPath).then(r => {
            if (r) {
                return localPath;
            }
        });
    } else {
        return Promise.resolve("");
    }
}

export const loadBuffer = (info: IFileDependInfo, path: string): Promise<ArrayBuffer> => {
    notifyResUsed(info.path);
    let p;
    if (Sign.checkLocalSign(info, info.path)) {  // 本地存在
        p = FileLoadParam.localStore.read(path).then(r => {
            if (!r) {
                return download(info, false);
            } else {
                const map = new Map;
                map.set(path, r)
                return map;
            }
        }).catch(e => {
            return download(info, false);
        });
    } else {
        p = download(info, false);
    }
    return p.then((map) => {
        return map.get(path);
    });
}


const downVideos: Set<string> = new Set; // 正在下载的video视频

/**
 * @param filePath 
 * @param args 
 * @returns 
 */
export const loadVideo: TloadVideo = (filePath: string, args: any[]): Promise<IPiVideo> => {
    return new Promise((resolve, reject) => {
        let info = Depend.getFile(filePath);
        let param = args[0];
        if (!param) {
            param = {};
        }
        if (info) {
            notifyResUsed(info.path);
            if (filePath.endsWith(".m3u8")) { // m3u8格式都走网络路径
                let url = getUrl(filePath);
                param.src = url;
                console.log("================= video url", param);
                let video = createVideo(param);
                resolve(video);
            } else {
                const waitVideoDownloaded = Env.get("waitVideoDownloaded");
                if (waitVideoDownloaded) { // 等待视频下载完成
                    loadLocalPath(info, filePath).then(localPath => {
                        param.src = localPath;
                        try {
                            let video = createVideo(param);
                            resolve(video);
                        } catch (err) {
                            reject(err);
                        }
                    })

                } else {
                    /**
                     * 由于视频有可能比较大,项目不希望等到视频下载完成才播放,体验较差
                     * 所以如果本地如果还没有缓存,则直接用网络url播放,同时去缓存视频(会消耗两份甚至更多流量)
                     */
                    getLocalPath(info, filePath).then(localPath => {
                        let url = localPath ? localPath : getUrl(filePath);
                        param.src = url;
                        console.log("================= video url", url);
                        try {
                            let video = createVideo(param);
                            if (!localPath && !downVideos.has(filePath)) {
                                downVideos.add(filePath);
                                const duration = !isNaN((<any>video).duration) ? Math.ceil((<any>video).duration) : 10;

                                // 等到视频播放完成后下载视频资源缓存
                                setTimeout(() => {
                                    const load = new BatchLoad([filePath]);
                                    load.startRes().then(() => {
                                        downVideos.delete(filePath);
                                    }).catch(e => {
                                        downVideos.delete(filePath);
                                    })
                                }, duration * 1000);
                            }
                            resolve(video);
                        } catch (err) {
                            reject(err);
                        }
                    });


                }
            }
        } else {
            reject(`loadVideo - Not Found File Depend Info ${filePath}`);
        }
    })
}



