import { IAudioPlayCfg, IPiAudio, StateType, StepType } from "pi_sys/media/audio";


class AudioPlayCfg implements IAudioPlayCfg {
    // 接口 成员 初始化
    public readonly context: wx.InnerAudioContext;
    public volume: number = 1;
    public playTime: number = 0;
    public pauseTime: number = 0;
    public onended: Function = null;
    public delay: number = 0;
    public repeat: number = 0;
    public repeatDelay: number = 0;
    public startTime: number = 0;
    public loop: boolean = false;

    constructor(context: wx.InnerAudioContext) {
        this.context = context;
    }
    public getState(): StateType {
        if (this.playTime) {
            return StateType.PLAY;
        }
        if (this.pauseTime) {
            return StateType.PAUSE;
        }

        return StateType.READY;
    }
    /**
     * 获取声音时长
     */
    public getDuration(): number {
        if (!this.context) {
            return -1;
        }

        return this.context.duration;
    }
    /**
     * 暂停
     */
    public pause(): void {
        if ((!this.context) && !this.playTime) {
            return;
        }
        this.context.pause();
        this.pauseTime = this.context.currentTime - this.playTime;
        if (this.pauseTime < 0) {
            this.pauseTime = 0;
        }
        this.playTime = 0;
    }
    /**
     * 停止
     */
    public stop(): void {
        if ((!this.context) && !this.playTime) {
            return;
        }
        this.context.pause();
        this.playTime = this.pauseTime = 0;
    }
    /**
     * 销毁
     */
    public destroy(): void {
        if (!this.context) {
            return;
        }
        if (this.context) {
            this.context.destroy();
        }
    }
    public play(volume: number = 1, delay: number = 0, offset: number = 0) {

        delay = delay ? this.context.currentTime + delay : 0;
        offset = offset || this.pauseTime;

        this.context.loop = this.loop;
        this.context.volume = volume || 1;
        this.context.startTime = offset;

        this.playTime = this.context.currentTime + delay - offset;
        this.pauseTime = 0;

        const currfunc = this.onended;

        if (this.onended) {
            this.context.onEnded((ev) => {
                this.playTime = 0;
                if (this.onended === currfunc) {
                    currfunc(this, ev);
                }
            });

            this.context.onError((ev) => {
                this.playTime = 0;
                if (this.onended === currfunc) {
                    currfunc(this, ev);
                }
            });
        }

        this.context.play();
    }

    public disconnect() { };
}

/**
 * 普通音效，外部只管发起播，不管后续,可能同时播多次同一音效
 * 背景音乐，外部控制 播放,暂停,停止，同时只能存在一个播放
 */
export class PiAudio implements IPiAudio {
    public onended: Function = null;
    public loop: boolean = false;
    // 当前播放的声音数组
    public arr: AudioPlayCfg[] = [];
    // 音量， 0-1之间
    public volume: number = 1;
    public readonly src: string;
    private static currBGM: PiAudio;
    public static enable(flag: boolean) {
        if (flag) {
            if (PiAudio.currBGM) {
                PiAudio.currBGM.playAsBGM();
            }
        } else {
            if (PiAudio.currBGM) {
                PiAudio.currBGM.pauseAsBGM();
            }
        }
    }
    /**
     * 
     * @param src localPath
     */
    constructor(src: string) {
        this.src = src;
    }
    public play(delay?: number, repeat?: number, repeatDelay?: number) {
        const cfg = new AudioPlayCfg(wx.createInnerAudioContext());

        cfg.startTime = cfg.context.currentTime;
        cfg.delay = delay || 0;
        cfg.repeat = repeat || 0;
        cfg.repeatDelay = repeatDelay || 0;
        cfg.context.src = this.src;

        this.arr.push(cfg);

        this.playAudio(cfg);
    }
    public playAsBGM(volume: number = 1, loop: boolean = true, stepTime: number = 0, stepType?: StepType) {
        PiAudio.currBGM = this;

        this.volume = volume;

        let cfg = this.arr[0];

        if (!cfg) {
            cfg = new AudioPlayCfg(wx.createInnerAudioContext());
        }

        cfg.startTime = cfg.context.currentTime;
        cfg.delay = 0;
        cfg.repeat = 0;
        cfg.repeatDelay = 0;
        cfg.loop = loop;
        cfg.context.src = this.src;

        this.arr.push(cfg);

        cfg.onended = () => {
            if (this.onended) {
                this.onended();
            }
        };

        const d = cfg.delay + cfg.startTime - cfg.context.currentTime;
        cfg.play(this.volume, d > 0 ? d : 0);
    }

    public stopAsBGM() {
        let cfg = this.arr[0];

        if (cfg) {
            cfg.stop();
        }
    }

    public pauseAsBGM() {
        let cfg = this.arr[0];

        if (cfg) {
            cfg.pause();
        }
    }

    // 播放声音
    private playAudio(cfg: AudioPlayCfg) {

        cfg.onended = () => {
            if (cfg.repeat < 1) {
                cfg.destroy();
                this.arr.splice(this.arr.indexOf(cfg), 1);
                if (this.onended) {
                    this.onended();
                }
                return '';
            }

            cfg.repeat--;
            cfg.play(this.volume, cfg.repeatDelay);
        };

        const d = cfg.delay + cfg.startTime - cfg.context.currentTime;
        cfg.play(this.volume, d > 0 ? d : 0);
    }
    public destroy() {
        if (PiAudio.currBGM === this) {
            PiAudio.currBGM = undefined;
        }
        this.arr.forEach((cfg) => {
            cfg.context.src = '';
            cfg.destroy();
        });
    }
}