import { AppName, BenchLevel, SystemInfo, TgetOAID, TgetDeviceID, TgetSystemInfo } from "pi_sys/native/base/system_info";
import { Env } from "pi_sys/util/env";


const _getSystemInfo = (res: WechatMinigame.SystemInfo) => {
	// 测试发现，vivo小游戏中描述屏幕宽高的字段使用的是物理像素值，这里强制将pixelRatio修改为1，使其与真实效果一致（华为小游戏就是这种做法）
	let pixelRatio = res.pixelRatio;
	if (window["_$platform_app_name"] === AppName.VVMinigame) {
		pixelRatio = 1
	}

	let benchLevel = BenchLevel.Unkown;

	if ((<any>res).deviceScore && typeof (<any>res).deviceScore === "object") {
		// 抖音小游戏
		const overall = (<any>res).deviceScore.overall;    // 综合评分
		if (overall >= 8.6) {
			benchLevel = BenchLevel.High;
		} else if (overall >= 7.6) {
			benchLevel = BenchLevel.Middle;
		} else {
			benchLevel = BenchLevel.Low;
		}
	} else {
		if (res.benchmarkLevel === -2 || res.benchmarkLevel === 0) {
			benchLevel = BenchLevel.CantRun;
		} else if (res.benchmarkLevel === -1) {
			benchLevel = BenchLevel.Unkown;
		} else if (res.benchmarkLevel <= 22) {
			benchLevel = BenchLevel.Low;
		} else if (res.benchmarkLevel <= 29) {
			benchLevel = BenchLevel.Middle;
		} else {
			benchLevel = BenchLevel.High;
		}
	}



	// 不支持 WebGL 或者 WebAssemly 的，全部是CantRun
	// 小游戏环境，肯定支持WebGL
	let isSupportWasm = false;
	try {
		if (window["WXWebAssembly"] && window["WXWebAssembly"].instantiate) {
			isSupportWasm = true;
		}

		if (!isSupportWasm) {
			if (window["WebAssembly"] && window["WebAssembly"].instantiate) {
				isSupportWasm = true;
			}
		}
	} catch (_e) {
		isSupportWasm = false;
	}
	if (!isSupportWasm) {
		benchLevel = BenchLevel.CantRun;
	} else {
		let model = res.model.toLocaleLowerCase();
		if (model.startsWith("iphone ")) {
			benchLevel = getIphoneLevel(model);
		}
	}

	return {
		safeArea: {
			left: res.safeArea?.left || 0,
			right: res.safeArea?.right || res.screenWidth,
			top: res.safeArea?.top || 0,
			bottom: res.safeArea?.bottom || res.screenHeight,
			width: res.safeArea?.width || res.screenWidth,
			height: res.safeArea?.height || res.screenHeight,
		},

		brand: res.brand,
		model: res.model,
		pixelRatio: pixelRatio,
		screenWidth: res.screenWidth,
		screenHeight: res.screenHeight,
		statusBarHeight: res.statusBarHeight,
		language: res.language,
		system: res.system,
		cameraAuthorized: (<any>res).cameraAuthorized,
		locationAuthorized: (<any>res).locationAuthorized,
		microphoneAuthorized: (<any>res).microphoneAuthorized,
		notificationAuthorized: (<any>res).notificationAuthorized,
		bluetoothEnabled: (<any>res).bluetoothEnabled,
		locationEnabled: (<any>res).locationEnabled,
		wifiEnabled: (<any>res).wifiEnabled,

		platform: res.platform,
		app: window["_$platform_app_name"], // 在code_adapter/index中，已经赋值到全局
		device: (res.platform === "windows" || res.platform === "mac") ? "pc" : "mobile",
		benchLevel,
		colorDepth: 24, // TODO 小游戏从哪里取到颜色深度？
	};
}



export const getSystemInfo: TgetSystemInfo = (): Promise<SystemInfo> => {
	return new Promise((resolve, reject) => {
		wx.getSystemInfo({
			success: async (res: WechatMinigame.SystemInfo) => {
				let r = _getSystemInfo(res);
				const rewriteSystemInfo = Env.get("rewriteSystemInfo") || window["winit"].env.rewriteSystemInfo;
				if (rewriteSystemInfo) {
					await rewriteSystemInfo(r, res);
				}
				resolve(r);
			},
			fail: reject,
			complete: () => {

			}
		})
	});
}

export const getDeviceID: TgetDeviceID = () => {
	return Promise.resolve(generateDeviceId());
}


function generateDeviceId(): string {
	let old = window.localStorage.getItem("pi_sys_minigame_deviceId")
	if (old) { return old }
	const characters: string = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789';
	let deviceId: string = '';
	for (let i = 0; i < 32; i++) {
		const randomIndex = Math.floor(Math.random() * characters.length);
		deviceId += characters.charAt(randomIndex);
	}
	window.localStorage.setItem("pi_sys_minigame_deviceId", deviceId)
	return deviceId;
}

/**
 * 取iphone的低中高档
 * 截止：2021.05月
 * 
 * 低档：
 * 	  iphone se
 *	  小于 iphone 8
 * 中档：
 *	  iphone 8
 *    iphone 8 plus
 *    iphone x
 * 高档：
 *    iphone xr
 *    iphone xs
 *    iphone xs max
 *    iphone 11
 *    iphone 11 pro
 *    iphone 11 pro max
 *    iphone 12 mini
 *    iphone 12
 *    iphone 12 pro
 *    iphone 12 pro max
 */
const getIphoneLevel = (model: string) => {
	if (model.startsWith("iphone se")) {
		return BenchLevel.Low;
	} else if (model.startsWith("iphone xr") || model.startsWith("iphone xs")) {
		return BenchLevel.High;
	} else if (model.startsWith("iphone x")) {
		return BenchLevel.Middle;
	}

	let level = BenchLevel.Unkown;

	// 下面都是数字
	let versionStr = /^iphone (\d+)/.exec(model);
	if (!versionStr) {
		level = BenchLevel.High;
	} else {
		let version = parseInt(versionStr[1]);
		if (version < 10) {
			level = BenchLevel.Low;
		} else if (version < 14) {
			level = BenchLevel.Middle;
		} else {
			level = BenchLevel.High;
		}
	}
	return level;
}



export const getOAID: TgetOAID = () => {
	return Promise.resolve("");
}