/**
 * 小游戏的 store 通过文件系统模拟 indexedDB ，
 * 文件存放到 dbName/tabName 目录下
 */
import { FileLoadParam } from "pi_sys/load/fileload_base";
import { IPiStoreBase, StoreType } from "pi_sys/store/store_base";
import { CompressAlgorithm, CompressLevel, EncryptType } from "pi_sys/util/constant";

import { Logger } from "pi_sys/util/logger";
import { utf8Decode, utf8Encode } from "pi_sys/util/utf8";
import { fileCache } from "../wx/file_cache";
declare var module;
let logger = new Logger(module.name, "sys");

/**
 * Store
 * 
 */
export class Store implements IPiStoreBase {
    /**
     * 项目名称
     */
    public readonly dbName: string;
    /**
     * 子文件夹名称
     */
    public readonly tabName: string;


    private style: StoreType;
    immediateWrite: boolean = true;  // 是否立即写入


    /**
     * 判断是否支持IndexedDB
     */
    public static check = () => {
        return false;
    }
    /**
     * 删除指定的子目录
     */
    public static deleteStore(childDirName: string): Promise<any> {
        return fileCache.clear(childDirName)
    }


    // public static create(projectName: string, childDirName?: string): Promise<Store> {
    public static createStore(projectName: string, childDirName?: string, style: StoreType = StoreType.Data) {
        // return new Store(projectName, childDirName);
        return new Promise((resolve: (store: Store) => void, reject) => {
            let store = new Store(projectName, childDirName, style);
            resolve(store);
        });
    }

    constructor(projectName: string, childDirName: string, style: StoreType) {
        this.dbName = projectName;
        this.tabName = childDirName;
        this.style = style;

    }

    /**
     * 获得主目录文件保存路径
     * @param path
     */
    public formatStorePath = (path?: string): string => {
        let p = "";
        if (this.tabName) {
            p = `${this.dbName}/${this.tabName}`;
        } else {
            p = `${this.dbName}`;
        }
        if (path) {
            p = `${p}/${path}`
        }
        return p;
    }

    public formatLocalStoragePath = (path?: string) => {
        path = this.formatStorePath(path);
        return path.replace("/", "_");
    }

    public read(key: string): Promise<any> {
        return new Promise((resolve, reject) => {
            FileLoadParam.storeWightQueuePool.read(this, key, () => { return this.__read(key) }, (data, err) => {
                if (err) {
                    reject(err)
                } else {
                    resolve(data);
                }
            })
        });
    }

    /**
     * 读取数据
     */
    public __read(key: string): Promise<any> {
        if (key === '') {
            key = '_';
        }
        if (this.style === StoreType.File) {
            return fileCache.get(this.formatStorePath(key)).catch(e => { return null });
        } else {
            // yuq
            return new Promise((resolve, reject) => {
                readStorage(this.formatLocalStoragePath(key)).then(resolve).catch((e) => resolve(null));
            });
        }
    }

    public write(key: string, data: Uint8Array, encryptType?: EncryptType, encryptKey?: string, compressAlgorithm?: CompressAlgorithm, compressLevel?: CompressLevel, immediate?: boolean): Promise<any> {
        if (immediate === undefined) {
            immediate = this.immediateWrite;
        }
        return new Promise((resolve, reject) => {
            FileLoadParam.storeWightQueuePool.write(this, key, () => { return this.__write(key, data) }, (_, err) => {
                if (err) {
                    reject(err)
                } else {
                    resolve(_);
                }
            }, immediate)
        });
    }

    /**
     * 写入数据，如果键名存在则替换
     * @param fileInfo 如果为空则启动文件写队列
     * @param data 文件数据
     */
    public __write(key: string, data: any): Promise<any> {
        if (key === '') {
            key = '_';
        }
        if (this.style === StoreType.File) {
            return fileCache.put(this.formatStorePath(key), data);
        } else {
            // yuq
            return new Promise((resolve, reject) => {
                let vtype, vData;
                if (data instanceof Uint8Array) {
                    vData = utf8Decode(data);
                    vtype = "Uint8Array";
                } else if (data instanceof ArrayBuffer) {
                    vData = utf8Decode(new Uint8Array(data));
                    vtype = "ArrayBuffer";
                } else if (typeof data === typeof "") {
                    vData = data;
                    vtype = "string";
                } else if (typeof data === typeof 0) {
                    vData = data + "";
                    vtype = "number";
                } else if (typeof data === typeof true) {
                    if (data === true) {
                        vData = "1";
                    } else {
                        vData = "0";
                    }
                    vtype = "boolean";
                } else if (data instanceof Object) {
                    if (data instanceof Map) {
                        vtype = "Map";
                        vData = JSON.stringify([...data]);
                    } else if (data instanceof Set) {
                        vtype = "Set";
                        vData = JSON.stringify([...data]);
                    } else {
                        vtype = "Object";
                        vData = JSON.stringify(data);
                    }
                } else {
                    throw new Error("Store.write failed!!, data format is invalid!!");
                }

                wx.setStorage({
                    key: this.formatLocalStoragePath(key),
                    data: JSON.stringify({ type: vtype, data: vData }),
                    fail(e) {
                        logger.debug("setStorage fail");
                        reject(e);
                    },
                    success() {
                        logger.debug("setStorage success");
                        resolve(null);
                    }


                });
            });
        }
    }
    /**
     * 删除数据
     */
    public delete(key: string): Promise<any> {
        if (this.style === StoreType.File) {
            return fileCache.delete(this.formatStorePath(key));
        } else {

            return new Promise((resolve, reject) => {
                wx.removeStorage({
                    key: this.formatLocalStoragePath(key),
                    success() {
                        resolve(null);
                    },
                    fail(e) {
                        reject(e)
                    }
                })
            });
        }
    }
    /**
     * 清除存储
     */
    public clear(): Promise<any> {
        if (this.style === StoreType.File) {
            return fileCache.clear(this.formatStorePath());
        } else {
            return new Promise((resolve, reject) => {
                wx.getStorageInfo({
                    success: (res) => {
                        let keys = res.keys as string[];
                        const dir = this.formatLocalStoragePath();
                        keys.forEach(key => {
                            if (key.startsWith(dir)) {
                                localStorage.removeItem(key);
                            }
                        });
                        resolve(null);
                    },
                    fail(e) {
                        reject(e);
                    }
                })
            });
        }
    }
    /**
     * 迭代, callback返回false表示停止迭代
     */
    public iterate(callback: (result: { key: any, value: any }) => boolean, errorCallback: (err: Event) => void) {
        setTimeout(() => {
            if (this.style === StoreType.File) {
                fileCache.iterate(this.formatStorePath(), callback, errorCallback)
            } else {
                wx.getStorageInfo({
                    success: async (res) => {
                        let keys = res.keys as string[];
                        const dir = this.formatLocalStoragePath();
                        let skip = false;
                        for (const key of keys) {
                            if (key.startsWith(dir)) {
                                try {
                                    const value = await readStorage(key);
                                    skip = callback && callback({ key, value });
                                } catch (e) {
                                    errorCallback && errorCallback(e);
                                }
                            }
                        }
                    },
                    fail(e) {
                        errorCallback(e);
                    }
                })
            }
        }, 0);
    }
}


function readStorage(key: string) {
    return new Promise((resolve, reject) => {
        wx.getStorage({
            key,
            fail(e) {
                logger.debug("getStorage fail", e, key);
                reject(e);
            },
            success(r) {
                logger.debug("getStorage success", r, key);
                try {
                    // 只要是项目传过来的，一定是 json字符串
                    if (typeof r.data !== typeof "") {
                        resolve(r.data);
                    } else {
                        let { type, data } = JSON.parse(r.data);
                        let vData = data;
                        if (type === "Uint8Array") {
                            vData = utf8Encode(data);
                        } else if (type === "ArrayBuffer") {
                            vData = utf8Encode(data).buffer;
                        } else if (type === "string") {
                            vData = data;
                        } else if (type === "number") {
                            vData = Number(data);
                        } else if (type === "boolean") {
                            vData = data === "1";
                        } else if (type === "Map") {
                            vData = new Map(JSON.parse(data))
                        } else if (type === "Set") {
                            vData = new Set(JSON.parse(data));
                        } else if (type === "Object") {
                            vData = JSON.parse(data);
                        } else {
                            throw new Error("Store.write failed!!, data format is invalid!!");
                        }
                        resolve(vData);
                    }
                } catch (e) {
                    resolve(r.data);
                }


            }
        })
    })

}