import { access, readFile, unlink, writeFile, mkdir } from "pi_sys/store/file_system";
import { dirname, writeFile as writeFile1 } from "../util/file_util";
import { Env } from "pi_sys/util/env";

class CacheNode {
    key: string;
    size: number;
    freq: number = 0;             // 访问频次
    constructor(key: string, size: number) {
        this.key = key;
        this.size = size;
    }

    toJson() {
        return [this.key, this.size];
    }
}

export class LFUCache {
    totalSize: number = 0;     // 总容量
    currentSize: number = 0;   // 当前已使用容量
    minFreq: number = 1;    // 最小使用频率
    freqMap: Map<number, Map<string, CacheNode>> = new Map;   // Map<频次,Map<key,CacheNode>>;
    cacheMap: Map<string, CacheNode> = new Map;          // Map<key,CacheNode>
    waitWriteMap: Map<string, { value: Uint8Array, success: Function, fail: Function, node?: CacheNode }> = new Map;
    writing: boolean = false;
    root: string = wx.env.USER_DATA_PATH;
    timer: any;
    MaxFreq: number = 20;
    existDirs: Set<string> = new Set;          // 已经存在的目录
    metaPath: string = `${wx.env.USER_DATA_PATH}/_meta`;         // meta信息存储文件名
    writeingMeta: Promise<any>;
    needWriteFreqMap: boolean = false;
    isOldVersion: boolean = false;        // 是否是旧版本


    async init(capacity: number, sdfTextSize: number): Promise<any> {
        this.totalSize = capacity - 10 * 1024 * 1024;    // 扣除10M用作冗余
        const useSdf = Env.get("use_sdf");
        if (useSdf) {
            this.totalSize = this.totalSize - sdfTextSize;
        }
        /**
         * 不用关心配置表是否已经变化,既然上次存储成功,就算变了也无所谓
         */
        try {
            let str = localStorage.getItem("lfuCache");
            if (str) {
                // 兼容旧版本
                this.isOldVersion = true;
            } else {
                str = await readFile(this.metaPath, "utf8") as string;
            }
            if (str) {
                const arr = JSON.parse(str);
                let maxFreq = 0;
                arr.forEach((item, index) => {
                    let freq = item[0];
                    if (index === 0) {
                        this.minFreq = freq;
                        maxFreq = freq;
                    }
                    if (this.minFreq > freq) {
                        this.minFreq = freq;
                    }
                    if (maxFreq < freq) {
                        maxFreq = freq;
                    }
                    let map = this.freqMap.get(freq);
                    if (!map) {
                        map = new Map;
                        this.freqMap.set(freq, map)
                    }
                    item[1].forEach(json => {
                        const node = new CacheNode(json[0], json[1]);
                        node.freq = freq;
                        this.cacheMap.set(node.key, node);
                        map.set(node.key, node);
                        this.currentSize += node.size;
                    });
                });
                if (maxFreq > this.MaxFreq) {
                    const newFreqMap: Map<number, Map<string, CacheNode>> = new Map;
                    // 降频
                    for (let i = 1; i <= maxFreq; i++) {
                        let oldMap = this.freqMap.get(i);
                        if (oldMap && oldMap.size) {
                            let newFreq = Math.ceil(i / 2);
                            let newMap = newFreqMap.get(newFreq);
                            if (!newMap) {
                                newMap = new Map;
                                newFreqMap.set(newFreq, newMap);
                            }
                            for (const [k, v] of oldMap) {
                                v.freq = newFreq;
                                newMap.set(k, v);
                                this.cacheMap.set(k, v);
                            }
                        }
                    }
                    this.freqMap = newFreqMap;
                }
            }
            console.log("lfu cache init success", this.totalSize, this.currentSize);
            if (this.isOldVersion) {
                this.needWriteFreqMap = true;
                this.checkWriteFreqMap();
            }
            return Promise.resolve(null);
        } catch (e) {
            return Promise.resolve(null);
        }


    }

    pathFormat(key: string) {
        return `${this.root}/${key}`;
    }

    writeFreqMap() {
        if (!this.timer) {
            this.timer = setTimeout(() => {
                this.needWriteFreqMap = true;
                this.checkWriteFreqMap();
            }, 200);
        }

    }

    checkWriteFreqMap() {
        if (!this.needWriteFreqMap) return;
        if (this.writeingMeta) return;
        this._writeFreqMap();
    }


    _writeFreqMap() {
        const arr = [];
        for (const [k, v] of this.freqMap) {
            if (v.size) {
                const jsons = [];
                for (const [_, c] of v) {
                    jsons.push(c.toJson())
                }
                arr.push([k, jsons]);
            }
        }
        this.timer = null;
        this.needWriteFreqMap = false;
        this.writeingMeta = writeFile1(this.metaPath, JSON.stringify(arr));
        return this.writeingMeta.then(() => {
            if (this.isOldVersion) {
                // 旧版本成功之后移除localStorage
                this.isOldVersion = false;
                localStorage.removeItem("lfuCache");
            }
            this.writeingMeta = null;
            this.checkWriteFreqMap();
        }).catch(e => {
            this.writeingMeta = null;
            this.checkWriteFreqMap();
        });
    }

    /**
     * 频率+1
     * ① 从freqMap的旧频率的key中移除
     * ② 旧频率为空 && 当前频率是最小频率，需要将最小频率+1
     * ③ 添加到新频率（旧频率+1）对应的map队列中 => 当前队列没有需要新建
     */
    increaseFreq(node: CacheNode, fixedFreq: number = 1) {
        let map = this.freqMap.get(node.freq)
        if (map) {
            map.delete(node.key)
            if (!map.size && node.freq === this.minFreq) {
                this.minFreq++;
            }
        }
        node.freq += fixedFreq;
        if (node.freq < this.minFreq) {
            this.minFreq = node.freq;
        }
        let newMap = this.freqMap.get(node.freq)
        if (newMap === undefined) {
            newMap = new Map()
            this.freqMap.set(node.freq, newMap);
        }
        newMap.set(node.key, node);
        this.cacheMap.set(node.key, node);

        this.writeFreqMap();
    }

    async freeSize(key: string, size: number) {
        let node = this.cacheMap.get(key);
        if (node) {
            await this.deleteNode(node)
        }
        while (this.currentSize + size > this.totalSize) {
            const k = this.getDeleteNodeKey();
            node = this.cacheMap.get(k);
            if (node) {
                await this.deleteNode(node);
            }

        }
    }

    getDeleteNodeKey(): string {
        let freqMap = this.freqMap.get(this.minFreq);
        let key: string;
        while (!freqMap) {
            this.minFreq++;
            freqMap = this.freqMap.get(this.minFreq);
        }
        if (freqMap) {
            key = freqMap.keys().next().value;
        }
        // 释放的时候忽略释放的文件,比如背景音乐,不能释放
        const freeIgnores = Env.get("freeIgnores") || new Set;
        if (key && freeIgnores.has(key)) {
            const node = this.cacheMap.get(key);
            if (node) {
                let maxFreq = 0;
                for (const freq of this.freqMap.keys()) {
                    if (freq > maxFreq) {
                        maxFreq = freq;
                    }
                }
                // 提升忽略文件的频率使其始终不被释放
                this.increaseFreq(node, maxFreq)
            }
            return this.getDeleteNodeKey();
        }
        if (!key) {
            this.minFreq++;
            return this.getDeleteNodeKey();
        } else {
            return key;
        }
    }

    async deleteNode(node: CacheNode) {
        let path = this.pathFormat(node.key);
        const exist = await access(path);
        if (exist) {
            await unlink(path);
        }
        this.currentSize -= node.size;
        this.cacheMap.delete(node.key);
        const freqMap = this.freqMap.get(node.freq);
        if (freqMap) {
            freqMap.delete(node.key);
            if (freqMap.size === 0) {
                this.freqMap.delete(node.freq);
            }
        }
    }

    writeNext() {
        if (this.writing) return;
        const key = this.waitWriteMap.keys().next().value;
        if (key) {
            this.writing = true;
            const obj = this.waitWriteMap.get(key);
            const size = obj.value.byteLength;
            this.waitWriteMap.delete(key);
            let node = this.cacheMap.get(key);
            this.freeSize(key, size).then(() => {
                let path = this.pathFormat(key);
                const dir = dirname(path);
                let p: Promise<any>;
                if (this.existDirs.has(dir)) {
                    p = Promise.resolve();
                } else {
                    p = access(dir).then(exist => {
                        if (!exist) {
                            return mkdir(dir);
                        }
                    }).then(() => {
                        this.existDirs.add(dir);
                    })
                }
                return p.then(() => {
                    return writeFile(path, obj.value.buffer).then(() => {
                        // console.log("writeNext111", Date.now() - time);
                        if (node) {
                            node.size = size;
                        } else {
                            node = new CacheNode(key, size);
                        }
                        this.increaseFreq(node);
                        this.currentSize += node.size;
                        obj.success();
                    });
                })
            }).catch(e => {
                console.log("writeFile fail", e);
                obj.fail(e);
            }).finally(() => {
                // console.log("writeFile complete");
                this.writing = false;
                this.writeNext();
            });
        }
    }


    async put(key: string, value: Uint8Array) {
        return new Promise((resolve, reject) => {
            if (!(value instanceof Uint8Array)) {
                console.log("==== put error value", value)
                return reject();
            }
            this.waitWriteMap.set(key, { value, success: resolve, fail: reject });
            this.writeNext();
        });
    }

    get(key: string) {
        const node = this.cacheMap.get(key);
        if (!node) return Promise.resolve(null);
        this.increaseFreq(node);
        let path = this.pathFormat(key);
        return readFile(path);
    }

    has(key: string) {
        return this.cacheMap.has(key);
    }

    delete(key: string) {
        const node = this.cacheMap.get(key);
        if (node) {
            return this.deleteNode(node);
        } else {
            return Promise.resolve();
        }
    }

    // 清除某个目录下的记录,默认清除跟路径下所有
    clear(dir: string = "") {
        const arr = [];
        for (const [key, node] of this.cacheMap) {
            if (key.startsWith(dir)) {
                arr.push(this.deleteNode(node));
            }
        }
        return Promise.all(arr);
    }

    async iterate(dir: string = "", callback: (result: { key: any, value: any }) => boolean, errorCallback: (err: Event) => void) {
        let skip = false;
        for (const [key, node] of this.cacheMap) {
            if (skip) return;
            if (key.startsWith(dir)) {
                try {
                    const value = await this.get(key);
                    skip = callback && callback({ key, value });
                } catch (e) {
                    errorCallback && errorCallback(e);
                }

            }
        }
    }
}


export const fileCache = new LFUCache();