import { WebGLInstance, WebGLInstanceOpt, ShaderCfg, Scene, DataBufferCfg, Mesh, TextureInstance } from "./webgl";
import { vs_texture, fs_texture } from "./shader_texture";
import { vs_progress, fs_progress } from "./shader_progress";

export type RenderFlag = 'progress';
/**
 * 图片渲染信息
 * @example 全屏背景        { url: "", t: "lt", x: 0, y: 0, w: 1, h: 1 }
 * @example 中心Icon        { url: "", t: "lt", x: 0.35, y: 0.2, w: 0.3, h: 0.3 }
 * @example 下方防沉迷提示  { url: "", t: "lt", x: 0.1, y: 0.75, w: 0.8, h: 0.2 }
 * 
 */
export interface PlaneInfo {
    /**
     * 图片路径
     */
    url: string;
    /**
     * 坐标系模式
     * @lt - left-top - css
     * @lb - left-bottom - webgl
     */
    t?: "lt" | "lb";
    /**
     * 矩形起点 x
     * 基于全屏的百分比坐标
     */
    x: number;
    /**
     * 矩形起点 y
     * 基于全屏的百分比坐标
     */
    y: number;
    /**
     * 矩形 宽度
     * 基于全屏的百分比
     */
    w: number;
    /**
     * 矩形 高度
     * 基于全屏的百分比
     */
    h: number;
}
export interface ProgressArgs {
    bgurl?: string;
    planes?: PlaneInfo[];
    /**
     * 进度条背景色 【0~1】
     */
    bgColor?: [number, number, number];
    /**
     * 进度条前景色 【0~1】
     */
    pgColor?: [number, number, number];
}

export class RenderLauncher {
    public static handler: number = 0;
    public static webgldemo: WebGLInstance;
    public static opt: any = {};
    public static active(canvas: HTMLCanvasElement, flag: RenderFlag, args: any) {
        switch (flag) {
            case 'progress': {
                if (args.image) {
                    RenderLauncher.createTextureLoad = (fname: string, engine: WebGLInstance, cb: (img: HTMLImageElement, fname: string, engine: WebGLInstance) => void) => {
                        RenderLauncher.loadImageSucc(args.image, fname);
                    };
                }
                RenderLauncher.progress(canvas, args);
            }
            default: {

            }
        }
    }
    public static puase() {

    }
    public static destroy() {
        if (RenderLauncher.webgldemo && !RenderLauncher.webgldemo.isDestroy) {
            cancelAnimationFrame(RenderLauncher.handler);
            RenderLauncher.webgldemo.destroy();
            (<any>RenderLauncher.webgldemo) = null;

            if (RenderLauncher.opt.meshProgress) {
                RenderLauncher.opt.meshProgress = undefined;
            }
        }
    }
    public static loadImageSucc = (img: HTMLImageElement, fname: string) => {
        TextureInstance.loaded(img, fname, RenderLauncher.webgldemo);
    }
    public static createTextureLoad = (fname: string, engine: WebGLInstance, cb: (img: HTMLImageElement, fname: string, engine: WebGLInstance) => void) => {
        const img = wx.createImage();

        img.onload = () => {
            RenderLauncher.loadImageSucc(img as any, fname);
            // setTimeout(() => { loadImageSucc(img, data.fname); }, 2000);
        };
        img.src = fname;
    }
    public static createPlane(scene: Scene, info: PlaneInfo, index: number, shader: ShaderCfg) {
        const webgldemo: WebGLInstance = RenderLauncher.webgldemo;

        if (!webgldemo.gl) {
            return;
        }

        const dataBuffer = new DataBufferCfg(`Plane${index}`);
        info.x = info.x * 2 - 1;
        info.y = info.y * 2 - 1;
        // info.w = info.w * 2;
        // info.h = info.h * 2;
        const halfW = info.w; //info.w / 2;
        const halfH = info.h; //info.h / 2;
        let cx = info.x + halfW;
        let cy = info.y + halfH;
        if (info.t == "lt") {
            cy = (-info.y) - halfH;
        }

        dataBuffer.addVertex(cx - halfW, cy - halfH, 0);
        dataBuffer.addUV(0, 0);
        dataBuffer.addVertex(cx + halfW, cy - halfH, 0);
        dataBuffer.addUV(1, 0);
        dataBuffer.addVertex(cx + halfW, cy + halfH, 0);
        dataBuffer.addUV(1, 1);
        dataBuffer.addVertex(cx - halfW, cy + halfH, 0);
        dataBuffer.addUV(0, 1);
        dataBuffer.addFace(0, 1, 2);
        dataBuffer.addFace(0, 2, 3);
        dataBuffer.update(webgldemo.gl);

        const mesh = new Mesh(`mesh_Plane${index}`, dataBuffer, shader);
        mesh.translate[0]   = 0.0;
        mesh.translate[1]   = 0.0;
        mesh.scale[0]       = 1;
        mesh.scale[1]       = 1;
        mesh.texture        = webgldemo.createTexture(info.url);
        scene.addMesh(mesh);
    }
    public static progress(canvas: HTMLCanvasElement, args: ProgressArgs) {
        const opt: WebGLInstanceOpt = <any>{};

        opt.canvas = canvas;

        RenderLauncher.webgldemo = new WebGLInstance(opt);
        TextureInstance.loadCall = RenderLauncher.createTextureLoad;

        const webgldemo: WebGLInstance = RenderLauncher.webgldemo;

        if (!webgldemo.gl) {
            return;
        }

        const scene = new Scene('06', webgldemo);

        const shader05 = new ShaderCfg('image', vs_texture, fs_texture);
        if (args.bgurl) {
            this.createPlane(scene, { url: args.bgurl, x: 0, y: 0, w: 1, h: 1 }, -1, shader05);
        }
        if (args.planes) {
            const count = args.planes.length;
            for (let i = 0; i < count; i++) {
                this.createPlane(scene, args.planes[i], i, shader05);
            }
        }

        const dataBuffer03 = new DataBufferCfg('03');
        dataBuffer03.addVertex(-1, -1, 0);
        dataBuffer03.addVertex(1, -1, 0);
        dataBuffer03.addVertex(1, -1 + 20 / canvas.height, 0);
        dataBuffer03.addVertex(-1, -1 + 20 / canvas.height, 0);
        dataBuffer03.addFace(0, 1, 2);
        dataBuffer03.addFace(0, 2, 3);
        dataBuffer03.update(webgldemo.gl);

        const shader06 = new ShaderCfg('06', vs_progress, fs_progress);
        const meshProgress = new Mesh('meshProgress', dataBuffer03, shader06);
        meshProgress.translate[0] = 0.0;
        meshProgress.translate[1] = 0.0;
        meshProgress.scale[0] = 1;
        meshProgress.scale[1] = 1;
        meshProgress.ufloat = 0.01;
        scene.addMesh(meshProgress);

        let bgColor = args.bgColor || [0.2, 0.2, 0.2];
        let pgColor = args.pgColor || [0.0, 0.8, 0.0];
        meshProgress.renderExtends = (shader: ShaderCfg, gl: WebGLRenderingContext) => {
            if ((<any>shader).u_color_bg == undefined) {
                (<any>shader).u_color_bg = <WebGLUniformLocation>gl.getUniformLocation((<any>shader).shader_program, `u_color_bg`);
            }
            if ((<any>shader).u_color_pg == undefined) {
                (<any>shader).u_color_pg = <WebGLUniformLocation>gl.getUniformLocation((<any>shader).shader_program, `u_color_pg`);
            }
            (<any>shader).u_color_bg       && gl.uniform3f((<any>shader).u_color_bg,     bgColor[0],     bgColor[1],     bgColor[2]);
            (<any>shader).u_color_pg       && gl.uniform3f((<any>shader).u_color_pg,     pgColor[0],     pgColor[1],     pgColor[2]);
        };

        webgldemo.renderLoop = (timestamp) => {
            // GUI 初始化后会修改canvas大小
            (<any>webgldemo).width = webgldemo.canvas.width;
            (<any>webgldemo).height = webgldemo.canvas.height;

            webgldemo.clearColor();
            scene.viewport[0] = 0;
            scene.viewport[1] = 0;
            scene.viewport[2] = webgldemo.width;
            scene.viewport[3] = webgldemo.height;
            scene.render(true);
        };

        function func(time) {
            RenderLauncher.webgldemo.loop(time);
            RenderLauncher.handler = requestAnimationFrame(func);
        }
        func(0);

        RenderLauncher.opt.meshProgress = meshProgress;
    }
    public static updateProgress(num: number) {
        if (RenderLauncher.opt.meshProgress) {
            RenderLauncher.opt.meshProgress.ufloat = num;
        }
    }
}