
import { BaseTools } from "../util/base_tools";
import { FileLoadParam, waitLoad, IFileLoad } from "./fileload_base";
import { handleBinMap } from "./cfg_tools";
import { CodeLoad } from "./code_load";
import { FileLoad } from "./file_load";
import { Download } from "./down_load";
import { LocalLoad } from "./local_load";
import * as Sign from "./sign";
import { loadFile, processValue } from "./batch_load_hook";
import { Depend } from "../util/depend";
import { IDirDependInfo, IFileDependInfo } from "../util/depend_base";
import { PackageLoad } from "./package_load";

import { Logger } from "pi_sys/util/logger";
import { Env } from "pi_sys/util/env";
import { formatError } from "pi_sys/util/util";
declare var module;
let logger = new Logger(module.name, "sys");


// 加载策略，用于batchload，做预处理到内存的策略
export enum LoadPolicy {
    No = 1,         // 不下载，不加载
    DownOnly = 2,   // 如果本地是最新的，不会加载到内存；
    All = 3,        // 无论是在服务器，还是本地，都要加载到内存
}



/**
 * @description 加载指定文件或目录(以/结尾)及子目录下的所有代码、配置、资源和对象。
 * 异步返回成功或失败。加载器上可以设置进度监听。
 * 次序是按照目录逐层加载，目录按序深度遍历加载
 * @example
 */
export class BatchLoad extends FileLoad {
    dirOrFiles: string[];
    fileFilters: RegExp[][] = [[], []]; // 第0位为匹配路径， 第1位为排除匹配
    dirFilters: RegExp[][] = [[], []]; // 第0位为匹配路径， 第1位为排除匹配
    loads: Set<IFileLoad> = new Set;
    resLoadPolicy: LoadPolicy = LoadPolicy.No;
    objectLoadPolicy: LoadPolicy = LoadPolicy.No;
    codeLoadPolicy: LoadPolicy = LoadPolicy.All;
    cfgLoadPolicy: LoadPolicy = LoadPolicy.All;
    preloadKey: string;    // 为了优化小游戏加载性能,代码包有可能提前加载
    waitWrite: boolean = false;     // 某些情况下需要等待下载完成且存储到本地成功后才算成功

    /**
     * @description 加载指定文件或目录(以/结尾)及子目录下的所有代码、配置、资源和对象
     */
    public constructor(dirOrFiles: string[]) {
        super();

        const len = dirOrFiles.length;

        this.dirOrFiles = [];

        for (let i = 0; i < len; i++) {
            const path = dirOrFiles[i];
            this.dirOrFiles.push(path);
        }
    }

    /**
     * @description 添加文件和目录的模式匹配。支持**, *。
     * 如果开头为!，表示为排除的模式匹配
     * 如果最后结尾为/，表示为目录的模式匹配
     * @example
     */
    public addFilter(s: string) {
        let filters = s.charAt(s.length - 1) === '/' ? this.dirFilters : this.fileFilters;
        let arr = filters[0];
        if (s.charAt(0) === '!') {
            arr = filters[1];
            s = s.slice(1);
        }
        arr.push(BaseTools.pattern(s));
    }

    /**
     * @description 加载, 是否对象资源及资源仅下载（如果已经在本地，则不加载）
     */
    public start(resLoadPolicy: LoadPolicy = LoadPolicy.No, objectLoadPolicy: LoadPolicy = LoadPolicy.No, secondLoad?: boolean) {
        return this.load(undefined, resLoadPolicy, objectLoadPolicy, LoadPolicy.All, LoadPolicy.All, secondLoad);
    }

    /**
     * @description 加载资源, 是否对象资源及资源仅下载
     */
    public startRes(resLoadPolicy: LoadPolicy = LoadPolicy.DownOnly, objectLoadPolicy: LoadPolicy = LoadPolicy.DownOnly) {
        return this.load(undefined, resLoadPolicy, objectLoadPolicy, LoadPolicy.No, LoadPolicy.No);
    }

    /**
     * @description 加载, 是否对象资源及资源仅下载（如果已经在本地，则不加载）
     * @param onlyDown  已过时
     */
    public load(onlyDown: boolean = false, resLoadPolicy: LoadPolicy = LoadPolicy.No, objectLoadPolicy: LoadPolicy = LoadPolicy.No, codeLoadPolicy: LoadPolicy, cfgLoadPolicy: LoadPolicy, secondLoad?: boolean) {
        const collect = Env.get("collect");
        if (collect) { // 如果是收集模式 资源和资源对象不下载,等到用到的时候再下载,这样更能体现资源优先级
            resLoadPolicy = LoadPolicy.No;
            objectLoadPolicy = LoadPolicy.No;
        }

        this.resLoadPolicy = resLoadPolicy;
        this.objectLoadPolicy = objectLoadPolicy;
        this.codeLoadPolicy = codeLoadPolicy;
        this.cfgLoadPolicy = cfgLoadPolicy;

        let localload = new LocalLoad();
        let download = new Download();
        download.waitWrite = this.waitWrite;
        let codeload = new CodeLoad();
        let packageLoad = new PackageLoad();
        let objload = undefined;

        let arr = [];

        let prePackageLoad: PackageLoad;
        let prePackageLoadPromise: Promise<any>;
        if (this.preloadKey !== undefined) { // 在winit里面提前加载小游戏代码
            prePackageLoad = Env.get(this.preloadKey + "load_packageLoad");
            prePackageLoadPromise = Env.get(this.preloadKey + "load_packageLoad_promise");
        }

        for (let s of this.dirOrFiles) {
            if (s.charAt(s.length - 1) === '/') {
                if (!prePackageLoad) {
                    packageLoad.addPackage(s);
                }
                this.loadDir(s, Depend.getDir(s), onlyDown, localload, download, codeload, objload, arr);
            } else {
                let info = Depend.getFile(s);
                if (info)
                    this.loadFile(info, onlyDown, localload, download, codeload, objload, arr);
                else {
                    logger.warn("batchload, invalid file, ", s);
                }
            }
        }
        if (localload.files.size) {
            let p = localload.start();
            arr.push(waitLoad(localload, FileLoadParam.localLoadSet, p.then(map => handleBinMap(map, secondLoad))));
            this.addLoad(localload);
        }
        if (download.files.size) {
            let startTime = Date.now();
            let p = download.start().then((r) => {
                console.log("============= download ", Date.now() - startTime);
                return r;
            });
            arr.push(waitLoad(download, FileLoadParam.downLoadSet, p.then(map => handleBinMap(map, secondLoad))));
            this.addLoad(download);
        }
        if (codeload.files.size) {
            let p = codeload.start();
            arr.push(waitLoad(codeload, FileLoadParam.codeLoadSet, p));
            this.addLoad(codeload);
        }
        // if (objload.files.size) {
        //     let p = objload.start();
        //     arr.push(waitLoad(objload, FileLoadParam.objLoadSet, p));
        //     this.addLoad(objload);
        // }

        if (prePackageLoad) {
            Promise.all(arr).then(() => {
                Env.set(this.preloadKey + "load_result", true);
            });
            arr.push(waitLoad(prePackageLoad, FileLoadParam.packageLoadSet, prePackageLoadPromise));
        } else {
            if (packageLoad.packages.size > 0) {
                let p = packageLoad.start();
                arr.push(waitLoad(packageLoad, FileLoadParam.packageLoadSet, p));
                this.addLoad(packageLoad);
            }
        }

        return Promise.all(arr)
            .then(
                (value: any) => {
                    this.onResult(value);
                }
            ).catch(
                (reason: any) => {
                    logger.error("pi_sys, BatchLoad failed, reason = ", formatError(reason));
                    this.onResult(null, reason);
                    throw reason;
                }
            );
    }
    loadDir(
        path: string,
        dir: IDirDependInfo,
        onlyDown: boolean,
        localload: LocalLoad,
        download: Download,
        codeload: CodeLoad,
        objload: any,
        result: Promise<any>[]) {
        if (!dir)
            return;
        // let s = dir.path.slice(path.length); // 要用从该目录起的相对路径
        let s = dir.path;
        for (let r of this.dirFilters[0]) {
            if (!r.test(s))
                return;
        }
        for (let r of this.dirFilters[1]) {
            if (r.test(s))
                return;
        }
        if (dir.files) {
            for (let f of dir.files) {
                // s = f.path.slice(path.length);
                s = f.path;
                if (BaseTools.filter(s, this.fileFilters[0], this.fileFilters[1])) {
                    this.loadFile(f, onlyDown, localload, download, codeload, objload, result);
                }
            }
        }
        if (dir.children) {
            for (let d of dir.children) {
                this.loadDir(path, d, onlyDown, localload, download, codeload, objload, result);
            }
        }
    }
    loadFile(
        file: IFileDependInfo,
        onlyDown: boolean,
        localload: LocalLoad,
        download: Download,
        codeload: CodeLoad,
        objload: any,
        result: Promise<any>[]) {
        return loadFile(this, file, onlyDown, localload, download, codeload, objload, result);
    }
    // 下载或加载
    downOrload(
        file: IFileDependInfo,
        onlyDown: boolean,
        localload: LocalLoad,
        download: Download,
        result: Promise<any>[]) {
        if (file.sign === Sign.getSign(file.path)) {
            if ((!onlyDown) && this.checkLoad(file, FileLoadParam.localLoadSet, result))
                localload.add(file);
        } else {
            if (this.checkLoad(file, FileLoadParam.downLoadSet, result)) {
                download.add(file);
            }
        }
    }
    // 检查是否正在加载
    checkLoad(file: IFileDependInfo, set: Set<IFileLoad>, result: Promise<any>[]) {
        for (let load of set) {
            if (load.files.has(file.path)) { // 文件正在加载
                if (!this.loads.has(load)) {
                    result.push(new Promise((resolve, reject) => {
                        load.addResult((val: any, err?: any) => {
                            err ? reject(err) : resolve(val)
                        });
                    }));
                    this.addLoad(load);
                }
                return false;
            }
        }
        return true
    }
    // 在下载器上添加进度通知
    addLoad(load: IFileLoad) {
        load.addProcess(this.processValue.bind(this))
        this.loads.add(load);
        this.total += load.total;
        this.loaded += load.loaded;
    }
    // 获得子加载器的进度通知，并计算当前进度，抛出进度事件
    processValue(url: string, type: string, total: number, loaded: number, data?: Uint8Array) {
        return processValue(this, url, type, total, loaded, data);
    }
}