
import { SuffixType, FileLoadParam, getLoadSuffixInfo } from "./fileload_base";
import { ResCacheMap } from "./res_cache";
import { cfgMap, cfgTempMap } from "./cfg_tools";
import { CodeLoad } from "./code_load";
import { Download } from "./down_load";
import { LocalLoad } from "./local_load";
import { BatchLoad, LoadPolicy } from "./batch_load";
import { IFileDependInfo } from "../util/depend_base";

import { Logger } from "pi_sys/util/logger";
import { Env } from "pi_sys/util/env";
import { BaseTools } from "pi_sys/util/base_tools";
declare var module;
let logger = new Logger(module.name, "sys");

export const loadFile = (
    batchLoad: BatchLoad,
    file: IFileDependInfo,
    onlyDown: boolean,
    localload: LocalLoad,
    download: Download,
    codeload: CodeLoad,
    _objload: any,
    result: Promise<any>[]
) => {
    // 文件长度为0，跳过
    if (!file.size) {
        return;
    }
    let [suffix, st] = getLoadSuffixInfo(file.path);

    if (!st) {
        const platform = Env.get("platform_type");
        // 这里测试微信小游戏js文件suffix取到的是null,所以需要取真实后缀
        if (platform === "minigame") {
            suffix = BaseTools.fileSuffix(file.path);
            if (suffix === "js") return;   // 微信小游戏并且是js 不警告(警告太多,卡)
        }
        return logger.warn(`batch load, invalid suffix, file: ${file.path}`);
    }

    switch (st) {
        case (SuffixType.CFG): {
            // 先检查是否在cfgMap
            let rr = cfgMap.get(file.path);
            if (rr) {
                result.push(
                    new Promise((resolve, reject) => {
                        rr.push(
                            (val: any, err?: any) => {
                                err
                                    ?
                                    reject(err)
                                    :
                                    resolve(val);
                            }
                        );
                    })
                );
                return;
            }

            // 是否在cfgTempMap
            const rd = cfgTempMap.get(file.path);
            if (rd) {
                return;
            }
            if (batchLoad.cfgLoadPolicy > LoadPolicy.No) {
                const onlyDown = batchLoad.cfgLoadPolicy === LoadPolicy.DownOnly;
                batchLoad.downOrload(file, onlyDown, localload, download, result);
            }
            break;
        }
        case (SuffixType.CODE): {
            /**
             * codeload只有debug的时候始终从网络下载,release当作cfg处理
             * 起始加载的时候强制LoadPolicy.All
             */
            if (batchLoad.codeLoadPolicy === LoadPolicy.All && batchLoad.checkLoad(file, FileLoadParam.codeLoadSet, result)) {
                codeload.add(file);
            }
            break;
        }
        case (SuffixType.RES): {
            let cache = ResCacheMap.getCache(suffix);
            if (!cache || !cache.map.get(file.path)) {
                if (batchLoad.resLoadPolicy > LoadPolicy.No) {
                    const onlyDown = batchLoad.resLoadPolicy === LoadPolicy.DownOnly;
                    batchLoad.downOrload(file, onlyDown, localload, download, result);
                }
            }
            break;
        }
        case (SuffixType.OBJ): {
            let cache = ResCacheMap.getCache(suffix);
            if (!cache || !cache.map.get(file.path)) {
                if (batchLoad.objectLoadPolicy > LoadPolicy.No) {
                    const onlyDown = batchLoad.objectLoadPolicy === LoadPolicy.DownOnly;
                    batchLoad.downOrload(file, onlyDown, localload, download, result);
                }
            }
            break;
        }
    }
    // if (st === SuffixType.RES) {
    //     // 如果缓存中有，则跳过
    //     let cache = ResCacheMap.getCache(suffix);
    //     if (cache.map.get(file.path)) {
    //         return;
    //     }

    //     batchLoad.downOrload(file, onlyDown, localload, download, result);
    // } else if (st === SuffixType.OBJ) {
    //     batchLoad.downOrload(file, onlyDown, localload, download, result);
    // } else if (st === SuffixType.CFG) {
    //     // 先检查是否在cfgMap
    //     let rr = cfgMap.get(file.path);
    //     if (rr) {
    //         result.push(new Promise((resolve, reject) => {
    //             rr.push((val: any, err?: any) => {
    //                 err ? reject(err) : resolve(val);
    //             });
    //         }));
    //         return;
    //     } else if (rr === null) {
    //         // 忽略已经处理过的配置
    //         return;
    //     }
    //     // 是否在cfgTempMap
    //     let rd = cfgTempMap.get(file.path);
    //     if (rd) {
    //         return;
    //     }
    //     batchLoad.downOrload(file, false, localload, download, result);
    // } else {
    //     if (batchLoad.checkLoad(file, FileLoadParam.codeLoadSet, result)) {
    //         codeload.add(file);
    //     }
    // }
}

// 获得子加载器的进度通知，并计算当前进度，抛出进度事件
export const processValue = (batchLoad: BatchLoad, url: string, type: string, total: number, loaded: number, data?: Uint8Array) => {
    batchLoad.loaded = 0;
    for (let load of batchLoad.loads) {
        // console.log("====== processValue ", load);
        batchLoad.loaded += load.loaded;
    }
    // console.log("======== batchLoad.onProcess ", type, batchLoad.total, batchLoad.loaded, batchLoad.process);
    batchLoad.onProcess(url, type, batchLoad.total, batchLoad.loaded);

    // 注：下面的代碼是为了 加快在batch下多个url下载时的进度，在 这里 提前 解析配置，目前先注释掉，没这个需求；

    // if (data) {
    //     let [suffix, st] = getLoadSuffixInfo(url);
    //     if (st === SuffixType.CFG) {
    //         handleCfg(url, data, suffix);
    //     }

    // }
}