import { SuffixType, FileLoadParam, getLoadSuffixInfo } from "./fileload_base";
import { BaseTools } from "../util/base_tools";
import { ResCacheMap } from "./res_cache"

import { Logger } from "pi_sys/util/logger";
import { Depend } from "pi_sys/util/depend";
import { BatchLoad, LoadPolicy } from "./batch_load";
import { Env } from "pi_sys/util/env";
import { getFileLang } from "pi_sys/util/util";
import { showRestartModel } from "pi_sys/setup/main";
import { ErrorCode } from "pi_sys/util/constant";
declare var module;
let logger = new Logger(module.name, "sys");

/**
 * 配置处理表
 */
export const handlerMap: Map<string, (file: string, data: Uint8Array) => Promise<any>> = new Map;

/**
 * 每配置状态表
 *      如果Promise有值，表示正在处理。
 *      为null，表示已处理完毕
 */
export const cfgMap: Map<string, Promise<any>> = new Map;

/**
 * 每配置的临时表
 */
export const cfgTempMap: Map<string, Uint8Array> = new Map;

/**
 * 每配置的永久存储表  会常驻内存  (功能：中英文切换 CSS使用不同版本)
 */
export const cfgPermanentMap: Map<string, Uint8Array> = new Map;

/**
 * 设置配置的后缀类型及处理函数， 如果处理函数为null, 表示将配置的二进制数据临时存放。 如果已经有临时存放的配置，则会立即进行处理
 * @param suffix 
 * @param handler 
 */
export let setCfgHandler = (suffix: string, handler: (file: string, data: Uint8Array) => Promise<any>): Promise<any> => {
    FileLoadParam.suffixMap.set(suffix, SuffixType.CFG);
    handlerMap.set(suffix, handler);
    if (!handler) {
        return Promise.resolve();
    }

    let arr = [];
    // console.warn("cfgTempMap === ", cfgTempMap, cfgTempMap.size);
    for (let [k, v] of cfgTempMap) {
        // console.warn("file ", k);
        let s = BaseTools.filePseudoSuffix(k);
        while (s) {
            if (s === suffix) {
                break;
            }
            s = BaseTools.filePseudoSuffix(s);
        }

        if (!s) {
            continue;
        }
        const needIgnore = filterLangCfg(k, v, suffix);
        if (needIgnore) continue;
        // 如果handler返回false，表示handler未处理该文件
        arr.push(handler(k, v).then((r) => {
            if (r !== false) {
                cfgTempMap.delete(k);
            }
        }));
    }
    return Promise.all(arr);
}

/**
 * 二进制文件 处理
 * 仅处理：res 和 cfg
 */
export const handleBinMap = (map: void | Map<string, Uint8Array>, secondLoad?: boolean) => {
    if (!map) {
        const reason = "pi_sys, handleBinMap, map = null";
        showRestartModel(ErrorCode.DATA_IS_NULL, reason);
        throw new Error(reason);
    }

    let arr = [];

    const loadFiles = [];
    /**
     * 这里只尝试下载一次
     * 有遇到一下情况：
     * 项目未推送dst_release目录,导致字节大小始终匹配不成功
     * 第二次如果还不匹配,弹框提示
     */
    for (let [k, v] of map) {
        let [suffix, st] = getLoadSuffixInfo(k);
        const depend = Depend.getFile(k);
        /**
         * 目前发现windows平台会出现写文件返回成功但实际文件落地失败的情况
         * 这里校验数据正确性,如果不正确,重新网络下载走一遍BatchLoad流程
         */
        if (!v || v.byteLength !== depend.size) {
            loadFiles.push(depend.path);

            if (FileLoadParam.assetsSign && FileLoadParam.assetsSign[depend.path]) {
                delete FileLoadParam.assetsSign[depend.path];
            }
            if (FileLoadParam.assetsCopySign && FileLoadParam.assetsCopySign[depend.path]) {
                delete FileLoadParam.assetsCopySign[depend.path];
            }
            if (FileLoadParam.dbSign && FileLoadParam.dbSign[depend.path]) {
                delete FileLoadParam.dbSign[depend.path];
            }
            if (secondLoad) {
                const error = `File ${k} byteLength mistake,depend size ${depend.size},but get length ${v?.byteLength || 0}`;
                logger.error(error);
            }
            // throw new Error(error);
        } else {
            if (st === SuffixType.RES || st === SuffixType.OBJ) {
                let cache = ResCacheMap.getCache(suffix);
                cache.add(k, v);
            } else if (st === SuffixType.CFG) {
                arr.push(handleCfg(k, v, suffix));
            }
        }

    }

    if (loadFiles.length > 0) {
        if (secondLoad) {
            const message = `File byteLength mistake,check depend and file length!!! Files = ${loadFiles.join(",")}`;
            showRestartModel(ErrorCode.BYTELENGTH_MISTAKE, message);
            return Promise.reject("Load res byteLength mistake!");
        }
        // 如果是网络下载大小不匹配,再次下载的时候需要从downLoadSet里面把对应下载文件删除掉
        if (FileLoadParam.downLoadSet) {
            for (let load of FileLoadParam.downLoadSet) {
                for (let file of loadFiles) {
                    if (load.files.has(file)) { // 文件正在加载
                        load.files.delete(file);
                    }
                }
            }
        }
        let load = new BatchLoad(loadFiles);
        // 这里需要显示指定代码和配置的加载策略为All
        let loadPromise = load.start(LoadPolicy.All, LoadPolicy.All, true);
        arr.push(loadPromise);
    }

    return Promise.all(arr);
}

// 配置处理
export const handleCfg = (file: string, data: Uint8Array, suffix: string) => {
    // 还没有注册处理器，直接返回
    // 注：这种情况下，要 setCfgHandler 保证 配置加载的同步性；
    let handler = handlerMap.get(suffix);
    if (handler === null) {
        cfgTempMap.set(file, data);
        return Promise.resolve();
    }

    // p值 为 undefined，代表没有处理过；
    // 否则就是正在处理；
    let p = cfgMap.get(file);

    if (!p) {
        const needIgnore = filterLangCfg(file, data, suffix);
        if (needIgnore) {
            return Promise.resolve();
        }
        p = handler(file, data).then((r) => {
            // 如果handler返回false，表示handler未处理该文件
            if (r !== false) {
                cfgMap.delete(file);
            }
        });
        cfgMap.set(file, p);
    }

    return p.then(() => { });
};

export const switchLang = (lang: string) => {
    Env.set("current_lang", lang);
    const langs = Env.get("langs");
    let arr = [];
    for (let [k, v] of cfgPermanentMap) {
        const fileLang = getFileLang(k, langs);
        if (fileLang === lang) {
            let [suffix, st] = getLoadSuffixInfo(k);
            arr.push(handleCfg(k, v, suffix));
        }
    }
    return Promise.all(arr);
}

// ============================== 本地

const filterLangCfg = (file: string, data: Uint8Array, suffix: string) => {
    const lang_cfg_suffixs = Env.get("lang_cfg_suffixs");
    if (lang_cfg_suffixs && lang_cfg_suffixs.includes(suffix)) {
        const current_lang = Env.get("current_lang");
        const langs = Env.get("langs");
        if (current_lang && langs && langs.length) {
            const fileLang = getFileLang(file, langs);
            if (fileLang) {
                cfgPermanentMap.set(file, data);
                if (fileLang !== current_lang) {
                    // console.warn("ignore exec ", file);
                    if (cfgTempMap.has(file)) {
                        cfgTempMap.delete(file);
                    }
                    return true;
                }

            }

        }
    }
    return false;
}

// ============================== 立即执行