
import { BaseTools } from "../util/base_tools"
import { IPiStoreBase } from "../store/store_base";
import { IFileDependInfo } from "../util/depend_base";
import { ProcessFunc, ResultFunc } from "../net/http_base";
import { NetWeightQueuePool } from "pi_sys/util/net_weight_queue_pool";
import { StoreWeightQueuePool } from "pi_sys/util/store_weight_queue_pool";

export type TFileLoadProgressType = 'BATCH_FILE_LOAD' | 'FILE_LOAD';

export interface IFileLoad {
    /**
     * 多个加载文件
     */
    files: Map<string, IFileDependInfo>;
    /**
     * 加载进度的回调函数
     */
    process: ProcessFunc[];
    /**
     * 结果的回调函数
     */
    result: ResultFunc[];
    /**
     * 加载的总文件长度
     */
    total: number;
    /**
     * 加载的当前进度
     */
    loaded: number;
    /**
     * @description 添加下载文件
     * @example
     */
    add(fname: IFileDependInfo): void;
    /**
     * @description 进度响应
     * @example
     */
    onProcess(url: string, type: string, total: number, loaded: number, data?: any): void;
    /**
     * @description 添加进度监听
     * @example
     */
    addProcess(p: ProcessFunc): void;
    /**
     * @description 结果响应
     * @example
     */
    onResult(val: any, err?: any): void;
    /**
     * @description 添加结果监听
     * @example
     */
    addResult(r: ResultFunc): void;
    /**
     * @description 开始
     * @example
     */
    start(): Promise<any>;
}

export enum SuffixType {
    CODE = 1,
    CFG,
    OBJ,
    RES,
}

export class Result {
    url: string = '';
    d: string = '';
    f: string = '';
    files: IFileDependInfo[] = [];
    size: number = 0;
    next: boolean = false;
}

export const FileLoadParam: IFileLoadParam = {
    urls: [],
    batchPath: "",
    limitLength: 1024 - 100,
    sizeLimit: 8 * 1024 * 1024,
    localStore: null,
    assetsSign: {},      // 存储在assets目录下的资源 androids windows ios平台有
    assetsCopySign: {},  // 拷贝的assets资源目录
    dbSign: {},          // 存储在indexdb中的资源

    SIGN_KEY: '',
    suffixMap: new Map(),
    codeLoadSet: new Set,
    objLoadSet: new Set(),
    localLoadSet: new Set(),
    downLoadSet: new Set(),
    packageLoadSet: new Set(),
    netWeightQueuePool: new NetWeightQueuePool(),
    downFileWeight: 2,
    storeWightQueuePool: new StoreWeightQueuePool()
};

/**
 * 获取 在后缀处理列表 的 后缀信息和类型
 * 返回 [满足配置的具体后缀，后缀类型]
 */
export const getLoadSuffixInfo = (path: string): [string, SuffixType] => {
    let suffix = BaseTools.filePseudoSuffix(path);
    while (suffix) {
        let st = FileLoadParam.suffixMap.get(suffix);
        if (st) {
            return [suffix, st];
        }
        suffix = BaseTools.filePseudoSuffix(suffix);
    }
    return [null, null];
}

/**
 * @description 获得正在下载的数量
 * @example
 */
export const downingCount = () => {
    return FileLoadParam.downLoadSet.size
};
/**
 * @description 获得正在加载的数量
 * @example
 */
export const loadingCount = () => {
    return FileLoadParam.localLoadSet.size
};
/**
 * @description 获得正在加载的代码数量
 * @example
 */
export const codeingCount = () => {
    return FileLoadParam.codeLoadSet.size
};
/**
 * @description 获得正在加载的对象数量
 * @example
 */
export const objingCount = () => {
    return FileLoadParam.objLoadSet.size
};

/**
 * 设置2类文件的后缀类型。 代码code, 对象资源(异步使用，对象形式)obj。
 * @param code 
 * @param obj 
 */
export const setCodeObjSuffix = (code: string[], obj: string[]) => {
    for (let s of code) {
        FileLoadParam.suffixMap.set(s, SuffixType.CODE);
    }
    for (let s of obj) {
        FileLoadParam.suffixMap.set(s, SuffixType.OBJ);
    }
}

/**
 * 等待加载
 */
export function waitLoad(load: IFileLoad, set: Set<any>, p: Promise<any>) {
    set.add(load);
    return p.then(r => {
        set.delete(load);
        return r;
    }).catch(e => {
        set.delete(load)
        throw e;
    });
}

/**
 * 检查等待加载
 * @param file
 * @param set
 */
export const checkWaitLoad = (file: IFileDependInfo, set: Set<IFileLoad>): Promise<ArrayBuffer> | undefined => {
    let resLoad: IFileLoad;

    for (let load of set) {
        if (load.files.has(file.path)) { // 文件正在加载
            resLoad = load;
            break;
        }
    }
    if (resLoad) {
        return new Promise((resolve, reject) => {
            resLoad.addResult(
                (val: Map<string, ArrayBuffer>, err?: any) => {
                    err ? reject(err) : resolve(val.get(file.path));
                }
            );
        });
    } else {
        return;
    }
}

interface IFileLoadParam {
    /**
     * 下载的多域名
     */
    urls: string[];
    /**
     * 批下载的路径，要含 "?"
     */
    batchPath: string;
    /**
     * 限制url的长度
     */
    limitLength: number;
    /**
     * 一次请求的byte限制
     */
    sizeLimit: number;
    /**
     * 本地存储
     */
    localStore: IPiStoreBase;

    /**
     * 底层assets资源签名表
     */
    assetsSign: { [index: string]: string };
    /**
    * 底层assets资源拷贝签名表
    * - 游戏会将assets中的资源拷贝到data目录下,这样即使更新了apk(预计更新apk只会带一点资源)，也能尽量使用可用的资源
    */
    assetsCopySign: { [index: string]: string };
    /**
     * 数据库签名表
     */
    dbSign: { [index: string]: string };


    SIGN_KEY: string;

    /**
     * 后缀类型表, 没有设置的都认为资源(异步使用，二进制数据形式)res
     */
    suffixMap: Map<string, SuffixType>;

    /**
     * 正在运行的代码加载器
     */
    codeLoadSet: Set<IFileLoad>;
    /**
     * 正在运行的对象资源加载器
     */
    objLoadSet: Set<IFileLoad>;
    /**
     * 正在运行的对象资源加载器
     */
    packageLoadSet: Set<IFileLoad>;
    /**
     * 正在运行的加载器
     */
    localLoadSet: Set<IFileLoad>;
    /**
     * 正在运行的下载器
     */
    downLoadSet: Set<IFileLoad>;

    /**
     * 网络权重队列池
     */
    netWeightQueuePool: NetWeightQueuePool;
    /**
     * 下载文件权重
     * 默认为2,比sdf文字下载权重低一点
     */
    downFileWeight: number;
    // store存储队列池
    storeWightQueuePool: StoreWeightQueuePool;
}