import { FileLoadParam, SuffixType } from "./fileload_base";


/**
 * 缓存：注，该缓存不是最近最少策略，所以不能命名为LRU
 * 有大小和时间两个维度来管理
 * 只要 size > limit 或 超时时间到了，就移除资源
 */
export class ResCache {
    map: Map<string, { time: number, data: Uint8Array }> = new Map();
    timeout: number; // 超过该时间就释放，单位：毫秒
    limit: number;   // 最大大小，超过了这个大小就释放，单位：字节
    size = 0;        // 实际大小，单位：字节

    constructor(timeout: number, limit: number) {
        this.timeout = timeout;
        this.limit = limit;
    }

    /**
     * 加入k-v，加入后，要触发一次整理操作。
     */
    add(key: string, value: Uint8Array) {
        this.map.set(key, {
            data: value,
            time: Date.now() + this.timeout,
        });
        this.size += value.byteLength;
        ResCacheMap.startCollectTimer();
    }

    /**
     * 移除，并返回缓存数据
     */
    remove(key: string) {
        let r = this.map.get(key);
        if (r) {
            this.size -= r.data.byteLength;
            this.map.delete(key);
            return r.data;
        }
    }

    /**
     * 是否有缓存
     * @param key 
     * @returns 
     */
    has(key: string) {
        return this.map.has(key);
    }
    /**
     * 按资源存在时间整理资源
     * 
     * 只要存在的时间比预定的久，久释放掉，哪怕高层一直没有用；
     * @param now 当前时间
     */
    collectTime(now: number) {
        for (let [k, v] of this.map) {
            // v.time 存的是 该资源 预期存的 截止时间
            if (now >= v.time) {
                // 当前时间超出了timeout，释放
                this.map.delete(k);
                this.size -= v.data.byteLength;
            }
        }
    }

    /**
     * 按资源大小 整理 缓存；
     * 
     * 只要缓存的资源比指定的大，就删除该资源；直到总量比指定的小为止；
     */
    collectSize() {
        if (this.size >= this.limit && this.map.size >= 2) {
            // 大小超过限制，而且数组元素多于1个
            for (let [k, v] of this.map) {
                this.map.delete(k);
                this.size -= v.data.byteLength;
                if (this.size < this.limit) {
                    // 这里比较的是总缓存不要超过限制，如果满足这一点，可以退出这次整理；
                    return;
                }
            }
        }
    }
}

export namespace ResCacheMap {
    /**
     * 资源缓存表
     */
    const resMap: Map<string, ResCache> = new Map();

    /**
     * 资源缓存表当前是否真正整理
     */
    let timerStart = false;

    export const getCache = (key: string) => {
        return resMap.get(key);
    }

    /**
     * 设置的后缀类型及缓存时间和大小
     */
    export const setResCache = (suffix: string, timeout: number, cacheSize: number) => {
        FileLoadParam.suffixMap.set(suffix, SuffixType.RES);
        resMap.set(suffix, new ResCache(timeout, cacheSize));
    };

    /**
     * 整理资源
     * 返回是否还有必要继续整理
     */
    export const collect = (time: number) => {
        let isAllSizeZero = true;
        for (let v of resMap.values()) {
            v.collectTime(time);
            v.collectSize();
            isAllSizeZero = isAllSizeZero && v.size === 0;
        }
        return isAllSizeZero;
    };

    /**
     * 启动整理定时器
     */
    export const startCollectTimer = () => {
        if (timerStart) {
            return;
        }
        timerStart = true;
        collectTimer();
    };

    /**
     * 启动整理定时器
     */
    export const collectTimer = () => {

        if (collect(Date.now())) {
            timerStart = false;
        }
        else {
            /**
             * 清理操作的频率不需要太高
             */
            setTimeout(collectTimer, 60 * 1000);
        }
    }
}