/*
 * 应用加载模块
 * 根据使用的不同将文件分为4类。
 *  代码code, 配置(必须同步使用)cfg, 对象资源(异步使用，对象形式)obj, 资源(异步使用，二进制数据形式)res
 * 可以设置代码加载器。 比如web版的release版将代码设置为配置，使用配置处理函数来加载代码。而微信小游戏，可能是分包下载。
 * 可以设置对象资源加载器。比如web版没有对象类型。而微信小游戏，使用download将对象资源下载到临时目录，并维护对应的depend。
 * 可以设置配置的处理函数表，并且记录每个配置文件的处理状态（下载中、加载中、处理中，处理完）。
 * 可以设置资源的下载缓冲时间和大小缓存，下载缓存的目标仅是为了下载后不用重新加载。
 */
import { FileLoadParam, SuffixType, checkWaitLoad, waitLoad, getLoadSuffixInfo } from "./fileload_base";
import { ResCacheMap } from "../load/res_cache";
import * as Sign from "../load/sign";
import { LocalLoad } from "../load/local_load";
import { Download } from "../load/down_load";
import { IFileDependInfo } from "../util/depend_base";
import { notifyResUsed } from "pi_sys/util/collect";

/**
 * @description 单个 资源 加载
 * @note 会进行延迟合并处理.
 * @note Web平台，只加载 资源
 */
export const loadRes = (file: IFileDependInfo, waitWrite: boolean = false): Promise<ArrayBuffer | Uint8Array> => {
    let [suffix, st] = getLoadSuffixInfo(file.path);
    notifyResUsed(file.path);
    // if (st === SuffixType.RES) {
    return loadResImpl(file, suffix, waitWrite);
    // } else {
    //     cc.warn() && log(`loadRes, only res type, file: ${file}, suffixType = ${st}`);
    //     return Promise.reject(`load, only res type, file: ${file}, suffixType = ${st}`);
    // }
};

// ============================== 本地



// 批下载器

const loadResImpl = (file: IFileDependInfo, suffix: string, waitWrite: boolean = false): Promise<ArrayBuffer | Uint8Array> => {

    /**
     * 假设高层有资源管理机制
     * 如果 缓存中有，则从缓存中删除，然后直接返回
     */
    let cache = ResCacheMap.getCache(suffix);
    if (cache) {
        let r = cache.remove(file.path);
        if (r) {
            return Promise.resolve(r);
        }
    }

    // 如果是本地加载
    if (Sign.checkLocalSign(file, file.path)) {
        // console.log("=======  loadResImpl from local ", file.path);
        let pr = checkWaitLoad(file, FileLoadParam.localLoadSet);
        if (pr) {
            return pr;
        }

        let load = new LocalLoad();
        load.add(file);
        let p = load.start();
        const waitLoadP = waitLoad(load, FileLoadParam.localLoadSet, p)
            .then((map: Map<string, any>) => {
                return map.values().next().value
            });

        return new Promise((resolve, reject) => {
            waitLoadP.then((data) => {
                // 防止误删 IndexDB - 如果恶意修改内容呢？？
                if (data) {
                    resolve(data);
                } else {
                    downLoad(file, waitWrite).then(resolve);
                }
            })
                .catch((e) => {
                    reject(e);
                });
        });
    } else {
        // console.log("=======  loadResImpl from net ", file.path);
        let pr = checkWaitLoad(file, FileLoadParam.downLoadSet);
        if (pr) {
            return pr;
        } else {
            return downLoad(file, waitWrite);
        }
    }
}

function downLoad(file: IFileDependInfo, waitWrite: boolean): Promise<Uint8Array> {
    let downWait: Download = new Download();
    downWait.waitWrite = waitWrite;
    downWait.add(file);
    let p = downWait.start();
    waitLoad(downWait, FileLoadParam.downLoadSet, p);
    return new Promise((resolve, reject) => {
        downWait.addResult(
            (map: Map<string, Uint8Array>, err?: any) => {
                err ? reject(err) : resolve(map.get(file.path));
            }
        );
    });
}
