/**
 * 媒体：音频
 */
export * from "./inner_audio_impl";


/**
 * 这里所有的事件都是按 小游戏模型 封装
 * 
 * 不同平台实现模型的区别：
 *      + Web端只有Ended事件，语义是调用了stop 或者 自然播放结束都会触发；
 */
export enum AudioEvent {
    Play = "play",
    Pause = "pause",
    Stop = "stop",
    Ended = "ended",
    Seeked = "seeked",
    TimeUpdate = "timeUpdate",  // 播放时间流逝时触发，大概 每 200ms 回调一次；
}

export type PiAudioBuffer = AudioBuffer | string;
/**
 * 音频
 * 接口都参考微信小游戏
 * https://developers.weixin.qq.com/minigame/dev/api/media/audio/InnerAudioContext.html
 * 小游戏开发者工具表现和真机有很多接口表现不一致
 */
export interface IPiInnerAudio {
    /**
     * 小游戏真机实测
     * 1. 第一次播放
     * 2. stop之后播放
     */
    startTime: number;      // 开始播放的位置（单位：s），默认为 0
    loop: boolean;          // 是否循环播放，默认为 false
    volume: number;         // 音量。范围 0~1。默认为 1
    playbackRate: number;   // 播放速度。范围 0.5-2.0，默认为 1。
    duration: number;       // 当前音频的长度（单位 s） （只读）
    currentTime: number;    // 当前音频的播放位置（单位 s）时间保留小数点后 6 位 （只读）
    paused: boolean;        // 当前是是否暂停或停止状态（只读）

    /**
     * 播放
     */
    play(): void;

    /**
     * 暂停；音频再播放会从暂停处开始播放
     */
    pause(): void;

    /**
     * 停止。停止后的音频再播放会从头开始播放
     */
    stop(): void;

    /**
     * 跳转到指定位置
     * @param position 跳转的时间，单位 s。精确到小数点后 3 位，即支持 ms 级别精确度
     * @note 只有当音频处于播放状态,此方法才有效  (微信真机实测结果)
     */
    seek(position: number): void;

    /**
     * 销毁当前实例
     */
    destroy(): void;        // 销毁当前实例

    /**
     * 监听音频进入可以播放状态的事件。但不保证后面可以流畅播放
     * @param callback 音频进入可以播放状态的事件的回调函数
     */
    onCanplay(callback: () => void): void;

    /**
     * 取消监听音频进入可以播放状态的事件
     * @param callback 音频进入可以播放状态的事件的回调函数
     */
    offCanplay(callback: () => void): void;

    /**
     * 监听音频播放事件
     * @param callback 音频播放事件的回调函数
     */
    onPlay(callback: () => void): void;

    /**
     * 取消监听音频播放事件
     * @param callback 音频播放事件的回调函数
     */
    offPlay(callback: () => void): void;

    /**
     * 监听音频暂停事件
     * @param callback 音频暂停事件的回调函数
     */
    onPause(callback: () => void): void;

    /**
     * 取消监听音频暂停事件
     * @param callback 音频暂停事件的回调函数
     */
    offPause(callback: () => void): void;

    /**
     * 监听音频停止事件
     * @param callback 音频停止事件的回调函数
     */
    onStop(callback: () => void): void;

    /**
     * 取消监听音频停止事件
     * @param callback 音频停止事件的回调函数
     */
    offStop(callback: () => void): void;

    /**
     * 监听音频自然播放至结束的事件
     * @param callback 音频自然播放至结束的事件的回调函数
     */
    onEnded(callback: () => void): void;

    /**
     * 取消监听音频自然播放至结束的事件
     * @param callback 音频自然播放至结束的事件的回调函数
     */
    offEnded(callback: () => void): void;

    /**
     * 监听音频播放进度更新事件
     * @param callback 音频播放进度更新事件的回调函数
     */
    onTimeUpdate(callback: () => void): void;

    /**
     * 取消监听音频播放进度更新事件
     * @param callback 音频播放进度更新事件的回调函数
     */
    offTimeUpdate(callback: () => void): void;

    /**
     * 监听音频进行跳转操作的事件
     * @param callback 音频进行跳转操作的事件的回调函数
     */
    onSeeking(callback: () => void): void;

    /**
     * 取消监听音频进行跳转操作的事件
     * @param callback 音频进行跳转操作的事件的回调函数
     */
    offSeeking(callback: () => void): void;

    /**
     * 监听音频完成跳转操作的事件
     * @param callback 音频完成跳转操作的事件的回调函数
     */
    onSeeked(callback: () => void): void;

    /**
     * 取消监听音频完成跳转操作的事件
     * @param callback 音频完成跳转操作的事件的回调函数
     */
    offSeeked(callback: () => void): void;
}

