import { ProcessFunc, RESP_TYPE } from "./http_base";
export * from "./http_impl";

export interface IHttpRequest {
	readonly type: string;
	readonly url: string;
	totalTimeout: number;   // 总超时
	responseTimeout: number;  // 响应超时
	noResponseCount: number;   // 连续未响应次数
	bpsMin: number;            // 每秒下载字节数最低值
	insufficientBufferCount: number;    // 连续下载数据不足次数
	onprocess: ProcessFunc;
	setHttpTimeout(httpTimeoutOption: HttpTimeoutOption): void;
	request(headers: any, reqData: string | ArrayBuffer | FormData, respType: string, totalSize: number): Promise<unknown>;
	abort(): void;
}



export interface HttpTimeoutOption {
	// 响应超时：在请求发出后,如果在该超时时间内没有任何数据响应,且连续noResponseCount次,则认为响应超时,默认2秒
	responseTimeout: number;
	// 连续未响应次数：连续触发响应超时的次数,到达该值,则认为响应超时,默认2次
	noResponseCount: number;
	// 每秒下载字节数最低值,默认100K;3G网络理论下载速率大概在每秒300K
	bpsMin: number;
	// 连续下载数据不足次数： 每秒下载数据不足连续次数,等于该次数则认为该次请求大概率会请求超时,中断请求,默认3次
	insufficientBufferCount: number;
	// 总超时： 一次请求在该超时时间内未全部完成,则认为请求超时;
	// 这个超时不会暴露给项目使用,只在引擎内部使用,一般不用,小游戏因为没有响应超时使用,根据下载大小和bpsMin计算出总超时
	// 项目不应该配置,且配置了也无用
	totalTimeout?: number;
}


export interface HttpTimeoutRetryOption {
	// 响应超时最大值：响应超时的最大值,超过该值后,统一使用该值作为响应超时时间,默认6秒
	responseTimeoutMax: number;
	// 超时增长因子：在请求重试时,会以该数值*上次requestTimeout作为当次的requestTimeout,默认1.6
	timeoutFactor: number;
	// 重试次数：请求失败重试次数,每次重试会增加超时时间,默认3次
	retryCount: number;
}



export interface NetOption extends HttpTimeoutOption, HttpTimeoutRetryOption {

}

/**
 * @description GET方法
 * @example
 */
export type Tget = (url: string, headers: any, reqData: any, respType: RESP_TYPE, timeout: number) => Promise<unknown>;

/**
 * @description POST方法
 * @example
 */
export type Tpost = (url: string, headers: any, reqData: string | ArrayBuffer | FormData, contentType: string, respType: RESP_TYPE, timeout: number) => Promise<unknown>;


export interface IHttpDownload {
	urls: string[];
	path: string;
	timeout: number;
	total: number;
	loaded: number;
	request: IHttpRequest;

	onprocess: ProcessFunc;

	setHeaders(headers: any): void;

	start(): Promise<ArrayBuffer | Uint8Array>;

	abort(): void;
}

