import { HttpTimeoutOption, HttpTimeoutRetryOption, IHttpDownload, IHttpRequest, NetOption, Tget, Tpost } from "pi_sys/net/http";
import { BaseTools } from "pi_sys/util/base_tools";
import { ProcessFunc, ERR_ABORT, ERR_NORMAL, ERR_LOCATION, HttpError, ERR_BPS, ERR_TOTAL_TIMEOUT, ERR_RESPONSE_TIMEOUT, RESP_TYPE } from "pi_sys/net/http_base";
import { Env } from "pi_sys/util/env";

declare var winit;

export class HttpRequest implements IHttpRequest {
	public readonly type: string;
	public readonly url: string;
	public totalTimeout: number = 0;   // 请求超时
	public responseTimeout: number = 0;  // 响应超时
	public noResponseCount: number = 0;   // 连续未响应次数
	public bpsMin: number = 0;           // 每秒下载字节数最低值
	public insufficientBufferCount: number = 0;    // 连续下载数据不足次数
	public onprocess: ProcessFunc;
	readonly xhr = new XMLHttpRequest();
	private loaded = 0;

	public constructor(type: string, url: string) {
		this.type = type;
		this.url = url;
	}

	public setHttpTimeout(httpTimeout: HttpTimeoutOption) {
		if (!httpTimeout) return;
		if (httpTimeout.totalTimeout !== undefined) {
			this.totalTimeout = httpTimeout.totalTimeout;
		}
		if (httpTimeout.responseTimeout !== undefined) {
			this.responseTimeout = httpTimeout.responseTimeout;
		}
		if (httpTimeout.noResponseCount !== undefined) {
			this.noResponseCount = httpTimeout.noResponseCount;
		}
		if (httpTimeout.bpsMin !== undefined) {
			this.bpsMin = httpTimeout.bpsMin;
		}
		if (httpTimeout.insufficientBufferCount !== undefined) {
			this.insufficientBufferCount = httpTimeout.insufficientBufferCount;
		}
	}

	public request(headers: any, reqData: string | ArrayBuffer | FormData, respType: string) {
		// console.log(`============ start request  url = ${this.url}`);
		let xhr = this.xhr;
		if (respType === RESP_TYPE.BIN) {
			xhr.responseType = 'arraybuffer';
		}
		let activeTime = 0;
		let responseTimerRef = 0;
		let noResponseCount = this.noResponseCount;
		let requestTimerRef = 0;
		let bpsTimerRef = 0;
		let bpsMin = this.bpsMin;  // 每秒最少应该下载的字节数
		let bps = 0;       // 每秒下载的字节数
		let insufficientBufferCount = this.insufficientBufferCount;
		let lastLoaded = 0;

		const now = Date.now();
		// NOTE: 一定不能设置withCredentials，否则跨域变得很严格
		xhr.withCredentials = false;
		return new Promise((resolve, reject) => {
			xhr.onabort = () => {
				if (!xhr) return;
				responseTimerRef && clearTimeout(responseTimerRef);
				requestTimerRef && clearTimeout(requestTimerRef);
				bpsTimerRef && clearTimeout(bpsTimerRef);
				reject(new HttpError(this.url, ERR_ABORT, "abort, " + this.url));
			};
			if (this.responseTimeout > 0) {
				// 有些(iOS)浏览器版本可能不支持xhr.timeout属性
				// 自己处理超时，超时指超过规定时间没有收到数据
				activeTime = Date.now();
				// 响应超时：在请求发出后,如果在该超时时间内没有任何数据响应,且连续noResponseCount次,则认为响应超时
				let timer = () => {
					let t = activeTime + this.responseTimeout - Date.now();
					if (t > 0) {
						noResponseCount = this.noResponseCount;
						responseTimerRef = <any>setTimeout(timer, this.responseTimeout);
					} else {
						// 如处理过就不用管了，否则就设置为处理过了。
						// 必须设置，否则，就会调用两次errorCallback
						if (!xhr) return;
						noResponseCount--;
						if (noResponseCount === 0) {

							xhr = null;
							reject(new HttpError(this.url, ERR_RESPONSE_TIMEOUT, `response timeout,responseTimeout = ${this.responseTimeout} , noResponseCount = ${this.noResponseCount},url = ${this.url}`));
							this.abort();
						} else {
							responseTimerRef = <any>setTimeout(timer, this.responseTimeout);
						}

					}
				};
				responseTimerRef = <any>setTimeout(timer, this.responseTimeout);
			}
			if (this.totalTimeout > 0) {
				// 总超时，一次请求在该超时时间内未全部完成,则认为请求超时
				let requestTimer = () => {
					if (!xhr) return;
					xhr = null;

					// console.log(`===== request timeout start = ${now},now = ${Date.now()},diff = ${Date.now() - now}`);
					reject(new HttpError(this.url, ERR_TOTAL_TIMEOUT, `request timeout,totalTimeout = ${this.totalTimeout},url = ${this.url}`));
					this.abort();
				}
				requestTimerRef = <any>setTimeout(requestTimer, this.totalTimeout);
			}

			if (bpsMin > 0) {
				let bpsTimer = () => {
					if (!xhr) return;
					bps = this.loaded - lastLoaded;
					// console.log(`url = ${this.url},loaded = ${this.loaded},lastLoaded = ${lastLoaded},bps = ${bps},bpsMin = ${bpsMin},totalTimeout = ${this.totalTimeout},responseTimeout = ${this.responseTimeout}`);
					if (bps < bpsMin) {
						insufficientBufferCount--;
						if (insufficientBufferCount === 0) {
							xhr = null;
							reject(new HttpError(this.url, ERR_BPS, `bps not enough,bps = ${bps},bpsMin = ${bpsMin},url =  ${this.url}`));
							this.abort();
						} else {
							bpsTimerRef = <any>setTimeout(bpsTimer, 1000);  // 每秒统计下载数据量
						}
					} else {
						insufficientBufferCount = this.insufficientBufferCount;
						bpsTimerRef = <any>setTimeout(bpsTimer, 1000);  // 每秒统计下载数据量
					}
					lastLoaded = this.loaded;
					bps = 0;
				}
				bpsTimerRef = <any>setTimeout(bpsTimer, 1000);  // 每秒统计下载数据量
			}
			xhr.onerror = (ev) => {
				// 避免定时器超时后重复调用
				if (!xhr) return;
				responseTimerRef && clearTimeout(responseTimerRef);
				requestTimerRef && clearTimeout(requestTimerRef);
				bpsTimerRef && clearTimeout(bpsTimerRef);
				reject(new HttpError(this.url, ERR_NORMAL, "error status: " + xhr.status + " " + xhr.statusText + ", " + this.url, ev));
			};
			// 华为不支持upload属性，TODO
			// 在IOS上,设置此属性有可能导致请求失败,未知,暂时毙掉
			// if (xhr.upload) {
			// 	xhr.upload.onprogress = (ev) => {
			// 		// 避免定时器超时后重复调用
			// 		if (!xhr) return;
			// 		activeTime = Date.now();
			// 		this.loaded = ev.loaded;
			// 		this.onprocess && this.onprocess(this.url, 'upload', ev.total, ev.loaded);
			// 	};
			// };
			xhr.onprogress = (ev) => {
				// 避免定时器超时后重复调用
				if (!xhr) return;
				activeTime = Date.now();
				this.loaded = ev.loaded;
				this.onprocess && this.onprocess(this.url, 'download', ev.total, ev.loaded);
			};

			xhr.onload = (ev) => {
				// 避免定时器超时后重复调用
				if (!xhr) return;
				responseTimerRef && clearTimeout(responseTimerRef);
				requestTimerRef && clearTimeout(requestTimerRef);
				bpsTimerRef && clearTimeout(bpsTimerRef);
				if (xhr.status === 300 || xhr.status === 301 || xhr.status === 302 || xhr.status === 303) {
					reject(new HttpError(this.url, ERR_LOCATION, xhr.getResponseHeader("Location")));
				} else if (xhr.status !== 0 && xhr.status !== 200 && xhr.status !== 304) {
					// iOS的file协议，成功的状态码是0
					reject(new HttpError(this.url, ERR_NORMAL, "error status: " + xhr.status + " " + xhr.statusText + ", " + this.url, ev));
				} else if (respType === RESP_TYPE.BIN) {
					resolve(xhr.response || new ArrayBuffer(0));
				} else if (respType === RESP_TYPE.JSON) {
					try {
						resolve(JSON.parse(xhr.responseText));
					} catch (e) {
						reject(new HttpError(this.url, ERR_NORMAL, "error json: " + xhr.responseText + ", " + this.url, ev))
					}
				} else {
					resolve(xhr.responseText);
				}
			};
			const h2 = Env.get("h2");    // 是否开启http2.0
			if (h2) {
				// android或者windows平台,并且是cdn,使用http2.0,使用2.0必须的是https
				const platform_type = Env.get("platform_type");
				if (platform_type === "android" || platform_type === "windows") {
					if (this.url.startsWith("https://c")) {
						(<any>xhr).setVersion(2.0);
					}
				}
			}

			xhr.open(this.type, this.url, true);

			//传输的文件HTTP头信息， 必须在open之后设置
			if (headers) {
				for (let i in headers) {
					if (headers.hasOwnProperty(i))
						xhr.setRequestHeader(i, headers[i]);
				}
			}
			xhr.send(reqData);
		});
	}
	public abort() {
		try {
			typeof this.xhr.abort === "function" && this.xhr.abort();
		} catch (e) {

		}

	}
}
/**
 * @description GET方法
 * @note  兼容旧版本,请使用新接口httpGet
 */
export const get: Tget = (url: string, headers: any, reqData: any, respType: RESP_TYPE, timeout: number) => {
	if (reqData !== null && typeof reqData === "object") {
		reqData = BaseTools.param(reqData);
		url = (url.indexOf("?") > 0) ? url + "&" + reqData : url + "?" + reqData;
	}
	let ar = new HttpRequest('GET', url);
	ar.totalTimeout = timeout;
	return ar.request(headers, undefined, respType);
}

/**
 * @description POST方法
 * @note  兼容旧版本,请使用新接口
 */
export const post: Tpost = (url: string, headers: any, reqData: string | ArrayBuffer | FormData, contentType: string, respType: RESP_TYPE, timeout: number) => {
	headers = headers || {};
	headers["Content-Type"] = contentType;
	let ar = new HttpRequest('POST', url);
	ar.totalTimeout = timeout;
	return ar.request(headers, reqData, respType);
}


/**
 * httpGet方法,带重试机制
 * @param urls 
 * @param headers 
 * @param reqData 
 * @param respType 
 * @param netOption 不传使用默认值
 * @param totalSize 不传使用默认值
 * @returns 
 */
export const httpGet = (urls: string | string[], headers: any, reqData: any, respType: RESP_TYPE, netOption?: NetOption, totalSize?: number) => {
	if (!Array.isArray(urls)) {
		urls = [urls];
	}
	if (reqData !== null && typeof reqData === "object") {
		reqData = BaseTools.param(reqData);
		urls = urls.map(url => {
			return (url.indexOf("?") > 0) ? url + "&" + reqData : url + "?" + reqData
		})
	}
	netOption = netOption || getDefaultNetOption(totalSize);
	const requestOption: RequestOption = {
		method: "GET",
		headers,
		reqData: undefined,
		respType,
		totalSize
	};
	return requestTimeoutRetry(urls, requestOption, netOption);
}


/**
 * httpPost方法,带重试机制
 * @param urls 
 * @param headers 
 * @param reqData 
 * @param contentType 
 * @param respType 
 * @param netOption 不传使用默认值
 * @param totalSize 不传使用默认值
 * @returns 
 */
export const httpPost = (urls: string | string[], headers: any, reqData: string | ArrayBuffer | FormData, contentType: string, respType: RESP_TYPE, netOption?: NetOption, totalSize?: number) => {
	if (!Array.isArray(urls)) {
		urls = [urls];
	}
	netOption = netOption || getDefaultNetOption(totalSize);
	const requestOption: RequestOption = {
		method: "POST",
		headers,
		reqData,
		respType,
		contentType,
		totalSize
	};
	return requestTimeoutRetry(urls, requestOption, netOption);
}


interface RequestOption {
	method: "GET" | "POST",
	headers: any,
	reqData: any,
	respType: string,
	contentType?: string,
	totalSize?: number,
	onsuccess?: (value: any) => void,
	onfail?: (err: HttpError) => void,
	onprocess?: ProcessFunc,
}



const requestTimeoutRetry = (urls: string[], option: RequestOption, netOption: NetOption): Promise<any> => {
	return new Promise((resolve, reject) => {
		let retryCount = netOption.retryCount || 1;
		const onfail = (err: HttpError) => {
			retryCount--;
			if (retryCount === 0) {
				return reject(err);
			} else {
				// 增加超时时间
				let newResponseTimeout = Math.ceil(netOption.responseTimeout * netOption.timeoutFactor);
				newResponseTimeout = newResponseTimeout > netOption.responseTimeoutMax ? netOption.responseTimeoutMax : newResponseTimeout;
				netOption.responseTimeout = newResponseTimeout;
				return requestRetry(urls, option, netOption, 0);
			}

		}
		option.onfail = onfail;
		option.onsuccess = resolve;
		requestRetry(urls, option, netOption, 0);
	});

}



/**
 * 请求重试
 * @note 第一个连接尝试两次,后续连接尝试一次,直到所有连接尝试完
 * @param urls 
 * @param option 
 * @param retry 
 * @param err 
 * @returns 
 */
const requestRetry = (urls: string[], option: RequestOption, httpTimeoutOption: HttpTimeoutOption, retry: number, err?: HttpError) => {
	if (retry > urls.length) {
		return option.onfail && option.onfail(err);
	}

	let headers = option.headers;
	if (option.contentType) {
		headers = headers || {};
		headers["Content-Type"] = option.contentType;
	}
	const url = urls[retry > 0 ? retry - 1 : 0];
	let ar = new HttpRequest(option.method, url);
	ar.setHttpTimeout(httpTimeoutOption);
	let p = ar.request(headers, option.reqData, option.respType);
	ar.onprocess = (url: string, type: string, total: number, loaded: number) => {
		if (type === 'upload')
			return;
		option.onprocess && option.onprocess(url, type, total, loaded);
	}
	p.then((value) => {
		option.onsuccess && option.onsuccess(value);
	}).catch((err: HttpError) => {
		requestRetry(urls, option, httpTimeoutOption, retry + 1, err);
	});
}



export const getDefaultNetOption = (totalSize: number = 500 * 1024) => {
	const netOption: NetOption = Env.get("netOption") || winit.env.netOption;
	const newNetOption: NetOption = {
		responseTimeout: 0,
		noResponseCount: 0,
		bpsMin: 0,
		insufficientBufferCount: 0,
		responseTimeoutMax: 0,
		timeoutFactor: 0,
		retryCount: 0,
		totalTimeout: 0
	};

	if (netOption.responseTimeout !== undefined) {
		newNetOption.responseTimeout = netOption.responseTimeout;
	}

	if (netOption.noResponseCount !== undefined) {
		newNetOption.noResponseCount = netOption.noResponseCount;
	}
	if (netOption.bpsMin !== undefined) {
		newNetOption.bpsMin = netOption.bpsMin;
	}
	if (netOption.insufficientBufferCount !== undefined) {
		newNetOption.insufficientBufferCount = netOption.insufficientBufferCount;
	}

	if (netOption.responseTimeoutMax !== undefined) {
		newNetOption.responseTimeoutMax = netOption.responseTimeoutMax;
	}

	if (netOption.timeoutFactor !== undefined) {
		newNetOption.timeoutFactor = netOption.timeoutFactor;
	}

	if (netOption.retryCount !== undefined) {
		newNetOption.retryCount = netOption.retryCount;
	}

	const platform = Env.get("platform_type");
	// 小游戏没有进度,忽略响应超时和BPS检测,需要明确声明总超时
	if (platform === "minigame") {
		newNetOption.totalTimeout = Math.ceil(totalSize / netOption.bpsMin * 1000);
		if (newNetOption.totalTimeout < 10 * 1000) {
			// 最低10秒超时
			newNetOption.totalTimeout = 10 * 1000;
		}
		newNetOption.responseTimeout = 0;
		newNetOption.bpsMin = 0;
	}
	return newNetOption;
}




export class HttpDownload implements IHttpDownload {
	urls: string[];
	path: string;
	timeout: number;
	total: number;
	loaded: number;
	request: HttpRequest;
	headers: any
	public onprocess: ProcessFunc;

	public constructor(urls: string[], path: string, total?: number) {
		this.urls = urls;
		this.path = path;
		this.total = total || 0;
		this.loaded = 0;
		this.request = null;
		this.onprocess = null;
	}
	setHeaders(headers: any) {
		this.headers = headers;
	}

	public start(): Promise<ArrayBuffer | Uint8Array> {
		const urls = this.urls.map(url => {
			return url + this.path;
		});

		const netOption = getDefaultNetOption(this.total);
		const requestOption: RequestOption = {
			method: "GET",
			headers: this.headers,
			reqData: undefined,
			respType: RESP_TYPE.BIN,
			totalSize: this.total,
			onprocess: (url: string, type: string, total: number, loaded: number) => {
				if (type === 'upload')
					return;
				this.total = total > this.total ? total : this.total;
				this.loaded = loaded;
				if (this.loaded > this.total) {
					this.total = this.loaded;
				}
				this.onprocess && this.onprocess(url, type, this.total, loaded);
			}
		};
		return requestTimeoutRetry(urls, requestOption, netOption);
	}
	public abort() {
		this.request && this.request.abort();
	}
}

// 下载, 如果是微信这类没有进度信息的，需要重写该代码，用统计的网速和设置的total模拟进度信息
// const download = (load: HttpDownload, resolve: (value?: unknown) => void, reject: (reason?: any) => void, retry: number, err?: HttpError) => {
// 	if (retry > load.urls.length) {
// 		return reject(err);
// 	}
// 	let ar = new HttpRequest('GET', load.urls[retry > 0 ? retry - 1 : 0] + load.path);
// 	let p = ar.request(load.headers, undefined, RESP_TYPE.BIN, load.timeout);
// 	load.request = ar;
// 	ar.onprocess = (url: string, type: string, total: number, loaded: number) => {
// 		if (type === 'upload')
// 			return;
// 		load.total = total;
// 		load.loaded = loaded;
// 		load.onprocess && load.onprocess(load.path, type, total, loaded);
// 	}
// 	p.then((value) => {
// 		resolve(value);
// 	}).catch((err: HttpError) => {
// 		download(load, resolve, reject, retry + 1, err)
// 	});
// };



