import "../code_adapter/index_init";

import { adaptive, selectCompressTexFormat } from '../util/screen_adapter';
import { ProgressBar } from '../load/progressbar';
import { setCfgHandler } from "../load/cfg_tools";
import { ResCacheMap } from "../load/res_cache";
import { FileLoadParam, setCodeObjSuffix } from "../load/fileload_base";
import { Env } from "../util/env";
import { BatchLoad, LoadPolicy } from "../load/batch_load";
import { Depend } from "../util/depend";
import { Load } from "../load/init";
import { requireAsync, setImmSuffixs } from "../util/mod"
import { AppName, BenchLevel, getDeviceID, getSystemInfo, SystemInfo } from "../native/base/system_info";
import { HttpDownload } from "../net/http";
import { utf8Decode } from "../util/utf8";
import { now } from "../util/time";
import { loadJSFromAB } from "pi_sys/load/code_load";
import { Logger, initLogger, addMeta, logStep, flushLog, LogStepId } from "pi_sys/util/logger";
import { getUserConfig, initDb } from "pi_sys/store/db";
import { Native } from "pi_sys/native/index";
import { initBiLogger, initGiLogger } from "pi_sys/util/bi_logger";
import { loadImage } from "pi_sys/load/object_load";
import { formatError } from "pi_sys/util/util";
import { showModel } from "pi_sys/native/ui/dialog";
import { restartApp } from "pi_sys/native/base/update";
import { ICfg, UpdateMgr } from "pi_sys/update/update_mgr";
import { initDataCopy, startCopy } from "pi_sys/update/assets_to_data";
import { IFileDependInfo } from "pi_sys/util/depend_base";
import { WeightQueueOption } from "pi_sys/util/net_weight_queue_pool";
import { ErrorCode } from "pi_sys/util/constant";

declare let module;
let logger = new Logger(module.name, "sys");


let hasShow = false;



/**
 * 展示严重错误模态框,全局只会展示一次,一般用于初始化流程中严重错误导致后续流程无法正常进行的提示
 * @param code ErrorCode
 * @returns 
 */
export const showRestartModel = (code: ErrorCode, errorMessage: string) => {
    let title = `Error：${ErrorCode[code]}`;
    logger.error(`Interrupt Error:${ErrorCode[code]},ErrorCode:${code},ErrorMessage:${errorMessage}`);
    if (hasShow) return;
    hasShow = true;
    return showModel(title, "", false, "Cancel", "#000000", "Confirm", "#576B95").then(() => {
        restartApp();
    });
}

/**
 * 执行初始化，返回是否可以执行
 * 目前：如果不支持WebGL，不支持WebAssembly，则是不可执行的；
 */
export const main = async (cfg: any, depend: any): Promise<any> => {
    logStep(LogStepId.SETUP_MAIN);
    flushLog();
    let info = cfg.info as SystemInfo;
    delete cfg.info;

    Env.Init(cfg);

    let antialias = (Env.get("antialias") as boolean) || false;
    let alpha = (Env.get("alpha") as boolean) || false;
    // 设置alpha,必须第一次执行,不能注释
    let gl = window["canvas"].getContext("webgl", { antialias: antialias, preserveDrawingBuffer: false, alpha });

    // 初始化Bi 日志上报
    initBiLogger();
    // 初始化Gi 日志上报
    initGiLogger();


    initPools(cfg.domains);

    // 初始化更新相关
    await initUpdate(cfg);

    try {
        await initDataCopy();
        logger.info("main(), initDataCopy success!");
    } catch (e) {
        logger.error("main(), initDataCopy failed, e = %o", formatError(e));
    }


    // 初始化用户设置得引擎配置
    await engineCfgInit(info, cfg);

    let dependContent = await ready(info, cfg, depend);
    logStep(LogStepId.READY_SUCCESS);
    flushLog();
    execLoad(cfg, dependContent);
}

/**
 * 初始化更新相关
 * @param cfg 
 * @returns 
 */
const initUpdate = async (cfg: any) => {
    let flag = false;
    try {
        logStep(LogStepId.UPDATE_CHECK);
        flushLog();
        UpdateMgr.init(cfg.cfg);
        flag = UpdateMgr.isMajorUpdate();
        logStep(LogStepId.UPDATE_CHECK_END);
        logger.info(`update = ${flag}`);
        flushLog();
    } catch (e) {
        const message = `UpdateMgr.init failed, e = ${formatError(e)}`;
        showRestartModel(ErrorCode.UPDATE_INIT_FAILED, message);
        return Promise.reject(e);
    }

    if (flag) {
        // 大版本
        // 1.清空本地数据,只清除资源数据
        // 2.开始更新
        // 3.更新完成继续下一步(exe,android平台更新完成会要求重启)

        try {
            await Load.getLocalStore(cfg.name);
            await FileLoadParam.localStore.clear();
            logStep(LogStepId.UPDATE_CLEAR_DATA_SUCCESS);
            flushLog();
        } catch (e) {
            const message = `UpdateMgr.clearLocalData failed, e = ${formatError(e)}`;
            showRestartModel(ErrorCode.UPDATE_CLEAR_DATA_FAILED, message);
            return Promise.reject(e);
        }
        return UpdateMgr.startUpdate().catch(e => {
            const message = `UpdateMgr.startUpdate failed, e = ${formatError(e)}`;
            showRestartModel(ErrorCode.UPDATE_FAILED, message);
            return Promise.reject(e);
        });

    }
    return Promise.resolve();
}

/**
 * 初始化日志Meta信息
 * 此方法在winit start方法中调用
 */
export const initLoggerMeta = (cfg: any) => {
    // 获取 系统信息
    let ps: [Promise<SystemInfo>, Promise<string>, Promise<string>, Promise<string>, Promise<number>, Promise<string>, Promise<string>] = [
        getSystemInfo(),
        Native.App.getProcessId(),
        Native.App.getPackageName(),
        Native.App.getVersionName(),
        Native.App.getVersionCode(),
        Native.App.getFlavorName(),
        Native.App.getApiVersion(),
    ];

    return Promise.all(ps).then((result) => {
        let [info, processId, packageName, versionName, versionCode, flavorName, apiVersion] = result;
        globalThis.systemInfo = info;   // 赋值到全局对象上,
        // 设置system_info
        Env.set("system_info", info);
        if (processId) {
            addMeta("process_id", processId);
        }
        if (packageName) {
            addMeta("project", packageName);
        }
        if (versionName) {
            addMeta("version", versionName);
        }
        if (flavorName) {
            addMeta("channel", flavorName);
        }

        // 日志：关键信息，UserAgent
        logger.info("navigator-userAgent: %o", navigator?.userAgent || {});

        const sysVersion = Env.get("sysVersion");

        logger.info("initial-info = %o", {
            packageName,
            versionName,
            versionCode: versionCode.toString(),
            flavorName,
            processId,
            apiVersion,
            sysVersion,
        });

        // 打印配置
        logForCfg(cfg);

        return info;
    }).catch(e => {
        const message = `main(), getSystem failed, e = ${formatError(e)}`;
        showRestartModel(ErrorCode.SYSTEM_INFO_FAILED, message);
        return Promise.reject(e);
    })
}




// 重置分辨率， 重新计算屏幕适配，并下载对应的depend文件
export const resetSize = (isLoadNewImgVersion: boolean): Promise<boolean> => {
    return new Promise((resolve) => {
        getSystemInfo().then((info) => {
            // let device = Env.get("device");
            let oldZoom = Env.get("imgZoom");
            // 初始化结束后resetSize才生效
            // if (!device) {
            // return resolve(null);
            // }

            // const oldWidth = info.screenWidth, oldHeight = info.screenHeight;

            // 重新设置env中的width和height
            // device.screen.width = window.innerWidth;
            // device.screen.height = window.innerHeight;
            Env.set("system_info", info);

            adaptive();

            // let existImgZooms = Env.get("existImgZooms");
            // 如果改分辨率对应的图片depend不存在，下载新的图片depend
            if (oldZoom !== Env.get("imgZoom")) {
                return extendDepend(null).then((depend) => {
                    if (depend) {
                        Depend.Extend(depend);
                    }

                    // 需要加载新的分辨率的图片
                    if (isLoadNewImgVersion && depend) {
                        let loadFiles = [];
                        for (let di of depend) {
                            const name = di[0];
                            if (name && (name.endsWith(".dcss") || name.endsWith(".widcfg"))) {
                                loadFiles.push(name);
                            }
                        }
                        if (loadFiles.length > 0) {
                            let load = new BatchLoad(loadFiles);
                            return load.start().then(() => {
                                resolve(true); // 返回true表示分分辨率改变，并且使用了新的图片版本
                            });
                        }
                    }
                    resolve(false); // 返回false表示，分辨率发生了改变，但图片版本不变
                });
            }
            // } else if (oldWidth !== device.screen.width || oldHeight !== device.screen.height) {
            return resolve(false); // 返回false表示，分辨率发生了改变，但图片版本不变
            // } else {
            //     return resolve(undefined); //undefined表示分辨率未改变
            // }
        })
    });
}

/**
 * 执行加载流程：load --> exec & next_load --> next_exec
 */
export const execLoad = (cfg: any, depend: any) => {

    // 初始化 Depend，注：这里的depend包含：image.depend & game.depend 的 合并版本；
    Depend.Init(depend);

    const batchLoadSizeLimit = Env.get("batchLoadSizeLimit");
    FileLoadParam.limitLength = 1024 - 100;
    FileLoadParam.sizeLimit = batchLoadSizeLimit || 400 * 1024; // 默认一次下载400K
    FileLoadParam.urls = cfg.domains;
    FileLoadParam.batchPath = cfg.batch_path;

    // 初始化加载模块
    const load = Load.init(cfg.name).catch(e => {
        logger.error("Load.init failed, e = %o", formatError(e));
    });

    load.then(() => {
        logStep(LogStepId.LOAD_INIT_SUCCESS);
        flushLog();
        loadBgImage(cfg).then(() => {
            // 正式 执行；
            loadExec();
        });
    })
    // 将 代码 和 对象 后缀 归类；
    setCodeObjSuffix(cfg.code_suffixs, cfg.obj_suffixs);

    for (let s of cfg.cfg_suffixs) {

        // 为js的资源 指定默认的 处理路径
        if (s === "js") {
            setCfgHandler(s, loadJSFromAB);
        } else {
            setCfgHandler(s, null);
        }
    }

    // 为 资源 和 对象 类型 指定 缓存 时间；

    for (let s of cfg.res_suffixs) {
        let size: number;
        let suffix: string;
        if (Array.isArray(s)) {
            suffix = s[0];
            size = s[1];
        } else {
            suffix = s;
            size = cfg.res_cache_size;
        }
        ResCacheMap.setResCache(suffix, cfg.res_timeout, size);
    }

    for (let s of cfg.obj_suffixs) {
        let size: number;
        let suffix: string;
        if (Array.isArray(s)) {
            suffix = s[0];
            size = s[1];
        } else {
            suffix = s;
            size = cfg.res_cache_size;
        }
        ResCacheMap.setResCache(suffix, cfg.res_timeout, size);
    }

    // window全局错误捕捉，记录次数后发送到服务器上
    if (cfg.catch && window.addEventListener && typeof window.addEventListener === "function") {
        window.addEventListener('unhandledrejection', onReject);
        (<any>window).onerror = onError;
    }
};

/**
 * 加载背景图
 * @param cfg 
 * @returns 
 */
const loadBgImage = (cfg: any) => {
    const platformType = Env.get("platform_type");
    if (platformType === "minigame") return Promise.resolve();
    if (cfg.load_bar.bgurl) {
        return loadImage(cfg.load_bar.bgurl, []).then(r => {
            const load_bar = Env.get("load_bar");
            load_bar.image = r;
            const next_load_bar = Env.get("next_load_bar");
            next_load_bar.image = r;
            Env.set("load_bar", load_bar);
            Env.set("next_load_bar", next_load_bar);
        }).catch(err => {
            console.log("======= load bgImage failed!!! reason = " + err.message);
        })
    } else {
        return Promise.resolve();
    }
}

// ============================== 本地

// 决定 高中低 配
const checkBenchLevel = (info: SystemInfo, cfg: any) => {
    // web和webview环境下，会有这个，在index.html中赋值
    if (info.benchLevel === BenchLevel.Unkown && cfg["benchLevelTest"]) {
        if (cfg["benchLevelTest"].cpu) {

            // 具体 测试 和 阈值 查看
            // http://ser.yinengyun.com:8181/docs/pi_demo/pi_demo-1d0oq97joshn8
            const middleThreshold = 2000;
            const highThreshold = (info.app === AppName.Browser) ? 3500 : 4200;
            let count = cfg["benchLevelTest"].cpu.count || 0;

            if (count < 0) {
                info.benchLevel = BenchLevel.Unkown;
            } else if (count < middleThreshold) {
                info.benchLevel = BenchLevel.Low;
            } else if (count < highThreshold) {
                info.benchLevel = BenchLevel.Middle;
            } else {
                info.benchLevel = BenchLevel.High;
            }
        }
    }

    let gpu_performance_impl = 1.0;
    switch (info.benchLevel) {
        case BenchLevel.Low:
            if (Array.isArray(cfg.gpu_performance)) {
                // 如果是数组，表示 [低配比例，中配比例，高配比例]
                gpu_performance_impl = cfg.gpu_performance[0] || 1.0;
            } else {
                gpu_performance_impl = cfg.gpu_performance || 1.0;
            }
            break;
        case BenchLevel.Middle:
            if (Array.isArray(cfg.gpu_performance)) {
                gpu_performance_impl = cfg.gpu_performance[1] || 1.0;
            } else {
                gpu_performance_impl = cfg.gpu_performance || 1.0;
            }
            break;
        case BenchLevel.High:
            if (Array.isArray(cfg.gpu_performance)) {
                gpu_performance_impl = cfg.gpu_performance[2] || 1.0;
            } else {
                gpu_performance_impl = cfg.gpu_performance || 1.0;
            }
            break;
        default:
            gpu_performance_impl = (typeof cfg.gpu_performance === typeof 0) ? cfg.gpu_performance : 1.0;
            break;
    }

    Env.set("gpu_performance_impl", gpu_performance_impl);
}

/**
 * 下载imgDepend
 * 可能需要重构该方法
 */
const extendDepend = (depend: any): Promise<any> => {
    let canvas = window["canvas"];
    initSize(canvas);
    let version = Env.get("version");
    if (!version.imageDepend) {
        return Promise.resolve(depend);
    }

    let antialias = (Env.get("antialias") as boolean) || false;
    let gl = canvas.getContext("webgl", { antialias: antialias, preserveDrawingBuffer: false });
    if (!gl) {
        let reason = "Can't Run App, Miss WebGL!";
        showRestartModel(ErrorCode.MISS_WEBGL, reason);
        return Promise.reject(reason);
    }

    let extensions = gl.getSupportedExtensions();

    let imgZoom = Env.get("imgZoom");

    // let existImgZooms = Env.get("existImgZooms");
    // if (!existImgZooms) {
    //     existImgZooms = [imgZoom];
    //     Env.set("existImgZooms", existImgZooms);
    // } else if (existImgZooms.indexOf(imgZoom) === -1) {
    //     existImgZooms.push(imgZoom);
    // }

    let ctf = selectCompressTexFormat(extensions, Env.get("compressed_tex_formats"));
    Env.set('compressed_tex_format', ctf);

    // let imgVersion = `img${imgZoom === 1 ? "" : "." + imgZoom * 100}${ctf ? ("." + ctf) : ""}`;

    let imgVersion = getAdaptiveImgVersion(version, ctf);
    let img_depend = version.imageDepend[imgVersion];

    // 日志：关键信息，使用哪个 imageDepend
    logger.info("extendDepend: img_depend = %s", img_depend);

    return loadImageDepend(img_depend).then(json => {
        return depend ? depend.concat(json) : json;
    })
}

/**
 * 释放多版本cfg上挂载的depend信息
 * 项目大了之后这个信息很大
 */
export const freeCfgDepend = (cfgname: string) => {
    const cfg = Env.get(cfgname) as ICfg;
    if (cfg) {
        const version_path = Env.get("version_path");
        const gameVersion = cfg[version_path];
        let paths = [];
        for (let key in gameVersion) {
            let path = gameVersion[key];
            if (typeof path === typeof "") {
                paths.push(path);
            } else {
                // 因为version有 一些对象，像 sys，imageDepend等，所以还要下一层取到所有的路径
                if (key === "imageDepend") {
                    for (let key in path) {
                        paths.push(path[key]);
                    }
                }
            }
        }
        let dependPaths = [];
        for (let path of paths) {
            // 路径得值是："path?sign"
            let len = path.indexOf("?");
            if (len >= 0) {
                path = path.slice(0, len);
            }
            // 只分析depend命名的变量；
            if (path.endsWith(".depend")) {
                dependPaths.push(path);
            }
        }
        dependPaths.forEach(key => {
            if (cfg[key]) {
                cfg[key] = null;
            }
        })
    }
}

const getAdaptiveImgVersion = (gameVersion: any, texture: string) => {
    const info = Env.get("system_info") as SystemInfo;
    let benchLevel = info.benchLevel === BenchLevel.Unkown ? BenchLevel.High : info.benchLevel;
    let grade = benchLevel.toLowerCase();
    const compressed_tex_grade = Env.get("compressed_tex_grade");
    if (compressed_tex_grade) {  // 如果指定使用哪个等级
        grade = compressed_tex_grade;
    }
    let imgVersion = ``;
    if (texture) {
        const imgVersions = [`img.${grade}.${texture}`, `img.${texture}`, `img`];
        for (let i = 0; i < imgVersions.length; i++) {
            imgVersion = imgVersions[i];
            if (gameVersion.imageDepend[imgVersion]) {
                break;
            }
        }
    } else {
        imgVersion = "img";
    }
    Env.set("imgVersion", imgVersion);
    return imgVersion;
}


/**
 * 设置canvas的像素
 * 此方法本来是在gui初始化完成后调用,这里将其摞进来
 * 因为在小游戏上首页图片会因为canvas像素问题导致开始图片模糊
 * @param canvas 
 */
const initSize = (canvas: HTMLCanvasElement) => {
    let adaptiveResult = Env.get("adaptiveResult")
    if (adaptiveResult.isRotate) {
        canvas.width = adaptiveResult.renderSize.height;
        canvas.height = adaptiveResult.renderSize.width;
        canvas.style.width = `${adaptiveResult.styleSize.height}px`;
        canvas.style.height = `${adaptiveResult.styleSize.width}px`;
    } else {
        canvas.width = adaptiveResult.renderSize.width;
        canvas.height = adaptiveResult.renderSize.height;
        canvas.style.width = `${adaptiveResult.styleSize.width}px`;
        canvas.style.height = `${adaptiveResult.styleSize.height}px`;
    }
}

const loadImageDepend = (img_depend: string) => {
    let cfg = Env.get("cfg");
    if (cfg) {
        let len = img_depend.indexOf("?");
        if (len >= 0) {
            img_depend = img_depend.slice(0, len);
        }
        return Promise.resolve(cfg[img_depend]);
    } else {
        let urls = Env.get("domains");
        let rootPath = Env.get("root_path");

        let path = `${rootPath}${img_depend}`;
        let load = new HttpDownload(urls, path);

        load.onprocess = () => { };

        // 下载 image.depend 对用的内容；
        return load.start().then((data: ArrayBuffer) => {
            let str = utf8Decode(data);
            let json = JSON.parse(str);

            return json;
        }).catch((e) => {
            const message = `loadImageDepend failed! e = ${formatError(e)}`;
            showRestartModel(ErrorCode.LOAD_IMAGE_DEPEND_FAILED, message)
            return Promise.reject(e);
        });
    }
}

/**
 * 根据 运行时特性，决定用哪个 image 版本（png | astc | dxt）
 * @param version game.version 的 内容；
 */
const initImgVersion = (version: any) => {
    let img_zoom_options = [1.0];
    let compressed_tex_formats = [];
    let compressed_tex_grades = [];

    /**********************
     * 
     * 在 game.version 的内容，键为 imageDepend 的对象；
     * 
     * 内容类似于：
     * {
     *      "img": "img.depend?UV45MTCfhnGop2BX7j5EkF",                  // png图
     *      "img.astc": "img.astc.depend?UV45MTCfhnGop2BX7j5EkF",        //  astc图
     *      "img.high.astc": "img.high.astc.depend?T4aF5Jw129bPX4KfXLmaXh",  // high astc图
     *      "img.middle.astc": "img.high.astc.depend?T4aF5Jw129bPX4KfXLmaXh",  // high astc图
     *      "img.low.astc": "img.high.astc.depend?T4aF5Jw129bPX4KfXLmaXh",  // high astc图
     * }
     * 
     **********************/
    let depend = version.imageDepend;

    if (depend) {
        for (let v in depend) {
            // v是字符串，类似于 "img.80.astc"
            for (let ai of v.split(".")) {
                // thumb 是 缩略图 的 意思；
                if (ai === "img" || ai === "thumb") {
                    continue;
                }

                let i = parseInt(ai);

                // 因为 这个键 是 img.缩放系数数字.压缩格式
                // 所以 这里 非数字，就代表压缩格式；
                if (isNaN(i)) {
                    if (["high", "middle", "low"].includes(ai)) {
                        compressed_tex_grades.push(ai);
                    } else {
                        compressed_tex_formats.push(ai);
                    }

                } else if (i > 0 && i < 100) {
                    // 这里代表 压缩系数，变成 0-1；
                    img_zoom_options.push(i / 100);
                }
            }
        }
    }

    // 项目 配置 在此之前 已经 设置到 Env去了；
    // 这里 的 规则，是 项目配置 优先；

    if (!Env.has("img_zoom_options")) {
        Env.set("img_zoom_options", img_zoom_options);
    }


    if (!Env.has("compressed_tex_formats")) {
        Env.set("compressed_tex_formats", compressed_tex_formats);
    }
}

/**
 * 下载并显示进度条
 */
const startLoad = (next: string): Promise<ProgressBar> => {
    let load = new BatchLoad(Env.get(next + "load"));
    let filters = Env.get(next + "load_filters");
    if (filters) {
        filters.forEach(f => load.addFilter(f));
    }
    load.preloadKey = next;
    let bar = new ProgressBar(Env.get(next + "load_bar"));
    load.addProcess(bar.onProcess.bind(bar));
    // 这里需要显示指定代码和配置的加载策略为All
    let loadPromise = load.start(LoadPolicy.No, LoadPolicy.No);
    bar.show(Env.get(next + "load_text"), load.total, load.loaded);
    return loadPromise.then(() => bar);
}

/**
 * 加载和执行流程
 * 
 * 依次是：
 * 
 * 1，加载：env.load 列出的目录，代码 和 资源 要 准备好；
 * 2，同时进行：加载 env.next_load, 执行 env.exec；
 * 3，等上面两个同时完毕后，执行：env.next_exec
 */
const loadExec = () => {
    let begin = now();

    // 日志：关键信息，BeginLoad
    logger.info("begin_startLoad");

    startLoad("").then((bar) => {

        // 第一次 的进度条下载完成时候取消；
        bar.clear();
        bar.flush();

        let end = now();
        let cost = Math.floor(end - begin)
        logStep(LogStepId.FIRST_LOAD_SUCCESS, cost);
        // 日志：关键信息，EndLoad
        logger.info("end_startLoad, time = %i ms; now begin start exec & nextLoad", cost);
        flushLog();


        let begin2 = end;
        let exec = execMfa(Env.get("exec")).then(() => {
            let end2 = now();
            // 日志：关键信息，EndExec
            logger.info("end_exec, time = %i ms", Math.floor(end2 - begin2));
            logStep(LogStepId.FIRST_EXEC_SUCCESS);
            flushLog();
        }).catch(err => {
            const message = `Exec first code failed! e = ${formatError(err)}`;
            showRestartModel(ErrorCode.FIRST_EXEC_FAILED, message);
            return Promise.reject(err);
        });

        let begin3 = now();
        let nextLoad = startLoad("next_").then((nextBar) => {
            let end3 = now();
            let cost = Math.floor(end3 - begin3);
            // 日志：关键信息，EndNextLoad
            logger.info("end_NextLoad, time = %i ms", cost);
            logStep(LogStepId.NEXT_LOAD_SUCCESS, cost);
            flushLog();

            return nextBar;
        }).catch((e) => {
            const message = `Load next res failed! e = ${formatError(e)}`;
            showRestartModel(ErrorCode.NEXT_LOAD_FAILED, message);
            return Promise.reject(e);
        });

        // 第二次执行 等待 第二次加载 和 第一次执行 结束
        let begin4 = now();
        Promise.all([nextLoad, exec]).then(([nextBar]) => {
            let end4 = now();

            // 日志：关键信息，EndNextLoad & Exec
            logger.info("end_NextLoad & Exec, time = %i ms; now start NextExec", Math.floor(end4 - begin4));

            let begin5 = end4;
            try {
                const updateMfa = Env.get("updateMfa");
                if (updateMfa && updateMfa[0]) {
                    import(updateMfa[0]).then(mod => {
                        if (mod && mod[updateMfa[1]]) {
                            mod[updateMfa[1]]();
                        }
                    })
                }
            } catch (e) {
                logger.error("UpdateMfa exec failed!", formatError(e));
            }

            logStep(LogStepId.START_NEXT_EXEC);
            flushLog();
            return execMfa(Env.get("next_exec")).then(() => {
                // 第二次执行 的进度条需要等到 执行next-exec 结束后才能关闭；
                nextBar.clear();
                nextBar.flush();
                hasShow = true;   // 进入项目后就不在弹重启框
                let end5 = now();
                // 释放dataCfg和netCfg,assetsCfg根据是否需要拷贝来决定释放
                freeCfgDepend("dataCfg");
                freeCfgDepend("netCfg");
                try {
                    const hasUpdate = Env.get("hasUpdate");
                    const assetsToData = Env.get("assetsToData");
                    if (assetsToData && !hasUpdate) {
                        startCopy();
                    } else {
                        freeCfgDepend("assetsCfg");
                    }
                } catch (e) {
                    logger.error("startCopy exec failed!", formatError(e));
                }

                // 日志：关键信息，EndNextExec
                logger.info("end_NextExec, time = %i ms", Math.floor(end5 - begin5));
                logStep(LogStepId.NEXT_EXEC_SUCCESS);
                flushLog();
            }).catch((e) => {
                // 第二次执行 的进度条需要等到 执行next-exec 结束后才能关闭；
                const message = `Exec next code failed! e = ${formatError(e)}`;
                showRestartModel(ErrorCode.NEXT_EXEC_FAILED, message);
                nextBar.clear();
                nextBar.flush();
            });
        });
    }).catch((e) => {
        const message = `Load first res failed! e = ${formatError(e)}`;
        showRestartModel(ErrorCode.FIRST_LOAD_FAILED, message);
    });
};

const execMfa = (exec: any): Promise<any> => {
    return new Promise((resolve, _reject) => {
        if (!exec || exec.length === 0) {
            return resolve(null);
        }
        let arr = [];
        let execCount = 0;
        for (let execi of exec) {
            import(execi[0]).then((mod) => {
                let r = mod[execi[1]](...(execi.slice(2)));
                // r instanceof Promise
                // 这里发现微信小游戏IOS端如果是async函数返回的Promise,那么r instanceof Promise返回false
                // 猜测微信把代码做了类似babel的转换,导致async最后包裹的不是Promise
                if (r && r.then) {
                    arr.push(r);
                }
                execCount += 1;
                if (execCount === exec.length) {
                    Promise.all(arr).then(() => {
                        resolve(null);
                    }).catch(_reject);
                }
                return mod;
            }).catch(_reject);
        }
    });

};

// ============================== 本地

const errMap: Map<string, { count: number }> = new Map;
const sid = Date.now().toString(36) + "X" + Math.floor(Math.random() * 0xffffffff).toString(36);

const count = (e: string) => {
    let r = errMap.get(e);
    if (!r) {
        r = { count: 0 };
        errMap.set(e, r);
    }
    r.count += 1;
    let c = 1;
    while (true) { // 仅在1,2,4,8...上发送错误信息
        if (c === r.count) { return r.count; }
        if (c > r.count) { return 0; }
        c += c;
    }
};

/**
 * Promise 的 reject，如果没人处理，则会走到这里来；
 */
const onReject = (ev: PromiseRejectionEvent) => {

    try {
        let stack = ev?.reason?.stack;
        let message = formatError(ev);
        if (stack) {
            logger.error("onReject: stack = " + stack, message);
        } else {
            logger.error("onReject: messege = " + ev?.reason, message);
        }
    } catch (e) {

    }
};

/**
 * 宏任务 本身 异常，如果没人处理，则会走到这里来；
 * 
 * @param message 错误信息
 * @param source  文件名
 * @param lineno  行号
 * @param colno   列号
 * @param error   Error
 */
const onError = (message, _source, lineno, colno, error) => {
    try {
        /**
         * 此接口是浏览器标准,message类型是string
         * 但是V8不是,V8 message是对象,并且后面几个参数都是undefiend
         * 另外:V8 中 unhandledrejection的捕获也会走到onerror中去
         */
        let stack = error?.stack || message?.stack;
        if (typeof message === "object") {
            message = formatError(message);
        }
        logger.error(`mssage = ${message}, line = ${lineno}, column = ${colno}, stack = ${stack}`);
    } catch (e) {

    }
};

/**
 * 初始化用户设置得引擎配置
 * @param info 
 * @param cfg 
 * @returns 
 */
const engineCfgInit = (info: SystemInfo, cfg: any) => {
    return initDb(cfg.name).then(() => {
        return getUserConfig().then((userConfig) => {
            logStep(LogStepId.ENGINE_CFG_INIT_SUCCESS);
            flushLog();
            if (userConfig && userConfig.engine) {
                Env.set("userConfig", userConfig);
                if (userConfig.engine && userConfig.engine.benchLevel) {
                    info.benchLevel = userConfig.engine.benchLevel;
                }
            }
        })
    }).catch(e => {
        const message = `main(), engineCfgInit failed, e = ${formatError(e)}`;
        showRestartModel(ErrorCode.ENGINE_CFG_INIT_FAILED, message);
        return Promise.reject(e);
    })
}

/**
 * 准备 相关 的 配置
 */
const ready = (info, cfg, depend) => {

    // 如果不能运行，则打印原因
    if (info.benchLevel === BenchLevel.CantRun) {
        let reason = "Can't Run App, Miss WebGL or WebAssembly";
        showRestartModel(ErrorCode.CANT_RUN, reason)
        return Promise.reject(reason);
    }

    // 决定高中低配
    checkBenchLevel(info, cfg);

    // 设置system_info
    Env.set("system_info", info);
    // `TODO` 时区偏移
    Env.set("timezone_offset", new Date().getTimezoneOffset());

    // 日志：关键信息，SystemInfo
    logger.info("System_Info: %o", info);

    // cfg.version 就是 game.version 的内容；
    // 根据 game.version 文件，分析当前存在的图片版本，将其设置在env中
    // 注：cfg已有的配置，优先于 动态分析 出来的image版本；
    initImgVersion(cfg.version);

    // 屏幕适配
    adaptive();

    // 设置 立即加载 的 模块 后缀；
    setImmSuffixs(Env.get("imm_load_suffixs") || []);

    // 将 相关的 image.depend 的 内容，扩充到 Depend模块去；
    return extendDepend(depend);
}

/**
 * 将配置 打印 到 日志
 * @param cfg 
 */
const logForCfg = (cfg: any) => {

    let project = {};
    let names = ["name", "debug", "mobile_asset_path", "platform_type", "benchLevelTest"];
    for (let name of names) {
        if (name in cfg) {
            project[name] = cfg[name];
        }
    }
    // 日志：关键信息，BeginMain
    logger.info("main: cfg.project = %o", project);

    let gui = {};
    names = [
        "antialias", "design_size", "gpu_performance", "max_extend", "show_rect", "rotate_fit", "compressed_tex_formats",
    ];
    for (let name of names) {
        if (name in cfg) {
            gui[name] = cfg[name];
        }
    }
    // 日志：关键信息，BeginMain
    logger.info("main: cfg.gui = %o", gui);

    let load = {};
    names = ["imm_load_suffixs", "res_timeout", "res_cache_max_size", "res_cache_total_size"];
    for (let name of names) {
        if (name in cfg) {
            load[name] = cfg[name];
        }
    }
    // 日志：关键信息，BeginMain
    logger.info("main: cfg.load = %o", load);

    let server = {};
    names = ["domains", "main", "remote", "remotePort"];
    for (let name of names) {
        if (name in cfg) {
            server[name] = cfg[name];
        }
    }
    // 日志：关键信息，BeginMain
    logger.info("main: cfg.server = %o", server);
}


const initPools = (domains: string[]) => {
    const option: WeightQueueOption = {
        weight: FileLoadParam.downFileWeight,
        maxCount: 6,
        desc: "快文件下载队列",
        domains,
        isMultiVersion: true,
        getDependInfo: (path: string) => {
            return Depend.getFile(path);
        }
    };
    const wq = FileLoadParam.netWeightQueuePool.createWeightQueue(option);

}