/**
 * 逐步将assets目录下的文件拷贝到data目录下
 * 方便更新apk后还可以使用有效资源
 */

import { readFromLocal } from "pi_sys/load/data_transfer";
import { Asset } from "pi_sys/store/asset";
import { access, readFile, writeFile } from "pi_sys/store/file_system";
import { Depend } from "pi_sys/util/depend";
import { Env } from "pi_sys/util/env"
import { getDecryptType, getEncryptKey } from "pi_sys/util/util";
import { ICfg } from "./update_mgr";
import { freeCfgDepend } from "pi_sys/setup/main";

const AssetsCopyDir = "assets_copy"; // 拷贝到data目录的目录名
const AssetsCopyDependName = "assets_copy_depend";  // 记录已经拷贝的文件
const FinishedName = "copy_finished";  // 如果已经全部拷贝完成,此文件存在
const WriteFileCount = 10;          // 10个文件落地一次depend
const CopyInterval = 1000;     // copy间隔

type DependInfo = [string, number, string, string?, string?, number?, string?, number?];
// assets所有depend    [name,size,sign,encryptType,encryptPath,encryptSize,compressAlgorithm,compressLevel]
let assetsDependsMap: Map<string, DependInfo> = new Map;
let copyDependsMap: Map<string, DependInfo> = new Map;
let needCopys: DependInfo[] = [];
let currentWriteFileCount = WriteFileCount;
let supported = true;


/**
 * 初始化资源拷贝
 * @returns 
 */
export const initDataCopy = async () => {
    const platform_type = Env.get("platform_type");
    if (platform_type === "web" || platform_type === "minigame" || platform_type === "webview") {
        supported = false;
        return;
    }
    initAssetsDepends();
    await initCopyDepends();
    await initNeedCopy();
}

/**
 * 获取拷贝的资源
 * @returns 
 */
export const getCopyDepends = () => {
    const depends = [];
    copyDependsMap.forEach(info => depends.push(info));
    return depends;
}


export const readCopyFile = (path: string) => {
    console.log("======== readCopyFile ", path);
    const info = Depend.getFile(path);
    return readFromLocal(`${AssetsCopyDir}/${path}`, false, getDecryptType(info.encryptType), getEncryptKey(path), info.compressAlgorithm);
}

export const getCopyFilePath = (path: string) => {
    return `${AssetsCopyDir}/${path}`;
}

const initAssetsDepends = () => {
    const version_path = Env.get("version_path");
    const assetsCfg = Env.get("assetsCfg") as ICfg;
    const gameVersion = assetsCfg[version_path];
    let paths = [];
    for (let key in gameVersion) {
        let path = gameVersion[key];
        if (typeof path === typeof "") {
            paths.push(path);
        } else {
            // 因为version有 一些对象，像 sys，imageDepend等，所以还要下一层取到所有的路径
            if (key === "imageDepend") {
                for (let key in path) {
                    paths.push(path[key]);
                }
            }
        }
    }
    let dependPaths = [];
    for (let path of paths) {
        // 路径得值是："path?sign"
        let len = path.indexOf("?");
        if (len >= 0) {
            path = path.slice(0, len);
        }
        // 只分析depend命名的变量；
        if (path.endsWith(".depend")) {
            dependPaths.push(path);
        }
    }

    let depends = []
    for (let dependPath of dependPaths) {
        depends = depends.concat(assetsCfg[dependPath] || []);
    }
    depends.forEach(depend => {
        assetsDependsMap.set(depend[0], depend);
    })
    return depends;
}


const initCopyDepends = async () => {
    const copyDependPath = `${AssetsCopyDir}/${AssetsCopyDependName}`;
    const isExist = await access(copyDependPath);
    if (isExist) {
        let content = await readFile(copyDependPath, "utf8") as string;
        let depends = JSON.parse(content);
        depends.forEach(depend => {
            copyDependsMap.set(depend[0], depend);
        })
    }
}

const initNeedCopy = async () => {
    let isFinished = await access(`${AssetsCopyDir}/${FinishedName}`);
    if (isFinished) {
        console.log("已经全部拷贝完成");
        return;
    }
    assetsDependsMap.forEach((v, k) => {
        if (!copyDependsMap.has(k)) {
            needCopys.push(v);
        }
    })

}

const writeCopyDepens = () => {
    const copyDependPath = `${AssetsCopyDir}/${AssetsCopyDependName}`;
    const depends = Array.from(copyDependsMap.values());
    writeFile(copyDependPath, JSON.stringify(depends));
}


const copyOne = (v: [string, number, string, string?, string?, number?, string?, number?]) => {
    let path = v[0];
    let encryptPath = v[4];
    return Asset.copyAssetToData(encryptPath || path, `${AssetsCopyDir}/${path}`).then(() => {
        currentWriteFileCount--;
        copyDependsMap.set(path, v);
        if (currentWriteFileCount === 0) {
            currentWriteFileCount = WriteFileCount;
            writeCopyDepens();
        }
    }).catch(e => {
        // 拷贝失败忽略
    })
}

/**
 * 开始拷贝
 */
export const startCopy = () => {
    if (!supported) return;
    setTimeout(() => {
        let v = needCopys.pop();
        // console.log("====== left file count ", needCopys.length);
        if (v) {
            copyOne(v).then(() => {
                startCopy();
            })
        } else { // 已经全部拷贝完成
            freeCfgDepend("assetsCfg");
            writeFile(`${AssetsCopyDir}/${FinishedName}`, "");
        }
    }, CopyInterval);
}