/**
 * 版本号有以下三种形式
 * 1. major.minor.patch                      // apk so 快照等都是
 * 2. major.minor.patch_major.minor.patch    // release打包pisys  pisys自身版本号+入口文件版本号组合
 * 3. sign                                   // debug打包pisys以文件签名定义
 */

export enum UpdateLevel {
    None = 0,      // 不需要更新
    Patch = 1,      // 小版本
    Minor = 2,     // 中版本
    Major = 3,     // 大版本
}

/**
* 项目对pi_sys的最低版本依赖,会影响pi_sys的更新策略.debug = true时值为*，使用默认更新策略
* 有两个大前提:1.远端版本一定大于等于配置的版本 2. 本地版本和远端版本必定有一个满足依赖
* - 配置具体major版本
*    + 如果本地版本满足,远端版本是minor或者patch更新,都视为静默更新,major更新,则不更新
*    + 如果本地版本不满足,则视为major更新,强制更新
* - 配置具体major & minor版本
*    + 如果本地版本满足,远端版本是minor或者patch更新,都视为静默更新,major更新,则不更新
*    + 如果本地版本不满足,则视为major更新,强制更新
* - 配置具体major & minor & patch版本
*    + 如果本地版本满足,远端版本是minor或者patch更新,都视为静默更新,major更新,则不更新
*    + 如果本地版本不满足,则视为major更新,强制更新
* 
* 总结：
* - 如果本地版本满足条件,minor或者patch更新统一视为patch更新,major更新则不更新
* - 此更新策略会受入口文件版本号的影响
* 
* 1.2.3 
* 1.2.*    
* 1.*      
* *        使用默认更新策略
*/
export const getUpdateLevel = (oldVersion: string, newVersion: string, dependVersion?: string): UpdateLevel => {
    if (!oldVersion || !newVersion) {
        return UpdateLevel.None;
    }
    if (oldVersion === newVersion) {
        return UpdateLevel.None;
    }
    if (oldVersion && !oldVersion.includes(".")) { // 以签名为版本号,认为是大版本更新
        return UpdateLevel.Major;
    }
    let [oldStr1, oldStr2] = oldVersion.split("_");
    let [newStr1, newStr2] = newVersion.split("_");
    let level1 = _getUpdateLevel(oldStr1, newStr1);
    let level2 = _getUpdateLevel(oldStr2, newStr2);
    if (level1 !== UpdateLevel.None && dependVersion && dependVersion !== "*") {
        let matched = checkVersionMatch(oldStr1, dependVersion);
        if (!matched) {
            return UpdateLevel.Major;
        } else {
            if (level1 === UpdateLevel.Major) {  // 如果本地满足依赖,且远端是Major更新,则不更新
                level1 = UpdateLevel.None;
            } else {                  // 否则视为patch更新
                level1 = UpdateLevel.Patch;
            }

        }
    }
    return level1 > level2 ? level1 : level2;  // 返回两个版本中更新较大的那个版本
}


/**
 * 比较版本大小
 * @param v1 版本1
 * @param v2 版本2
 * @returns 0 相等  1 版本1大 -1 版本2大
 */
export const compareVersion = (v1: string, v2: string) => {
    if (v1 === v2) return 0;
    if (!v1) return -1;
    if (!v2) return 1;
    let arr1 = v1.split(".");
    let arr2 = v2.split(".");
    let major1 = Number(arr1[0]);
    let major2 = Number(arr2[0]);

    if (major1 !== major2) {
        return major1 > major2 ? 1 : -1;
    }

    let minor1 = Number(arr1[1]);
    let minor2 = Number(arr2[1]);
    if (minor1 !== minor2) {
        return minor1 > minor2 ? 1 : -1;
    }
    let patch1 = Number(arr1[2]);
    let patch2 = Number(arr2[2]);
    if (patch1 !== patch2) {
        return patch1 > patch2 ? 1 : -1;
    }
    return 0;
}

const _getUpdateLevel = (oldVersion: string, newVersion: string) => {
    if (oldVersion === newVersion || !newVersion) {
        return UpdateLevel.None;
    }
    if (!oldVersion) {
        // 历史原因,oldVersion可能为空串,全部视为小版本更新
        return UpdateLevel.Patch;
    }
    let oldArr = oldVersion.split(".");
    let newArr = newVersion.split(".");
    let oldMajor = Number(oldArr[0]);
    let newMajor = Number(newArr[0]);
    if (newMajor > oldMajor) {
        return UpdateLevel.Major;
    } else if (newMajor < oldMajor) {
        throw new Error(`new major version must greater than or equal old major version`)
    }
    let oldMinor = Number(oldArr[1]);
    let newMinor = Number(newArr[1]);
    if (newMinor > oldMinor) {
        return UpdateLevel.Minor;
    } else if (newMinor < oldMinor) {
        throw new Error(`new minor version must greater than or equal old minor version`)
    }
    let oldPatch = Number(oldArr[2]);
    let newPatch = Number(newArr[2]);
    if (newPatch > oldPatch) {
        return UpdateLevel.Patch;
    } else if (newPatch < oldPatch) {
        throw new Error(`new patch version must greater than or equal old patch version`)
    }
}

/**
 * 检查本地版本是否满足依赖版本
 * @param oldVersion 
 * @param dependVersion 
 */
const checkVersionMatch = (oldVersion: string, dependVersion: string) => {
    if (!oldVersion) return true;
    let oldArr = oldVersion.split(".");
    let dependArr = dependVersion.split(".");
    let oldMajor = Number(oldArr[0]);
    let dependMajor = Number(dependArr[0]);
    if (oldMajor < dependMajor) return false;
    // 1.*
    if (dependArr[1] === "*") return true;
    // 1.1.?
    let oldMinor = Number(oldArr[1]);
    let dependMinor = Number(dependArr[1]);
    if (oldMinor < dependMinor) return false;
    // 1.1.*
    if (dependArr[2] === "*") return true;
    // 1.1.1
    let oldPatch = Number(oldArr[2]);
    let dependPatch = Number(dependArr[2]);
    if (oldPatch < dependPatch) return false;
    return true;
}