import { TDependJson, IFileDependInfo, IDirDependInfo } from './depend_base';
import { BaseTools } from './base_tools';
import { getLoadSuffixInfo } from 'pi_sys/load/fileload_base';
import { Env } from './env';
import { getFilePlatform } from './util';
import { CompressAlgorithm, CompressLevel, EncryptType } from './constant';

/**
 * 保存有效 depend 信息
 * + 提供保存接口
 * + 提供查询接口
 */
export namespace Depend {
    const DependFileMap: Map<string, IFileDependInfo> = new Map();
    const DependDirMap: Map<string, IDirDependInfo> = new Map();
    const RuntimeFullNameMap: Map<string, string> = new Map();
    const RootDirName = '';

    export class FileDependInfo implements IFileDependInfo {
        path: string;
        size: number;
        sign: string;
        depend: string[] | undefined;
        suffix: string;
        encryptType?: EncryptType;
        encryptPath?: string;
        encryptSize?: number;
        compressAlgorithm?: CompressAlgorithm;
        compressLevel?: CompressLevel;

        constructor(args: TDependJson) {
            this.path = args[0];
            this.size = args[1];
            this.sign = args[2];
            this.encryptType = args[3];
            this.encryptPath = args[4];
            this.encryptSize = args[5];
            this.compressAlgorithm = args[6];
            this.compressLevel = args[7];
            this.suffix = BaseTools.fileSuffix(this.path);

            recordForRuntimeFullName(this.path);
        }
    }

    export class DirDependInfo implements IDirDependInfo {
        parent: DirDependInfo | undefined = undefined;
        path: string = '';
        size: number = 0;
        level: number = 0;
        public readonly children: DirDependInfo[] = [];
        public readonly files: FileDependInfo[] = [];
        public readonly suffixMap: Map<string, IFileDependInfo[]> = new Map();

        constructor(path: string) {
            this.path = path;
        }
        add(info: IFileDependInfo, isSuffix: boolean) {
            let r: IFileDependInfo[];
            if (isSuffix) {
                r = this.suffixMap.get(info.suffix);
                if (!r) {
                    r = [];
                    this.suffixMap.set(info.suffix, r);
                }
            } else {
                r = this.files;
            }

            r.push(info);

            // tslint:disable-next-line:no-this-assignment
            let p: DirDependInfo = this;

            while (p) {
                p.level += 1;
                p.size += info.size;
                p = p.parent;
            }
        }
    }

    /**
     * 根据文件表初始化依赖表
     * 初始化depend文件数据 - 解析-记录
     * @param files depend文件数据 - (已通过字符串转JSON)
     */
    export function Init(files: TDependJson[]) {
        DependDirMap.set(RootDirName, new DirDependInfo(RootDirName));

        Extend(files);
    }

    /**
     * 扩展depend
     * 忽略掉非当前平台的文件
     * @param files depend文件数据 - (已通过字符串转JSON)
     */
    export function Extend(files: TDependJson[]) {
        const cfg_platform_suffixs = Env.get("cfg_platform_suffixs") || [];
        const curPlatform = Env.get("platform_type");
        const supportedPlatforms = Env.get("supportedPlatforms");
        const ignorePaths = new Set;
        for (let args of files) {
            let path = args[0];
            let suffix = BaseTools.fileSuffix(path);
            if (cfg_platform_suffixs.includes(suffix)) {
                const filePlatform = getFilePlatform(path, supportedPlatforms);
                if (filePlatform === curPlatform) {  // 有平台相关文件,同名平台无关文件应该被忽略
                    // console.warn("===== ", path);
                    ignorePaths.add(path.replace(`.${curPlatform}`, ""));
                } else if (filePlatform && filePlatform !== curPlatform) {
                    ignorePaths.add(path);
                }
            }
        }

        console.log("======= ignorePaths", ignorePaths);

        for (let args of files) {
            let f = new FileDependInfo(args);
            if (ignorePaths.has(f.path)) continue;
            DependFileMap.set(f.path, f);
            InitDir(f, DependDirMap, false);
        }
    }

    /**
     * 将文件放入到目录表中
     * + 目录表可以是临时的
     * @param f
     * @param dirDependMap
     * @param isSuffix
     */
    export function InitDir(f: FileDependInfo, dirDependMap: Map<string, DirDependInfo>, isSuffix: boolean) {
        const charA = '/';

        let fileName = f.path;
        let charALastIndex = fileName.lastIndexOf(charA);
        let dirName = charALastIndex > 0 ? fileName.slice(0, charALastIndex + 1) : "";
        let dirDependInfo = dirDependMap.get(dirName);

        if (!dirDependInfo) {
            dirDependInfo = new DirDependInfo(dirName);
            dirDependMap.set(dirName, dirDependInfo);

            let sub = dirDependInfo;
            while (true) {
                charALastIndex = dirName.lastIndexOf(charA, charALastIndex - 1);

                if (charALastIndex <= 0) {
                    if (!dirDependMap.get(RootDirName)) {
                        dirDependMap.set(RootDirName, new DirDependInfo(RootDirName));
                    }
                    let root = dirDependMap.get(RootDirName);
                    root.children.push(sub);
                    sub.parent = root;
                    break;
                }

                dirName = dirName.slice(0, charALastIndex + 1);

                let parent = dirDependMap.get(dirName);
                if (parent) {
                    parent.children.push(sub);
                    sub.parent = parent;
                    break;
                }

                parent = new DirDependInfo(dirName);
                parent.children.push(sub);
                dirDependMap.set(dirName, parent);

                sub.parent = parent;
                sub = parent;
            }
        }
        dirDependInfo.add(f, isSuffix);
    }

    /**
     * 获得文件depend信息
     * @param path
     */
    export function getFile(path: string) {
        return DependFileMap.get(path);
    }

    /**
     * 获得目录depend信息
     * ***\/  表示为目录，包含所有的文件和子目录
     * @param path
     */
    export function getDir(path: string) {
        return DependDirMap.get(path);
    }

    function getRootDirDependInfo() {
        let root = DependDirMap.get(RootDirName);

        if (!root) {
            root = new DirDependInfo(RootDirName);
            root.children = [];
            DependDirMap.set(RootDirName, root);
        }

        return root;
    }

    export const KEY_PURE_FILE_RECORD_FILTER = "KEY_PURE_FILE_RECORD_FILTER";
    export const DEFAULT_PURE_FILE_RECORD_FILTER = /(\.jpg)|(\.png)|((\.high)|(\.middle)|(\.low))?((((\.s3tc)|(\.astc))?)\.ktx)$/;
    export const KEY_PURE_FILE_REPLACE = "KEY_PURE_FILE_REPLACE";
    export const DEFAULT_PURE_FILE_REPLACE = /(\.jpg)|(\.png)|((\.high)|(\.middle)|(\.low))?((((\.s3tc)|(\.astc))?)\.ktx)$/;
    export function recordForRuntimeFullName(path: string) {
        const purePath = getPureResPath(path);
        let filter = Env.get(KEY_PURE_FILE_RECORD_FILTER) || DEFAULT_PURE_FILE_RECORD_FILTER;
        filter.lastIndex = -1;
        if (filter.test(path)) {
            RuntimeFullNameMap.set(purePath, path);
        }
    }

    /**
     * 获取文件名中以下后缀以外内容 
     * * /(.60)?((.jpg)|(.png)|((.astc)?.ktx))$/
     * 下面正则有点问题,无法匹配到数字
     * 另外3d部分如果要压缩,相对应的配置文件也应该做处理,目前还不支持
     * @param fullPath 文件名
     */
    export function getPureResPath(fullPath: string) {
        // let dir = BaseTools.fileDirname(fullPath);
        // let name = BaseTools.fileName(fullPath);
        // let index = name.indexOf(".");
        // if (index >= 0) {
        //     name = name.slice(0, index);
        // }
        // return `${dir}/${name}`;
        let filter = Env.get(KEY_PURE_FILE_REPLACE) || DEFAULT_PURE_FILE_REPLACE;
        filter.lastIndex = -1;
        return fullPath.replace(filter, '');
    }

    /**
     * 获取 getPureResPath 处理后文件名对应 实际运行时使用的文件名
     * @param pureName 
     */
    export function queryRuntimeFullName(pureName: string) {
        return RuntimeFullNameMap.get(pureName);
    }

    export function forEach(func: (indo: IFileDependInfo) => void) {
        DependFileMap.forEach(v => func(v));
    }
}