/**
 * 日志框架， 需要在构建时，将模块名和cc.xxx()的参数进行混合。
 * 
 * 可以在外部通过路径匹配来配置来决定指定模块及其等级。
 *
 *  Example
 * 
 * ```
 *  import {cc, log} from './log';
 *  cc.debug() && log("user: ", user, 2);
 *  cc.debug("modify") && log("user: ", user, 2, arr);
 * ```
*/

import { BaseTools } from "./base_tools";

/**
 * Log等级
 */
export enum LogLevel {
    /**
     * 调试 
     */
    debug = 1,
    /**
     * 信息
     */
    info,
    /**
     * 警告
     */
    warn,
    /**
     * 错误
     */
    error,
    /**
     * 不打印
     */
    none
}

/**
 * 外部设置的全局日志级别
 */
export let logLevel = LogLevel.info;

/**
 * 根据配置设置日志级别
 */
export const init = (cfg: any) => {
    for (const k in cfg) {
        if (cfg.hasOwnProperty(k)) {
            set(k, cfg[k]);
        }
    }
};

/**
 * 设置模块名的模式匹配及对应的日志级别
 * 
 * @param p 模块模式，支持：*，**；后设置会覆盖前面的设置
 * @param level 对应的日志级别;
 */
export const set = (p: string, level: LogLevel) => {
    const r = BaseTools.pattern(p);
    patternMap.set(p, { reg: r, level: level });
    for (const [k, v] of logMap) {
        if (r.test(k)) {
            v.pattern = p;
            v.level = level;
        }
    }
};

/**
 * 检查 模块是否可以 debug及其以上等级的 日志打印
 * @param mod 模块名
 * @returns bool 是否允许 输出日志；
 */
export const debug = (mod?: string) => {
    return check(LogLevel.debug, mod);
};

/**
 * 检查 模块是否可以 info及其以上等级的 日志打印
 * @param mod 模块名
 * @returns bool 是否允许 输出日志；
 */
export const info = (mod?: string) => {
    return check(LogLevel.info, mod);
};

/**
 * 检查 模块是否可以 warn 及其以上等级的 日志打印
 * @param mod 模块名
 * @returns bool 是否允许 输出日志
 */
export const warn = (mod?: string) => {
    return check(LogLevel.warn, mod);
};

/**
 * 检查 模块是否可以 error 及其以上等级的 日志打印
 * @param mod 模块名
 * @returns bool 是否允许 输出日志
 */
export const error = (mod?: string) => {
    return check(LogLevel.error, mod);
};

/**
 * 整个模块就叫cc；
 */
export const cc = {
    debug: debug,
    info: info,
    warn: warn,
    error: error
};

/**
 * 检查 模块是否可以 level 及其以上等级的 日志打印
 * @param level 日志等级
 * @param mod 模块名
 */
export const check = (level: LogLevel, mod?: string) => {
    if (logLevel > level || !console || !mod) {
        return false;
    }

    let r = logMap.get(mod);
    if (!r) {
        // 匹配最后的设置
        r = { pattern: '', level: LogLevel.none };
        for (const [k, v] of patternMap) {
            if (v.reg.test(mod)) {
                r.pattern = k;
                r.level = v.level;
            }
        }
        logMap.set(mod, r);
    }
    if (r.level > level) {
        return false;
    }
    lastLevel = level;
    lastMod = mod;
    return true;
};

/**
 * 该方法支持重新设置，已支持扩展
 */
export let log = (msg: any, ...args: any[]): void => {
    consoleLog[lastLevel - LogLevel.debug](lastMod, Date.now() / 1000, msg, ...args);
};

// =========================== 本地

/**
 * @hidden
 */
const logMap: Map<string, { pattern: string; level: LogLevel }> = new Map; // 日志表

/**
 * @hidden
 */
const patternMap: Map<string, { reg: RegExp; level: number }> = new Map; // 匹配表

/**
 * @hidden
 */
const consoleLog = [console.debug, console.info, console.warn, console.error];

/**
 * @hidden
 */
let lastLevel = LogLevel.debug;

/**
 * @hidden
 */
let lastMod = '';