
/**
 * 类似于 log4js, log4rs 的 日志系统
 *
 * @note 项目关于日志采集，请使用 pi_util/util/logger 的 getLogger
 * 
 * @note 概念
 * 
 * + Meta: 为了让项目更方便搜索日志，需要绑定一些项目信息到日志系统
 * 
 * + Level 等级，目前有6种，从大到小：5 verbose, 4 debug, 3 info, 2 warn, 1 error, 0 none
 * 
 * + Target 标签，有的日志库是tag，有的日志库是 category
 *      - Filter 和 Append 是按标签分类；
 * 
 * + Logger   采集器
 *      - 实际采集时，logV代表采用verbose等级进行采集，以此类推；
 * 
 * + Filter 目标过滤器
 *      - 如果 采集等级 高于 target对应的最大等级，将忽略该日志事件；
 * 
 * + Appender 附加程序
 *      - 获取日志记录并将其记录在某处，例如，到文件、控制台、http服务器；
 * 
 * + Encoder  编码器 (内部概念)
 *      - 用于 Appender内部，负责对日志的输出格式进行编码；
 * 
 * @note 该模块 不会接管 console.log, console.debug, console.info, console.warn, console.error
 * 
 * @note 如果是 web 和 minigame 平台， 则 wasm 中 的 log库的宏，也会走这里的 HttpAppender
 * 
 * @example
 * 
 * 日志 初始化, game_env.js
 * 
 *      let logger = {
 *          appenders: [{
 *              name: "console",            // 下面filter使用的Appender的名字
 *              kind: "ConsoleAppender",    // 底层通过注册的 Appender实现类
 *          },
 *          {
 *              name: "http",               // 下面filter使用的Appender的名字
 *              kind: "HttpAppender",       // 底层通过注册的 Appender实现类
 *              
 *              args: {
 *                  refresh_interval: 3,            // 3秒 刷新一次
 *                  refresh_max_size: 4 * 1024,     // 达到 4K字节 刷新一次
 *                  url: "http://47.104.135.216:8111",
 *              }
 *          }],
 *       
 *          meta: {
 *              platform: "client", // 客户端 还是 服务器端
 *              project: "pi_demo",    // 项目名
 *              channel: "default",    // 渠道名
 *              process_id: performance.now().toString(),
 *          },
 *   
 *          root: { // 默认 target
 *              level: 4,  // 5 verbose, 4 debug, 3 info, 2 warn, 1 error, 0 none
 *              appenders: ["console"], // ["console", "http"],
 *          },
 *   
 *          loggers: [{
 *              target: "wasm_engine",
 *              level: 4,  // 5 verbose, 4 debug, 3 info, 2 warn, 1 error, 0 none
 *              appenders: ["console"], // ["console", "http"], 
 *          }]
 *      };
 *      
 *      winit.setLogger(logger);
 * 
 * 库 初始化 (以GUI为例，假设 不管debug还是release都用info，不上传 http 的话)
 * 
 *      addLogFilter("gui", LogLevel.info, ["console"]);
 * 
 * 模块 初始化（以GUI的某个模块为例）
 * 
 *      模块变量：
 *          文件 "pi_gui/engine/vdoucment.ts"，对应的 module.name === "pi_gui/engine/vdoucment"
 *          
 *          // 等级从大到小：verbose, debug, info, warn, error, none
 *          let {logV, logD, logI, logW, logE} = getLogger(module.name, "gui");
 * 
 *      函数里面使用: 
 *          
 *          let a = 123;
 *          let b = {
 *              a: 123,
 *              b: "string"
 *          };
 *          
 *          // 输出: "this is a info, a = 123, b = {a: 3434},  c = NAN, "abcdefg", 3343434, 564"
 *          logI("this is a info, a = %d, b = %o, c = %f", 123, {a: 3434}, "abcd", "abcdefg", 3343434, 564);
 *          
 *          // 按上面的过滤规则，没有 输出
 *          logD("this is a debug");
 * 
 */

import { post } from "../net/http";
import { Env } from "./env";
import { formatStringImpl } from "./util";
// ======================== 接口
declare let module;
/**
 * 隐私信息上报
 * - 有些渠道要求某些隐私信息(比如:设备id)必须在隐私弹框之后才能获取
 * - 这里把这些隐私信息上报的时机交给项目自己管理
 */
export const privacyInfoUpload = () => {
	if (window._$pi) {
		_$pi.require('pi_sys/native/index', (mod) => {
			mod.Native.SystemInfo.getDeviceID().then(deviceId => {
				// 设备ID
				deviceId = deviceId || "";
				if (deviceId) {
					addMeta("device_id", deviceId);
				}
				let logger = new Logger(module.name, "sys");
				logger.info("privacy info = %o", {
					deviceId,
				});
				Env.set("deviceId", deviceId);
			})

		});
	}

}

/**
 * 关键日志上报
 */
export enum LogStepId {
	WINIT_START = 10,     // winit start开始
	CFG_DOWNLOAD = 20,    // 开始下载cfg
	CFG_DOWNLOAD_SUCCESS = 30,   // cfg下载成功
	SETUP_MAIN = 40,      // 进入setup-main方法执行主逻辑
	UPDATE_CHECK = 50,      // 更新检查
	UPDATE_CHECK_END = 60,   // 更新检查完成
	UPDATE_CLEAR_DATA_SUCCESS = 70,  // 有更新清除数据
	UPDATE_DOAWLOAD_SUCCESS = 80,     // 更新下载成功
	ENGINE_CFG_INIT_SUCCESS = 90,    // 引擎配置初始化成功
	READY_SUCCESS = 100,           // 相关准备工作初始化成功
	LOAD_INIT_SUCCESS = 110,         // load模块init初始化成功
	FIRST_LOAD_SUCCESS = 120,        // 第一次下载成功
	FIRST_EXEC_SUCCESS = 130,       // 第一次执行成功
	NEXT_LOAD_SUCCESS = 140,         // 第二次下载成功
	START_NEXT_EXEC = 150,          // 开始第二次执行
	NEXT_EXEC_SUCCESS = 160,         // 第二次执行成功
}

/**
 * 关键日志上报name
 */
export const LogStepName = {
	[LogStepId.WINIT_START]: "winit_start",
	[LogStepId.CFG_DOWNLOAD]: "cfg_download",
	[LogStepId.CFG_DOWNLOAD_SUCCESS]: "cfg_download_success",
	[LogStepId.SETUP_MAIN]: "setup_main",
	[LogStepId.UPDATE_CHECK]: "update_check",
	[LogStepId.UPDATE_CHECK_END]: "update_check_end",
	[LogStepId.UPDATE_CLEAR_DATA_SUCCESS]: "update_clear_data_success",
	[LogStepId.UPDATE_DOAWLOAD_SUCCESS]: "update_download_success",
	[LogStepId.ENGINE_CFG_INIT_SUCCESS]: "engine_cfg_init_success",
	[LogStepId.READY_SUCCESS]: "ready_success",
	[LogStepId.LOAD_INIT_SUCCESS]: "load_init_success",
	[LogStepId.FIRST_LOAD_SUCCESS]: "first_load_success",
	[LogStepId.FIRST_EXEC_SUCCESS]: "first_exec_success",
	[LogStepId.NEXT_LOAD_SUCCESS]: "next_load_success",
	[LogStepId.START_NEXT_EXEC]: "start_next_exec",
	[LogStepId.NEXT_EXEC_SUCCESS]: "next_exec_success",
}


/**
 * 关键日志上报
 * @param id 
 */
export const logStep = (id: LogStepId, useTime?: number) => {
	let logger = new Logger(module.name, "sys");
	let obj: any = { step: id, stepname: LogStepName[id] };
	if (!isNaN(useTime)) {
		obj.useTime = useTime;
	}
	logger.info(JSON.stringify(obj))
}

/**
 * 打印到控制台
 */
export const ConsoleAppenderKind = "ConsoleKind";

/**
 * 上传到日志服务器
 */
export const HttpAppenderKind = "HttpKind";

/**
 * 日志等级
 */
export enum LogLevel {
	None = 0,       // 什么都不打印
	Error = 1,      // 错误 信息
	Warn = 2,       // 警告 信息
	Info = 3,       // 关键 信息       
	Debug = 4,      // 调试 信息
	Verbose = 5,    // 最详细 信息，类似于底层的 Trace
}

/**
 * 强制刷新对应的日志
 */
export const flushLog = () => {
	let ps = [];
	for (let [_, appender] of _logAppenderMap) {
		if (appender) {
			ps.push(appender.flush());
		}
	}
	return Promise.all(ps);
}

/**
 * 加 日志 元数据
 * 
 * project: 包名名
 * channel: 渠道名
 * version: 版本
 * project_version:项目版本
 * device_id: 设备ID
 * process_id: 进程ID，用于区分不同时候的进程
 * user_id: 用户ID，用于过滤特定用户
 */
export const addMeta = (key: "project" | "channel" | "version" | "project_version" | "process_id" | "user_id" | "device_id", value: string) => {
	if (!value) {
		console.warn(`addMeta key = ${key},value can not be null or undefined!!!`);
		return;
	}
	// iOS 和 Android 渠道，某些值 从底层取出来的时候，不是字符串类型，也不会报错
	// 故而加这个 判断
	if (typeof value !== typeof "") {
		try {
			value = value.toString();
		} catch (_) {
			value = "";
		}
	}
	switch (key) {
		case "project":
			_metaPackageName = value;
			break;
		case "channel":
			_metaChannel = value;
			break;
		case "version":
			_metaVersion = value;
			break;
		case "project_version":
			_metaProjectVersion = value;
			break;
		case "process_id":
			_metaProcessID = value;
			break;
		case "user_id":
			_metaUserID = value;
			break;
		case "device_id":
			_metaDeviceID = value;
			break;
		default:
			throw new Error("not impl !");
	}
}

/**
 * 强制 全部使用 默认 目标策略
 * 
 * @note 用于特殊场合，比如 服务器配置 某个特定用户 全部用 verbose 打印
 */
export const useRootTargetFilter = (enable: boolean) => {
	_logUseRootDefault = enable;
}

/**
 * 注册 日志 附加程序的 类型器
 * @param kind 用于日志附加程序构造的名字
 * @param constructor 构造函数，都是空参
 * 
 * @note 这里默认会注册两个：ConsoleKind, HttpKind,
 */
export const registerAppenderKind = (kind: string, constructor: AppenderConstructor) => {
	_logAppenderKindMap.set(kind, constructor);
}

/**
 * 添加 appender
 * @param name 附加程序名
 * @param kind AppendKind类型
 * @param args 构造函数参数
 * 
 */
export const addAppender = (name: string, kind: string, args: any) => {
	let constructor = _logAppenderKindMap.get(kind);
	if (!constructor) {
		throw new Error("not find logger appender kind, kind = " + kind);
	}
	if (_logAppenderMap.get(name)) {
		throw new Error("logger appender name has already added, name = " + name);
	}

	let appender = new constructor(args);
	_logAppenderMap.set(name, appender);
}

/**
 * 添加 日志 过滤器
 * @param target 目标，如果值为 root，则 指定 所有target的 默认过滤器；
 * @param maxLevel 最大等级，如果采集日志时小于该等级，则不输出
 * @param appenderNames 该目标的日志 输出地 在哪里
 * @param additive TODO 还没实现可加性；
 */
export const addLogFilter = (target: string, maxLevel: LogLevel, appenderNames: string[]) => {

	let additive = false;

	const filter = { target, maxLevel, appenders: appenderNames, additive };

	_logFilterMap.set(target, filter);
}

/**
 * 给pi_util使用的wasm模块的粘合层
 * 
 * @note 仅 wasm模块使用，在 pi_util/src/wasm/，其他模块不要用；
 */
export const _writeRecord = (target: string, level: LogLevel, file: string, line: number, message: string) => {
	let logger = new Logger(file, target, line);
	switch (level) {
		case LogLevel.Verbose:
			logger.verbose(message);
			break;
		case LogLevel.Debug:
			logger.debug(message);
			break;
		case LogLevel.Info:
			logger.info(message);
			break;
		case LogLevel.Warn:
			logger.warn(message);
			break;
		case LogLevel.Error:
			logger.error(message);
			break;
		default:
			break;
	}
}

/**
 * 日志编码器
 */
export interface ILogEncoder {

	/**
	 * 日志内容：如果最后一个参数 有值，以 最后一个参数为准
	 * 
	 */
	encode(level: LogLevel, logger: Logger, format: string, args: any[], callLocation?: string, message?: string): any;

	getMessage(format: string, args: any[], callLocation?: string): string;
}

/**
 * 日志 输出地
 */
export interface ILogAppender {

	/**
	 * 设置 对应 的 编码器
	 */
	setEncoder(encoder: ILogEncoder);

	/**
	 * 写日志
	 * @param content 编码后的字符串
	 */
	write(level: LogLevel, logger: Logger, format: string, args: any[], callLocation?: string);

	/**
	 * 强制 刷新，等待 返回
	 */
	flush(): Promise<void>;
}

/**
 * 初始化配置
 * 
 */
export interface ILogInitCfg {

	appenders: {
		name: string,   // filter使用的Appender的名字
		kind: string,    // 底层通过注册的 Appender实现类
		args?: any[],    // 参数
	}[],

	meta: {
		version: string,    // 版本
		project: string,    // 项目名
		channel: string,    // 渠道名
		process_id?: string,   // 进程ID
	},

	root: {     // 默认 target
		level: LogLevel,      // 最小过滤等级
		appenders: string[],  // 输出的地方，可以填 "console", "http"
	},

	loggers: {

		// wasm 里面叫做："wasm_engine"
		target: string,       // 目标名，相当于其他日志库的 tag 或者 category

		level: LogLevel,      // 最小过滤等级

		appenders: string[],  // 输出的地方，可以填 "console", "http"

		// additive: boolean, // TODO, 还没实现，是否可追加，false会替换，目前全部按替换搞；
	}[],
};

let initlogger = false;

/**
 * 是否已经初始化日志配置
 * @returns 
 */
export const hasInitLogger = () => {
	return initlogger;
}


/**
 * 如果没有 设置配置，设置一个最简单的
 */
export const initLogger = (cfg: ILogInitCfg) => {
	if (initlogger) return;
	initlogger = true;
	if (!cfg) {
		cfg = {
			appenders: [{
				name: "console",
				kind: "ConsoleKind",
			}],
			meta: {
				project: "",
				channel: "default",
				version: "1.0"
			},

			root: {
				level: LogLevel.Debug,
				appenders: ["console"],
			},
			loggers: [],
		}
	}

	for (let { kind, name, args } of cfg.appenders || []) {
		addAppender(name, kind, args);
	}

	addMeta("version", cfg.meta?.version);
	addMeta("project", cfg.meta?.project);
	addMeta("channel", cfg.meta?.channel);
	if (cfg.meta?.process_id) {
		addMeta("process_id", cfg.meta?.process_id);
	}

	if (cfg.root) {
		addLogFilter("root", cfg.root.level, cfg.root.appenders);
	}

	for (let { target, level, appenders } of cfg.loggers || []) {
		addLogFilter(target, level, appenders);
	}
}

/**
 * 日志采集器
 */
export class Logger {

	from: string;     // 从哪里来，一般填 模块ID
	line: number;     // 行号，可选
	target: string;   // 策略，默认：""

	/**
	 * 构造函数
	 * @param from 从那个模块过来，一般填模块名
	 * @param target 日志 策略
	 */
	constructor(from: string, target = "root", line?: number) {
		this.from = from;
		this.target = target || "root";
		this.line = line;
	}

	/**
	 * 详细 信息，类似于 js -- console.log | rust -- trace!
	 * 
	 * @param format 格式化字符串，类似于 NodeJS 环境的 #[util.format](http://nodejs.cn/api/util.html#util_util_format_format_args)
	 * 
	 * @note 格式化字符如下
	 *      + %s 字符串，实现：String(value);
	 *      + %d 数字，实现： Number(value);
	 *      + %i 十进制整数，实现： parseInt(value, 10);
	 *      + %f 浮点数，实现： parseFloat(value);
	 *      + %o JSON字符串，实现：JSON.stringify(value);
	 *      + %% 输出一个百分号；
	 */
	verbose(format: string, ...args: any[]) {
		this.handleLog(LogLevel.Verbose, format, args, getCallLocation());
	}

	/**
	 * 调试 信息，类似于 js -- console.debug | rust -- debug!
	 * 
	 * @param format 格式化字符串，类似于 NodeJS 环境的 #[util.format](http://nodejs.cn/api/util.html#util_util_format_format_args)
	 * 
	 * @note 格式化字符如下
	 *      + %s 字符串，实现：String(value);
	 *      + %d 数字，实现： Number(value);
	 *      + %i 十进制整数，实现： parseInt(value, 10);
	 *      + %f 浮点数，实现： parseFloat(value);
	 *      + %o JSON字符串，实现：JSON.stringify(value);
	 *      + %% 输出一个百分号；
	 */
	debug(format: string, ...args: any[]) {
		this.handleLog(LogLevel.Debug, format, args, getCallLocation());
	}

	/**
	 * 关键 信息，类似于 js -- console.info | rust -- info!
	 * 
	 * @param format 格式化字符串，类似于 NodeJS 环境的 #[util.format](http://nodejs.cn/api/util.html#util_util_format_format_args)
	 * 
	 * @note 格式化字符如下
	 *      + %s 字符串，实现：String(value);
	 *      + %d 数字，实现： Number(value);
	 *      + %i 十进制整数，实现： parseInt(value, 10);
	 *      + %f 浮点数，实现： parseFloat(value);
	 *      + %o JSON字符串，实现：JSON.stringify(value);
	 *      + %% 输出一个百分号；
	 */
	info(format: string, ...args: any[]) {
		this.handleLog(LogLevel.Info, format, args, getCallLocation());
	}

	/**
	 * 警告 信息，类似于 js -- console.warn | rust -- warn!
	 * 
	 * @param format 格式化字符串，类似于 NodeJS 环境的 #[util.format](http://nodejs.cn/api/util.html#util_util_format_format_args)
	 * 
	 * @note 格式化字符如下
	 *      + %s 字符串，实现：String(value);
	 *      + %d 数字，实现： Number(value);
	 *      + %i 十进制整数，实现： parseInt(value, 10);
	 *      + %f 浮点数，实现： parseFloat(value);
	 *      + %o JSON字符串，实现：JSON.stringify(value);
	 *      + %% 输出一个百分号；
	 */
	warn(format: string, ...args: any[]) {
		this.handleLog(LogLevel.Warn, format, args, getCallLocation());
	}

	/**
	 * 错误 信息，类似于 js -- console.error | rust -- trace!
	 * 
	 * @param format 格式化字符串，类似于 NodeJS 环境的 #[util.format](http://nodejs.cn/api/util.html#util_util_format_format_args)
	 * 
	 * @note 格式化字符如下
	 *      + %s 字符串，实现：String(value);
	 *      + %d 数字，实现： Number(value);
	 *      + %i 十进制整数，实现： parseInt(value, 10);
	 *      + %f 浮点数，实现： parseFloat(value);
	 *      + %o JSON字符串，实现：JSON.stringify(value);
	 *      + %% 输出一个百分号；
	 */
	error(format: string, ...args: any[]) {
		this.handleLog(LogLevel.Error, format, args, getCallLocation(true));
	}


	getFilter(currentLevel: LogLevel) {
		let filter;

		// 不强制用 默认的，就先看看target的
		if (!_logUseRootDefault) {
			filter = _logFilterMap.get(this.target);
		}

		// 没有，用默认的
		if (!filter) {
			filter = _logFilterMap.get("root");
		}

		// 连默认过滤器都没有，不需要日志
		if (!filter) {
			console.error("warn: default log filter isn't set");
			return;
		}

		// 如果 当前level 比 期望的最大Level 还大，不需要日志
		if (currentLevel > filter.maxLevel) {
			return;
		}
		return filter;
	}

	private handleLog(currentLevel: LogLevel, format: string, args: any[], callLocation = "") {
		let filter = this.getFilter(currentLevel);
		if (!filter) return;
		for (let name of filter.appenders) {
			let appender = _logAppenderMap.get(name);
			if (!appender) {
				throw new Error("log appender not found, appender = " + name);
			}
			// 一个一个输出写
			appender.write(currentLevel, this, format, args, callLocation);
		}
	}
}

// ======================== 实现

/**
 * 将日志 编码为 Json对象
 * 
 * 整个日志编码为：
 *  
 * {
 *      target: 分类标签,
 *      time: "年/月/日-时:分:秒.毫秒",
 *
 *      level: "verbose" | "debug" | "info" | "warn" | "error",
 *      process_id: string,    // 进程ID，Web或小游戏，需要 模拟一个独一无二的进程ID
 *      device_id: string,     // 设备ID
 *      user_id: string,       // 用户ID，用于过滤特定用户的日志；
 *      file: string,          // 文件路径，可选,
 *       
 *      // Rust层 需要，Web 或 小游戏 没有
 *      line: number,          // 行号，可选
 *      thread_id: number,     // 线程ID，可选
 *      thread_name: string,   // 线程名，可选
 *    
 *      message: string;      // 具体 日志 信息
 * }
 */
export class JsonEncoder implements ILogEncoder {

	constructor() {

	}

	/**
	 * 编码 成 json 对象
	 * @param log 
	 * @param args 
	 */
	encode(level: LogLevel, logger: Logger, format: string, args: any[], callLocation = "", message = "") {
		let meta = this.getMeta(level, logger);
		if (!message) {
			message = this.getMessage(format, args, callLocation);
		}

		let json = {
			...meta,
			message,
		}

		return json;
	}

	/**
	 * process_id: string,
	 * device_id: string,
	 * user_id: string,
	 * create_time: string, 日志生成时间，ISO-String, 英国格林威治时间，类似 '2022-01-07T02:50:43.495Z'
	 * target: string,
	 * level: "verbose" | "debug" | "info" | "warn" | "error",
	 * file: string,     // 文件路径，可选,
	 * line: number,
	 */
	private getMeta(level: LogLevel, logger: Logger) {
		let target = logger.target;
		let create_time = getNow();
		let levelName = getLevelName(level);
		let file = logger.from;

		if (!_metaProcessID) {
			let currentTime = new Date().getTime();
			_metaProcessID = currentTime.toString();
		}

		let processID = _metaProcessID;

		let r = {
			target,
			create_time,
			file,
			level: levelName,
			user_id: _metaUserID,
			process_id: processID,
		} as any;

		if (typeof logger.line === typeof 0) {
			r.line = logger.line.toString();
		}

		if (_metaDeviceID) {
			r.device_id = _metaDeviceID;
		}

		if (_metaProjectVersion) {
			r.multi_version = _metaProjectVersion;
		}
		return r;
	}

	getMessage(format: string, args: any[], callLocation = "") {
		let [str, currentArg] = formatStringImpl(format, args);

		let r = [str];
		for (; currentArg < args.length; ++currentArg) {
			let arg = args[currentArg];
			if (arg === undefined) {
				arg = null;
			}
			r.push(arg);

			// 插入调用者定位信息
			if (callLocation) {
				r.push(callLocation);
			}

		}
		let message = "";
		try {
			message = JSON.stringify(r);
		} catch (e) {

		}
		return message;
	}
}

/**
 * 最简单的 编码器，用于 ConsoleAppender
 */
class SimpleEncoder implements ILogEncoder {

	getMessage(format: string, args: any[], callLocation?: string): string {
		return undefined;
	}

	/**
	 * 编码 成 json 对象
	 * @param logger
	 * @param args
	 */
	encode(level: LogLevel, logger: Logger, format: string, args: any[], callLocation = "", message = "") {

		// ======== 日志体

		let r;

		if (navigator.userAgent.indexOf("PI_NATIVE") > -1) {
			// 原生开发平台
			r = [this.formatMessage(format, args)];
		} else {
			// 不管 是 浏览器 还是 小游戏，都有 控制台
			r = [format, ...args];
		}

		// =========== 调用者定位
		r.push("    " + callLocation);

		// ======== 日志头

		let create_time = getNow();
		let levelName = getLevelName(level);

		let header = `${levelName}/${logger.target}; `;

		header += `from = ${logger.from}; `;

		if (!_metaProcessID) {
			let currentTime = new Date().getTime();
			_metaProcessID = currentTime.toString();
		}
		header += `process_id = ${_metaProcessID}; `;

		if (_metaUserID) {
			header += `user_id = ${_metaUserID}; `;
		}
		if (_metaDeviceID) {
			header += `device_id = ${_metaDeviceID}; `;
		}
		if (typeof logger.line === typeof 0) {
			header += `line = ${logger.line}; `;
		}
		header += `create_time = ${create_time}; `;

		r.push("\n    " + header);

		return r;
	}

	private formatMessage(format: string, args: any[]) {
		let [str, currentArg] = formatStringImpl(format, args);

		let r = [str];
		for (; currentArg < args.length; ++currentArg) {
			let arg = args[currentArg];

			try {
				if (arg === undefined) {
					arg = "null";
				} else {
					let str = JSON.stringify(arg);
					arg = str;
				}
			} catch (e) {
			}

			r.push(arg);
		}
		return r.join(", ");
	}
}

/**
 * 打印 到 控制台
 */
class ConsoleAppender implements ILogAppender {

	encoder: ILogEncoder;

	constructor() {
		this.encoder = new SimpleEncoder();
	}

	flush() {
		return Promise.resolve();
	}

	/**
	 * 设置 对应 的 编码器
	 * @param encoder，默认 空
	 */
	setEncoder(encoder: ILogEncoder) {
		this.encoder = encoder;
	}

	/**
	 * 打印到控制台，挨个打印即可
	 */
	write(level: LogLevel, logger: Logger, format: string, args: any[], callLocation = "") {

		if (!this.encoder) {
			return;
		}

		let logFun = this.geConsoleLogFunc(level);

		let r = this.encoder.encode(level, logger, format, args, callLocation);
		if (logFun) {
			if (Array.isArray(r)) {
				let body = r.slice(0, r.length - 2);
				logFun(...body);

				let meta = r.slice(r.length - 2);
				console.groupCollapsed && console.groupCollapsed();
				logFun(...meta);
				console.groupEnd && console.groupEnd();

				logFun("");
			} else {
				logFun(r);
			}
		}
	}

	private geConsoleLogFunc(level: LogLevel) {
		let logFun = null;
		switch (level) {
			case LogLevel.Verbose:
				logFun = console.log;
				break;
			case LogLevel.Debug:
				logFun = console.log;
				break;
			case LogLevel.Info:
				logFun = console.log;
				break;
			case LogLevel.Warn:
				logFun = console.warn || console.log;
				break;
			case LogLevel.Error:
				logFun = console.error || console.log;
				break;
			default:
				break;
		}
		if (logFun) {
			logFun = logFun.bind(console);
		}
		return logFun;
	}
}

interface IHttpAppenderArgs {
	url: string;

	refresh_interval: number;   // 几秒 刷新一次，单位：秒
	refresh_max_count: number;  // 达到 多少条日志 刷新一次；单位：条
	post_timeout: number;       // Post 请求 超时 时间，单位：秒
};

/**
 * 将日志上传到 Http 服务器
 */
class HttpAppender implements ILogAppender {

	url: string;
	refreshInterval: number;     // 多久 刷新一次，默认值：100 ms
	refreshMaxCount: number;    // 达到 多少条日志 刷新一次
	postTimeout: number;    // Http-Post 超时，默认值：10 * 1000 ms
	uploadPromise: Promise<void>; // 上传任务的Promise 
	uploadTimer: number;          // 上传定时器；
	lastUploadInfo: any;     // 上次上报日志,用作上报失败时重新上报

	encoder: ILogEncoder;
	logEvents: any[];

	// 日志信息 hash值，用于 去重
	msgSet: Set<number>;

	constructor(cfg: IHttpAppenderArgs) {
		this.logEvents = [];
		this.encoder = new JsonEncoder();
		this.msgSet = new Set();

		if (cfg) {
			this.url = cfg.url;

			this.postTimeout = (cfg.post_timeout || 10) * 1000;
			this.refreshMaxCount = cfg.refresh_max_count || 3;
			this.refreshInterval = (cfg.refresh_interval || 0.1) * 1000;
		}
	}

	flush() {
		return this.uploadWhenReady(true);
	}

	setEncoder(encoder: ILogEncoder) {
		this.encoder = encoder;
	}

	write(level: LogLevel, logger: Logger, format: string, args: any[], callLocation = "") {
		if (!this.url || !this.encoder) {
			return;
		}

		// 去重
		let msg = this.encoder.getMessage(format, args, callLocation);
		if (typeof msg === typeof "") {
			let h = this.getHash(msg);
			if (this.msgSet.has(h)) {
				return;
			}
			this.msgSet.add(h);
		}
		let r = this.encoder.encode(level, logger, format, args, callLocation, msg);
		this.logEvents.push(r);

		this.uploadWhenReady();
	}

	// 计算字符串的 hash
	private getHash(msg: string): number {
		let hash = 0;
		if (!msg) {
			return hash;
		}

		for (let i = 0; i < msg.length; ++i) {
			let c = msg.charCodeAt(i);

			// 乘31，然后使用了 “|” 运算符，转换参数为 32bit，确保结果是个32位整数
			hash = (((hash << 5) - hash) + c) | 0;
		}

		return hash;
	};

	/** 
	 * 超时 或 超过指定条目，上传一次
	 * 正在上传中，不能再次上传，节省网络负载；
	 * @param isForce 是否强制刷新
	 */
	private uploadWhenReady(isForce = false): Promise<void> {

		// 正在进行，返回
		if (this.uploadPromise) {
			return this.uploadPromise;
		}

		if (isForce || this.logEvents.length >= this.refreshMaxCount) {
			// 到数量后，清掉定时器；
			if (this.uploadTimer) {
				clearTimeout(this.uploadTimer);
				this.uploadTimer = null;
			}
			// 到数量后，上传
			return this.uploadLog(this.postTimeout);
		} else if (!this.uploadTimer) {
			// 设置定时器
			return new Promise((resolve, reject) => {
				this.uploadTimer = setTimeout(() => {
					this.uploadTimer = null;
					// 到时间后，上传
					this.uploadLog(this.postTimeout).then(() => {
						resolve(undefined);
					}).catch((e) => {
						reject(e);
					});
				}, this.refreshInterval);
			});
		}

		return Promise.resolve();
	}

	private uploadLog(postTimeout: number) {
		// 开始上传
		return this.uploadPromise = this.uploadLogImpl(postTimeout).then(() => {
			// 上传结束，清变量
			this.uploadPromise = null;

			if (this.logEvents.length > 0) {
				// 如果还有内容，继续 上传
				this.uploadWhenReady();
			}
		});
	}

	/**
	 * 
	 * 调用 HttpDownload 上传到 阿里云的 日志服务器
	 * 
	 *  https://help.aliyun.com/document_detail/120218.htm?spm=a2c4g.11186623.2.12.557144f0RZFp2D#reference-354467
	 * 
	 * 注：这里 __logs__ 数组每个元素都是 对象，对象的 所有的值 都要是 字符串，否则 日志服务器会返回 400 错误，说 Invalid Body
	 * 
	 * 例1：下面发送会报错
	 * 
	 * __logs__: [{
	 *     process_id: 124546,
	 * }]
	 * 
	 * 例2：下面发送不会报错
	 * 
	 * __logs__: [{
	 *     process_id: "124546",
	 * }]
	 * 
	 * 注：如果 客户端要 严格跨域，那么 最终发送的 请求头 必须包含 Origin 和 Referer 这两个字段（都是宿主加），否则会报 跨域 错误。
	 * 
	 * {
	 *   __source__: "${project}-${channel}-${version}-${project_version}",
	 *   __logs__: [日志json, ...]
	 * }
	 */
	private async uploadLogImpl(postTimeout: number) {
		if (this.logEvents.length === 0 && !this.lastUploadInfo) {
			return;
		}

		const header = {
			"x-log-apiversion": "0.6.0",
			"x-log-bodyrawsize": "0",
		};

		let info;
		if (!this.lastUploadInfo) {
			let slice = this.logEvents.slice(0, this.refreshMaxCount);
			this.logEvents.splice(0, this.refreshMaxCount);
			// __source__  包名-渠道名-包版本
			info = {
				__source__: `${_metaPackageName}-${_metaChannel}-${_metaVersion}`,
				__logs__: slice,
			};
		}

		try {
			let str;
			if (!this.lastUploadInfo) {
				str = JSON.stringify(info);
				this.lastUploadInfo = str;
			} else {
				str = this.lastUploadInfo;
				this.lastUploadInfo = null;   // 只尝试重新发送一次
			}

			// 发出去 或者 错误，才发 下一个，减少 网络负载；
			await post(
				this.url, header, str,
				"application/json", "text",
				postTimeout);
			this.lastUploadInfo = null;
		} catch (e) {
			// TODO：处理 异常，只有超时，才会重新发送；
		}
	}
}

/**
 * 日志过滤器
 */
interface ILogFilter {
	target: string,         // 目标名
	maxLevel: LogLevel,     // 最小过滤等级
	appenders: string[],    // 输出地的名字，对应于 _logAppenderMap 的 键
	additive: boolean,      // 为true代表可追加，为false代表覆盖（后设置的覆盖前面设置的）
}

// 策略 过滤器，如果找不到，默认为 None
// root 为 默认所有的日志 过滤器
let _logFilterMap: Map<string, ILogFilter> = new Map();

// 用默认target过滤器代替所有的过滤器；
let _logUseRootDefault = false;

interface AppenderConstructor {
	new(args?: any): ILogAppender;
};

// 日志类型注册器
let _logAppenderKindMap: Map<string, AppenderConstructor> = new Map();

// 日志 生成器
let _logAppenderMap: Map<string, ILogAppender> = new Map();

// 元信息，包名，自动获取底层包名
let _metaPackageName = "";

// 元信息，渠道名,自动获取底层渠道
let _metaChannel = "";

// 元信息，设备ID
let _metaDeviceID = "";

// 元信息，版本,自动获取底层版本
let _metaVersion = "";

// 元信息,项目版本-多版本有
let _metaProjectVersion = "";

// 元信息，进程ID，用于查找这个进程的日志
// 默认情况下，是 第一次采集日志的时间；
let _metaProcessID = "";

// 元信息，用户ID，用于客户端日志搜索这个用户
let _metaUserID = "";

/**
 * 返回时间字符串
 * ISO-String, 英国格林威治时间，类似 '2022-01-07T02:50:43.495Z'
 */
const getNow = () => {
	let date = new Date();
	return date.toISOString();
}

/**
 * 取 LogLevel 对应的名字
 */
const getLevelName = (level: LogLevel) => {
	let name = "";
	switch (level) {
		case LogLevel.Verbose:
			name = "Verbose";
			break;
		case LogLevel.Debug:
			name = "Debug";
			break;
		case LogLevel.Info:
			name = "Info";
			break;
		case LogLevel.Warn:
			name = "Warn";
			break;
		case LogLevel.Error:
			name = "Error";
			break;
		case LogLevel.None:
			name = "none";
			break;
		default:
			throw new Error("not impl!, level = " + level);
	}
	return name;
}

/**
 * 取日志调用者的定位点；
 * 
 * @param isGetAll 是否獲取 所有堆棧，如果為false，則獲取最後一行
 */
const getCallLocation = (isGetAll = false) => {
	// ============= 堆栈 定位 到 调用 的 地方；
	let location = "";
	try {
		// 取堆栈，注：并不是每个宿主都有Error对象的stack属性；
		let stack = new Error().stack;
		if (stack) {
			if (isGetAll) {
				location = stack;
			} else {
				// arr = [Error, getCallLocation, Logger.verbose/debug/..., 有可能是pi_utils的位置, 调用者位置, ...]
				let arr = stack.split("\n");
				location = arr[3].trim();
				if (/log[V|D|I|W|E]/.test(location)) {
					location = arr[4] || "";
				}
			}
		}
		location = location.trim();
	} catch (e) {

	}
	return location.trim();
}

// 注册两个Appender
(function () {

	registerAppenderKind(ConsoleAppenderKind, ConsoleAppender);

	registerAppenderKind(HttpAppenderKind, HttpAppender);
})();