import { decodeBuffer } from "pi_sys/load/data_transfer";
import { stringify } from "pi_sys/load/down_load_hook";
import { FileLoadParam, Result } from "pi_sys/load/fileload_base";
import { HttpDownload } from "../net/http";
import { BaseTools } from "./base_tools";
import { Depend } from "./depend";
import { IDirDependInfo, IFileDependInfo } from "./depend_base";
import { Env } from "./env";
import { utf8Decode } from "./utf8";
import { getDecryptType, getEncryptKey } from "./util";

// 解密后的数据,下载的原始数据(解密后的数据),错误
type Callback = (data: Uint8Array, downloadData?: Uint8Array, err?: Error) => void;
type ProcessCallback = (url: string, type: string, total: number, loaded: number) => void

let networkSpeed = 500 * 1024;
/**
 * 网络权重任务池
 * - 由应用层判断从本地加还是网络下.这里只管网络下载的部分
 * - 以队列维护,每个队列有权重,大体可以分为快下载文字队列,快下载资源队列,慢下载文字队列,慢下载资源队列
 * - 需要维护正在下载的情况
 * - 需要维护在等待队列的情况
 * - 下载有并发数限制,可以设计总并发数,每个队列的并发数,比如快下载并发数可以设置为总并发数大小,慢下载的并发数设置为1,可以避免慢下载占用全部网络资源的情况
 * - 计算每个文件的下载进度,使用下载百分比来模拟进度
 */
export class NetWeightQueuePool {
    maxCount = 6;  // 总并发数
    private weightQueueMap: Map<number, WeightQueue> = new Map;       // 权重列表map
    callbackMap: Map<string, Callback[]> = new Map;           // 回调map
    private processMap: Map<string, ProcessCallback[]> = new Map;                 // 进度条回调map
    private downloading: Set<string> = new Set;                        // 正在下载的列表
    private dataCache: Map<string, Uint8Array> = new Map;         // 数据缓存,等到外部写完成后通知移除
    private downDataCache: Map<string, Uint8Array> = new Map;   // 原始数据(下载的加密数据)缓存,等到外部写完成后通知移除
    private timer: any;

    get curCount() {
        let count = 0;
        this.weightQueueMap.forEach(wq => {
            count += wq.curCount;
        });
        return count;
    }


    public createWeightQueue(option: WeightQueueOption) {
        if (this.weightQueueMap.has(option.weight)) {
            throw new Error(`WeightQueue has already exsit!`);
        }
        const queue = new WeightQueue(this, option);
        this.weightQueueMap.set(option.weight, queue);
        return queue;
    }

    public put(weight: number, key: string, callback: Callback, processCallback?: ProcessCallback) {
        if (this.dataCache.has(key)) {
            let data = this.dataCache.get(key);
            let downData = this.downDataCache.get(key)
            setTimeout(() => {
                // console.log("====== get from dataCache ", key);
                callback && callback(data, downData);
            }, 0);
            return;
        }
        if (this.downloading.has(key)) {
            return this.pushCallback(key, callback, processCallback);
        }
        const queue = this.weightQueueMap.get(weight);
        if (!queue) {
            throw new Error(`WeightQueue not find,you should call createWeightQueue first!`);
        }
        // 检查该资源是否已经在等待下载队列
        const inQueue = this.checkInWeightQueue(key);
        if (inQueue) {
            if (inQueue.weight > weight) {  // 原来的队列权重更低,需要将资源提升到更高权重队列
                inQueue.delete(key);   // 从原队列移除
                queue.put(key);       // put到更高权重的队列
            }
        } else {
            queue.put(key);
        }
        this.pushCallback(key, callback, processCallback);
        if (!this.timer) {
            this.timer = setTimeout(() => {
                this.timer = null;
                this.run();
            }, 10);
        }
    }

    public onSuccess(infos: IFileDependInfo[], fileMap: Map<string, Uint8Array>, downloadFileMap: Map<string, Uint8Array>) {
        this.removeDownloading(infos);
        this.removeProcessCallback(infos);
        infos.forEach(info => {
            let key = info.path;
            const callbacks = this.callbackMap.get(key);
            if (callbacks) {
                this.callbackMap.delete(key);
                callbacks.forEach(cb => {
                    let data = fileMap.get(key);
                    let downData = downloadFileMap.get(key);
                    this.dataCache.set(key, data);
                    this.downDataCache.set(key, downData);
                    cb && cb(data, downData);
                })
            }
        })

    }

    public onFail(infos: IFileDependInfo[], err: Error) {
        this.removeDownloading(infos);
        this.removeProcessCallback(infos);
        infos.forEach(info => {
            let key = info.path;
            const callbacks = this.callbackMap.get(key);
            if (callbacks) {
                this.callbackMap.delete(key);
                callbacks.forEach(cb => {
                    cb && cb(null, null, err);
                })
            }
        })

    }
    public onProcess(infos: IFileDependInfo[], url: string, type: string, percent: number) {
        const now = Date.now()
        infos.forEach(info => {
            const callbacks = this.processMap.get(info.path);
            if (callbacks) {
                let size = info.encryptSize || info.size
                let downSize = Math.floor(size * percent);
                callbacks.forEach(cb => {
                    cb(url, type, size, downSize);
                })

            }
        })
    }

    public addDownloading(infos: IFileDependInfo[], wq: WeightQueue) {
        infos.forEach(info => {
            let key = info.path;
            this.downloading.add(key);
            wq.delete(key);
        })
    }

    public removeDownloading(infos: IFileDependInfo[]) {
        infos.forEach(info => {
            let key = info.path;
            this.downloading.delete(key);
        });
    }

    public removeProcessCallback(infos: IFileDependInfo[]) {
        infos.forEach(info => {
            let key = info.path;
            this.processMap.delete(key);
        });
    }

    // 移除缓存,一般在落地成功后移除
    public removeDataCache(key: string) {
        this.dataCache.delete(key);
        this.downDataCache.delete(key);
    }

    private pushCallback(key: string, callback: Callback, processCallback?: ProcessCallback) {
        let callbackList = this.callbackMap.get(key);
        if (!callbackList) {
            callbackList = [];
            this.callbackMap.set(key, callbackList);
        }
        callbackList.push(callback);

        if (processCallback) {
            let processCallbackList = this.processMap.get(key);
            if (!processCallbackList) {
                processCallbackList = [];
                this.processMap.set(key, processCallbackList);
            }
            processCallbackList.push(processCallback);
        }

    }

    /**
     * 检查资源是否已经在权重队列
     * @param key 
     * @returns 资源所在的权重队列,null表示不在任何权重队列
     */
    private checkInWeightQueue(key: string) {
        let queue: WeightQueue = null;
        for (const [w, q] of this.weightQueueMap) {
            if (q.has(key)) {
                queue = q;
                break;
            }
        }
        return queue;
    }

    /**
     * 找到权重最大的队列
     * @returns 
     */
    private findMaxWeightQueue() {
        let queue: WeightQueue = null;
        for (const [w, q] of this.weightQueueMap) {
            if (!q.isPaused() && q.size()) {
                if (!queue || queue.weight > q.weight) {
                    queue = q;
                }
            }
        }
        return queue;
    }

    run() {
        if (this.curCount >= this.maxCount) return;
        const wq = this.findMaxWeightQueue();
        if (!wq) return;
        wq.startDownload();
        // 比如慢下载需要有间隔时间
        if (wq.maxCount === 1 && wq.runTimer) return;
        // 如果权重队列的并发数还未用完,尝试下载其他队列
        if (wq.curCount < wq.maxCount) {
            this.run();
        }
    }

}


export interface WeightQueueOption {
    weight: number;   // 权重,Int类型,1开始,数字越小权重越大
    maxCount: number;    // 总并发数
    desc?: string;    // 简要描述,说明当前队列的作用
    domains: string[];   // 下载域名
    isMultiVersion: boolean;                 // 是否是多版本资源,SDF文字就不是多版本资源
    getDependInfo: (file: string) => IFileDependInfo;   // 通过文件名获取文件Depend信息,外部使用的时候需要实现此方法
}

export class WeightQueue {
    weight: number;  // 权重,Int类型,1开始,数字越小权重越大
    maxCount: number;    // 总并发数
    curCount: number = 0;    // 当前并发数
    private desc: string;      // 简要描述,说明当前队列的作用
    private queue: Set<string> = new Set;      // 队列列表 
    private pool: NetWeightQueuePool;
    private getDependInfo: (file: string) => IFileDependInfo;  // 通过文件名获取文件Depend信息,外部使用的时候需要实现此方法
    private domains: string[];                         // 下载域名
    private isMultiVersion: boolean = true;                 // 是否是多版本资源,SDF文字就不是多版本资源
    // 一次URL下载成功回调,慢下载会用到
    private downSuccessCallback: (fileMap: Map<string, Uint8Array>, downloadFileMap: Map<string, Uint8Array>) => void;
    private delay = 0;   // 下载成功后的延迟时间
    private sizeLimit = 0;              // 一次下载大小限制
    private timerMap: Map<string, any> = new Map;
    private paused: boolean = false;          // 是否暂停下载

    runTimer;

    constructor(pool: NetWeightQueuePool, option: WeightQueueOption) {
        this.pool = pool;
        this.weight = option.weight;
        this.maxCount = option.maxCount;
        this.desc = option.desc;
        this.domains = option.domains;
        this.isMultiVersion = option.isMultiVersion;
        this.getDependInfo = option.getDependInfo;
    }

    setOption(delay: number, sizeLimit: number) {
        this.delay = delay;
        this.sizeLimit = sizeLimit;
    }

    // 设置是否暂停
    public setPaused(paused: boolean) {
        this.paused = paused;
        if (!paused) {
            setTimeout(() => {
                this.pool.run();
            }, 10);
        }
    }

    // 是否暂停
    isPaused() {
        return this.paused;
    }


    registerDownSuccessCallback(func: (fileMap: Map<string, Uint8Array>, downloadFileMap: Map<string, Uint8Array>) => void) {
        this.downSuccessCallback = func;
    }

    removeDownSuccessCallback() {
        this.downSuccessCallback = null;
    }


    put(key: string) {
        this.queue.add(key);
    }

    pop() {
        const iter = this.queue.values();
        const { done, value } = iter.next();
        if (!done) {
            this.queue.delete(value);
        }
        return value;
    }

    prePop() {
        const iter = this.queue.values();
        const { done, value } = iter.next();
        return value;
    }

    has(key: string) {
        return this.queue.has(key);
    }

    delete(key: string) {
        return this.queue.delete(key);
    }

    size() {
        return this.queue.size;
    }

    // 字节大小
    byteSize() {
        let size = 0;
        for (const k of this.queue) {
            const info = this.getDependInfo(k);
            if (info) {
                size += info.size;
            }
        }
        return size;
    }

    canDownload() {
        if (this.curCount < this.maxCount && this.pool.curCount < this.pool.maxCount) {
            return true;
        } else {
            return false;
        }
    }

    // 开始下载,直到并发数限制或者没有可下载的
    startDownload() {
        let now = Date.now();
        if (!this.canDownload()) return;
        if (this.delay && this.runTimer) {
            return
        }

        // 这里需要提前计算,否则stringify没有顺序,不会按照我们期望的顺序下载
        let results: string[][];
        let sizeLimit = this.sizeLimit || FileLoadParam.sizeLimit;
        // 慢下载必须按照其设置顺序下载
        if (this.maxCount === 1) {
            const infos: IFileDependInfo[] = [];
            let totalSize = 0;
            for (const path of this.queue) {
                const info = this.getDependInfo(path);
                if (info) {
                    let size = info.encryptSize || info.size;
                    infos.push(info);
                    if (size + totalSize > sizeLimit) {
                        break;
                    }
                    totalSize += size;
                }
            }
            results = this.mergeDownFiles(infos, sizeLimit);
        } else {
            const files = [...this.queue.values()];
            const infos = files.map(f => this.getDependInfo(f));
            results = this.mergeDownFiles(infos, sizeLimit);
        }

        for (let i = 0; i < results.length; i++) {
            const files = results[i];
            // 加载文件的数据
            let fileMap: Map<string, Uint8Array> = new Map;
            let downloadFileMap: Map<string, Uint8Array> = new Map;
            let map: Map<string, IDirDependInfo> = new Map;
            // 先构建一个目录树，目录内再构建一个后缀表
            for (let key of files) {
                const info = this.getDependInfo(key);
                Depend.InitDir(info, map, true);
            }
            let fileDirTree = map.get("");
            let result = new Result();
            // 检查url长度, 限制长度1k，超过1k，需要分多次下载
            // size大小在前面限制,这里不做限制,设置一个较大值
            stringify(fileDirTree, FileLoadParam.limitLength, result, 8 * 1024 * 1024);
            this.curCount++;
            this.pool.addDownloading(result.files, this);
            const finallyDo = () => {
                this.curCount--;
                if (!this.runTimer) {
                    this.runTimer = setTimeout(() => {
                        this.runTimer = null;
                        this.pool.run();
                    }, this.delay);
                }
            }
            const resultFiles = result.files;
            this.startURL(result.d, result.url, resultFiles, fileMap, downloadFileMap).then(() => {
                this.pool.onSuccess(resultFiles, fileMap, downloadFileMap);
                this.downSuccessCallback && this.downSuccessCallback(fileMap, downloadFileMap);
                finallyDo();   // 为了兼容低版本webview不支持finally
            }).catch(e => {
                this.pool.onFail(result.files, e);
                finallyDo();   // 为了兼容低版本webview不支持finally
            });
            while (result.next) {
                // size大小已经在刚开始做了限制,这里只可能是url长度大小限制造成的,如果这种情况,
                // 不管当前是否已经超并发数,都发出去
                result = new Result();
                stringify(fileDirTree, FileLoadParam.limitLength, result, 8 * 1024 * 1024);
                this.curCount++;
                this.pool.addDownloading(result.files, this);
                const resultFiles = result.files;
                this.startURL(result.d, result.url, resultFiles, fileMap, downloadFileMap).then(() => {
                    this.pool.onSuccess(resultFiles, fileMap, downloadFileMap);
                    this.downSuccessCallback && this.downSuccessCallback(fileMap, downloadFileMap);
                    finallyDo();   // 为了兼容低版本webview不支持finally
                }).catch(e => {
                    this.pool.onFail(result.files, e);
                    finallyDo();   // 为了兼容低版本webview不支持finally
                });
            }
            if (!this.canDownload()) {
                break;
            }
        }
    }

    /**
     * 依据size大小从大到小排序,返回合并后需要下载的文件合并结果
     * @param infos 
     * @returns 
     */
    mergeDownFiles(infos: IFileDependInfo[], sizeLimit: number) {
        infos.sort((info1, info2) => {
            let size1 = info1.encryptSize || info1.size;
            let size2 = info2.encryptSize || info2.size;
            return size2 - size1;
        });
        const results: string[][] = [];
        let arr = [];
        let mergeSize = 0;
        for (let i = 0; i < infos.length; i++) {
            const info = infos[i];
            let size = info.encryptSize || info.size;
            arr.push(info.path);
            mergeSize += size;
            if (mergeSize > sizeLimit) {
                results.push(arr);
                arr = [];
                mergeSize = 0;
            }
        }
        if (arr.length) {
            results.push(arr);
        }
        return results;
    }

    startURL(durl: string, furl: string, files: IFileDependInfo[], fileMap: Map<string, Uint8Array>, downloadFileMap: Map<string, Uint8Array>) {

        let now = Date.now();
        let len = files.length, size = 0, h = 0; // hash值
        if (len === 0) {
            return;
        }
        for (let i = 0; i < len; i++) {
            let info = files[i];
            size += info.encryptSize || info.size;
            h = BaseTools.strHash(info.sign, h);
        }
        // let path = FileLoadParam.batchPath + "s=" + size + (durl ? "&d=" + durl : "") + (furl ? "&f=" + furl : "") + "&h=" + h;
        if (this.isMultiVersion) {
            const currentVersion = Env.get("currentVersion");
            const resDirname = Env.get("resDirname"); // 版本目录下的二级目录
            if (currentVersion) { // 多版本,需要拼上版本号
                if (durl) {
                    durl = `:${currentVersion}(:${resDirname}(${durl}))`;
                }
                if (furl) {
                    furl = `:${currentVersion}(:${resDirname}(${furl}))`;
                }
            }
        }

        let path = FileLoadParam.batchPath + "s=" + size + (durl ? "&d=" + durl : "") + (furl ? "&f=" + furl : "") + "&h=" + h;
        let down = new HttpDownload(this.domains, path, size);
        // https://juejin.cn/post/6960988505816186894
        // 走load流程的请求头禁用缓存
        down.setHeaders({ "Cache-Control": "no-cache" });


        down.onprocess = (url: string, type: string, total: number, loaded: number,) => {
            // console.log("=== onprocess ", url, total, loaded);
            let percent = loaded / total;
            this.pool.onProcess(files, url, type, percent);
        };
        let timer;
        let totalSize = size;
        let platform = Env.get("platform_type");
        if (platform === "minigame") { // 小游戏没有进度回调,这里模拟
            let totalLoaded = 0;

            timer = setInterval(() => {
                let loaded = Math.floor(networkSpeed / 1000 * 16);
                if (totalLoaded + loaded >= totalSize) {
                    // console.log("============== xxxx ", this.total, this.loaded, loaded)
                    let needLoaded = totalSize - totalLoaded;
                    loaded = Math.floor(needLoaded * 0.1);
                }
                if (loaded / totalLoaded > 0.3) {
                    loaded = Math.floor(totalSize * 0.3); // 一次最多30进度条
                }
                totalLoaded += loaded;
                let percent = totalLoaded / totalSize;
                this.pool.onProcess(files, path, "fileDownload", percent);
            }, 16);
            this.timerMap.set(path, timer);
        }

        const finallyDo = () => {
            let timer = this.timerMap.get(path);
            if (timer) {
                clearInterval(timer);
                this.timerMap.delete(path)
            }
        }
        return down.start().then((value) => {
            networkSpeed = Math.floor(totalSize / ((Date.now() - now) / 1000));
            this.pool.onProcess(files, path, 'download', 1);
            finallyDo(); // 为了兼容低版本webview不支持finally
            return this.dataSplit(value, fileMap, downloadFileMap);
        }).catch((e) => {
            finallyDo(); // 为了兼容低版本webview不支持finally
            return Promise.reject(e)
        });
    }

    async dataSplit(buff: ArrayBuffer, fileMap: Map<string, Uint8Array>, downloadFileMap: Map<string, Uint8Array>) {
        const currentVersion = Env.get("currentVersion");
        const resDirname = Env.get("resDirname"); // 版本目录下的二级目录
        let view = new DataView(buff);
        let offset = 0;
        while (offset < buff.byteLength) {
            let len = view.getInt16(offset, true);
            offset += 2;
            let path = utf8Decode(new Uint8Array(buff, offset, len));
            if (this.isMultiVersion && currentVersion) {
                path = path.replace(`${currentVersion}/${resDirname}/`, "");
            }
            // let info = Depend.getFile(path);
            const info = this.getDependInfo(path);
            offset += len;
            let datalen = view.getInt32(offset, true);
            offset += 4;
            // 使用视图切分数据，降低内存消耗
            let data = new Uint8Array(buff, offset, datalen);
            offset += datalen;
            const originData = await decodeBuffer(data, getDecryptType(info.encryptType), getEncryptKey(path), info.compressAlgorithm);
            fileMap.set(path, originData);
            downloadFileMap.set(path, data);
            data = undefined;
        }
    }

}

