import { SystemInfo } from "../native/base/system_info";
import { Env } from "./env";

import { Logger } from "pi_sys/util/logger";
declare var module;
let logger = new Logger(module.name, "sys");

export interface Size {
	width: number;
	height: number;
}

/**
 * 屏幕 适配
 */
export const adaptive = () => {

	let showRect = Env.get("show_rect") as {
		top: number,
		bottom: number,
		left: number,
		right: number
	};

	let systemInfo: SystemInfo = Env.get("system_info");
	let oldAdaptiveResult: AdaptiveResult = Env.get("adaptiveResult");

	let deviceSize = {
		width: systemInfo.screenWidth,
		height: systemInfo.screenHeight
	};
	if (!deviceSize.width || !deviceSize.height) {
		// exe最小化
		logger.warn("deviceSize is zero");
		return;
	}

	// 进行刘海适配
	if (systemInfo && systemInfo.safeArea) {
		if (!showRect) {
			// -1表示适配到安全区域
			showRect = { top: -1, bottom: -1, right: -1, left: -1 };
		}

		if (showRect.top === -1 || showRect.top === null || showRect.top === undefined) {
			showRect.top = systemInfo.safeArea.top;
		}

		if (showRect.bottom === -1 || showRect.bottom === null || showRect.bottom === undefined) {
			showRect.bottom = deviceSize.height - systemInfo.safeArea.height - systemInfo.safeArea.top;
		}

		if (showRect.left === -1 || showRect.left === null || showRect.left === undefined) {
			showRect.left = systemInfo.safeArea.left;
		}

		if (showRect.right === -1 || showRect.right === null || showRect.right === undefined) {
			showRect.right = deviceSize.width - systemInfo.safeArea.width - systemInfo.safeArea.left;
		}
	}

	showRect = showRect || { top: 0, bottom: 0, left: 0, right: 0 };

	// 如果是横屏，目前无法知道是左横屏还是右横屏，因此临时将上下、左右留出的区域对称
	if (systemInfo.screenWidth > systemInfo.screenHeight) {
		showRect.left = showRect.right = Math.max(showRect.left, showRect.right);
		showRect.bottom = showRect.top = Math.max(showRect.bottom, showRect.top);
	}

	// 竖屏变横屏或横屏变竖屏，则交换showRect的left、top和right、bottom
	if (
		oldAdaptiveResult &&
		oldAdaptiveResult.oldScreenInfo &&
		oldAdaptiveResult.oldScreenInfo.width !== oldAdaptiveResult.oldScreenInfo.height &&
		oldAdaptiveResult.oldScreenInfo.width === systemInfo.screenHeight &&
		oldAdaptiveResult.oldScreenInfo.height === systemInfo.screenWidth
	) {
		let t = showRect.left;
		showRect.left = showRect.right = showRect.top;
		showRect.top = showRect.bottom = t;
	}

	let adaptiveResult = getAdaptiveResult(
		systemInfo.pixelRatio,
		Env.get("design_size"),
		deviceSize,
		Env.get("max_extend"),
		Env.get("gpu_performance_impl"),
		showRect,
		Env.get("layout_flex"),
	);

	// 只记录首次systemInfo
	if (!oldAdaptiveResult) {
		adaptiveResult.oldScreenInfo = {
			width: systemInfo.screenWidth,
			height: systemInfo.screenHeight,
			safeArea: {
				top: systemInfo.safeArea.top,
				right: systemInfo.safeArea.right,
				bottom: systemInfo.safeArea.bottom,
				left: systemInfo.safeArea.left,
			}
		}
	} else {
		adaptiveResult.oldScreenInfo = oldAdaptiveResult.oldScreenInfo;
	}

	Env.set("adaptiveResult", adaptiveResult);

	// 日志：关键信息，屏幕适配结果；
	logger.info("adaptive_result: %o", adaptiveResult);

	// 决定 缩放 尺寸
	let zoom = selectImgZoom(
		adaptiveResult.layoutSize,
		adaptiveResult.renderSize,
		Env.get("img_zoom_options")
	);

	Env.set("imgZoom", zoom);

	// 日志：关键信息，图片 缩放系数
	logger.info("adaptive_imgZoom: %i", zoom);
}

/***
 * 适配屏幕
 * @param designSize 设计稿尺寸
 * @param deviceSize 设备尺寸（单位为css像素）
 * @param maxExtend 最大可扩展比率
 * @param gpuPerformance 设备的gui性能参考值，默认为1，有效范围（0-1）， gpuPerformance值小于1时，返回更小的布局尺寸
 * @param top 可用于刘海屏适配，表示刘海高度
 * @param rotateFit 当设备尺寸与设计尺寸宽高比差距很大时，是否允许旋转
 * @param layoutFlex 布局伸缩，如果设置为true，界面中设置为绝对像素的div也将被缩放（不论宽高比是否与设计稿一样）
 */
export const getAdaptiveResult = (
	devicePixelRatio: number,
	designSize: Size,
	deviceSize: Size,
	maxExtend = 0,
	gpuPerformance = 1,
	showRect = { left: 0, top: 0, right: 0, bottom: 0 },
	layoutFlex = false,
): AdaptiveResult => {
	// 容错，避免gpuPerformance不合法， gpuPerformance应该在0~1之间
	gpuPerformance = gpuPerformance < 0 ? 1 : (gpuPerformance > 1 ? 1 : gpuPerformance);
	// top = top < 0 ? 0 : top;

	// let availableDeviceSize = {width: deviceSize.width, height: deviceSize.height};

	let availableDeviceSize: Size = { width: deviceSize.width - showRect.left - showRect.right, height: deviceSize.height - showRect.bottom - showRect.top };
	// if (holdRotate) {
	// 	availableDeviceSize = { width: deviceSize.width, height: deviceSize.height}
	// } else {
	// 	availableDeviceSize = { width: deviceSize.width, height: deviceSize.height}
	// }
	// 可用的设备尺寸(可用与适配刘海屏， top值代表刘海高度)
	// let availableDeviceSize = {width: deviceSize.width, height: deviceSize.height - (top * 2)};

	// 旋转似乎没什么作用，先注释掉
	// // 代表是否需要旋转显示
	// let isRotate: boolean;
	// if (rotateFit) {
	// 	isRotate = calcIsRotate(designSize, deviceSize);
	// 	if (isRotate) {
	// 		availableDeviceSize = { width: availableDeviceSize.height, height: availableDeviceSize.width };
	// 		deviceSize = {width: deviceSize.height, height: deviceSize.width};
	// 	}
	// }

	let ratio = calcShowRatio(designSize, availableDeviceSize, maxExtend);

	let safeSize = calcLayoutSize(designSize, ratio);
	// 适配尺寸
	let adaptiveSize = calcRenderSize(availableDeviceSize, ratio);


	// 布局尺寸，尽量与设计尺寸保持一致，当设计比率和屏幕比率有差异时，保持布局宽高比与屏幕宽高比一致，大小与设计尺寸靠近
	let layoutSize;

	if (layoutFlex) {
		// 如果需要自动伸缩，则layout尺寸就是设计尺寸
		layoutSize = { width: Math.round(designSize.width), height: Math.round(designSize.height) }
	} else {
		layoutSize = { width: Math.round(safeSize.width / adaptiveSize.width * deviceSize.width), height: Math.round(safeSize.height / adaptiveSize.height * deviceSize.height) };
	}


	// // 显示尺寸
	// let showSize = calcRenderSize(availableDeviceSize, ratio);
	let showSize = { width: deviceSize.width, height: deviceSize.height };
	console.log("------", safeSize, adaptiveSize, layoutSize);

	// 显示尺寸
	let renderSize = { width: showSize.width * devicePixelRatio, height: showSize.height * devicePixelRatio };

	// 根据gui性能指标，合理调整渲染尺寸
	renderSize.width = Math.round(renderSize.width * gpuPerformance);
	renderSize.height = Math.round(renderSize.height * gpuPerformance);


	let remainW = (availableDeviceSize.width - adaptiveSize.width) / 2.0, remainH = (availableDeviceSize.height - adaptiveSize.height) / 2.0;

	console.log("remainW", remainW, "remainH", remainH, "deviceSize", deviceSize, "adaptiveSize", adaptiveSize);
	return {
		isRotate: false,
		renderSize: renderSize,
		styleSize: showSize,
		layoutSize: layoutSize,
		adaptiveSize: adaptiveSize,
		styleLeft: showRect.left + remainW,
		styleTop: showRect.top + remainH,
		styleRight: deviceSize.width - (showRect.left + remainW + adaptiveSize.width),
		styleBottom: deviceSize.height - (showRect.top + remainH + adaptiveSize.height),
	}
};

/**
 * 选择图片缩放版本
 * @param layoutSize 设计尺寸
 * @param renderSize 渲染尺寸
 * @param isRotate 是否旋转
 * @param zoomOptions 缩放比率选项
 */
export const selectImgZoom = (layoutSize: Size, renderSize: Size, zoomOptions: number[] = [1]): number => {
	// 该值不会大于1
	let zoom = renderSize.width / layoutSize.width;
	let curz = 1, oldDiff = Number.MAX_SAFE_INTEGER;
	for (let z of zoomOptions) {
		if (z > 1) {
			continue;
		}

		let diff = Math.abs(z - zoom);
		if (diff > oldDiff) {
			continue;
		}

		if (diff < oldDiff) { // 当前diff更小时，取当前缩放
			curz = z;
			oldDiff = diff;
		} else if (z > curz) { // 当diff相等时，取更高的缩放
			curz = z
		}
	}
	return curz;
}

export const selectCompressTexFormat = (
	deviceExtensions: string[],
	compressTextureFormatOption: string[]
): string => {

	if (!compressTextureFormatOption) {
		return "";
	}

	// 目前 支持 astc dxt
	let selects = [

		{ simpleName: "astc", extension: ["WEBGL_compressed_texture_astc", "WEBKIT_WEBGL_compressed_texture_astc"] },

		{ simpleName: "s3tc", extension: ["WEBGL_compressed_texture_s3tc", "WEBKIT_WEBGL_compressed_texture_s3tc"] },

	];

	for (let s of selects) {
		if (compressTextureFormatOption.indexOf(s.simpleName) > -1) {
			for (let e of s.extension) {
				if (deviceExtensions.indexOf(e) > -1) {
					return s.simpleName
				}
			}
		}
	}

	return "";
}

export const calcShowRatio = (design: Size, screen: Size, maxExtend: number): number => {
	if (!design.width || !design.height || !screen.width || !screen.height) {
		throw new Error("width or height is not exist in design or screen");
	}
	let d_ratio = design.width / design.height;
	let s_ratio = screen.width / screen.height;
	let ratio: number;

	if (d_ratio === s_ratio) {
		return d_ratio;
	}

	if (s_ratio > d_ratio) {
		ratio = d_ratio * (1 + maxExtend);
		// 如果屏幕比率与设计比率相差大于最大可扩展比率， 直接将设计比率扩展最大比率,否则使用屏幕比率
		ratio = s_ratio > ratio ? ratio : s_ratio;
	} else {
		ratio = d_ratio / (1 + maxExtend);
		ratio = ratio > s_ratio ? ratio : s_ratio;
	}
	return ratio;
}

/**
 * 计算布局尺寸
 * 设计尺寸可能与屏幕的实际尺寸差异较大， 为达到最佳显示效果， 屏幕显示宽高比应与设计宽高比保持一致， 有的手机宽高比与设计宽高比可能差距很大， 该方法允许设置一个最大允许扩展比率， 扩展宽度或高度
 * @param designSize 设计尺寸
 * @param screen 屏幕尺寸
 * @param maxExtend 最大可扩展比率
 */
export const calcLayoutSize = (designSize: Size, ratio: number): Size => {
	let d_ratio = designSize.width / designSize.height;

	if (d_ratio === ratio) {
		return designSize;
	}

	if (ratio > d_ratio) {
		return { width: Math.floor(designSize.height * ratio), height: designSize.height };
	} else {
		return { width: designSize.width, height: Math.floor(designSize.width / ratio) };
	}
}

/**
 * 计算显示区域
 * 设计尺寸可能与屏幕的实际尺寸差异较大， 为达到最佳显示效果， 屏幕显示宽高比应与设计宽高比保持一直， 有的手机宽高比与设计宽高比可能差距很大， 该方法允许设置一个最大允许扩展比率， 扩展宽度或高度
 * @param design 设计尺寸
 * @param screen 屏幕尺寸
 * @param maxExtend 最大可扩展比率
 */
export const calcRenderSize = (screenSize: Size, ratio: number): Size => {
	let s_ratio = screenSize.width / screenSize.height;
	if (ratio > s_ratio) {
		// 如果屏幕宽高比大于设计宽高比， 显示区域的高度应该于屏幕高度相等， 显示区域的宽度应该是保持设计宽高比不变, 但允许对其宽度扩展（最多扩展原宽度的max_extend），
		// let width = (screen.height * d_ratio) * (1 + maxExtend);
		screenSize = { width: screenSize.width, height: Math.floor(screenSize.width / ratio) };
	} else if (ratio < s_ratio) {
		// 否则屏幕宽高比小于设计宽高比， 显示区域的宽度应该于屏幕宽度相等， 显示区域的高度应该是保持设计宽高比不变, 但允许对其高度扩展（最多扩展原宽度的max_extend），
		// let height = (screen.width / ratio) * (1 + maxExtend);
		screenSize = { width: Math.floor(screenSize.height * ratio), height: screenSize.height };;
	}
	return screenSize;
}

/**
 * 是否旋转
 * 当设计尺寸为横屏设计、 设备尺寸为竖屏， 或设计尺寸为竖屏设计、 设备尺寸为横屏时， 应该旋转， 以达到最佳显示效果
 * @param design
 * @param screen
 */
export const calcIsRotate = (design: Size, screen: Size): boolean => {
	if ((screen.width < screen.height && design.width > design.height) || (screen.width > screen.height && design.width < design.height)) {
		return true;
	} else {
		return false;
	}
}

/**
 * 适配结果
 */
export interface AdaptiveResult {
	oldScreenInfo?: {
		width: number,
		height: number,
		safeArea: {
			top: number,
			bottom: number,
			left: number,
			right: number,
		},
	}
	/**
	 * 布局尺寸（单位： css像素）
	 */
	layoutSize: Size,

	/**
	 * 适配尺寸
	 */
	adaptiveSize: Size,
	/**
	 * 显示区域尺寸（单位：css像素）
	 */
	styleSize: Size,

	/**
	 * 渲染尺寸（单位：真实像素， renderSize与showSize的关系可以看作是window.devicePixelRatio，只是根据设备性能，将该值做了调整）
	 */
	renderSize: Size,

	/**
	 * 是否需要旋转画面
	 */
	isRotate: boolean,
	/**
	 * 显示区域到屏幕顶部的距离
	 */
	styleTop: number,
	/**
	 * 显示区域到屏幕左侧的距离
	 */
	styleLeft: number,

	/**
	 * 显示区域到屏幕顶部的距离
	 */
	styleBottom: number,
	/**
	 * 显示区域到屏幕左侧的距离
	 */
	styleRight: number,
}

// ============================== 本地

let maxPixel = 2048;

const limitRenderSize = (size: Size): Size => {
	let ratio = size.width / size.height;
	// 将渲染尺寸限定在最大像素范围内
	let w = size.width, h = size.height;
	if (w > maxPixel) {
		w = maxPixel;
		h = Math.floor(w / ratio);
	}
	if (h > maxPixel) {
		h = maxPixel;
		w = Math.floor(h * ratio);
	}
	return { width: w, height: h };
}

