import { IPiStoreBase } from "pi_sys/store/store_base";


type Callback = (data: Uint8Array, err?: Error) => void;


/**
 * store存储权重任务池
 * - 以队列维护,每个队列有权重,大体可以分为快读队列,慢写队列
 * - 需要维护正在读,正在写状态,如果正在写,然后马上要复写,需要等到写完在写,否则可能文件写坏,如果还在等待写,那就合并两次写
 * - 读写文件有并发数限制，可以设计总并发数,每个队列的并发数,比如快读并发数可以设置为总并发数大小,慢写的并发数设置为1,可以避免慢写占用全部的读写资源
 * - TODO 统一管理文件源信息,允许设置文件源信息(目前为文件签名)
 */
export class StoreWeightQueuePool {
    maxCount = 4;    // 总并发数
    readQueue: FileQueue = new FileQueue(this, this.maxCount, true);            // 读队列
    // 快写队列,小游戏读图片等资源对象只能从本地加载,实现是先从网络下,然后写入本地,所以这类使用需要快速写入,优先级低于读
    quickWriteQueue: FileQueue = new FileQueue(this, this.maxCount, false);
    slowWriteQueue: FileQueue = new FileQueue(this, 1, false);           // 慢写队列
    private reading: Map<IPiStoreBase, Set<string>> = new Map;      // 正在读列表
    readCallbackMap: Map<IPiStoreBase, Map<string, Callback[]>> = new Map;           // 读回调map
    private writeing: Map<IPiStoreBase, Set<string>> = new Map;      // 正在写列表
    writeCallbackMap: Map<IPiStoreBase, Map<string, Callback[]>> = new Map;           // 写回调map
    private slowWritePaused: boolean = false;       // 是否暂停慢写入,小游戏中,仙书在战斗得时候很卡

    get curCount() {
        return this.readQueue.curCount + this.quickWriteQueue.curCount + this.slowWriteQueue.curCount;
    }


    public read(store: IPiStoreBase, key: string, readFunc: () => Promise<any>, callback: Callback) {
        const readingSet = this.reading.get(store);
        if (readingSet && readingSet.has(key)) {
            return this.pushCallback(this.readCallbackMap, store, key, callback);
        }
        if (!this.readQueue.has(store, key)) {  // 不在正在等待读列表
            this.readQueue.put(store, key, readFunc);
        }
        this.pushCallback(this.readCallbackMap, store, key, callback);
        this.run();
    }




    public write(store: IPiStoreBase, key: string, writeFunc: () => Promise<any>, callback: Callback, immediate: boolean = false) {
        const writeingSet = this.writeing.get(store);
        if (writeingSet && writeingSet.has(key)) {
            // 等待写完后再次写
            return this.pushCallback(this.writeCallbackMap, store, key, () => {
                this.write(store, key, writeFunc, callback);
            });
        }
        if (immediate) { // 需要立即写入
            if (this.quickWriteQueue.has(store, key)) {  // 正在等待写列表
                this.quickWriteQueue.replace(store, key, writeFunc);
            } else {
                this.quickWriteQueue.put(store, key, writeFunc)
            }
            if (this.slowWriteQueue.has(store, key)) {  // 慢写如果有,删除
                this.slowWriteQueue.delete(store, key);
            }
        } else {
            if (this.quickWriteQueue.has(store, key)) {
                this.quickWriteQueue.replace(store, key, writeFunc);
            } else {
                if (this.slowWriteQueue.has(store, key)) {  // 正在等待写列表
                    this.slowWriteQueue.replace(store, key, writeFunc);
                } else {
                    this.slowWriteQueue.put(store, key, writeFunc)
                }
            }

        }

        this.pushCallback(this.writeCallbackMap, store, key, callback);
        this.run();
    }

    // 暂停慢写
    public pauseSlowWrite() {
        this.slowWritePaused = true;
    }

    // 恢复慢写
    public resumeSlowWrite() {
        if (!this.slowWritePaused) return;
        this.slowWritePaused = false;
        setTimeout(() => {
            this.run();
        }, 0);
    }


    private pushCallback(fileCallbackMap: Map<IPiStoreBase, Map<string, Callback[]>>, store: IPiStoreBase, key: string, callback: Callback) {
        let callbackMap = fileCallbackMap.get(store);
        if (!callbackMap) {
            callbackMap = new Map;
            fileCallbackMap.set(store, callbackMap);
        }
        let list = callbackMap.get(key);
        if (!list) {
            list = [];
            callbackMap.set(key, list);
        }
        list.push(callback)
    }


    public onResult(store: IPiStoreBase, key: string, queue: FileQueue, value: any, err?: Error) {
        this.removeDoing(store, key, queue);
        let fileCallbackMap = queue.isRead ? this.readCallbackMap : this.writeCallbackMap;
        let callbackMap = fileCallbackMap.get(store);
        if (callbackMap) {
            let callbacks = callbackMap.get(key);
            if (callbacks) {
                callbackMap.delete(key);
                callbacks.forEach(cb => {
                    cb && cb(value, err);
                })
            }
        }
    }


    public addDoing(store: IPiStoreBase, key: string, queue: FileQueue) {
        let doing = queue.isRead ? this.reading : this.writeing;
        let set = doing.get(store);
        if (!set) {
            set = new Set;
            doing.set(store, set);
        }
        set.add(key);
        queue.delete(store, key);
    }

    public removeDoing(store: IPiStoreBase, key: string, queue: FileQueue) {
        let doing = queue.isRead ? this.reading : this.writeing;
        let set = doing.get(store);
        if (set) {
            set.delete(key);
        }
    }

    run() {
        if (this.curCount >= this.maxCount) return;
        this.readQueue.start();
        // 如果快读队列未用完并行数
        if (this.readQueue.curCount < this.readQueue.maxCount && this.curCount < this.maxCount) {
            this.quickWriteQueue.start();
        }
        // 如果快写队列未用完并行数
        if (!this.slowWritePaused && this.quickWriteQueue.curCount < this.quickWriteQueue.maxCount && this.curCount < this.maxCount) {
            this.slowWriteQueue.start();
        }
    }

}


export interface FileItem {
    store: IPiStoreBase;
    key: string;
    func: () => Promise<any>;
}

export class FileQueue {
    isRead: boolean;          // 是否是读队列
    maxCount: number;    // 总并发数
    curCount: number = 0;    // 当前并发数
    private queue: Set<FileItem> = new Set;      // 队列列表 
    private indexMap: Map<IPiStoreBase, Map<string, FileItem>> = new Map;     // 索引Map
    private pool: StoreWeightQueuePool;


    constructor(pool: StoreWeightQueuePool, maxCount: number, isRead: boolean) {
        this.pool = pool;
        this.maxCount = maxCount;
        this.isRead = isRead;
    }

    get(store: IPiStoreBase, key: string) {
        let map = this.indexMap.get(store);
        if (map) {
            return map.get(key);
        }
    }

    put(store: IPiStoreBase, key: string, func: () => Promise<any>) {
        let item: FileItem = { store, key, func };
        this.queue.add(item);
        let map = this.indexMap.get(store);
        if (!map) {
            map = new Map;
            this.indexMap.set(store, map);
        }
        map.set(key, item);
    }

    replace(store: IPiStoreBase, key: string, func: () => Promise<any>) {
        let item = this.get(store, key);
        if (item) {
            item.func = func;
        }
    }

    has(store: IPiStoreBase, key: string) {
        let map = this.indexMap.get(store);
        if (map) {
            return map.has(key);
        } else {
            return false;
        }
    }

    delete(store: IPiStoreBase, key: string) {
        let map = this.indexMap.get(store);
        if (map) {
            let item = map.get(key);
            if (item) {
                this.queue.delete(item);
                map.delete(key);
                return item;
            }
        }
    }

    size() {
        return this.queue.size;
    }

    canContinue() {
        if (this.curCount < this.maxCount && this.pool.curCount < this.pool.maxCount) {
            return true;
        } else {
            return false;
        }
    }

    // 开始下载,直到并发数限制或者没有可下载的
    start() {
        if (this.queue.size === 0) return;
        const iter = this.queue.values();
        while (this.canContinue()) {
            let { done, value } = iter.next();
            if (done) break;
            let item = value as FileItem;
            this.curCount++;
            this.pool.addDoing(item.store, item.key, this);
            // console.log("============ start write ", item.key);
            item.func().then(v => {
                this.pool.onResult(item.store, item.key, this, v);
                this.finallyDo(); // 为了兼容低版本webview不支持finally
            }).catch(e => {
                this.pool.onResult(item.store, item.key, this, null, e);
                this.finallyDo(); // 为了兼容低版本webview不支持finally
            });
        }
    }

    finallyDo() {
        this.curCount--;
        this.pool.run();
    }
}

