import { fromByteArray } from "./base64js";
import { BaseTools } from "./base_tools";
import { EncryptType } from "./constant";
import { Env } from "./env";

/**
 * 格式化字符串
 * 
 * @param format 格式化字符串，类似于 NodeJS 环境的 #[util.format](http://nodejs.cn/api/util.html#util_util_format_format_args)
 * @param args 其余 参数
 * 
 * @return string
 * 
 * @example 
 * 		// r值："a = 12.34, b = 555, c = {a: 345}"
 * 		let r = formatString("a = %d, b = %i, c = %o", 12.34, 555, {a: 345});
 * 
 * @note 格式化字符如下(为了兼容 后端和服务器版本，不支持 %o）
 *      + %s 字符串，实现：String(value);
 *      + %d 数字，实现： Number(value);
 *      + %i 十进制整数，实现： parseInt(value, 10);
 *      + %f 浮点数，实现： parseFloat(value);
 *      + %j JSON字符串，实现：JSON.stringify(value);
 *      + %% 输出一个百分号；
 */
export const formatString = (format: string, ...args: any[]): string => {
	return formatStringImpl(format, args)[0];
}

/**
 * 实现字符串格式化，返回有用到的 args的个数；
 * @note 除了用于字符串格式化，还用于 logger 模块
 */
export const formatStringImpl = (format: string, args: any[]): [string, number] => {
	if (typeof format !== typeof "") {
		return [format, 0];
	}

	let currArg = 0;
	let r = format.replace(/%[s|d|i|f|o|%]/g, v => {
		let r = v;

		try {
			// 参数不够的情况，当字符串打印；
			if (currArg >= args.length) {
				return v;
			}

			let argument = args[currArg++];
			switch (v) {
				case "%s":
					r = argument.toString();
					break;
				case "%d":
					r = (new Number(argument)).toString();
					break;
				case "%i":
					r = parseInt(argument, 10).toString();
					break;
				case "%f":
					r = parseFloat(argument).toString();
					break;
				case "%o":
					r = JSON.stringify(argument);
					break;
				case "%%":
					r = "%";
					break;
				default:
					break;
			}
		} catch (e) {
		}

		return r;
	});

	return [r, currArg];
}

/**
 * @description 函数调用
 * @example
 */
export const call = (func: Function, args: any[]) => {
	if (Array.isArray(args)) {
		switch (args.length) {
			case 0:
				return func();
			case 1:
				return func(args[0]);
			case 2:
				return func(args[0], args[1]);
			case 3:
				return func(args[0], args[1], args[2]);
			case 4:
				return func(args[0], args[1], args[2], args[3]);
			case 5:
				return func(args[0], args[1], args[2], args[3], args[4]);
			case 6:
				return func(args[0], args[1], args[2], args[3], args[4], args[5]);
			case 7:
				return func(args[0], args[1], args[2], args[3], args[4], args[5], args[6]);
			case 8:
				return func(args[0], args[1], args[2], args[3], args[4], args[5], args[6], args[7]);
			default:
				func.apply(undefined, args);
		}
	} else {
		return func(args);
	}
};

/**
 * 获取该文件属于哪个平台
 * @param path 文件路径
 * @param platforms  支持的所有平台 
 * @returns 
 */
export const getFilePlatform = (path: string, platforms: string[]) => {
	const arr = path.split(".");
	for (let i = 1; i < arr.length - 1; i++) {
		if (platforms.includes(arr[i])) {
			return arr[i];
		}
	}
	return "";
}


/**
 * 获取该文件属于哪个语言
 * @param path 文件路径
 * @param langs 支持的所有语言
 * @returns 
 */
export const getFileLang = (path: string, langs: string[]) => {
	const arr = path.split(".");
	for (let i = 1; i < arr.length - 1; i++) {
		if (langs.includes(arr[i])) {
			return arr[i];
		}
	}
	return "";
}

/**
 * 获取错误信息
 * @param e 
 * @returns 
 */
export const formatError = (e: any) => {
	if (typeof e === "object") {
		let message = e.message || e.reason;
		if (!message) {
			try {
				message = JSON.stringify(e);
			} catch (e) {
				message = "GetErrorMessage Parse Json Error";
			}
		}
		return message;
	} else {
		return e;
	}
}


/**
 * 将path的目录结构拆分成数组形式
 * @param path 
 * @returns 
 */
export const getDirnames = (path: string) => {
	const dir = BaseTools.fileDirname(path);
	const arr = dir.split("/");
	const name = BaseTools.fileBasename(path);
	arr.push(name);
	return arr;
}


/**
 * 异或加密
 * @param message 需要加密的文本 
 * @param key 加密密钥
 */
const xor = (message: Uint8Array, key: Uint8Array) => {
	const cipherMessage = new Uint8Array(message.length);
	message.forEach((t, i) => {
		const index = i % key.length;
		cipherMessage[i] = t ^ key[index];
	});
	return cipherMessage;
}

/**
 * 对路径进行异或加密
 * 1. 对路径的每一层进行异或加密
 * 2. 对加密后的每一层进行base64编码
 * 3. 对base64编码后的文本进行替换 / ==> $  (正斜杠代表目录)
 * @param path 
 * @param keyStr   异或值
 */
export const pathXor = (path: string, keyStr: string) => {
	if (!path) return path;
	const dirnames = path.split("/");
	const encoder = new TextEncoder();
	const key = encoder.encode(keyStr);
	const messages: Uint8Array[] = [];
	// 转uint8
	dirnames.forEach(dir => {
		messages.push(encoder.encode(dir));
	});
	const cipherMessages: Uint8Array[] = [];
	// 和密钥进行异或加密
	messages.forEach(message => {
		cipherMessages.push(xor(message, key));
	})

	// uint8进行base64编码,并替换斜杠(base64编码会出现斜杠)
	const cipherDirnames = cipherMessages.map(r => {
		return fromByteArray(r).replace("/", "$");
	});

	return `${cipherDirnames.join("/")}`;
}


/**
 * 获取加密密钥
 * @param fixKey 
 * @param path 
 * @returns 
 */
export const getEncryptKey = (path: string) => {
	const fixedKey = Env.get("encryptKey");
	return `${fixedKey}_${path}`;
}


/**
 * 根据加密类型获取对应的解密类型
 * @param encryptType 
 * @returns 
 */
export const getDecryptType = (encryptType: EncryptType) => {
	let decryptType = EncryptType.None;
	if (encryptType === EncryptType.CompressEncrypt) {
		decryptType = EncryptType.DecompressDecrypt;
	} else if (encryptType === EncryptType.Encrypt) {
		decryptType = EncryptType.Decrypt;
	} else if (encryptType === EncryptType.Compress) {
		decryptType = EncryptType.Decompress;
	}
	return decryptType;
}


/**
 * 获取网络地址
 * @param path 
 * @returns 
 */
export const getUrl = (path: string) => {
	const domains = Env.get("domains");
	const currentVersion = Env.get("currentVersion");
	const resDirname = Env.get("resDirname"); // 版本目录下的二级目录
	let url = domains[0];
	if (currentVersion !== undefined) {
		url = `${url}/${currentVersion}`;
	}
	if (resDirname !== undefined) {
		url = `${url}/${resDirname}`;
	}
	url = `${url}/${path}`;
	return url;
}