

/**
 * 资源加载工具
 */

import { FileLoadParam, getLoadSuffixInfo } from "pi_sys/load/fileload_base";
import type { TloadAudio, TloadAudioBuffer, TloadFile, TloadFont, TloadImage, TloadVideo } from "pi_sys/load/interface";
import { ResCacheMap } from "pi_sys/load/res_cache";
import { loadRes } from "pi_sys/load/res_load";
import { PiAudio } from "pi_sys/media/audio";
import { IPiInnerAudio, PiAudioBuffer, PiInnerAudio } from "pi_sys/media/inner_audio";
import { BlobType, createURL, revokeURL } from "pi_sys/net/bloburl";
import { IParamPiFont, PiFont } from "pi_sys/render/font";
import { createImage, PiImage } from "pi_sys/render/image";
import { createVideo, IPiVideo } from "pi_sys/render/video";
import { notifyResUsed } from "pi_sys/util/collect";
import { Depend } from "pi_sys/util/depend";
import { Env } from "pi_sys/util/env";
import * as Sign from "pi_sys/load/sign";
import { Logger } from "pi_sys/util/logger";
import { getUrl } from "pi_sys/util/util";
declare var module;
let logger = new Logger(module.name, "sys");

export const loadImage: TloadImage = (filePath: string, args: any[]): Promise<PiImage> => {
    return new Promise((resolve, reject) => {
        let info = Depend.getFile(filePath);

        if (info) {
            loadRes(info).then((data) => {
                let ab: ArrayBuffer = data;
                if (ArrayBuffer.isView(data)) {
                    ab = data.slice().buffer;
                }

                const blobUrl = createURL(ab, BlobType.jpg);

                let img = createImage();
                if ((<HTMLImageElement>img).decode) {
                    imageDecodeCall(img, blobUrl, resolve, reject);
                }
                else {
                    imageOnloadCall(img, blobUrl, resolve, reject);
                }
            });
        } else {
            reject(`loadImage - Not Found File Depend Info ${filePath}`);
        }
    })
}

export function imageDecodeCall(img: PiImage, blobUrl: string, resolve: Function, reject: Function) {
    img.src = blobUrl;
    (<HTMLImageElement>img).decode()
        .then(() => {
            resolve(img);
            revokeURL(blobUrl);
        })
        .catch((err) => {
            let image = createImage();
            imageOnloadCall(image, blobUrl, resolve, reject);
        });
}

export function imageOnloadCall(img: PiImage, blobUrl: string, resolve: Function, reject: Function) {
    img.onload = () => {
        resolve(img);
        revokeURL(blobUrl);
    }
    img.onerror = (err) => {
        reject(err);
        revokeURL(blobUrl);
    }
    img.src = blobUrl;
}

const downVideos: Set<string> = new Set; // 正在下载的video视频
export const loadVideo: TloadVideo = (filePath: string, args: any[]): Promise<IPiVideo> => {
    return new Promise((resolve, reject) => {
        let info = Depend.getFile(filePath);

        if (info) {
            if (filePath.endsWith(".m3u8")) { // m3u8格式都走网络路径,因为变bloburl后无法处理
                let url = getUrl(filePath);
                let video = createVideo(url);
                resolve(video);
            } else {
                const func = () => {
                    loadRes(info).then((data) => {
                        let ab: ArrayBuffer = data;
                        if (ArrayBuffer.isView(data)) {
                            ab = data.slice().buffer;
                        }

                        const blobUrl = createURL(ab, BlobType.mp4);

                        let video = createVideo();
                        video.oncanplaythrough = function () {
                            resolve(video);
                            // revokeURL(blobUrl);
                        }
                        video.onerror = (err) => {
                            reject(err);
                            revokeURL(blobUrl);
                        }
                        video.src = blobUrl;
                        if (video.load) {
                            video.load();     // safari浏览器需要主动调用才会下载资源
                        }
                    });
                }
                const waitVideoDownloaded = Env.get("waitVideoDownloaded");
                if (waitVideoDownloaded) { // 等待视频下载完成
                    func();
                } else {
                    let [suffix, st] = getLoadSuffixInfo(filePath);
                    notifyResUsed(filePath);
                    let cache = ResCacheMap.getCache(suffix);
                    if (cache) {
                        let r = cache.has(filePath);
                        if (r) {
                            // 有缓存
                            return func();
                        }
                    }
                    if (Sign.checkLocalSign(info, filePath)) {
                        return func();
                    } else {
                        // 边下边播
                        let url = getUrl(filePath);
                        let video = createVideo();
                        video.oncanplaythrough = function () {
                            if (!downVideos.has(filePath)) {
                                downVideos.add(filePath);
                                const duration = !isNaN((<any>video).duration) ? Math.ceil((<any>video).duration) : 10;
                                // 等到视频播放完成后下载视频资源缓存
                                setTimeout(() => {
                                    loadRes(info).then(() => {
                                        downVideos.delete(filePath);
                                    }).catch(e => {
                                        downVideos.delete(filePath);
                                    })
                                }, duration * 1000);
                            }
                            resolve(video);
                        }
                        video.onerror = (err) => {
                            reject(err);
                        }
                        video.src = url;
                        if (video.load) {
                            video.load();     // safari浏览器需要主动调用才会下载资源
                        }
                    }
                }
            }


        } else {
            reject(`loadVideo - Not Found File Depend Info ${filePath}`);
        }
    })
}


export const loadFont: TloadFont = (filePath: string, args: [IParamPiFont?]): Promise<PiFont> => {
    return new Promise((resolve, reject) => {
        let info = Depend.getFile(filePath);

        if (info) {
            loadRes(info)
                .then((data) => {
                    let ab: ArrayBuffer = data;
                    if (ArrayBuffer.isView(data)) {
                        data = data.slice().buffer;
                    }

                    const blobUrl = createURL(ab, BlobType.ttf);

                    let font = new PiFont();
                    font.onload = () => {
                        resolve(font);
                        revokeURL(blobUrl);
                    }
                    font.onerror = (err) => {
                        reject(err);
                        revokeURL(blobUrl);
                    }
                    font.arg = args[0];
                    font.src = blobUrl;
                })
                .catch(reject);
        } else {
            reject(`loadFont - Not Found File Depend Info ${filePath}`);
        }
    })
}

export const loadAudio: TloadAudio = (filePath: string, args: any[]): Promise<PiAudio> => {
    return new Promise((resolve, reject) => {
        let info = Depend.getFile(filePath);

        if (info) {
            loadRes(info)
                .then((data) => {
                    let ab: ArrayBuffer = data;
                    if (ArrayBuffer.isView(data)) {
                        ab = data.slice().buffer;
                    }

                    PiAudio.context.decodeAudioData(
                        ab,
                        (buffer) => {
                            let audio = new PiAudio(buffer);
                            resolve(audio);
                        }, (err) => {
                            logger.warn(name, {
                                error: 'SOUND_DECODE_ERROR',
                                reason: `decode fail: ${err}`
                            });
                            reject(err);
                        }
                    );
                })
                .catch(reject);
        } else {
            reject(`loadAudio - Not Found File Depend Info ${filePath}`);
        }
    })
}

export const loadAudioBuffer: TloadAudioBuffer = (filePath: string, args: any[]): Promise<PiAudioBuffer> => {
    return new Promise((resolve, reject) => {
        let info = Depend.getFile(filePath);

        if (info) {
            loadRes(info)
                .then((data) => {
                    let ab: ArrayBuffer = data;
                    if (ArrayBuffer.isView(data)) {
                        ab = data.slice().buffer;
                    }

                    (<any>PiInnerAudio).context.decodeAudioData(
                        ab,
                        (buffer) => {
                            resolve(buffer);
                        }, (err) => {
                            logger.warn(module.name, {
                                error: 'SOUND_DECODE_ERROR',
                                reason: `decode fail: ${err}`
                            });
                            reject(err);
                        }
                    );
                })
                .catch(reject);
        } else {
            reject(`loadAudio - Not Found File Depend Info ${filePath}`);
        }
    })
}

export const loadFile: TloadFile = (filePath: string, args: any[], waitWrite: boolean = false): Promise<Uint8Array> => {
    return new Promise((resolve, reject) => {
        let info = Depend.getFile(filePath);

        if (info) {
            loadRes(info, waitWrite)
                .then(r => {
                    if (!ArrayBuffer.isView(r)) {
                        r = new Uint8Array(r);
                    }
                    resolve(r as Uint8Array);
                })
                .catch(reject);
        } else {
            reject(`loadFile - Not Found File Depend Info ${filePath}`);
        }
    })
}


