/**
 * 媒体：音频
 */

import { IAudioPlayCfg, IPiAudio, StateType, StepType } from "pi_sys/media/audio";

import { Logger } from "pi_sys/util/logger";
declare var module;
let logger = new Logger(module.name, "sys");

class AudioPlayCfg implements IAudioPlayCfg {
    // 接口成员 初始化
    public readonly context: AudioContext;
    public volume: GainNode = null;
    public playTime: number = 0;
    public pauseTime: number = 0;
    public onended: Function = null;
    public delay: number = 0;
    public repeat: number = 0;
    public repeatDelay: number = 0;
    public startTime: number = 0;
    public loop: boolean = false;

    // 实现成员 初始化
    private buffer: AudioBuffer;
    private source: AudioBufferSourceNode;

    constructor(buffer: AudioBuffer, context: AudioContext) {
        this.context = context;
        this.buffer = buffer;
    }
    public getState(): StateType {
        if (!this.buffer) {
            return StateType.INIT;
        }
        if (this.playTime) {
            return StateType.PLAY;
        }
        if (this.pauseTime) {
            return StateType.PAUSE;
        }

        return StateType.READY;
    }
    /**
     * 获取声音时长
     */
    public getDuration(): number {
        if (!this.source) {
            return -1;
        }

        return this.buffer.duration;
    }
    /**
     * 暂停
     */
    public pause(): void {
        if ((!this.source) && !this.playTime) {
            return;
        }
        this.source.stop();
        this.pauseTime = this.context.currentTime - this.playTime;
        if (this.pauseTime < 0) {
            this.pauseTime = 0;
        }
        this.playTime = 0;
    }
    public disconnect() {
        if (this.volume) {
            this.volume.disconnect();
        }
    }
    /**
     * 停止
     */
    public stop(): void {
        if ((!this.source) && !this.playTime) {
            return;
        }
        this.source.stop();
        this.playTime = this.pauseTime = 0;
    }
    /**
     * 销毁
     */
    public destroy(): void {
        if (!this.source) {
            return;
        }
        if (this.volume) {
            this.volume.disconnect();
        }
        this.buffer = null;
        this.source = null;
        this.volume = null;
    }
    public play(volume: number = 1, delay: number = 0, offset: number = 0) {

        if (!this.source) {
            this.source = this.context.createBufferSource();
            this.source.buffer = this.buffer;
            this.volume = this.context.createGain();

            this.volume.gain.value = volume || 1;
            this.source.connect(this.volume);
            this.volume.connect(this.context.destination);

            delay = delay ? this.context.currentTime + delay : 0;
            offset = offset || this.pauseTime;

            this.source.start(delay, offset);
            this.source.loop = this.loop;
        } else {

            this.volume.connect(this.context.destination);
            logger.error("audio source create too")
        }


        this.playTime = this.context.currentTime + delay - offset;
        this.pauseTime = 0;

        const currfunc = this.onended;

        if (this.onended) {
            this.source.onended = (ev) => {
                this.playTime = 0;
                if (this.onended === currfunc) {
                    currfunc(this, ev);
                }
            }
        }
    }
}



/**
 * 普通音效，外部只管发起播，不管后续,可能同时播多次同一音效
 * 背景音乐，外部控制 播放,暂停,停止，同时只能存在一个播放
 */
export class PiAudio implements IPiAudio {
    public static readonly context: AudioContext = new ((window as any).AudioContext || (window as any).webkitAudioContext)();
    public static readonly contextBGM: AudioContext = new ((window as any).AudioContext || (window as any).webkitAudioContext)();
    public static enable(flag: boolean) {
        if (!flag) {
            PiAudio.context.suspend();
            PiAudio.contextBGM.suspend();
        } else {
            PiAudio.context.resume();
            PiAudio.contextBGM.resume();
        }
    }
    private buffer: AudioBuffer;
    public onended: Function = null;
    public loop: boolean = false;
    // 当前播放的声音数组
    public arr: AudioPlayCfg[] = [];
    // 音量， 0-1之间
    public volume: number = 1;
    constructor(buffer: AudioBuffer) {
        this.buffer = buffer;
    }
    public play(delay?: number, repeat?: number, repeatDelay?: number) {
        const cfg = new AudioPlayCfg(this.buffer, PiAudio.context);

        cfg.startTime = cfg.context.currentTime;
        cfg.delay = delay || 0;
        cfg.repeat = repeat || 0;
        cfg.repeatDelay = repeatDelay || 0;

        this.arr.push(cfg);

        this.playAudio(cfg);
    }
    public playAsBGM(volume: number = 1, loop: boolean = true, stepTime: number = 0, stepType?: StepType) {
        this.volume = volume;

        let cfg = this.arr[0];

        if (!cfg) {
            cfg = new AudioPlayCfg(this.buffer, PiAudio.contextBGM);
        }

        cfg.startTime = cfg.context.currentTime;
        cfg.delay = 0;
        cfg.repeat = 0;
        cfg.repeatDelay = 0;
        cfg.loop = loop;

        this.arr.push(cfg);

        cfg.onended = () => {
            if (this.onended) {
                this.onended();
            }
        };

        const d = cfg.delay + cfg.startTime - cfg.context.currentTime;
        cfg.play(this.volume, d > 0 ? d : 0);
    }

    public stopAsBGM() {
        let cfg = this.arr[0];

        if (cfg) {
            cfg.disconnect();
        }
    }

    public pauseAsBGM() {
        let cfg = this.arr[0];

        if (cfg) {
            cfg.disconnect();
        }
    }

    // 播放声音
    private playAudio(cfg: AudioPlayCfg) {

        cfg.onended = () => {
            if (cfg.repeat < 1) {
                cfg.destroy();
                this.arr.splice(this.arr.indexOf(cfg), 1);
                if (this.onended) {
                    this.onended();
                }
                return '';
            }

            cfg.repeat--;
            cfg.play(this.volume, cfg.repeatDelay);
        };

        const d = cfg.delay + cfg.startTime - cfg.context.currentTime;
        cfg.play(this.volume, d > 0 ? d : 0);
    }
    public destroy() {
        this.arr.forEach((cfg) => {
            cfg.destroy();
        });
        this.buffer = undefined;
    }
}