/**
 * 媒体：音频
 */

import { AudioEventTarget } from "pi_sys/media/event_target";
import { AudioEvent, IPiInnerAudio } from "pi_sys/media/inner_audio";
import { Env } from "pi_sys/util/env";



export class PiInnerAudio extends AudioEventTarget implements IPiInnerAudio {
    public static readonly context: AudioContext = new ((window as any).AudioContext || (window as any).webkitAudioContext)();

    private buffer: AudioBuffer;  // 解码后的 音频数据，通过构造函数传进来；外部使用是扔到ResTab去管理的；

    // 这些节点的创建成本都很低；
    private sourceNode: AudioBufferSourceNode;          // 音频源节点
    private gainNode: GainNode;                         // 音量处理节点

    private playbackRateHistory: [number, number][];    // 记录播放速率历史 [播放速率,持续时间][] 
    private startAt: number;                            // 音频以速率x播放时在AudioContext上的时间(绝对时间)
    private timerId: number;                            // 

    private _loop: boolean = false;
    private _volume: number = 1;
    private _playbackRate: number = 1;
    private _startTime: number = 0;
    private _paused: boolean = false;
    private _stoped: boolean = true;


    public get startTime() {
        return this._startTime;
    }

    public set startTime(startTime: number) {
        this._startTime = startTime;
    }

    public get loop() {
        return this._loop;
    }

    public set loop(loop: boolean) {
        this._loop = loop;
        if (this.sourceNode) {
            this.sourceNode.loop = loop;

        }
    }

    public get volume() {
        return this._volume;
    }

    public set volume(volume: number) {
        if (volume > 1) {
            volume = 1;
        } else if (volume < 0) {
            volume = 0;
        }
        if (volume === this.volume) return;
        if (this.gainNode) {
            this.gainNode.gain.value = volume;
        }
        this._volume = volume;
    }

    public get playbackRate() {
        return this._playbackRate;
    }

    public set playbackRate(rate: number) {
        if (rate > 2) {
            rate = 2;
        } else if (rate < 0.5) {
            rate = 0.5;
        }
        if (rate === this.playbackRate) return;
        if (this.sourceNode) {
            const currentTime = PiInnerAudio.context.currentTime;
            this.playbackRateHistory.push([this.playbackRate, currentTime - this.startAt]);
            this.startAt = currentTime;
            this.sourceNode.playbackRate.value = rate;
        }
        this._playbackRate = rate;
    }


    public get duration() {
        return this.buffer.duration;
    }

    public get currentTime() {
        let time = 0;
        if (this.playbackRateHistory) {
            const len = this.playbackRateHistory.length - 1;
            for (let i = len; i >= 0; i--) {
                const item = this.playbackRateHistory[i];
                time += item[0] * item[1];
            }
        }
        if (!this.paused) {
            time += this.playbackRate * (PiInnerAudio.context.currentTime - this.startAt);
        }
        return time % this.duration;
    }

    public get paused() {
        return this._paused || this._stoped;
    }


    constructor(buffer: AudioBuffer) {
        super();
        this.buffer = buffer;
        // 创建音量节点
        this.gainNode = PiInnerAudio.context.createGain();
        // 将音量节点和输出节点连接
        this.gainNode.connect(PiInnerAudio.context.destination);
    }

    private stopSourceNode() {
        if (this.sourceNode) {
            this.sourceNode.onended = undefined;  // 手动停掉的话就不触发ended
            this.sourceNode.disconnect();
            this.sourceNode.stop();
            this.sourceNode = null;         // 没必要在握住,不能重复调用start()
        }
        if (this.timerId) {
            clearTimeout(this.timerId);
            this.timerId = undefined;
        }
    }

    private playAudio(offset: number) {
        this.sourceNode = PiInnerAudio.context.createBufferSource();
        this.sourceNode.buffer = this.buffer;
        this.sourceNode.connect(this.gainNode);
        this.sourceNode.loop = this.loop;
        this.sourceNode.playbackRate.value = this.playbackRate;
        this.sourceNode.onended = () => {
            if (this.timerId) {
                clearTimeout(this.timerId);
                this.timerId = undefined;
            }
            this._paused = true;
            this._stoped = true;
            this.startAt = undefined;
            this.playbackRateHistory = undefined;
            this.dispatchEvent(AudioEvent.Ended);
        }
        this.startAt = PiInnerAudio.context.currentTime;
        this.playbackRateHistory = [];
        this.playbackRateHistory.push([1, offset]); // 偏移量默认以速率1播放
        this.sourceNode.start(0, offset);
        this._paused = false;
        this._stoped = false;
        this.dispatchEvent(AudioEvent.Play);
        /**
         * android平台目前无法触发onended事件,导致无法释放定时器
         * 且项目目前还未使用到该事件,暂时注释掉
         */
        // if (!this.timerId) {
        //     /**
        //      * 浏览器这个事件的触发频率由系统决定，但是会保证每秒触发4-66次
        //      * https://developer.mozilla.org/zh-CN/docs/Web/API/HTMLMediaElement/timeupdate_event
        //      * 这里200ms触发一次
        //      */
        //     const func = () => {
        //         this.dispatchEvent(AudioEvent.TimeUpdate);
        //         this.timerId = setTimeout(func, 200);
        //     }
        //     this.timerId = setTimeout(func, 200);
        // }
    }

    // 标准浏览器支持暂停恢复播放
    private resume() {
        if (!this._paused) return;
        this._paused = false;
        this._stoped = false;
        PiInnerAudio.context.resume();
        this.dispatchEvent(AudioEvent.Play);
    }

    // 自己处理暂停恢复播放
    private _play() {
        let offset: number = 0;
        if (this.playbackRateHistory) { // 说明是暂停之后重新播放
            offset = this.currentTime;
        } else {   // 要么是第一次播放,要么是stop之后重新从头开始播放
            offset = this.startTime;
        }
        this.playAudio(offset);
    }

    /**
    * https://developer.mozilla.org/zh-CN/docs/Web/API/AudioBufferSourceNode
    * 一个 AudioBufferSourceNode 只能被播放一次，也就是说，每次调用 start() (en-US) 之后，
    * 如果还想再播放一遍同样的声音，那么就需要再创建一个 AudioBufferSourceNode。
    * 庆幸的是，创建该节点的代价并不大，并且想要多次播放声音的话，实际上 AudioBuffer 也可以被重用。
    * 事实上，你可以用一种“阅后即焚”的方式来使用这些节点：
    * 创建节点，
    * 调用 start() 开始播放声音，
    * 然后，你甚至不需要再保留这个节点的引用了。
    * 声音播放完成之后，垃圾收集器会找个恰当的时机回收资源。
    */
    play() {
        if (!this._paused && !this._stoped) return;
        const platform_type = Env.get("platform_type");

        if (platform_type === "web" || platform_type === "app_webview") {
            if (this._paused) {  // 说明是暂停之后重新播放
                this.resume();
            } else {
                this._play();
            }
        } else {
            this._play();
        }

    }

    /**
     * 暂停；音频再播放会从暂停处开始播放
     */
    pause() {
        if (this._paused) return;
        this._paused = true;
        const platform_type = Env.get("platform_type");
        if (platform_type === "web" || platform_type === "app_webview") {
            PiInnerAudio.context.suspend();  // 标准浏览器支持暂停
        } else {
            this.playbackRateHistory.push([this.playbackRate, PiInnerAudio.context.currentTime - this.startAt]);
            this.stopSourceNode();
        }
        this.dispatchEvent(AudioEvent.Pause);
    }




    /**
     * 停止。停止后的音频再播放会从头开始播放
     */
    stop() {
        if (this._stoped) return;
        this.startAt = undefined;
        this.playbackRateHistory = undefined;
        this._stoped = true;
        this.stopSourceNode();
        this.dispatchEvent(AudioEvent.Stop);
    }

    /**
     * 跳转到指定位置
     * @param position 跳转的时间，单位 s。精确到小数点后 3 位，即支持 ms 级别精确度
     */
    seek(position: number) {
        if (this.paused) return;   // 处于暂停/停止态,此方法无效
        this.stopSourceNode();      // 停止正在播放的sourceNode
        this.dispatchEvent(AudioEvent.Seeked);
        this.playAudio(position);   // 指定位置播放音频
    }

    /**
     * 销毁当前实例
     */
    destroy() {
        this.stopSourceNode();
        if (this.gainNode) {
            this.gainNode.disconnect();
            this.gainNode = null;
        }
        this.buffer = null;
    }

    /**
     * 监听音频进入可以播放状态的事件。但不保证后面可以流畅播放
     * @param callback 音频进入可以播放状态的事件的回调函数
     */
    onCanplay(callback: () => void) {
        // TODO
    }

    /**
     * 取消监听音频进入可以播放状态的事件
     * @param callback 音频进入可以播放状态的事件的回调函数
     */
    offCanplay(callback: () => void) {
        // TODO
    }

    /**
     * 监听音频播放事件
     * @param callback 音频播放事件的回调函数
     */
    onPlay(callback: () => void) {
        this.addListener(AudioEvent.Play, callback);
    }

    /**
     * 取消监听音频播放事件
     * @param callback 音频播放事件的回调函数
     */
    offPlay(callback: () => void) {
        this.removeListener(AudioEvent.Play, callback);
    }

    /**
     * 监听音频暂停事件
     * @param callback 音频暂停事件的回调函数
     */
    onPause(callback: () => void) {
        this.addListener(AudioEvent.Pause, callback);
    }

    /**
     * 取消监听音频暂停事件
     * @param callback 音频暂停事件的回调函数
     */
    offPause(callback: () => void) {
        this.removeListener(AudioEvent.Pause, callback);
    }

    /**
     * 监听音频停止事件
     * @param callback 音频停止事件的回调函数
     */
    onStop(callback: () => void) {
        this.addListener(AudioEvent.Stop, callback);
    }

    /**
     * 取消监听音频停止事件
     * @param callback 音频停止事件的回调函数
     */
    offStop(callback: () => void) {
        this.removeListener(AudioEvent.Stop, callback);
    }

    /**
     * 监听音频自然播放至结束的事件
     * @param callback 音频自然播放至结束的事件的回调函数
     */
    onEnded(callback: () => void) {
        this.addListener(AudioEvent.Ended, callback);
    }

    /**
     * 取消监听音频自然播放至结束的事件
     * @param callback 音频自然播放至结束的事件的回调函数
     */
    offEnded(callback: () => void) {
        this.removeListener(AudioEvent.Ended, callback);
    }

    /**
     * 监听音频播放进度更新事件
     * @param callback 音频播放进度更新事件的回调函数
     */
    onTimeUpdate(callback: () => void) {
        this.addListener(AudioEvent.TimeUpdate, callback);
    }

    /**
     * 取消监听音频播放进度更新事件
     * @param callback 音频播放进度更新事件的回调函数
     */
    offTimeUpdate(callback: () => void) {
        this.removeListener(AudioEvent.TimeUpdate, callback);
    }

    /**
     * 监听音频进行跳转操作的事件
     * @param callback 音频进行跳转操作的事件的回调函数
     */
    onSeeking(callback: () => void) {
        // TODO
    }

    /**
     * 取消监听音频进行跳转操作的事件
     * @param callback 音频进行跳转操作的事件的回调函数
     */
    offSeeking(callback: () => void) {
        // TODO
    }

    /**
     * 监听音频完成跳转操作的事件
     * @param callback 音频完成跳转操作的事件的回调函数
     */
    onSeeked(callback: () => void) {
        this.addListener(AudioEvent.Seeked, callback);
    }

    /**
     * 取消监听音频完成跳转操作的事件
     * @param callback 音频完成跳转操作的事件的回调函数
     */
    offSeeked(callback: () => void) {
        this.removeListener(AudioEvent.Seeked, callback);
    }

}