/**
 * 存储， 会全局管理并行写的次数，这样提高性能和读的响应性。
 * 当有写请求时， 如果并行写次数容许，则立刻写入，否则挂起等待。 写入完毕后检查挂起的存储，继续写入。
 */

import { FileLoadParam } from "pi_sys/load/fileload_base";
import { IPiStoreBase } from "pi_sys/store/store_base";
import { CompressAlgorithm, CompressLevel, EncryptType } from "pi_sys/util/constant";

export class Store implements IPiStoreBase {
    public readonly dbName: string;
    public readonly tabName: string;
    tab: IDBDatabase;
    immediateWrite: boolean = true;  // 是否立即写入

    /**
     * @description 判断是否支持IndexedDB
     * @example
     */
    static check = () => {
        return !!iDB;
    }
    /**
     * 删除指定的数据库
     */
    static deleteStore = (dbName: string): Promise<any> => {
        return new Promise((resolve, reject) => {
            let request = iDB.deleteDatabase(dbName);
            request.onsuccess = resolve;
            request.onerror = reject;
        });
    }
    /**
     * @description 创建指定名称的存储
     * @example
     */
    static createStore = (dbName: string, tabName = "_db"): Promise<Store> => { // 返回值类型1|类型2
        return new Promise((resolve, reject) => {
            if (!iDB) {
                let s = new Store(dbName, tabName);
                return resolve(s);
            }
            try {
                let request = iDB.open(dbName, 1);
                request.onupgradeneeded = (e) => {
                    // 创建table
                    (<any>e.currentTarget).result.createObjectStore(tabName, {
                        autoIncrement: false
                    });
                };
                request.onsuccess = (e) => {
                    let s = new Store(dbName, tabName);
                    s.tab = (<any>e.currentTarget).result;
                    return resolve(s);
                };
                request.onerror = reject;
            } catch (e) {
                iDB = undefined;
                let s = new Store(dbName, tabName);
                return resolve(s);
            }
        });
    }
    constructor(dbName: string, tabName: string) {
        this.dbName = dbName;
        this.tabName = tabName;
    }

    public read(key: string): Promise<any> {
        return new Promise((resolve, reject) => {
            FileLoadParam.storeWightQueuePool.read(this, key, () => { return this.__read(key) }, (data, err) => {
                if (err) {
                    reject(err)
                } else {
                    resolve(data);
                }
            })
        });
    }

    /**
     * @description 读取数据
     * @example
     */
    public __read(key: number | string): Promise<any> {
        return new Promise((resolve, reject) => {
            let request = this.tab.transaction(this.tabName, "readonly").objectStore(this.tabName).get(key);
            request.onsuccess = (e) => resolve((<any>e.target).result);
            request.onerror = reject;
        });
    }

    public write(key: string, data: Uint8Array, encryptType?: EncryptType, encryptKey?: string, compressAlgorithm?: CompressAlgorithm, compressLevel?: CompressLevel, immediate?: boolean): Promise<any> {
        if (immediate === undefined) {
            immediate = this.immediateWrite;
        }
        return new Promise((resolve, reject) => {
            FileLoadParam.storeWightQueuePool.write(this, key, () => { return this.__write(key, data) }, (_, err) => {
                if (err) {
                    reject(err)
                } else {
                    resolve(_);
                }
            }, immediate)
        });
    }

    /**
     * @description 写入数据，如果键名存在则替换
     * @example
     */
    public __write(key: string, data: Uint8Array): Promise<any> {
        return new Promise((resolve, reject) => {
            let tx = this.tab.transaction(this.tabName, "readwrite");
            tx.objectStore(this.tabName).put(data, key);
            tx.oncomplete = () => {
                resolve(null);
            };
            tx.onerror = (err) => {
                reject(err);
            };
        });
    }
    /**
     * @description 删除数据
     * @example
     */
    public delete(key: number | string): Promise<any> {
        return new Promise((resolve, reject) => {
            let tx = this.tab.transaction(this.tabName, "readwrite");
            tx.objectStore(this.tabName).delete(key);
            tx.oncomplete = resolve;
            tx.onerror = reject;
        });
    }
    /**
     * @description 清除存储
     * @example
     */
    public clear(): Promise<any> {
        return new Promise((resolve, reject) => {
            let tx = this.tab.transaction(this.tabName, "readwrite");
            tx.objectStore(this.tabName).clear();
            tx.oncomplete = resolve;
            tx.onerror = reject;
        });
    }
    /**
     * @description 迭代, callback返回false表示停止迭代
     * @example
     */
    public iterate(callback: (result: { key: any, value: any }) => boolean, errorCallback: (err: Event) => void) {
        let cursor = this.tab.transaction(this.tabName, "readonly").objectStore(this.tabName).openCursor();
        cursor.onsuccess = () => {
            let r = cursor.result;
            if (r) {
                if (callback(r) === false) {
                    return;
                }
                r.continue();
            } else {
                callback(null);
            }
        };
        cursor.onerror = errorCallback;
    }

}

// ============================== 立即执行
let iDB = self.indexedDB; // || self.webkitIndexedDB || self.mozIndexedDB || self.msIndexedDB;
