/**
 * 系统信息
 */

import { AppName, BenchLevel, SystemInfo, TgetOAID, TgetDeviceID, TgetSystemInfo } from "pi_sys/native/base/system_info";
import { Native } from "../core/index";
import { Env } from "pi_sys/util/env";

/**
 * 获取系统信息
 */
export const getSystemInfo: TgetSystemInfo = (): Promise<SystemInfo> => {
    return new Promise((resolve, reject) => {
        Native.App.init().then(() => {
            Native.App.getInfo(
                async function (res) {
                    let left = Math.ceil(res.safearea_left / res.pixelRatio);
                    let right = Math.floor(res.safearea_right / res.pixelRatio);
                    let top = Math.ceil(res.safearea_top / res.pixelRatio);
                    let bottom = Math.floor(res.safearea_bottom / res.pixelRatio);
                    let screenWidth = Math.ceil(res.screenWidth / res.pixelRatio);
                    // IOS上这个值有可能取值错误,有时候会返回逻辑像素,有时候会返回逻辑像素*设备像素比。
                    // 取值较早的时候会出现上述情况,就感觉其本身初始化未完成的感觉
                    // let screenWidth = window.innerWidth; 
                    let screenHeight = Math.ceil(res.screenHeight / res.pixelRatio);
                    // let screenHeight = window.innerHeight;
                    let statusBarHeight = Math.ceil(res.statusBarHeight / res.pixelRatio);

                    let ua = navigator.userAgent;
                    ua = ua.toLowerCase();

                    let app: AppName;      // browser, webview, native
                    let platform; // windows, android, ios;
                    let device;   // pc or mobile;

                    if (ua.indexOf("pi_webview") >= 0) {
                        app = AppName.Webview;
                    } else if (ua.indexOf("pi_native") >= 0) {
                        app = AppName.Native;
                    } else {
                        app = AppName.Browser;
                    }

                    if (ua.indexOf("pi_android") >= 0) {
                        device = "mobile";
                        platform = "android";
                    } else if (ua.indexOf("pi_ios") >= 0) {
                        device = "mobile";
                        platform = "ios";
                    } else if (ua.indexOf("pi_windows") >= 0) {
                        device = "pc";
                        platform = "windows";
                    } else {
                        device = ua.indexOf("mobile") >= 0 ? "mobile" : "pc";

                        if (ua.indexOf('windows nt') > -1) {
                            platform = "windows";
                        } else if (ua.indexOf('iphone os') > -1) {
                            platform = "ios";
                        } else if (ua.indexOf('android') > -1) {
                            platform = "android";
                        } else if (ua.indexOf('linux') > -1) {
                            platform = "linux";
                        }
                    }

                    let benchLevel = BenchLevel.Unkown;

                    // 某些 PC模拟器 不支持WebGL
                    let isSupportWebGL = false;
                    try {
                        let canvas = document.createElement("canvas");
                        if (canvas) {
                            canvas.width = canvas.height = 2;
                            let context = canvas.getContext("webgl", { alpha: true, antialias: false, preserveDrawingBuffer: false });
                            isSupportWebGL = !!context;
                        }
                    } catch (_e) {
                        isSupportWebGL = false;
                    }
                    if (!isSupportWebGL) {
                        benchLevel = BenchLevel.CantRun;
                    }

                    if (isSupportWebGL) {
                        // 不支持 WebAssembly的，统统是CantRun
                        let isSupportWasm = false;
                        try {
                            if (window["WebAssembly"] && window["WebAssembly"].instantiate) {
                                isSupportWasm = true;
                            }
                        } catch (_e) {
                            isSupportWasm = false;
                        }
                        if (!isSupportWasm) {
                            benchLevel = BenchLevel.CantRun;
                        }
                    }

                    if (benchLevel !== BenchLevel.CantRun && platform === "ios") {
                        benchLevel = getIOSLevel(res.model);
                    }

                    let r: SystemInfo = {
                        safeArea: {
                            left,
                            right,
                            top,
                            bottom,
                            width: right - left,
                            height: bottom - top,
                        },

                        screenWidth,
                        screenHeight,
                        statusBarHeight,

                        brand: res.brand,
                        model: res.model,
                        // pixelRatio: res.pixelRatio,// IOS上底层的设备像素比是写死的,这里取window上的
                        pixelRatio: window.devicePixelRatio,
                        language: res.language,
                        cameraAuthorized: (<any>res).cameraAuthorized,
                        locationAuthorized: (<any>res).locationAuthorized,
                        microphoneAuthorized: (<any>res).microphoneAuthorized,
                        notificationAuthorized: (<any>res).notificationAuthorized,
                        bluetoothEnabled: (<any>res).bluetoothEnabled,
                        locationEnabled: (<any>res).locationEnabled,
                        wifiEnabled: (<any>res).wifiEnabled,

                        colorDepth: 24,

                        app,
                        platform,
                        device,
                        system: res.system,
                        benchLevel,
                    };
                    const rewriteSystemInfo = Env.get("rewriteSystemInfo") || window["winit"].env.rewriteSystemInfo;
                    if (rewriteSystemInfo) {
                        await rewriteSystemInfo(r, res);
                    }
                    resolve(r);
                },
                reject
            );
        })
            .catch((e) => {
                reject(e);
            });
    });
}



/**
 * 取手机的设备ID
 */
export const getDeviceID: TgetDeviceID = (): Promise<string> => {
    return new Promise((resolve, reject) => {
        Native.App.init().then(() => {
            Native.App.getDeviceID(
                function (pName) {
                    resolve(pName);
                },
                reject
            );
        })
            .catch((e) => {
                reject(e);
            });
    });
}

export const getOAID: TgetOAID = () => {
    return new Promise((resolve, reject) => {
        Native.App.init().then(() => {
            Native.App.getOAID(
                function (pName) {
                    resolve(pName);
                },
                reject
            );
        })
            .catch((e) => {
                reject(e);
            });
    });
}

// https://blog.csdn.net/blog_jihq/article/details/80590767
const getIOSLevel = (model: string): BenchLevel => {
    let level = BenchLevel.Unkown;

    // iPad，iPod，AppleTV3，AppleWatch，AirPods，XCode模拟器 都不用管；
    if (model.toLocaleLowerCase().startsWith("iphone")) {
        switch (model) {
            case "iPhone1,1": // "iPhone 1G";
            case "iPhone1,2": // "iPhone 3G"
            case "iPhone2,1": // "iPhone 3GS"
            case "iPhone3,1": // iPhone 4
            case "iPhone3,2": // iPhone 4
            case "iPhone3,3": // iPhone 4
            case "iPhone4,1": // iPhone 4S
            case "iPhone5,1": // iPhone 5
            case "iPhone5,2": // iPhone 5
            case "iPhone5,3": // iPhone 5c
            case "iPhone5,4": // iPhone 5c
            case "iPhone6,1": // iPhone 5s
            case "iPhone6,2": // iPhone 5s
            case "iPhone7,1": // iPhone 6 Plus
            case "iPhone7,2": // iPhone 6
            case "iPhone8,1": // iPhone 6s
            case "iPhone8,2": // iPhone 6s Plus
            case "iPhone8,4": // iPhone SE
            case "iPhone9,1": // iPhone 7
            case "iPhone9,2": // iPhone 7 Plus
            case "iPhone9,3": // iPhone 7
            case "iPhone9,4": // iPhone 7 Plus
                level = BenchLevel.Low;
                break;
            case "iPhone10,1": // iPhone 8
            case "iPhone10,2": // iPhone 8 Plus
            case "iPhone10,4": // iPhone 8
            case "iPhone10,5": // iPhone 8 Plus
            case "iPhone10,3": // iPhone X
            case "iPhone10,6": // iPhone X
                level = BenchLevel.Middle;
                break;
            case "iPhone11,2":  // iPhone XS
            case "iPhone11,4":	// iPhone XS Max
            case "iPhone11,6":	// iPhone XS Max
            case "iPhone11,8":	// iPhone XR
            case "iPhone12,1":	// iPhone 11
            case "iPhone12,3":	// iPhone 11 Pro
            case "iPhone12,5":	// iPhone 11 Pro Max
            case "iPhone12,8":	// iPhone SE 2
            case "iPhone13,1":	// iPhone 12 mini
            case "iPhone13,2":	// iPhone 12
            case "iPhone13,3":	// iPhone 12 Pro
            case "iPhone13,4":	// iPhone 12 Pro Max
                level = BenchLevel.High;
                break;
        }
    }

    return level;
}