
import { NativeHook } from "./nativehook";

import { Logger } from "pi_sys/util/logger";
import { nativeHookCall, nativeHookClose, nativeHookInit, nativeHookCallNoRes } from "./base";
declare var module;
let logger = new Logger(module.name, "sys");

export namespace Ad {
    const className = 'Ad';

    const paramCfg: NativeHook.FuntionParameCfg = {
        initAd: [4],
        load: [3],
        show: [3],
        destroy: [3],
        onClose: [3],
        offClose: [3]
    };

    let promise: Promise<void> | undefined;
    let uniqueID: number | undefined;
    let isInited: boolean = false;

    function error(msg: string) {
        throw new Error(msg);
    }

    /**
     * init
     */
    export function init() {

        const funcName = 'init';

        const paramList: [] = [];

        if (promise === undefined && uniqueID === undefined) {

            isInited = true;
            promise = nativeHookInit(className, funcName, paramList).then(r => {
                uniqueID = r;
            }).catch((msg) => {
                promise = undefined;
                logger.warn(msg);
            })
        }

        return promise;
    }

    /**
     * close
     */
    export function close() {

        const funcName = 'close';

        const paramList: [] = [];

        if (uniqueID !== undefined && isInited) {

            nativeHookClose(() => {
                uniqueID = undefined;
                isInited = false;
            }, (msg: string) => {
                logger.warn(msg);
            }, paramList, className, funcName, uniqueID);

            uniqueID = undefined;
        }
    }

    /**
     * 初始化广告顺序,对广告进行排序(load前调用有效).
     * @param josn - Json格式的广告优先级
     * @example - .eg: [ {"adType" : "splash", "ads" : ["csj", "gdt"] } ]
     */
    export function initAd(json: any) {

        const funcName = 'initAd';

        const paramList: [any] = [json];

        if (uniqueID === undefined) {

            logger.warn('Ad not init');
        }
        else {
            nativeHookCallNoRes(paramList, paramCfg, className, funcName, uniqueID)
        }
    }

    /**
     * 加载广告(调用一次即可).
     * @param adType - 广告类型 - "splash","interstitial","reward","fullscreen"
     */
    export function load(adType: string, succCall: (info: any) => void, failCall: (msg: string) => void) {

        const funcName = 'load';

        const paramList: [string] = [adType];

        if (uniqueID === undefined) {

            logger.warn('Ad not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 展示广告.
     */
    export function show(adType: string, succCall: (info: any) => void, failCall: (msg: string) => void) {

        const funcName = 'show';

        const paramList: [string] = [adType];

        if (uniqueID === undefined) {

            logger.warn('Ad not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 销毁广告实例.
     */
    export function destroy(adType: string) {

        const funcName = 'destroy';

        const paramList: [string] = [adType];

        if (uniqueID === undefined) {

            logger.warn('Ad not init');
        }
        else {
            nativeHookCallNoRes(paramList, paramCfg, className, funcName, uniqueID)
        }
    }

    /**
     * 监听广告关闭事件(load后,调用才有效)..
     */
    export function onClose(adType: string, succCall: (flag: boolean) => void, failCall: (msg: string) => void) {

        const funcName = 'onClose';

        const paramList: [string] = [adType];

        if (uniqueID === undefined) {

            logger.warn('Ad not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 取消广告关闭事件的监听.
     */
    export function offClose(adType: string) {

        const funcName = 'offClose';

        const paramList: [string] = [adType];

        if (uniqueID === undefined) {

            logger.warn('Ad not init');
        }
        else {
            nativeHookCallNoRes(paramList, paramCfg, className, funcName, uniqueID)
        }
    }

}
