
import { NativeHook } from "./nativehook";

import { Logger } from "pi_sys/util/logger";
import { nativeHookCall, nativeHookClose, nativeHookInit, nativeHookCallNoRes } from "./base";
declare var module;
let logger = new Logger(module.name, "sys");

export namespace App {
    const className = 'App';

    const paramCfg: NativeHook.FuntionParameCfg = {
        exitApp: [],
        getDeviceID: [],
        getOAID: [],
        getInfo: [],
        getProcessId: [],
        getPackageName: [],
        getFlavorName: [],
        getApiVersion: [],
        getVersionName: [],
        getVersionCode: [],
        getRunconfigValue: [3],
        getChannelInfo: [],
        getMethodSign: [],
        registerAppLifeListener: [],
        unRegisterAppLifeListener: [],
        registerSoftInputChangedListener: [],
        unregisterSoftInputChangedListener: [],
        hideFullScreenPopupWindow: [],
        openWebView: [3],
        closeWebView: [3],
        installApk: [3],
        openAppstore: [],
        getTemperature: [],
        getIdfv: [],
        getIpAdress: [],
        getIdfa: [],
        getChannelQuick: [],
        PrivacyAgreement: [2],
        restartApp: [],
        clearCache: [],
        showWithEntranceId: [],
        updateUserInfo: [3],
        resetUserInfo: [],
        showWithEntranceIdLogin: []
    };

    let promise: Promise<void> | undefined;
    let uniqueID: number | undefined;
    let isInited: boolean = false;

    function error(msg: string) {
        throw new Error(msg);
    }

    /**
     * init
     */
    export function init() {

        const funcName = 'init';

        const paramList: [] = [];

        if (promise === undefined && uniqueID === undefined) {

            isInited = true;
            promise = nativeHookInit(className, funcName, paramList).then(r => {
                uniqueID = r;
            }).catch((msg) => {
                promise = undefined;
                logger.warn(msg);
            })
        }

        return promise;
    }

    /**
     * close
     */
    export function close() {

        const funcName = 'close';

        const paramList: [] = [];

        if (uniqueID !== undefined && isInited) {

            nativeHookClose(() => {
                uniqueID = undefined;
                isInited = false;
            }, (msg: string) => {
                logger.warn(msg);
            }, paramList, className, funcName, uniqueID);

            uniqueID = undefined;
        }
    }

    /**
     * 退出APP.
     */
    export function exitApp() {

        const funcName = 'exitApp';

        const paramList: [] = [];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCallNoRes(paramList, paramCfg, className, funcName, uniqueID)
        }
    }

    /**
     * 获取设备ID(不完全可信,无完美方案).
     */
    export function getDeviceID(succCall: (result: string) => void, failCall: (msg: string) => void) {

        const funcName = 'getDeviceID';

        const paramList: [] = [];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 获取设备ANID
     * @param succCall 
     * @param failCall 
     */
    export function getOAID(succCall: (result: string) => void, failCall: (msg: string) => void) {

        const funcName = 'getOAID';

        const paramList: [] = [];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 获取系统信息.
     */
    export function getInfo(succCall: (info: any) => void, failCall: (msg: string) => void) {

        const funcName = 'getInfo';

        const paramList: [] = [];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 获取 进程ID，失败则使用 当前毫秒数 作为 字符串
     */
    export function getProcessId(succCall: (id: number) => void, failCall: (msg: string) => void) {

        const funcName = 'getProcessId';

        const paramList: [] = [];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCall(succCall, (msg: string) => {
                console.warn(`warning: getProcessId failed, reason = ${msg}, now get Date as processID`);

                succCall && succCall(new Date().getTime());
            }, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 获取包名.
     */
    export function getPackageName(succCall: (pName: string) => void, failCall: (msg: string) => void) {

        const funcName = 'getPackageName';

        const paramList: [] = [];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 获取底层描述的渠道名.
     */
    export function getFlavorName(succCall: (pName: string) => void, failCall: (msg: string) => void) {

        const funcName = 'getFlavorName';

        const paramList: [] = [];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 获取API版本号..
     */
    export function getApiVersion(succCall: (pName: string) => void, failCall: (msg: string) => void) {

        const funcName = 'getApiVersion';

        const paramList: [] = [];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 获取版本名.
     */
    export function getVersionName(succCall: (pName: string) => void, failCall: (msg: string) => void) {

        const funcName = 'getVersionName';

        const paramList: [] = [];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 获取版本编号.
     */
    export function getVersionCode(succCall: (pName: string) => void, failCall: (msg: string) => void) {

        const funcName = 'getVersionCode';

        const paramList: [] = [];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    export function getRunconfigValue(key: string, succCall: (pName: string) => void, failCall: (msg: string) => void) {

        const funcName = 'getRunconfigValue';

        const paramList: [string] = [key];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    export function getChannelInfo(succCall: (pName: string) => void, failCall: (msg: string) => void) {

        const funcName = 'getChannelInfo';

        const paramList: [] = [];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    export function openWebView(jsonString: string, succCall: (tag: string) => void, failCall: (msg: string) => void) {

        const funcName = 'openWebView';

        const paramList: [string] = [jsonString];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    export function closeWebView(tag: string, succCall: () => void, failCall: (msg: string) => void) {

        const funcName = 'closeWebView';

        const paramList: [string] = [tag];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    export function getTemperature(succCall: (pName: string) => void, failCall: (msg: string) => void) {

        const funcName = 'getTemperature';

        const paramList: [] = [];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    export function getIdfv(succCall: (pName: string) => void, failCall: (msg: string) => void) {

        const funcName = 'getIdfv';

        const paramList: [] = [];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    export function getIpAdress(succCall: (pName: string) => void, failCall: (msg: string) => void) {

        const funcName = 'getIpAdress';

        const paramList: [] = [];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    export function getIdfa(succCall: (pName: string) => void, failCall: (msg: string) => void) {

        const funcName = 'getIdfa';

        const paramList: [] = [];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    export function getChannelQuick(succCall: (pName: string) => void, failCall: (msg: string) => void) {

        const funcName = 'getChannelQuick';

        const paramList: [] = [];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    export function clearCache(succCall: () => void, failCall: (msg: string) => void) {

        const funcName = 'clearCache';

        const paramList: [] = [];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    export function restartApp() {

        const funcName = 'restartApp';

        const paramList: [] = [];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCallNoRes(paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    export function privacyAgreement(isAccepted: boolean, succCall: () => void, failCall: (msg: string) => void) {

        const funcName = 'PrivacyAgreement';

        const paramList: [boolean] = [isAccepted];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    export function installApk(version: string, succCall: () => void, failCall: (msg: string) => void) {

        const funcName = 'installApk';

        const paramList: [string] = [version];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 获取底层API库的方法签名. 以Json格式获取API库的类信息(包括类名,方法名,方法参数类型,方法返回值类型).
     */
    export function getMethodSign(succCall: (pName: string) => void, failCall: (msg: string) => void) {

        const funcName = 'getMethodSign';

        const paramList: [] = [];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 注册APP生命周期监听.
     */
    export function registerAppLifeListener(succCall: (msg: string) => void, failCall: (msg: string) => void) {

        const funcName = 'registerAppLifeListener';

        const paramList: [] = [];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            return nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 取消注册APP生命周期监听.
     */
    export function unRegisterAppLifeListener(cbID: number) {

        const funcName = 'unRegisterAppLifeListener';

        const paramList: [] = [];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCallNoRes(paramList, paramCfg, className, funcName, uniqueID, cbID)
        }
    }

    /**
     * 注册软键盘变化监听.
     */
    export function registerSoftInputChangedListener(succCall: (msg: number) => void, failCall: (msg: string) => void) {

        const funcName = 'registerSoftInputChangedListener';

        const paramList: [] = [];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            return nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 取消注册软键盘变化监听
     */
    export function unregisterSoftInputChangedListener(cbID: number) {

        const funcName = 'unregisterSoftInputChangedListener';

        const paramList: [] = [];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCallNoRes(paramList, paramCfg, className, funcName, uniqueID, cbID)
        }
    }

    /**
     * 隐藏NativeActivity的全屏开屏图PopupWindow.
     */
    export function hideFullScreenPopupWindow() {
        const funcName = 'hideFullScreenPopupWindow';

        const paramList: [] = [];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCallNoRes(paramList, paramCfg, className, funcName, uniqueID)
        }
    }

    export function openAppstore() {

        const funcName = 'openAppstore';

        const paramList: [] = [];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCallNoRes(paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    export function showWithEntranceId() {

        const funcName = 'showWithEntranceId';

        const paramList: [] = [];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCallNoRes(paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    export function updateUserInfo(str: string) {

        const funcName = 'updateUserInfo';

        const paramList: [Object] = [str];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCallNoRes(paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    export function resetUserInfo() {

        const funcName = 'resetUserInfo';

        const paramList: [] = [];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCallNoRes(paramList, paramCfg, className, funcName, uniqueID);
        }
    }


    export function showWithEntranceIdLogin() {

        const funcName = 'showWithEntranceIdLogin';

        const paramList: [] = [];

        if (uniqueID === undefined) {

            logger.warn('App not init');
        }
        else {
            nativeHookCallNoRes(paramList, paramCfg, className, funcName, uniqueID);
        }
    }
}
