
import { NativeHook } from "./nativehook";

import { Logger } from "pi_sys/util/logger";
import { nativeHookCall, nativeHookClose, nativeHookInit } from "./base";
declare var module;
let logger = new Logger(module.name, "sys");

export namespace Asset {
    const className = 'Asset';

    const paramCfg: NativeHook.FuntionParameCfg = {
        getStringFromAsset: [3],
        getStringFromAssetEcrypt: [3, 3, 3, 3],
        getBase64FromAsset: [3],
        getBase64FromAssetEcrypt: [3, 3, 3, 3],
        getSystemPath: [3],
        copyAssetToData: [3, 3]
    };

    let promise: Promise<void> | undefined;
    let uniqueID: number | undefined;
    let isInited: boolean = false;

    function error(msg: string) {
        throw new Error(msg);
    }

    /**
     * init
     */
    export function init() {

        const funcName = 'init';

        const paramList: [] = [];

        if (promise === undefined && uniqueID === undefined) {

            isInited = true;
            promise = nativeHookInit(className, funcName, paramList).then(r => {
                uniqueID = r;
            }).catch((msg) => {
                promise = undefined;
                logger.warn(msg);
            })
        }

        return promise;
    }

    /**
     * close
     */
    export function close() {

        const funcName = 'close';

        const paramList: [] = [];

        if (uniqueID !== undefined && isInited) {

            nativeHookClose(() => {
                uniqueID = undefined;
                isInited = false;
            }, (msg: string) => {
                logger.warn(msg);
            }, paramList, className, funcName, uniqueID);

            uniqueID = undefined;
        }
    }

    /**
     * 读取asset文件夹下文件的String内容.
     * @param fileName - 文件名
     */
    export function getStringFromAsset(fileName: string, succCall: (data: string) => void, failCall: (msg: string) => void) {

        const funcName = 'getStringFromAsset';

        const paramList: [string] = [fileName];

        if (uniqueID === undefined) {

            logger.warn('Asset not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    export function getStringFromAssetEcrypt(fileName: string, decryptType: string, encryptKey: string, compressAlgorithm: string, succCall: (data: string) => void, failCall: (msg: string) => void) {

        const funcName = 'getStringFromAssetEcrypt';

        const paramList: [string, string, string, string] = [fileName, decryptType, encryptKey, compressAlgorithm];

        if (uniqueID === undefined) {

            logger.warn('Asset not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 读取asset文件夹下文件的Base64.
     * @param fileName - 文件名
     */
    export function getBase64FromAsset(fileName: string, succCall: (data: string) => void, failCall: (msg: string) => void) {

        const funcName = 'getBase64FromAsset';

        const paramList: [string] = [fileName];

        if (uniqueID === undefined) {

            logger.warn('Asset not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    export function getBase64FromAssetEcrypt(fileName: string, decryptType: string, encryptKey: string, compressAlgorithm: string, succCall: (data: string) => void, failCall: (msg: string) => void) {

        const funcName = 'getBase64FromAssetEcrypt';

        const paramList: [string, string, string, string] = [fileName, decryptType, encryptKey, compressAlgorithm];

        if (uniqueID === undefined) {

            logger.warn('Asset not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 获取文件在底层的绝对路径
     * @param fileName 
     * @param succCall 
     * @param failCall 
     */
    export function getSystemPath(fileName: string, succCall: (data: string) => void, failCall: (msg: string) => void) {

        const funcName = 'getSystemPath';

        const paramList: [string] = [fileName];

        if (uniqueID === undefined) {

            logger.warn('Asset not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }



    export function copyAssetToData(assetsPath: string, dataPath: string, succCall: () => void, failCall: (msg: string) => void) {

        const funcName = 'copyAssetToData';

        const paramList: [string, string] = [assetsPath, dataPath];

        if (uniqueID === undefined) {

            logger.warn('Asset not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

}
