
import { NativeHook } from "./nativehook";

import { Logger } from "pi_sys/util/logger";
import { nativeHookCall, nativeHookClose, nativeHookInit, nativeHookCallNoRes } from "./base";
declare var module;
let logger = new Logger(module.name, "sys");

export namespace AudioRecorder {
    const className = 'AudioRecorder';

    const paramCfg: NativeHook.FuntionParameCfg = {
        initAudioRecorder: [1, 1, 1, 1, 3, 3],
        start: [],
        pause: [],
        stop: [],
        registerStart: [],
        registerPause: [],
        registerResume: [],
        registerStop: [],
        registerError: []
    };

    let promise: Promise<void> | undefined;
    let uniqueID: number | undefined;
    let isInited: boolean = false;

    function error(msg: string) {
        throw new Error(msg);
    }

    /**
     * init
     */
    export function init() {

        const funcName = 'init';

        const paramList: [] = [];

        if (promise === undefined && uniqueID === undefined) {

            isInited = true;
            promise = nativeHookInit(className, funcName, paramList).then(r => {
                uniqueID = r;
            }).catch((msg) => {
                promise = undefined;
                logger.warn(msg);
            })
        }

        return promise;
    }

    /**
     * close
     */
    export function close() {

        const funcName = 'close';

        const paramList: [] = [];

        if (uniqueID !== undefined && isInited) {

            nativeHookClose(() => {
                uniqueID = undefined;
                isInited = false;
            }, (msg: string) => {
                logger.warn(msg);
            }, paramList, className, funcName, uniqueID);

            uniqueID = undefined;
        }
    }

    /**
     * 开启录音.
     * @desc 采样率和编码码率需要有个对应的范围，设置不合法的采样率或编码码率会导致录音失败
     * @desc 采样率 编码码率 8000 16000 ~ 48000 11025 16000 ~ 48000 12000 24000 ~ 64000 16000 24000 ~ 96000 22050 32000 ~ 128000 24000 32000 ~ 128000 32000 48000 ~ 192000 44100 64000 ~ 320000 48000 64000 ~ 320000
     * @param duration - 录音的时长，单位 ms，最大值 600000（10 分钟）
     * @param sampleRate - 采样率: 8000, 11025, 12000, 16000, 22050, 24000, 32000, 44100, 48000
     * @param numberOfChannels - (暂无实现!默认单声道AudioFormat.CHANNEL_IN_MONO!)录音通道数: 1, 2
     * @param encodeBitRate - (暂无实现!)编码率，和采样率之间有个对应关系，见上表格
     * @param format - (暂无实现!默认mp3!)音频格式: "mp3", "aac", "wav", "PCM"
     * @param audioSource - (暂无实现!默认MediaRecorder.AudioSource.MIC!)指定录音的音频输入源，可通过 getAvailableAudioSources() 获取当前可用的音频源 auto 自动设置，默认使用手机麦克风，插上耳麦后自动切换使用耳机麦克风，所有平台适用 buildInMic 手机麦克风，仅限 iOS headsetMic 耳机麦克风，仅限 iOS mic 麦克风（没插耳麦时是手机麦克风，插耳麦时是耳机麦克风），仅限 Android camcorder 同 mic，适用于录制音视频内容，仅限 Android voice_communication 同 mic，适用于实时沟通，仅限 Android voice_recognition 同 mic，适用于语音识别，仅限 Android
     */
    export function initAudioRecorder(duration: number, sampleRate: number, numberOfChannels: number, encodeBitRate: number, format: string, audioSource: string, succCall: (info: any) => void, failCall: (msg: string) => void) {

        const funcName = 'initAudioRecorder';

        const paramList: [number, number, number, number, string, string] = [duration, sampleRate, numberOfChannels, encodeBitRate, format, audioSource];

        if (uniqueID === undefined) {
            logger.warn('AudioRecorder not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 继续录音.
     */
    export function start() {

        const funcName = 'start';

        const paramList: [] = [];

        if (uniqueID === undefined) {
            logger.warn('AudioRecorder not init');
        }
        else {
            nativeHookCallNoRes(paramList, paramCfg, className, funcName, uniqueID)
        }
    }

    /**
     * 暂停录音
     */
    export function pause() {

        const funcName = 'pause';

        const paramList: [] = [];

        if (uniqueID === undefined) {
            logger.warn('AudioRecorder not init');
        }
        else {
            nativeHookCallNoRes(paramList, paramCfg, className, funcName, uniqueID)
        }
    }

    /**
     * 停止录音
     */
    export function stop() {

        const funcName = 'stop';

        const paramList: [] = [];

        if (uniqueID === undefined) {
            logger.warn('AudioRecorder not init');
        }
        else {
            nativeHookCallNoRes(paramList, paramCfg, className, funcName, uniqueID)
        }
    }

    /**
     * 监听录音开始事件
     */
    export function registerStart(succCall: (info: any) => void, failCall: (msg: string) => void) {

        const funcName = 'registerStart';

        const paramList: [] = [];

        if (uniqueID === undefined) {
            logger.warn('AudioRecorder not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 监听录音暂停事件
     */
    export function registerPause(succCall: (info: any) => void, failCall: (msg: string) => void) {

        const funcName = 'registerPause';

        const paramList: [] = [];

        if (uniqueID === undefined) {
            logger.warn('AudioRecorder not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 监听录音继续
     */
    export function registerResume(succCall: (info: any) => void, failCall: (msg: string) => void) {

        const funcName = 'registerResume';

        const paramList: [] = [];

        if (uniqueID === undefined) {
            logger.warn('AudioRecorder not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 监听录音结束事件
     */
    export function registerStop(succCall: (path: string) => void, failCall: (msg: string) => void) {

        const funcName = 'registerStop';

        const paramList: [] = [];

        if (uniqueID === undefined) {
            logger.warn('AudioRecorder not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 监听录音错误事件
     */
    export function registerError(succCall: (path: string) => void, failCall: (msg: string) => void) {

        const funcName = 'registerError';

        const paramList: [] = [];

        if (uniqueID === undefined) {
            logger.warn('AudioRecorder not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

}
