
import { NativeHook } from "./nativehook";

import { Logger } from "pi_sys/util/logger";
import { nativeHookCall, nativeHookClose, nativeHookInit, nativeHookCallNoRes } from "./base";
declare var module;
let logger = new Logger(module.name, "sys");

export namespace Dialog {
    const className = 'Dialog';

    const paramCfg: NativeHook.FuntionParameCfg = {
        showToast: [3, 3, 3, 1, 2],
        hideToast: [],
        showModel: [3, 3, 2, 3, 3, 3, 3],
        showLoading: [3, 2],
        hideLoading: []
    };

    let promise: Promise<void> | undefined;
    let uniqueID: number | undefined;
    let isInited: boolean = false;

    function error(msg: string) {
        throw new Error(msg);
    }

    /**
     * init
     */
    export function init() {

        const funcName = 'init';

        const paramList: [] = [];

        if (promise === undefined && uniqueID === undefined) {

            isInited = true;
            promise = nativeHookInit(className, funcName, paramList).then(r => {
                uniqueID = r;
            }).catch((msg) => {
                promise = undefined;
                logger.warn(msg);
            })
        }

        return promise;
    }

    /**
     * close
     */
    export function close() {

        const funcName = 'close';

        const paramList: [] = [];

        if (uniqueID !== undefined && isInited) {

            nativeHookClose(() => {
                uniqueID = undefined;
                isInited = false;
            }, (msg: string) => {
                logger.warn(msg);
            }, paramList, className, funcName, uniqueID);

            uniqueID = undefined;
        }
    }

    /**
     * 显示消息提示框
     * @param title 提示的内容
     * @param icon 图标 "success" 显示成功图标，此时 title 文本最多显示 7 个汉字长度 //"loading" 显示加载图标，此时 title 文本最多显示 7 个汉字长度 "none" 不显示图标，此时 title 文本最多可显示两行
     * @param image - 自定义图标的本地路径，image 的优先级高于 icon 1.1.0
     * @param duration - 提示的延迟时间
     * @param mask - 是否显示透明蒙层，防止触摸穿透
     */
    export function showToast(title: string, icon: string, image: string, duration: number, mask: boolean) {

        const funcName = 'showToast';

        const paramList: [string, string, string, number, boolean] = [title, icon, image, duration, mask];

        if (uniqueID === undefined) {
            logger.warn('Dialog not init');
        }
        else {
            nativeHookCallNoRes(paramList, paramCfg, className, funcName, uniqueID)
        }
    }

    /**
     * 隐藏正在显示的对话框
     */
    export function hideToast() {

        const funcName = 'hideToast';

        const paramList: [] = [];

        if (uniqueID === undefined) {
            logger.warn('Dialog not init');
        }
        else {
            nativeHookCallNoRes(paramList, paramCfg, className, funcName, uniqueID)
        }
    }

    /**
     * 显示模态对话框
     * @param title 提示的标题
     * @param content - 提示的内容
     * @param showCancel - 是否显示取消按钮
     * @param cancelText - 取消按钮的文字，最多 4 个字符
     * @param cancelColor - 取消按钮的文字颜色，必须是 16 进制格式的颜色字符串,如:#FFFFFF
     * @param confirmText - 确认按钮的文字，最多 4 个字符
     * @param confirmColor - 确认按钮的文字颜色，必须是 16 进制格式的颜色字符串,如:#FFFFFF
     */
    export function showModel(title: string, content: string, showCancel: boolean, cancelText: string, cancelColor: string, confirmText: string, confirmColor: string, succCall: (flag: boolean) => void, failCall: (msg: string) => void) {

        const funcName = 'showModel';

        const paramList: [string, string, boolean, string, string, string, string] = [title, content, showCancel, cancelText, cancelColor, confirmText, confirmColor];

        if (uniqueID === undefined) {
            logger.warn('Dialog not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 显示 loading 提示框. 注意:离开界面时候,记得hideLoading,防止内存泄漏.
     * @param title 提示的内容
     * @param mask - 是否显示透明蒙层，防止触摸穿透
     */
    export function showLoading(title: string, mask: boolean) {

        const funcName = 'showLoading';

        const paramList: [string, boolean] = [title, mask];

        if (uniqueID === undefined) {
            logger.warn('Dialog not init');
        }
        else {
            nativeHookCallNoRes(paramList, paramCfg, className, funcName, uniqueID)
        }
    }

    /**
     * 隐藏 正在显示 的 loading 提示框
     */
    export function hideLoading() {

        const funcName = 'hideLoading';

        const paramList: [] = [];

        if (uniqueID === undefined) {
            logger.warn('Dialog not init');
        }
        else {
            nativeHookCallNoRes(paramList, paramCfg, className, funcName, uniqueID)
        }
    }

}
