
import { NativeHook } from "./nativehook";

import { Logger } from "pi_sys/util/logger";
import { nativeHookCall, nativeHookClose, nativeHookInit, nativeHookCallNoRes } from "./base";
declare var module;
let logger = new Logger(module.name, "sys");

export namespace File {
    const className = 'File';

    const paramCfg: NativeHook.FuntionParameCfg = {
        getBase64FromLocalFile: [3, 1],
        deleteMore: [3],
        writeFile_byte: [3, 3],
        writeFile_string: [3, 3],
        writeLz4File_byte: [3, 3],
        writeLz4File_string: [3, 3]
    };

    let promise: Promise<void> | undefined;
    let uniqueID: number | undefined;
    let isInited: boolean = false;

    function error(msg: string) {
        throw new Error(msg);
    }

    /**
     * init
     */
    export function init() {

        const funcName = 'init';

        const paramList: [] = [];

        if (promise === undefined && uniqueID === undefined) {

            isInited = true;
            promise = nativeHookInit(className, funcName, paramList).then(r => {
                uniqueID = r;
            }).catch((msg) => {
                promise = undefined;
                logger.warn(msg);
            })
        }

        return promise;
    }

    /**
     * close
     */
    export function close() {

        const funcName = 'close';

        const paramList: [] = [];

        if (uniqueID !== undefined && isInited) {

            nativeHookClose(() => {
                uniqueID = undefined;
                isInited = false;
            }, (msg: string) => {
                logger.warn(msg);
            }, paramList, className, funcName, uniqueID);

            uniqueID = undefined;
        }
    }

    /**
     * 根据本地路径,获取文件的base64字符串.(适配Android10分区路径).
     * @param localPath - 本地路径
     * @param length - 所取文件的最大尺寸
     */
    export function getBase64FromLocalFile(localPath: string, maxLength: number, succCall: (data: string) => void, failCall: (msg: string) => void) {

        const funcName = 'getBase64FromLocalFile';

        const paramList: [string, number] = [localPath, maxLength];

        if (uniqueID === undefined) {
            logger.warn('File not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 清除多个文件.(未适配Android10分区存储路径,需要填写沙盒中的路径)
     * @param pathArr 路径数组的字符串
     */
    export function deleteMore(pathArr: string) {

        const funcName = 'deleteMore';

        const paramList: [string] = [pathArr];

        if (uniqueID === undefined) {
            logger.warn('File not init');
        }
        else {
            nativeHookCallNoRes(paramList, paramCfg, className, funcName, uniqueID)
        }
    }

    /**
     * 写二进制文件  需要把二进制转成base64
     * @param path 
     * @param data 
     * @param succCall 
     * @param failCall 
     */
    export function writeFileByte(path: string, data: string, succCall: () => void, failCall: (msg: string) => void) {

        const funcName = 'writeFile_byte';

        const paramList: [string, string] = [path, data];

        if (uniqueID === undefined) {
            logger.warn('File not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 写文本文件
     * @param path 
     * @param data 
     * @param succCall 
     * @param failCall 
     */
    export function writeFileString(path: string, data: string, succCall: () => void, failCall: (msg: string) => void) {

        const funcName = 'writeFile_string';

        const paramList: [string, string] = [path, data];

        if (uniqueID === undefined) {
            logger.warn('File not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 写二进制文件,需要把二进制转成base64,底层会根据配置判断是否需要lz4加密
     * @param path 
     * @param data 
     * @param succCall 
     * @param failCall 
     */
    export function writeLz4FileByte(path: string, data: string, succCall: () => void, failCall: (msg: string) => void) {

        const funcName = 'writeLz4File_byte';

        const paramList: [string, string] = [path, data];

        if (uniqueID === undefined) {
            logger.warn('File not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }


    /**
     * 写文本文件  底层会根据配置判断是否需要lz4加密
     * @param path 
     * @param data 
     * @param succCall 
     * @param failCall 
     */
    export function writeLz4FileString(path: string, data: string, succCall: () => void, failCall: (msg: string) => void) {

        const funcName = 'writeLz4File_string';

        const paramList: [string, string] = [path, data];

        if (uniqueID === undefined) {
            logger.warn('File not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }
}
