
import { NativeHook } from "./nativehook";

import { Logger } from "pi_sys/util/logger";
import { nativeHookCall, nativeHookClose, nativeHookInit, nativeHookCallNoRes } from "./base";
declare var module;
let logger = new Logger(module.name, "sys");

export namespace Location {
    const className = 'Location';

    const paramCfg: NativeHook.FuntionParameCfg = {
        get: [3, 3, 2, 1],
        register: [2, 1, 1],
        unregister: []
    };

    let promise: Promise<void> | undefined;
    let uniqueID: number | undefined;
    let isInited: boolean = false;

    function error(msg: string) {
        throw new Error(msg);
    }

    /**
     * init
     */
    export function init() {

        const funcName = 'init';

        const paramList: [] = [];

        if (promise === undefined && uniqueID === undefined) {

            isInited = true;
            promise = nativeHookInit(className, funcName, paramList).then(r => {
                uniqueID = r;
            }).catch((msg) => {
                promise = undefined;
                logger.warn(msg);
            })
        }

        return promise;
    }

    /**
     * close
     */
    export function close() {

        const funcName = 'close';

        const paramList: [] = [];

        if (uniqueID !== undefined && isInited) {

            nativeHookClose(() => {
                uniqueID = undefined;
                isInited = false;
            }, (msg: string) => {
                logger.warn(msg);
            }, paramList, className, funcName, uniqueID);

            uniqueID = undefined;
        }
    }

    /**
     * GPS准确度高但耗电多，网络定位耗电少但准确度低
     * type - wgs84 返回 gps 坐标；gcj02 返回可用于 wx.openLocation 的坐标
     * altitude - 传入 true 会返回高度信息，由于获取高度需要较高精确度，会减慢接口返回速度
     * isHighAccuracy - 开启高精度定位
     * highAccuracyExpireTime - 高精度定位超时时间(ms)，指定时间内返回最高精度，该值3000ms以上高精度定位才有效果
     */
    export function get(type: string, altitude: string, isHighAccuracy: boolean, highAccuracyExpireTime: number, succCall: (info: any) => void, failCall: (msg: string) => void) {

        const funcName = 'get';

        const paramList: [string, string, boolean, number] = [type, altitude, isHighAccuracy, highAccuracyExpireTime];

        if (uniqueID === undefined) {
            logger.warn('Location not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * isHighAccuracy - 需要高精度吗
     * interval - 时间间隔，单位：ms
     * distance - 触发距离，单位：米
     */
    export function register(isHighAccuracy: boolean, interval: number, distance: number, succCall: (info: any) => void, failCall: (msg: string) => void) {

        const funcName = 'register';

        const paramList: [boolean, number, number] = [isHighAccuracy, interval, distance];

        if (uniqueID === undefined) {
            logger.warn('Location not init');
        }
        else {
            return nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 取消位置监听.
     */
    export function unregister(cbID: number) {

        const funcName = 'unregister';

        const paramList: [] = [];

        if (uniqueID === undefined) {
            logger.warn('Location not init');
        }
        else {
            nativeHookCallNoRes(paramList, paramCfg, className, funcName, uniqueID, cbID)
        }
    }

}
