
import { nativeHookCall, nativeHookInit, nativeHookCallNoRes } from "./base";
import { NativeHook } from "./nativehook";

export namespace Log {
    const className = 'Log';

    const paramCfg: NativeHook.FuntionParameCfg = {
        notifySendLog: [1, 3],
        setLocalLogLevel: [1],
        setConsoleLogLevel: [1],
        setLogLocal: [2]
    };

    let promise: Promise<void> | undefined;
    let uniqueID: number | undefined;
    let isInited: boolean = false;

    function error(msg: string) {
        throw new Error(msg);
    }

    /**
     * init
     */
    export function init() {

        const funcName = 'init';

        const paramList: [] = [];

        if (promise === undefined && uniqueID === undefined) {

            isInited = true;
            promise = nativeHookInit(className, funcName, paramList).then(r => {
                uniqueID = r;
            }).catch((msg) => {
                promise = undefined;
                console.warn(msg);
            })
        }

        return promise;
    }


    /**
     * notifySendLog
     * @param indexOfLogFile 日志文件索引(0表示最新日志,1表示上一天日志文件..)
     * @param userName 标识符(后台用来区分是那个用户的日志,建议填写用户名)
     */
    export function notifySendLog(indexOfLogFile: number, userName: string, succCall: () => void, failCall: (msg: string) => void) {

        const funcName = 'notifySendLog';

        const paramList: [number, string] = [indexOfLogFile, userName];

        if (uniqueID === undefined) {
            // console.warn('App 未初始化');
            console.warn('App not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }


    /**
     * 设置日志落地的最低等级,包含该等级.
     *
     * @param level 落地等级
     *              [0:Verbose]
     *              [1:Debug]
     *              [2:Info]
     *              [3:Warn]
     *              [4:Error]
     */
    export function setLocalLogLevel(level: number) {

        const funcName = 'setLocalLogLevel';

        const paramList: [number] = [level];

        if (uniqueID === undefined) {
            // console.warn('APP 未初始化');
            console.warn('App not init');
        }
        else {
            nativeHookCallNoRes(paramList, paramCfg, className, funcName, uniqueID)
        }
    }

    /**
     * 设置AndroidStudio控制台日志显示的最低等级,包含该等级.
     *
     * @param level 落地等级
     *              [0:Verbose]
     *              [1:Debug]
     *              [2:Info]
     *              [3:Warn]
     *              [4:Error]
     */
    export function setConsoleLogLevel(level: number) {

        const funcName = 'setConsoleLogLevel';

        const paramList: [number] = [level];

        if (uniqueID === undefined) {
            // console.warn('APP 未初始化');
            console.warn('App not init');
        }
        else {
            nativeHookCallNoRes(paramList, paramCfg, className, funcName, uniqueID)
        }
    }

    /**
     * 是否开启日志落地.
     *
     * @param isOpen 是否开启
     */
    export function setLogLocal(isOpen: boolean) {

        const funcName = 'setLogLocal';

        const paramList: [boolean] = [isOpen];

        if (uniqueID === undefined) {
            // console.warn('APP 未初始化');
            console.warn('App not init');
        }
        else {
            nativeHookCallNoRes(paramList, paramCfg, className, funcName, uniqueID)
        }
    }
}
