export namespace NativeHook {

    export interface FuntionParameCfg {
        [funcName: string]: FuntionParameType[];
    }

    export interface WaitParameCfg {
        [funcName: string]: (number | boolean | string)[];
    }

    export const FuntionParameNumber = 1;
    export const FuntionParameBoolean = 2;
    export const FuntionParameString = 3;
    export const FuntionParameAny = 4;
    export type FuntionParameType = 1 | 2 | 3 | 4;

    /**
     * 底层回调方法的接口
     */
    export interface NativeListener {
        // 成功回调
        success?: Function;
        // 失败回调
        fail?: Function;
        // 持续回调
        error?: Function;
        remove?: (cdID: number) => void;
    }
    export function removeNativeListen(cbID: number) {
        listenerMap.delete(cbID);
    }
    /**
     * 回调函数对应的id map
     */
    const listenerMap: Map<number, NativeListener> = new Map();
    let listenerCounter = 1;
    let listenerNoCounter = -1;  // 下载模块底层会关心listenerID,这里需要递减
    // 以 register 或 on 开头的函数名，就认为是 订阅-发布 模式
    // 否则都是 请求-响应式，以Promise接口类似；
    let registerPattern = /^(?:register|on)/i;


    /**
     * 调用 底层 方法
     * @param className 底层类名；每个类，底层都有两个特殊的方法：
     *     "init", 静态方法，无参，返回的是该对象的 nativeID
     *     "close", 用于将该nativeID从底层哈希表中移除，以便于可以 gc 或者 引用 回收；
     * @param methodName 底层方法名
     * @param nativeID 底层对象对应的ID，如为0，则表示 methodName是静态方法；
     * @param cb 高层回调函数，至少有两个回调：success, fail，如果cb=null就意味着高层不关心底层返回结果；
     * @param args 高层给底层的参数
     * 
     * @return cbID 回调ID
     */
    export function call(className: string, methodName: string, nativeID: number, cb: NativeListener, args: (number | boolean | string)[] = []) {
        const cbID = recordListener(cb);
        if (cb) {
            /**
             * @note 高层 on/off 监听对应 底层 regist/unRegist
             * @tip Ad onClose 是例外
             */
            registerPattern.lastIndex = 0;
            if (!registerPattern.test(methodName)) {
                cb.remove = removeNativeListen;
            }
        }
        nativeCall(className, methodName, nativeID, cbID, args);
        return cbID;
    }

    /**
     * 不调用底层方法  只是注册回调
     * @param cb 
     * @returns 
     */
    export function fakeCall(cb: NativeListener) {
        const cbID = recordListener(cb);
        return cbID;
    }

    export function checkParams(args: (number | boolean | string)[] = [], cfg: FuntionParameType[]) {
        let res = true;
        if (args) {
            if (args.length !== cfg.length) {
                res = false;
            }
            else {
                const count = args.length;

                for (let i = 0; i < count; i++) {
                    res = checkParam(args[i], cfg[i]);
                    if (res === false) {
                        break;
                    }
                }
            }
        }

        if (!res) {
            console.error(`Native Call Param Error`);
        }

        return res;
    }

    function checkParam(arg: number | boolean | string, _type: FuntionParameType) {
        let flag = true;
        switch (_type) {
            case (FuntionParameNumber): {
                if (typeof arg !== 'number') {
                    flag = false;
                }
                break;
            }
            case (FuntionParameBoolean): {
                if (typeof arg !== 'boolean') {
                    flag = false;
                }
                break;
            }
            case (FuntionParameString): {
                if (typeof arg !== 'string') {
                    flag = false;
                }
                break;
            }
            case (FuntionParameAny): {
                if (typeof arg !== 'object') {
                    flag = false;
                }
                break;
            }
        }

        return flag;
    }

    function recordListener(cb: NativeListener) {
        if (cb) {
            listenerCounter++;
            listenerMap.set(listenerCounter, cb);
            return listenerCounter;
        } else {
            return listenerNoCounter--;
            // return -1;
        }
    }

    /**
     * 
     * @param className 
     * @param methodName 
     * @param nativeID 
     * @param listenerID 
     * @param arg 
     */
    let channelNativeCallFunc = undefined
    // let isFirst = true;
    function nativeCall(className: string, methodName: string, nativeID: number, listenerID: number, args?: any[]) {
        if (channelNativeCallFunc === undefined) {

            if (!window["handle_native_message"]) {
                throw new Error("window.handle_native_message undefined，Android/iOS Can't call the js function. It is suspected that index.html calls winit.clear(). Please delete the corresponding code");
            }

            let str = navigator.userAgent;
            if (str.indexOf("PI_ANDROID") >= 0) {
                channelNativeCallFunc = (className, methodName, nativeID, listenerID, args) => {
                    (<any>window).JSBridge.postMessage(className, methodName, nativeID, listenerID, JSON.stringify(args));
                }
            } else if (str.indexOf("PI_IOS") >= 0) {
                channelNativeCallFunc = (className, methodName, nativeID, listenerID, args) => {
                    (<any>window).webkit.messageHandlers.Native.postMessage([className, methodName, nativeID, listenerID, ...args]);
                }
            } else if (str.indexOf("PI_HARMONY") >= 0) {
                channelNativeCallFunc = (className, methodName, nativeID, listenerID, args) => {

                    (<any>window).JSBridge.postMessage(className, methodName, nativeID, listenerID, JSON.stringify(args));

                }
            } else if (str.indexOf("PI_WINDOWS") >= 0) {
                if (str.indexOf("PI_WEBVIEW") >= 0) {
                    channelNativeCallFunc = (className, methodName, nativeID, listenerID, args) => {
                        let msg = JSON.stringify({
                            className,
                            methodName,
                            nativeID,
                            listenerID,
                            args
                        });
                        // console.log("========= msg" + msg);
                        (<any>window).chrome.webview.postMessage(msg);
                    }
                    // if (isFirst) {
                    //     (<any>window).chrome.webview.addEventListener('message', (event) => {
                    //         console.log("========= event.data: " + event.data);
                    //         let msg = JSON.parse(event.data);
                    //         handle_native_message(msg.cbID, msg.code, msg.arg);
                    //     });
                    // }
                } else {
                    channelNativeCallFunc = (className, methodName, nativeID, listenerID, args) => {
                        (<any>window).JSBridge.postMessage(className, methodName, nativeID, listenerID, JSON.stringify(args));
                    }
                }

            }
        }

        channelNativeCallFunc(className, methodName, nativeID, listenerID, args)
        // const str = navigator.userAgent;
        // if (str.indexOf('PI_ANDROID') >= 0) {
        //     (<any>window).JSBridge.postMessage(className, methodName, nativeID, listenerID, JSON.stringify(args));
        // } 
        // else if (str.indexOf('PI_IOS') >= 0) {
        //     (<any>window).webkit.messageHandlers.Natie.postMessage([className, methodName, nativeID, listenerID, ...args]);
        // }
    }
    /**
     * 接口调用成功/监听响应
     * @param cbID 
     * @param code 0: 成功, 否则失败
     * @param arg 
     */
    function handle_native_message(cbID: number, code: number, arg: string) {
        const cb = listenerMap.get(cbID);
        if (!cb) {
            // console.warn(`${cbID} not found!`);/
            return;
        }
        if (code === 0) {
            // console.log(`+++++++++ handle_native_message(${cbID}, ${code}, ${arg})` + "方法=", cb.success);
            cb.success && cb.success(arg);
        }
        else {
            cb.fail && cb.fail(arg);
        }
        cb.remove && cb.remove(cbID);
    }

    /**
     * 运行出现异常
     * @param cbID 
     * @param className 
     * @param methodName 
     * @param msg 
     */
    function handle_native_throw_error(cbID: number, className: string, methodName: string, msg: string) {
        const info = `handle_native_throw_error, ${className}.${methodName} failed: ${msg}`;
        const cb = listenerMap.get(cbID);

        if (cb) {
            cb.fail && cb.fail(info);
            cb.error && cb.error(info);
            cb.remove && cb.remove(cbID);
        }
        throw new Error(`handle_native_throw_error, ${className}.${methodName} failed: ${msg}`);
    }

    export function init() {
        (<any>window).handle_native_message = handle_native_message;
        (<any>window).handle_native_throw_error = handle_native_throw_error;
    }
}

NativeHook.init();
