
import { NativeHook } from "./nativehook";

import { Logger } from "pi_sys/util/logger";
import { nativeHookCall, nativeHookClose, nativeHookInit } from "./base";
declare var module;
let logger = new Logger(module.name, "sys");

export namespace Pay {
    const className = 'Pay';

    const paramCfg: NativeHook.FuntionParameCfg = {
        pay: [3, 3],
        checkLocalReceipt: [],
        payDelete: [3],
        confirmPay: [3, 3],
        getUnconfirmPays: [3, 3],
    };

    let promise: Promise<void> | undefined;
    let uniqueID: number | undefined;
    let isInited: boolean = false;

    function error(msg: string) {
        throw new Error(msg);
    }

    /**
     * init
     */
    export function init() {
        const funcName = 'init';

        const paramList: [] = [];

        if (promise === undefined && uniqueID === undefined) {

            isInited = true;
            promise = nativeHookInit(className, funcName, paramList).then(r => {
                uniqueID = r;
            }).catch((msg) => {
                promise = undefined;
                logger.warn(msg);
            })
        }

        return promise;
    }

    /**
     * close
     */
    export function close() {

        const funcName = 'close';

        const paramList: [] = [];

        if (uniqueID !== undefined && isInited) {

            nativeHookClose(() => {
                uniqueID = undefined;
                isInited = false;
            }, (msg: string) => {
                logger.warn(msg);
            }, paramList, className, funcName, uniqueID);

            uniqueID = undefined;
        }
    }

    /**
     * 发起支付.
     * @param sdkName 三方平台名称 - "zfb","wx","quicksdk"，"apple"
     * @param jsonConfig Json格式的请求参数
     * @example - zfb - { order_info: '支付请求参数字符串,从服务端传来' } 
     * @example - wx - { ... }
     * @example - wx - partnerid:     商户号
     * @example - wx - prepayid:      预支付交易会话ID
     * @example - wx - nonceStr:      随机字符串.
     * @example - wx - timestamp:     时间戳.
     * @example - wx - sign:          签名.
     * @example - quicksdk - { ... }
     * @example - quicksdk - server_id:       服务器id,数字字符串,不能含有中文字符
     * @example - quicksdk - server_name:     服务器名称
     * @example - quicksdk - game_role_name:  游戏角色名称
     * @example - quicksdk - game_role_id:    游戏角色ID
     * @example - quicksdk - game_user_level: 游戏角色等级
     * @example - quicksdk - vip_level:       当前用户vip等级，必须为数字整型字符串,请勿传"vip1"等类似字符串
     * @example - quicksdk - game_balance:    角色用户余额
     * @example - quicksdk - party_name:      帮派名称
     * @example - quicksdk - goods_id:        产品ID,用来识别购买的产品
     * @example - quicksdk - goods_name:      产品名称
     * @example - quicksdk - cp_order_id:     产品订单号（游戏方的订单号）
     * @example - quicksdk - count:           游戏币数量
     * @example - quicksdk - amount:          总金额（元）.
     * @example - quicksdk - goods_desc:      商品描述.
     * @example - quicksdk - price:           产品价格
     * @example - quicksdk - extras_params:   额外参数.透传参数，服务器发送异步通知时原样回传.
     */
    export function pay(sdkName: string, jsonConfig: string, succCall: (info: any) => void, failCall: (msg: string) => void) {

        const funcName = 'pay';

        const paramList: [string, string] = [sdkName, jsonConfig];

        if (uniqueID === undefined) {
            logger.warn('Pay not init');
        }
        else {
            nativeHookCall((params: [any]) => {
                succCall && succCall(params);
            }, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 获取未确认支付列表
     */
    export function getUnconfirmPays(sdkName: string, jsonConfig: string, succCall: (info: any) => void, failCall: (msg: string) => void) {
        const funcName = 'getUnconfirmPays';

        const paramList: [string, string] = [sdkName, jsonConfig];

        if (uniqueID === undefined) {
            logger.warn('Pay not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 确认订单
     */
    export function confirmPay(sdkName: string, jsonConfig: string, succCall: (info: any[]) => void, failCall: (msg: string) => void) {
        const funcName = 'confirmPay';

        const paramList: [string, string] = [sdkName, jsonConfig];

        if (uniqueID === undefined) {
            logger.warn('Pay not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * @deprecated
     * 检查苹果本地漏单凭证，仅Apple可用
     * TODO 会改成 上面的 get...confirm 系列
     */
    export function checkLocalReceipt(succCall: (info: any) => void, failCall: (msg: string) => void) {
        let funcName = "checkLocalReceipt"
        nativeHookCall((params: [any]) => {
            params[0]["ApplePay"] = new ApplePay(uniqueID)
            succCall && succCall(...params);
        }, failCall, error, [], paramCfg, className, funcName, uniqueID);
    }

    /**
     * @deprecated 苹果支付
     * TODO 会改成 上面的 get...confirm 系列
     */
    class ApplePay {
        uniqueID: number;

        /**
         * @deprecated 苹果支付
         * TODO 会改成 上面的 get...confirm 系列
         */
        constructor(uniqueID) {
            this.uniqueID = uniqueID;
        }

        /**
         * @deprecated 删除存在底层的支付凭证
         * TODO 会改成 上面的 get...confirm 系列
         */
        delete(pt_oid: string) {
            const funcName = 'payDelete';

            const paramList = [pt_oid];

            if (this.uniqueID === undefined) {
                logger.warn('Pay not init');
            } else {
                nativeHookCall((arg: string) => {
                }, (msg: string) => {
                }, error, paramList, paramCfg, className, funcName, this.uniqueID);
            }
        }
    }
}