
import { NativeHook } from "./nativehook";

import { Logger } from "pi_sys/util/logger";
import { nativeHookCall, nativeHookClose, nativeHookInit, nativeHookCallNoRes } from "./base";
declare var module;
let logger = new Logger(module.name, "sys");

export namespace Screen {
    const className = 'Screen';

    const paramCfg: NativeHook.FuntionParameCfg = {
        setBrightness: [1],
        getBrightness: [],
        setKeepOn: [2],
        registerCaptureScreen: [],
        unRegisterCaptureScreen: []
    };

    let promise: Promise<void> | undefined;
    let uniqueID: number | undefined;
    let isInited: boolean = false;

    function error(msg: string) {
        throw new Error(msg);
    }

    /**
     * init
     */
    export function init() {

        const funcName = 'init';

        const paramList: [] = [];

        if (promise === undefined && uniqueID === undefined) {

            isInited = true;
            promise = nativeHookInit(className, funcName, paramList).then(r => {
                uniqueID = r;
            }).catch((msg) => {
                promise = undefined;
                logger.warn(msg);
            })
        }

        return promise;
    }

    /**
     * close
     */
    export function close() {

        const funcName = 'close';

        const paramList: [] = [];

        if (uniqueID !== undefined && isInited) {

            nativeHookClose(() => {
                uniqueID = undefined;
                isInited = false;
            }, (msg: string) => {
                logger.warn(msg);
            }, paramList, className, funcName, uniqueID);

            uniqueID = undefined;
        }
    }

    /**
     * 设置屏幕亮度
     * @param level - 屏幕亮度值，范围 0 ~ 1。【0 最暗，1 最亮】
     */
    export function setBrightness(level: number, succCall: (level: number) => void, failCall: (msg: string) => void) {

        const funcName = 'setBrightness';

        const paramList: [number] = [level];

        if (uniqueID === undefined) {
            logger.warn('Screen not init');
        }
        else {
            return nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 获取屏幕亮度
     */
    export function getBrightness(succCall: (level: number) => void, failCall: (msg: string) => void) {

        const funcName = 'getBrightness';

        const paramList: [] = [];

        if (uniqueID === undefined) {
            logger.warn('Screen not init');
        }
        else {
            return nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 设置保持常亮状态
     */
    export function setKeepOn(isKeepOn: boolean, succCall: (flag: boolean) => void, failCall: (msg: string) => void) {

        const funcName = 'setKeepOn';

        const paramList: [boolean] = [isKeepOn];

        if (uniqueID === undefined) {
            logger.warn('Screen not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 监听用户主动截屏事件。 用户使用系统截屏按键截屏时触发，只能注册一个监听
     */
    export function registerCaptureScreen(cbID: number, succCall: (msg: string) => void, failCall: (msg: string) => void) {

        const funcName = 'registerCaptureScreen';

        const paramList: [] = [];

        if (uniqueID === undefined) {
            logger.warn('Screen not init');
        }
        else {
            return nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID, cbID);
        }
    }

    /**
     * 取消监听用户主动截屏事件.
     */
    export function unRegisterCaptureScreen(cbID: number) {

        const funcName = 'unRegisterCaptureScreen';

        const paramList: [] = [];

        if (uniqueID === undefined) {
            logger.warn('Screen not init');
        }
        else {
            nativeHookCallNoRes(paramList, paramCfg, className, funcName, uniqueID, cbID)
        }
    }

}
