
import { NativeHook } from "./nativehook";

import { Logger } from "pi_sys/util/logger";
import { nativeHookCall, nativeHookClose, nativeHookInit } from "./base";
declare var module;
let logger = new Logger(module.name, "sys");

export namespace SpecialCall {
    const className = 'SpecialCall';

    const paramCfg: NativeHook.FuntionParameCfg = {
        call: [3, 3, 3],
        getChannelType: [3],
        showToolbar: [3],
        hideToolbar: [3],
        verifyRealName: [3],
    };

    let promise: Promise<void> | undefined;
    let uniqueID: number | undefined;
    let isInited: boolean = false;

    function error(msg: string) {
        throw new Error(msg);
    }

    /**
     * init
     */
    export function init() {

        const funcName = 'init';

        const paramList: [] = [];

        if (promise === undefined && uniqueID === undefined) {

            isInited = true;
            promise = nativeHookInit(className, funcName, paramList).then(r => {
                uniqueID = r;
            }).catch((msg) => {
                promise = undefined;
                logger.warn(msg);
            })
        }

        return promise;
    }

    /**
     * close
     */
    export function close() {

        const funcName = 'close';

        const paramList: [] = [];

        if (uniqueID !== undefined && isInited) {

            nativeHookClose(() => {
                uniqueID = undefined;
                isInited = false;
            }, (msg: string) => {
                logger.warn(msg);
            }, paramList, className, funcName, uniqueID);

            uniqueID = undefined;
        }
    }

    /**
     * 获取渠道类型
     * @param sdkName 三方平台名称
     * @param succCall info {channelType: number} 
     */
    export function getChannelType(sdkName: string, succCall: (info: any) => void, failCall: (msg: string) => void) {

        const funcName = 'getChannelType';

        const paramList: [string] = [sdkName];

        if (uniqueID === undefined) {
            logger.warn('getChannelType not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 显示浮框
     * @param sdkName 三方平台名称
     * @param succCall void 
     */
    export function showToolbar(sdkName: string, succCall: (info: any) => void, failCall: (msg: string) => void) {

        const funcName = 'showToolbar';

        const paramList: [string] = [sdkName];

        if (uniqueID === undefined) {
            logger.warn('showToolbar not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 隐藏浮框
     * @param sdkName 三方平台名称
     * @param succCall void 
     */
    export function hideToolbar(sdkName: string, succCall: (info: any) => void, failCall: (msg: string) => void) {

        const funcName = 'hideToolbar';

        const paramList: [string] = [sdkName];

        if (uniqueID === undefined) {
            logger.warn('hideToolbar not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 实名认证
     * @param sdkName 三方平台名称
     * @param succCall void 
     */
    export function verifyRealName(sdkName: string, succCall: (info: any) => void, failCall: (msg: string) => void) {

        const funcName = 'verifyRealName';

        const paramList: [string] = [sdkName];

        if (uniqueID === undefined) {
            logger.warn('verifyRealName not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 特殊调用.
     * @param sdkName 三方平台名称
     * @param methodName 想要调用的方法名
     * @param paramsJson - 被调用的方法的参数(JsonArray格式)
     * @example - [0, 0.1, true, "Hi", "{"name":"好嗨"}"]
     */
    export function call(sdkName: string, methodName: string, paramJson: string, succCall: (info: any) => void, failCall: (msg: string) => void) {

        const funcName = 'call';

        const paramList: [string, string, string] = [sdkName, methodName, paramJson];

        if (uniqueID === undefined) {
            logger.warn('SpecialCall not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }
}
