
import { NativeHook } from "./nativehook";

import { Logger } from "pi_sys/util/logger";
import { nativeHookCall, nativeHookClose, nativeHookInit, nativeHookCallNoRes } from "./base";
declare var module;
let logger = new Logger(module.name, "sys");

export namespace VoIPChat {
    const className = 'VoIPChat';

    const paramCfg: NativeHook.FuntionParameCfg = {
        login: [1, 3, 3],
        logout: [],
        joinVideoRoom: [1, 3, 3, 2, 2, 1],
        join: [3, 2, 2],
        exit: [],
        updateMuteConfig: [2, 2],
        registerMembersChanged: [],
        unregisterMembersChanged: [],
        registerSpeakersChanged: [],
        unregisterSpeakersChanged: []
    };

    let promise: Promise<void> | undefined;
    let uniqueID: number | undefined;
    let isInited: boolean = false;

    function error(msg: string) {
        throw new Error(msg);
    }

    /**
     * init
     */
    export function init() {

        const funcName = 'init';

        const paramList: [] = [];

        if (promise === undefined && uniqueID === undefined) {

            isInited = true;
            promise = nativeHookInit(className, funcName, paramList).then(r => {
                uniqueID = r;
            }).catch((msg) => {
                promise = undefined;
                logger.warn(msg);
            })
        }

        return promise;
    }

    /**
     * close
     */
    export function close() {

        const funcName = 'close';

        const paramList: [] = [];

        if (uniqueID !== undefined && isInited) {

            nativeHookClose(() => {
                uniqueID = undefined;
                isInited = false;
            }, (msg: string) => {
                logger.warn(msg);
            }, paramList, className, funcName, uniqueID);

            uniqueID = undefined;
        }
    }

    /**
     * 登录实时音视频系统.
     * @parameter sdkAppId - 腾讯实时音视频三方SDK的AppID
     * @parameter userId - 腾讯实时音视频三方SDK中设置的userId
     * @parameter userSig - 用上述userId以腾讯签名规则得到的userSig
     */
    export function login(sdkAppId: number, userId: string, userSig: string, succCall: (info: any) => void, failCall: (msg: string) => void) {

        const funcName = 'login';

        const paramList: [number, string, string] = [sdkAppId, userId, userSig];

        if (uniqueID === undefined) {
            logger.warn('VoIPChat not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 退出登录.
     */
    export function logout(succCall: (info: any) => void, failCall: (msg: string) => void) {

        const funcName = 'logout';

        const paramList: [] = [];

        if (uniqueID === undefined) {
            logger.warn('VoIPChat not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 进入实时视频通话.
     * @parameter roomId - 房间号
     * @parameter userName - 用户名
     * @parameter userAvatar - 头像地址
     * @parameter openCamera - 是否开启摄像头
     * @parameter openAudio - 是否开启麦克风
     * @parameter audioQuality - 音质(1:语音 2:默认 3:音乐)
     */
    export function joinVideoRoom(roomId: number, userName: string, userAvatar: string, openCamera: boolean, openAudio: boolean, audioQuality: number, succCall: () => void, failCall: (msg: string) => void) {

        const funcName = 'joinVideoRoom';

        const paramList: [number, string, string, boolean, boolean, number] = [roomId, userName, userAvatar, openCamera, openAudio, audioQuality];

        if (uniqueID === undefined) {
            logger.warn('VoIPChat not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 进入实时语音通话.
     * @parameter roomId - 房间号
     * @parameter muteMicrophone - 是否静音麦克风
     * @parameter muteEarphone - 是否静音
     */
    export function join(roomId: string, muteMicrophone: boolean, muteEarphone: boolean, succCall: () => void, failCall: (msg: string) => void) {

        const funcName = 'join';

        const paramList: [string, boolean, boolean] = [roomId, muteMicrophone, muteEarphone];

        if (uniqueID === undefined) {
            logger.warn('VoIPChat not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 退出实时语音通话.
     */
    export function exit(succCall: () => void, failCall: (msg: string) => void) {

        const funcName = 'exit';

        const paramList: [] = [];

        if (uniqueID === undefined) {
            logger.warn('VoIPChat not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 更新实时语音通话的声音和麦克设置.
     * @parameter muteMicrophone - 是否静音麦克风
     * @parameter muteEarphone - 是否静音
     */
    export function updateMuteConfig(muteMicrophone: boolean, muteEarphone: boolean, succCall: () => void, failCall: (msg: string) => void) {

        const funcName = 'updateMuteConfig';

        const paramList: [boolean, boolean] = [muteMicrophone, muteEarphone];

        if (uniqueID === undefined) {
            logger.warn('VoIPChat not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 监听实时语音通话成员在线状态变化事件。有成员加入/退出通话时触发回调.
     */
    export function registerMembersChanged(succCall: (info: any) => void, failCall: (msg: string) => void) {

        const funcName = 'registerMembersChanged';

        const paramList: [] = [];

        if (uniqueID === undefined) {
            logger.warn('VoIPChat not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 取消 监听实时语音通话成员通话状态变化事件.
     */
    export function unregisterMembersChanged() {

        const funcName = 'unregisterMembersChanged';

        const paramList: [] = [];

        if (uniqueID === undefined) {
            logger.warn('VoIPChat not init');
        }
        else {
            nativeHookCallNoRes(paramList, paramCfg, className,funcName, uniqueID)
        }
    }

    /**
     * 监听实时语音通话成员通话状态变化事件。有成员开始/停止说话时触发回调.
     */
    export function registerSpeakersChanged(succCall: (info: any) => void, failCall: (msg: string) => void) {

        const funcName = 'registerSpeakersChanged';

        const paramList: [] = [];

        if (uniqueID === undefined) {
            logger.warn('VoIPChat not init');
        }
        else {
            nativeHookCall(succCall, failCall, error, paramList, paramCfg, className, funcName, uniqueID);
        }
    }

    /**
     * 取消 监听实时语音通话成员通话状态变化事件。有成员开始/停止说话时触发回调.
     */
    export function unregisterSpeakersChanged() {

        const funcName = 'unregisterSpeakersChanged';

        const paramList: [] = [];

        if (uniqueID === undefined) {
            logger.warn('VoIPChat not init');
        }
        else {
            nativeHookCallNoRes(paramList, paramCfg, className,funcName, uniqueID)
        }
    }

}
