import { MotionData, ToffChange, TonChange, TregisterRotation, TsetAutoRotation, TsetManualScreenOrientation, Tstart, Tstop, TunRegisterRotation } from "pi_sys/native/device/motion";
import { Native } from "../core/index";

import { Logger } from "pi_sys/util/logger";
declare var module;
let logger = new Logger(module.name, "sys");

let funcs: Set<Function>;      // 监听回调
let cbID: number;              // NativeHook监听回调ID
let oldInterval: string;       // 频率  

export const start: Tstart = (interval: string = 'normal') => {
    if (oldInterval === interval) return Promise.resolve();
    oldInterval = interval;
    return Native.DeviceMotion.init().then(() => {
        const id = Native.DeviceMotion.register(
            interval,
            cbID,
            (info: MotionData) => {
                if (!funcs || !funcs.size) {
                    logger.warn(`Motion started,But no callback Function ,Please call "stop" Function to improve performance!!!`);
                } else {
                    funcs.forEach(func => func && func(info))
                }
            },
            undefined
        );
        cbID = id;
    });

}


export const stop: Tstop = () => {
    if (!oldInterval) return Promise.resolve();
    oldInterval = undefined;
    return Native.DeviceMotion.init().then(() => {
        Native.DeviceMotion.unregister(cbID);
        cbID = undefined;
    });
}

export const onChange: TonChange = (callback: (info: MotionData) => void) => {
    if (!funcs) {
        funcs = new Set;
    }
    funcs.add(callback);
}

export const offChange: ToffChange = (callback?: (info: MotionData) => void) => {
    if (funcs) {
        if (callback) {
            funcs.delete(callback);
        } else {
            funcs.clear();
        }
        if (!funcs.size) funcs = undefined;
    }
}





let rotationFuncs: Set<Function>;      // 监听回调
let rotationCbID: number;              // NativeHook监听回调ID

/**
 * 注册横竖屏切换监听
 */
export const registerRotation: TregisterRotation = (callback: (orientation: "portrait" | "landscape") => void) => {
    if (!rotationFuncs) {
        rotationFuncs = new Set;
    }
    if (rotationFuncs.has(callback)) return;
    rotationFuncs.add(callback);
    if (rotationFuncs.size === 1) {
        Native.DeviceMotion.init().then(() => {
            const id = Native.DeviceMotion.registerRotation(
                null,
                (orientation: string) => {
                    funcs && funcs.forEach(func => func && func(orientation))
                },
                undefined
            );
            rotationCbID = id;
        });
    }

}

/**
 * 取消注册横竖屏切换监听
 */
export const unRegisterRotation: TunRegisterRotation = (callback?: (orientation: "portrait" | "landscape") => void) => {
    if (!callback) {
        rotationFuncs.clear();
    } else {
        rotationFuncs.delete(callback);
    }
    if (rotationFuncs.size === 0) {
        Native.DeviceMotion.init().then(() => {
            Native.DeviceMotion.unRegisterRotation(rotationCbID);
            rotationCbID = undefined;
        });
    }

}

/**
 * 设置自动横竖屏切换
 */
export const setAutoRotation: TsetAutoRotation = (auto: boolean) => {
    return new Promise((resolve, reject) => {
        Native.DeviceMotion.init().then(() => {
            Native.DeviceMotion.setAutoRotation(auto, resolve as any, reject);
        });
    })

}

/**
 * 手动切换横竖屏
 */
export const setManualScreenOrientation: TsetManualScreenOrientation = () => {
    return new Promise((resolve, reject) => {
        Native.DeviceMotion.init().then(() => {
            Native.DeviceMotion.setManualScreenOrientation((orientation: string) => {
                resolve(orientation as "portrait" | "landscape")
            }, reject);
        });
    })
}