import { Native } from "../core/index"
import { Audio as NativeAudio } from "../core/audio"
import { ErrCode, IInnerAudioContext, TcreateInnerAudioContext, TgetAvailableAudioSources, TsetInnerAudioOption } from "pi_sys/native/media/inner_audio";

/**
 * 媒体：音频
 */


/**
 * 设置 InnerAudioContext 的播放选项
 * @param mixWithOther	     是否与其他音频混播，设置为 true 之后，不会终止其他应用或微信内的音乐
 * @param obeyMuteSwitch	（仅在 iOS 生效）是否遵循静音开关，设置为 false 之后，即使是在静音模式下，也能播放声音
 * @note 设置之后对当前小程序全局生效
 */
export const setInnerAudioOption: TsetInnerAudioOption = (mixWithOther = true, obeyMuteSwitch = true) => {
    return new Promise((resolve, reject) => {

    });
}

/**
 * 获取当前支持的音频输入源
 * @return Promise<支持的音频输入源列表>
 * @note 可在 RecorderManager.start() 接口中使用。
 *      返回值定义参考 https://developer.android.com/reference/kotlin/android/media/MediaRecorder.AudioSource
 */
export const getAvailableAudioSources: TgetAvailableAudioSources = () => {
    return new Promise((resolve, reject) => {

    });
}

/**
 * 创建内部 audio 上下文 InnerAudioContext 对象
 */
export const createInnerAudioContext: TcreateInnerAudioContext = () => {
    return Promise.resolve(new InnerAudioContext);
}

/**
 * InnerAudioContext 实例
 * @note 可通过 createInnerAudioContext 接口获取实例
 * @note 支持格式
 *           格式	iOS	Android
 *           flac	x	√
 *           m4a	√	√
 *           ogg	x	√
 *           ape	x	√
 *           amr	x	√
 *           wma	x	√
 *           wav	√	√
 *           mp3	√	√
 *           mp4	x	√
 *           aac	√	√
 *           aiff	√	x
 *           caf	√	x
 */
export class InnerAudioContext implements IInnerAudioContext {
    private nativeAudio: NativeAudio;
    private srcPromise: Promise<void>;
    constructor() {
        this.nativeAudio = new Native.Audio();
        this.nativeAudio._onError = (info: any) => {
            this.nativeAudio.ErrorListenList && this.nativeAudio.ErrorListenList.forEach((f) => {
                f && f(info.errMsg, (info).errCode - 0);
            });
        }
    }
    private _src: string;
    public get src() {
        return this._src;
    }
    public set src(value: string) {
        this._src = value;
        this.srcPromise = new Promise((resolve, reject) => {
            this.nativeAudio.initAudio(
                value,
                (info) => {
                    resolve();
                },
                (msg) => {
                    reject(msg)
                }
            )
        })
    }
    /**
     * 销毁当前实例
     */
    destroy(): Promise<void> {
        return this.nativeAudio.promise.then(() => {
            this.nativeAudio.destroy();
            this.nativeAudio.dispose();
            return undefined;
        });
    }

    /**
     * 暂停
     * @note 暂停后的音频再播放会从暂停处开始播放
     */
    pause(): Promise<void> {
        if (this.srcPromise) {
            return this.nativeAudio.promise.then(() => {
                return this.srcPromise.then(() => {
                    this.nativeAudio.pause(undefined, undefined);
                    return undefined;
                });
            });
        } else {
            return Promise.reject('not load')
        }
    }

    /**
     * 播放
     */
    play(): Promise<void> {
        if (this.srcPromise) {
            return this.nativeAudio.promise.then(() => {
                return this.srcPromise.then(() => {
                    this.nativeAudio.play(undefined, undefined);
                    return undefined;
                });
            })
        } else {
            return Promise.reject('not load')
        }
    }

    /**
     * 跳转到指定位置
     * @param position  跳转的时间，单位 s。精确到小数点后 3 位，即支持 ms 级别精确度
     */
    seek(position: number): Promise<void> {
        if (this.srcPromise) {
            return this.nativeAudio.promise.then(() => {
                return this.srcPromise.then(() => {
                    this.nativeAudio.seek(position, undefined, undefined);
                    return undefined;
                });
            });
        } else {
            return Promise.reject('not load')
        }
    }

    /**
     * 停止
     * @note 停止后的音频再播放会从头开始播放
     */
    stop(): Promise<void> {
        if (this.srcPromise) {
            return this.nativeAudio.promise.then(() => {
                return this.srcPromise.then(() => {
                    this.nativeAudio.stop(undefined, undefined);
                    return undefined;
                });
            });
        } else {
            return Promise.reject('not load')
        }
    }

    /**
     * 取消监听音频进入可以播放状态的事件
     * @param cb 回调
     */
    offCanplay(cb: () => void) {
        return this.nativeAudio.promise.then(() => {
            this.nativeAudio.offCanplay(cb);
            return undefined;
        });
    }

    /**
     * 取消监听音频自然播放至结束的事件
     * @param cb 回调
     */
    offEnded(cb: () => void) {
        return this.nativeAudio.promise.then(() => {
            this.nativeAudio.offEnded(cb);
            return undefined;
        });
    }

    /**
     * 取消监听音频播放错误事件
     * @param cb 回调
     */
    offError(cb: () => void) {
        return this.nativeAudio.promise.then(() => {
            this.nativeAudio.offError(cb);
            return undefined;
        });
    }

    /**
     * 取消监听音频暂停事件
     * @param cb 回调
     */
    offPause(cb: () => void) {
        return this.nativeAudio.promise.then(() => {
            this.nativeAudio.offPause(cb);
            return undefined;
        });
    }

    /**
     * 取消监听音频播放事件
     * @param cb 回调
     */
    offPlay(cb: () => void) {
        return this.nativeAudio.promise.then(() => {
            this.nativeAudio.offPlay(cb);
            return undefined;
        });
    }

    /**
     * 取消监听音频完成跳转操作的事件
     * @param cb 回调
     */
    offSeeked(cb: () => void) {
        return this.nativeAudio.promise.then(() => {
            this.nativeAudio.offSeeked(cb);
            return undefined;
        });
    }

    /**
     * 取消监听音频进行跳转操作的事件
     * @param cb 回调
     */
    offSeeking(cb: () => void) {
        return this.nativeAudio.promise.then(() => {
            this.nativeAudio.offSeeking(cb);
            return undefined;
        });
    }

    /**
     * 取消监听音频停止事件
     * @param cb 回调
     */
    offStop(cb: () => void) {
        return this.nativeAudio.promise.then(() => {
            this.nativeAudio.offStop(cb);
            return undefined;
        });
    }

    /**
     * 取消监听音频播放进度更新事件
     * @param cb 回调
     */
    offTimeUpdate(cb: () => void) {
        return this.nativeAudio.promise.then(() => {
            this.nativeAudio.offTimeUpdate(cb);
            return undefined;
        });
    }

    /**
     * 取消监听音频加载中事件
     * @param cb 回调
     */
    offWaiting(cb: () => void) {
        return this.nativeAudio.promise.then(() => {
            this.nativeAudio.offWaiting(cb);
            return undefined;
        });
    }

    /**
     * 监听音频进入可以播放状态的事件。但不保证后面可以流畅播放
     * @param cb 回调
     */
    onCanplay(cb: () => void) {
        return this.nativeAudio.promise.then(() => {
            this.nativeAudio.onCanplay(cb);
            return undefined;
        });
    }

    /**
     * 监听音频自然播放至结束的事件
     * @param cb 回调
     */
    onEnded(cb: () => void) {
        return this.nativeAudio.promise.then(() => {
            this.nativeAudio.onEnded(cb);
            return undefined;
        });
    }

    /**
     * 监听音频播放错误事件
     * @param cb 回调,参数:
     *      errMsg	string	
     *      errCode	number
     */
    onError(cb: (errMsg: string, errCode: ErrCode) => void) {
        return this.nativeAudio.promise.then(() => {
            this.nativeAudio.onError(cb);
            return undefined;
        });
    }

    /**
     * 监听音频暂停事件
     * @param cb 回调
     */
    onPause(cb: () => void) {
        return this.nativeAudio.promise.then(() => {
            this.nativeAudio.onPause(cb);
            return undefined;
        });
    }

    /**
     * 监听音频播放事件
     * @param cb 回调
     */
    onPlay(cb: () => void) {
        return this.nativeAudio.promise.then(() => {
            this.nativeAudio.onPlay(cb);
            return undefined;
        });
    }

    /**
     * 监听音频完成跳转操作的事件
     * @param cb 回调
     */
    onSeeked(cb: () => void) {
        return this.nativeAudio.promise.then(() => {
            this.nativeAudio.onSeeked(cb);
            return undefined;
        });
    }

    /**
     * 监听音频进行跳转操作的事件
     * @param cb 回调
     */
    onSeeking(cb: () => void) {
        return this.nativeAudio.promise.then(() => {
            this.nativeAudio.onSeeking(cb);
            return undefined;
        });
    }

    /**
     * 监听音频停止事件
     * @param cb 回调
     */
    onStop(cb: () => void) {
        return this.nativeAudio.promise.then(() => {
            this.nativeAudio.onStop(cb);
            return undefined;
        });
    }

    /**
     * 监听音频播放进度更新事件
     * @param cb 回调
     */
    onTimeUpdate(cb: () => void) {
        return this.nativeAudio.promise.then(() => {
            this.nativeAudio.onTimeUpdate(cb);
            return undefined;
        });
    }

    /**
     * 监听音频加载中事件
     * @param cb 回调
     * @note 当音频因为数据不足，需要停下来加载时会触发
     */
    onWaiting(cb: () => void) {
        return this.nativeAudio.promise.then(() => {
            this.nativeAudio.onWaiting(cb);
            return undefined;
        });
    }
}