
import { getAssetFile } from "../util/js_intercept"
import { PlatformType, getPlatformType } from "../util/platform"
import { Env } from "../../pi_sys/util/env";
import { CommonKey } from "../../pi_sys/util/env_base";
import { utf8Decode } from "../../pi_sys/util/utf8";
import { Asset as NatvieAsset } from "../native_impl/core/asset";
import { Logger } from "pi_sys/util/logger";
import { CompressAlgorithm, EncryptType } from "pi_sys/util/constant";
import { Depend } from "pi_sys/util/depend";
import { getDecryptType, getEncryptKey } from "pi_sys/util/util";
import { readFile } from "pi_sys/store/file_system";

declare var module;
let logger = new Logger(module.name, "sys");

/**
 * 本地资源加载
 */

export namespace Asset {
    /**
     * @description 获取资源信息数组Json, [[file, time, size, sign], ...]
     * 读不到返回空数组
     */
    export const get = (): Promise<string[][]> => {
        let assetPath = getAssetPath();
        if (!assetPath) {
            return Promise.resolve([]);
        }

        let dependPaths = [];

        // 这里的version对应 es6.version
        let version = Env.get("version");
        let paths = [];
        for (let key in version) {
            let path = version[key];
            if (typeof path === typeof "") {
                paths.push(path);
            } else {
                // 因为version有 一些对象，像 sys，imageDepend等，所以还要下一层取到所有的路径
                for (let key in path) {
                    paths.push(path[key]);
                }
            }
        }

        for (let path of paths) {
            // 路径得值是："path?sign"
            let len = path.indexOf("?");
            if (len >= 0) {
                path = path.slice(0, len);
            }
            // 只分析depend命名的变量；
            if (path.endsWith(".depend")) {
                dependPaths.push(path);
            }
        }

        let ps = [];
        for (let dependPath of dependPaths) {
            ps.push(new Promise((resolve, reject) => {
                let filePath = `${assetPath}/${dependPath}`;
                let platformType = getPlatformType();
                if (platformType === PlatformType.Harmony) {
                    readFile(filePath).then((content) => {
                        let json = [];
                        if (content) {
                            try {
                                let str = utf8Decode(content as Uint8Array);
                                json = JSON.parse(str);
                            } catch (e) {
                                json = [];
                            }
                        }
                        resolve(json);
                    }).catch(e => {
                        reject(e)
                    })
                } else {
                    getAssetFile(filePath, EncryptType.None, "", CompressAlgorithm.NONE, (content) => {
                        let json = [];
                        if (content) {
                            try {
                                let str = utf8Decode(content as Uint8Array);
                                json = JSON.parse(str);
                            } catch (e) {
                                json = [];
                            }
                        }
                        resolve(json);
                    });
                }

            }));
        }

        return Promise.all(ps).then(arr => {
            // 取完之后，需要合并所有的 []，这数组就是所有底层的资源；
            return [].concat(...arr);
        });
    };

    /**
     * @description 读取资源
     * @example
     */
    export const read = (path: string): Promise<ArrayBuffer> => {
        return new Promise((resolve, reject) => {
            let assetPath = getAssetPath();
            if (!assetPath) {
                resolve(null);
                return;
            }
            const info = Depend.getFile(path);
            let decryptType: EncryptType = EncryptType.None, encryptKey: string = "", compressAlgorithm: CompressAlgorithm = CompressAlgorithm.NONE;
            if (info) {
                decryptType = getDecryptType(info.encryptType);
                encryptKey = getEncryptKey(info.path);
                compressAlgorithm = info.compressAlgorithm || CompressAlgorithm.NONE;
                path = info.encryptPath || info.path;
            }

            let filePath = `${assetPath}/${path}`;
            let platformType = getPlatformType();
            if (platformType === PlatformType.Harmony) {
                readFile(filePath).then((content) => {
                    let c = content as Uint8Array;
                    resolve(c ? c.buffer : null);
                }).catch(e => {
                    reject(e)
                })
            } else {
                getAssetFile(filePath, decryptType, encryptKey, compressAlgorithm, (content) => {
                    let c = content as Uint8Array;
                    resolve(c ? c.buffer : null);
                });
            }

        });
    };

    let assetPathImpl = undefined;
    /**
     * 取 asset包路径
     *     Windows系统：就是 emv.windows_asset_path
     *     Android 和 iOS：就是 env.mobile_asset_path
     */
    const getAssetPath = () => {
        if (assetPathImpl === undefined) {

            let platformType = getPlatformType();

            if (platformType === PlatformType.Win) {
                assetPathImpl = Env.get(CommonKey.WindowsAssetPath);
                if (!assetPathImpl) {
                    logger.warn("windows, env can't find asset path");
                }
            } else if (platformType === PlatformType.Android || platformType === PlatformType.IOS || platformType === PlatformType.Harmony) {
                assetPathImpl = Env.get(CommonKey.MobileAssetPath);
                if (!assetPathImpl) {
                    logger.warn("mobile, env can't find asset path");
                }
            }

            if (!assetPathImpl) {
                assetPathImpl = null;
            }
        }

        return assetPathImpl;
    }

    /**
     * 获取文件在底层的绝对路径
     * @param path 
     * @returns 
     */
    export const getSystemPath = (path: string): Promise<string> => {
        return new Promise((resolve, reject) => {
            NatvieAsset.init().then(() => {
                NatvieAsset.getSystemPath(path,
                    resolve,
                    reject
                );
            }).catch(reject);
        });
    }


}