import { BatchLoad, LoadPolicy } from "pi_sys/load/batch_load";
import { Download } from "pi_sys/load/down_load";
import { FileLoadParam } from "pi_sys/load/fileload_base";
import { TloadVideo } from "pi_sys/load/interface";
import { createVideo, IPiVideo } from "pi_sys/render/video";
import { getCopyFilePath } from "pi_sys/update/assets_to_data";
import { notifyResUsed } from "pi_sys/util/collect";
import { Depend } from "pi_sys/util/depend";
import { IFileDependInfo } from "pi_sys/util/depend_base";
import { Env } from "pi_sys/util/env";
import { getUrl } from "pi_sys/util/util";

const downVideos: Set<string> = new Set; // 正在下载的video视频

/**
 * android和exe的src实际传文件的路径
 * assets中传assets的路径
 * db中传相对于data目录的路径
 * @param filePath 
 * @param args 
 * @returns 
 */
export const loadVideo: TloadVideo = (filePath: string, args: any[]): Promise<IPiVideo> => {
    return new Promise((resolve, reject) => {
        let info = Depend.getFile(filePath);
        if (info) {
            notifyResUsed(info.path);
            if (filePath.endsWith(".m3u8")) { // m3u8格式认为资源都在本地,通过高层处理资源下载,保证资源存在
                let localPath = getLocalPath(info, filePath);
                console.log("================== video url ", localPath);
                let video = createVideo();
                video.oncanplaythrough = function () {
                    resolve(video);
                }
                video.onerror = (err) => {
                    reject(err);
                }
                video.src = localPath;
            } else {
                const waitVideoDownloaded = Env.get("waitVideoDownloaded");
                if (waitVideoDownloaded) { // 等待视频下载完成
                    loadLocalPath(info, filePath).then(localPath => {
                        let video = createVideo();
                        video.oncanplaythrough = function () {
                            resolve(video);
                        }
                        video.onerror = (err) => {
                            reject(err);
                        }
                        video.src = localPath;
                    })

                } else {
                    /**
                     * 由于视频有可能比较大,项目不希望等到视频下载完成才播放,体验较差
                     * 所以如果本地如果还没有缓存,则直接用网络url播放,同时去缓存视频(会消耗两份甚至更多流量)
                     */
                    let localPath = getLocalPath(info, filePath);
                    let url = localPath ? localPath : getUrl(filePath);
                    let video = createVideo();
                    video.oncanplaythrough = function () {
                        if (!localPath && !downVideos.has(filePath)) {
                            downVideos.add(filePath);
                            const duration = !isNaN((<any>video)._duration) ? Math.ceil((<any>video)._duration) : 10;
                            // 等到视频播放完成后下载视频资源缓存
                            setTimeout(() => {
                                const load = new BatchLoad([filePath]);
                                load.startRes().then(() => {
                                    downVideos.delete(filePath);
                                }).catch(e => {
                                    downVideos.delete(filePath);
                                })
                            }, duration * 1000);
                        }
                        resolve(video);
                    }
                    video.onerror = (err) => {
                        reject(err);
                    }
                    video.src = url;
                }

            }


        } else {
            reject(`loadVideo - Not Found File Depend Info ${filePath}`);
        }
    })
}

const download = (file: IFileDependInfo, waitWrite: boolean = true) => {
    const down = new Download();
    (<any>down).waitWrite = waitWrite;
    down.add(file);
    return down.start();
}


export const loadLocalPath = (info: IFileDependInfo, path: string): Promise<string> => {
    let p;
    // assets目录
    if (FileLoadParam.assetsSign[path] === info.sign) {
        const finalPath = info.encryptPath || info.path;
        p = Promise.resolve(finalPath);
    } else if (FileLoadParam.dbSign[path] === info.sign) {  // db目录
        p = Promise.resolve((<any>FileLoadParam.localStore).getFullPath(path));
    } else if (FileLoadParam.assetsCopySign[path]) {   // assetsCopy目录
        const finalPath = getCopyFilePath(path)
        p = Promise.resolve(finalPath);
    } else {
        p = download(info).then(() => {
            return (<any>FileLoadParam.localStore).getFullPath(path);
        });
    }
    return p;
}


/**
 * 获取视频本地地址
 * @param info 
 * @param path 
 * @returns 
 */
const getLocalPath = (info: IFileDependInfo, path: string) => {
    // assets目录
    if (FileLoadParam.assetsSign[path] === info.sign) {
        return info.encryptPath || info.path;
    } else if (FileLoadParam.dbSign[path] === info.sign) {  // db目录
        return (<any>FileLoadParam.localStore).getFullPath(path);
    } else if (FileLoadParam.assetsCopySign[path] === info.sign) {   // assetsCopy目录
        return getCopyFilePath(path)
    }
}

