/**
 * 数据转换器
 * 有些数据加密了,有些数据未加密,在这里统一处理
 */

import { CompressAlgorithm, CompressLevel, EncryptType } from "pi_sys/util/constant";


declare interface IOStream {
    parent: IOStream;
    onlyOutput: boolean;
    output(): Promise<any>;
    pipe(node: IOStream): IOStream;
}

declare class AsyncBufferReader implements IOStream {
    parent: IOStream;
    onlyOutput: boolean;
    output(): Promise<any>;
    pipe(node: IOStream): IOStream;
    constructor(buffer: ArrayBuffer)
}

declare class AsyncFileReader implements IOStream {
    parent: IOStream;
    onlyOutput: boolean;
    output(): Promise<any>;
    pipe(node: IOStream): IOStream;
    constructor(path: string, isAsset: boolean)
}

declare class AsyncEncoder implements IOStream {
    parent: IOStream;
    onlyOutput: boolean;
    output(): Promise<any>;
    pipe(node: IOStream): IOStream;
    constructor(method: string, key?: string, format?: string, level?: number)
}


declare class AsyncDecoder implements IOStream {
    parent: IOStream;
    onlyOutput: boolean;
    output(): Promise<any>;
    pipe(node: IOStream): IOStream;
    constructor(method: string, key?: string, format?: string)
}

declare class AsyncWriteFile implements IOStream {
    parent: IOStream;
    onlyOutput: boolean;
    output(): Promise<any>;
    pipe(node: IOStream): IOStream;
    constructor(path: string)
}


export const decodeBuffer = async (data: Uint8Array | ArrayBuffer, decryptType?: EncryptType, encryptKey?: string, compressAlgorithm?: CompressAlgorithm) => {
    if (data instanceof Uint8Array) {
        data.buffer.slice(data.byteOffset, data.byteOffset + data.byteLength)
        data = data.buffer.slice(data.byteOffset, data.byteOffset + data.byteLength);
    }
    let bufferReader = new AsyncBufferReader(data);
    let originData: ArrayBuffer;
    if (decryptType !== EncryptType.None) {
        originData = await bufferReader.pipe(new AsyncDecoder(decryptType, encryptKey, compressAlgorithm)).output();
    } else {
        originData = data;
    }
    return new Uint8Array(originData);
}


/**
 * 从本地读数据  
 * @param path 
 * @param isAsset 是否是assets目录,否则从data目录读 
 * @param encryptType 
 * @param compressAlgorithm 
 * @returns 
 */
export const readFromLocal = async (path: string, isAsset: boolean, decryptType?: EncryptType, encryptKey?: string, compressAlgorithm?: CompressAlgorithm): Promise<Uint8Array> => {
    let fileReader = new AsyncFileReader(path, isAsset);
    let originData: ArrayBuffer;
    if (decryptType !== EncryptType.None) {
        originData = await fileReader.pipe(new AsyncDecoder(decryptType, encryptKey, compressAlgorithm)).output();
    } else {
        originData = await fileReader.output();
    }
    return new Uint8Array(originData);
}


/**
 * 写数据到data目录
 * @param path 
 * @param decryptType 
 * @param encryptKey 
 * @param compressAlgorithm 
 * @param compressLevel 
 * @returns 
 */
export const writeToData = async (path: string, data: Uint8Array | ArrayBuffer, encryptType?: EncryptType, encryptKey?: string, compressAlgorithm?: CompressAlgorithm, compressLevel?: CompressLevel): Promise<void> => {
    if (data instanceof Uint8Array) {
        data = data.buffer;
    }
    let bufferReader = new AsyncBufferReader(data);
    if (encryptType !== EncryptType.None) {
        return bufferReader.pipe(new AsyncEncoder(encryptType, encryptKey, compressAlgorithm, compressLevel)).pipe(new AsyncWriteFile(path)).output();
    } else {
        return bufferReader.pipe(new AsyncWriteFile(path)).output();
    }

}