// declare var NativeObject;
// declare var _$cid;
/**
 * 接口文档参考:https://developer.mozilla.org/en-US/docs/Web/API/Web_Audio_API
 */

/**
* 解码音频数据
* 因为接口只是单纯的将音频数据包装了一下，未涉及到真正的解码，所以未扔到异步运行时
*/
async function decode_audio_data(audio_data: ArrayBuffer): Promise<object> {
    let result = NativeObject.async_static_call(0, audio_data) as Promise<object>;
    return await result;
}

/**
 * 音频数据
 */
export class AudioBufferBase {
    /**
     * 本地对象
     */
    private self: object;

    /**
     * 类的私有构造方法
     */
    private constructor(self: object) {
        this.self = self;
        if (NativeObject.registry != undefined) {
            NativeObject.registry.register(self, [self]);
        }
    }

    /**
     * 获取本地对象方法
     */
    public get_self() {
        return this.self;
    }

    /**
     * 释放本地对象的方法
     */
    public destroy() {
        if (this.self == undefined) {
            throw new Error("AudioBufferBase already destroy");
        }

        NativeObject.release(_$cid, this.self);
        this.self = undefined;
    }

    /**
     * 从指定本地对象构建当前类方法，此方法是不安全的，使用错误的本地对象将会导致调用时异常
     */
    static from(obj: object): AudioBufferBase {
        return new AudioBufferBase(obj);
    }

    /**
     * 返回当前帧结束前的样本数。None表示“无限”或“直到声音结束”。除非没有更多数据，否则永远不要返回0。
     */
    public current_frame_len(): bigint | undefined {
        if (this.self == undefined) {
            throw new Error("AudioBufferBase object already destroy");
        }

        let result = NativeObject.call(0, this.self) as bigint;
        return result;
    }

    /**
     * 返回通道数。通道总是交错的。
     */
    public channels(): number {
        if (this.self == undefined) {
            throw new Error("AudioBufferBase object already destroy");
        }

        let result = NativeObject.call(1, this.self) as number;
        return result;
    }

    /**
     * 返回播放源的速率。每秒的样本数。
     */
    public sample_rate(): number {
        if (this.self == undefined) {
            throw new Error("AudioBufferBase object already destroy");
        }

        let result = NativeObject.call(2, this.self) as number;
        return result;
    }

    /**
     * 返回此源的总持续时间（如果已知）。
     * None 同时表示“无限”或“未知”。
     */
    public total_duration(): bigint | undefined {
        if (this.self == undefined) {
            throw new Error("AudioBufferBase object already destroy");
        }

        let result = NativeObject.call(3, this.self) as bigint;
        return result;
    }

    /**
     * 返回数据指针
     */
    public get_ptr(): string {
        if (this.self == undefined) {
            throw new Error("AudioBufferBase object already destroy");
        }

        let result = NativeObject.call(4, this.self) as string;
        return result;
    }

}

/**
 * 音频播放器，音频效果的处理
 */
class AudioBufferSourceNodeBase {
    /**
     * 本地对象
     */
    private self: object;

    /**
     * 类的私有构造方法
     */
    private constructor(self: object) {
        this.self = self;
        if (NativeObject.registry != undefined) {
            NativeObject.registry.register(self, [self]);
        }
    }

    /**
     * 获取本地对象方法
     */
    public get_self() {
        return this.self;
    }

    /**
     * 释放本地对象的方法
     */
    public destroy() {
        if (this.self == undefined) {
            throw new Error("AudioBufferSourceNodeBase already destroy");
        }

        NativeObject.release(_$cid, this.self);
        this.self = undefined;
    }

    /**
     * 从指定本地对象构建当前类方法，此方法是不安全的，使用错误的本地对象将会导致调用时异常
     */
    static from(obj: object): AudioBufferSourceNodeBase {
        return new AudioBufferSourceNodeBase(obj);
    }

    /**
     * 构建音频播放器
     */
    static new(): AudioBufferSourceNodeBase {
        let result = NativeObject.static_call(0) as object;
        if (result instanceof Error) {
            throw result;
        } else {
            return new AudioBufferSourceNodeBase(result);
        }
    }

    /**
     * 将音频数据传给音频播放器
     */
    public set_audio_buffer(node: string): void {
        if (this.self == undefined) {
            throw new Error("AudioBufferSourceNodeBase object already destroy");
        }

        NativeObject.call(5, this.self, node);
    }

    /**
     * 设置音量
     */
    public set_volume(volume: number): void {
        if (this.self == undefined) {
            throw new Error("AudioBufferSourceNodeBase object already destroy");
        }

        NativeObject.call(6, this.self, volume);
    }

    /**
     * 是否循环播放
     */
    public set_loop_player(value: boolean): void {
        if (this.self == undefined) {
            throw new Error("AudioBufferSourceNodeBase object already destroy");
        }

        NativeObject.call(7, this.self, value);
    }

    /**
     * 设置播放设备
     */
    public set_destination(destination: string): void {
        if (this.self == undefined) {
            throw new Error("AudioBufferSourceNodeBase object already destroy");
        }

        NativeObject.call(8, this.self, destination);
    }

    /**
     * 设置播放速度
     */
    public set_speed(_ratio: number): void {
        if (this.self == undefined) {
            throw new Error("AudioBufferSourceNodeBase object already destroy");
        }

        NativeObject.call(9, this.self, _ratio);
    }

    /**
     * 暂停播放
     */
    public pausable(_initially_paused: boolean): void {
        if (this.self == undefined) {
            throw new Error("AudioBufferSourceNodeBase object already destroy");
        }

        NativeObject.call(10, this.self, _initially_paused);
    }

    /**
     * 跳到持续时间之后播放
     */
    public skip_duration(_time: number): void {
        if (this.self == undefined) {
            throw new Error("AudioBufferSourceNodeBase object already destroy");
        }

        NativeObject.call(11, this.self, _time);
    }

    /**
     * 延迟到持续时间之后播放
     */
    public set_delay(_time: number): void {
        if (this.self == undefined) {
            throw new Error("AudioBufferSourceNodeBase object already destroy");
        }

        NativeObject.call(12, this.self, _time);
    }

    /**
     * 获取播放持续时间
     */
    public get_current_time(): number | undefined {
        if (this.self == undefined) {
            throw new Error("AudioBufferSourceNodeBase object already destroy");
        }

        let result = NativeObject.call(13, this.self) as number;
        return result;
    }

    /**
     * 开始播放
     */
    public start(): string | Error {
        if (this.self == undefined) {
            throw new Error("AudioBufferSourceNodeBase object already destroy");
        }

        let result = NativeObject.call(14, this.self) as string;
        return result;
    }

    /**
     * 暂停播放
     */
    public pause(): void {
        if (this.self == undefined) {
            throw new Error("AudioBufferSourceNodeBase object already destroy");
        }

        NativeObject.call(15, this.self);
    }

    /**
     * 恢复播放
     */
    public play(): void {
        if (this.self == undefined) {
            throw new Error("AudioBufferSourceNodeBase object already destroy");
        }

        NativeObject.call(16, this.self);
    }

    /**
     * 恢复播放
     */
    public play_volume(volume: number): void {
        if (this.self == undefined) {
            throw new Error("AudioBufferSourceNodeBase object already destroy");
        }

        NativeObject.call(17, this.self, volume);
    }

    /**
     * 恢复播放
     */
    public detach(): void {
        if (this.self == undefined) {
            throw new Error("AudioBufferSourceNodeBase object already destroy");
        }

        NativeObject.call(18, this.self);
    }

}

/**
 * 播放设备
 */
class AudioDestinationNodeBase {
    /**
     * 本地对象
     */
    private self: object;

    /**
     * 类的私有构造方法
     */
    private constructor(self: object) {
        this.self = self;
        if (NativeObject.registry != undefined) {
            NativeObject.registry.register(self, [self]);
        }
    }

    /**
     * 获取本地对象方法
     */
    public get_self() {
        return this.self;
    }

    /**
     * 释放本地对象的方法
     */
    public destroy() {
        if (this.self == undefined) {
            throw new Error("AudioDestinationNodeBase already destroy");
        }

        NativeObject.release(_$cid, this.self);
        this.self = undefined;
    }

    /**
     * 从指定本地对象构建当前类方法，此方法是不安全的，使用错误的本地对象将会导致调用时异常
     */
    static from(obj: object): AudioDestinationNodeBase {
        return new AudioDestinationNodeBase(obj);
    }

    /**
     * 构建一个默认播放设备
     */
    static new(): string | Error {
        let result = NativeObject.static_call(1) as string;
        return result;
    }

}







/**
 * 文档参考：https://developer.mozilla.org/zh-CN/docs/Web/API/AudioBuffer
 */
class AudioBuffer {

    public base: AudioBufferBase;

    constructor(base?: object) {
        if (base) {
            this.base = AudioBufferBase.from(base);
        } else {
            this.base = undefined;
        }

    }

    /**
     *  存储在缓存区的PCM数据的采样率：浮点数，单位为 sample/s。 
     */
    get sampleRate(): number {
        return this.base.sample_rate();
    }

    /**
     * 返回存储在缓存区的PCM数据的采样帧率：整形。
     */
    get length(): number {
        return Number(this.base.current_frame_len()) ;
    }

    /**
     * 返回存储在缓存区的PCM数据的时长：双精度型（单位为秒），。
     */
    get duration(): number {
        return Number(this.base.total_duration());
    }

    /**
     * 返回存储在缓存区的PCM数据的通道数：整形。
     */
    get numberOfChannels(): number {
        return this.base.channels();
    }


    /**
     * 返回一Float32Array ，其中包含与通道关联的PCM数据，通道参数定义(0表示第一个通道)。
     * 
     * @param channel 属性是要获取特定通道数据的索引。0代表第一个通道。 如果索引值大于或等于AudioBuffer.numberOfChannels, 会抛出一个索引大小异常（INDEX_SIZE_ERR ）的错误。
     */
    public getChannelData(channel: number) {

    }

    /**
     * 法将样本从AudioBuffer的指定通道复制到目标数组中
     * 
     * @param destination 将通道数据复制到的Float32Array
     * @param channelNumber 当前AudioBuffer的通道号，用于复制通道数据。键入channelNumber大于或等于AudioBuffer.numberOfChannels，将会抛出INDEX_SIZE_ERR 的错误
     * @param startInChannel 可选,用于复制数据的可选偏移量。假如startInChannel比AudioBuffer.length大，将会抛出INDEX_SIZE_ERR 的错误
     */
    public copyFromChannel(destination: Float32Array, channelNumber: number, startInChannel: number) {

    }

    /**
     * 复制样品到原数组的AudioBuffer的指定频道
     */
    public copyToChannel() {

    }

}

/**
 * 文档参考：https://developer.mozilla.org/zh-CN/docs/Web/API/AudioNode
 */
class AudioNode {
    constructor(contxet: BaseAudioContext) {
        this.context = contxet;
        this.last = undefined;
        this.next = undefined;
    }

    next: AudioNode;
    last: AudioNode;
    /**
     *  链接到关联的 AudioContext，处理图中模块的上下文对象。
     */
    public context: BaseAudioContext;


    /**
     *  返回这个node需要的输入数量。源节点被定义为具有numberOfInputs值为的属性的节点0。
     */
    public numberOfInputs: number;

    /**
     * 返回这个node的输出数量。目的地节点（例如AudioDestinationNode）具有0此属性的值。
     */
    public numberOfOutputs: number

    /**
     * 表示用于确定输出必须包含多少个通道的整数。
     * 其用法和精确定义取决于的值AudioNode.channelCountMode：如果该值是"max"，
     * 用作的最大值"clamped-max"或用作的有效值，则将忽略该值"explicit"。
     */
    public channelCount: number

    /**
     * 表示一个枚举值，描述必须在输入和输出之间匹配通道的方式。
     */
    public channelCountMode: string

    /**
     * 表示描述通道含义的枚举值。这种解释将定义上混合和下混合的发生方式。
     * 可能的值为"speakers"或"discrete"。
     */
    public channelInterpretation: string;

    /**
     * 允许将此节点的一个输出连接到另一个节点的一个输入。
     * 允许将此节点的一个输出连接到音频参数的一个输入。
     * 
     * @param node 需要连接的AudioNode或者AudioParam
     */
    public connect(node: AudioNode) {
        // if ((node as AudioNode).state) {
        //     this.player.pausable(false)
        //     return;
        // }
        // if (this.player) {
        //     if (node instanceof AudioNode) {
        //         node.player = this.player;
        //         node.state = this.state;
        //     }

        //     if (node instanceof AudioDestinationNode) {
        //         this.player.set_destination(node.ptr);
        //     } else if (node instanceof GainNode) {
        //         this.player.set_volume(node.gain.value);
        //     }

        //     return node;
        // }
        this.next = node;
        node.last = this;
        let tempNode: AudioNode = this;
        while (tempNode) {
            if (tempNode instanceof AudioBufferSourceNode) {
                switch (tempNode.state) {
                    case State.Pause:
                        tempNode.stop();
                        break;
                    default:
                        break;
                }
                break;
            }
            tempNode = tempNode.last;
        }
    }

    public disconnect() {
        let tempNode: AudioNode = this;
        let nextNode = this.next;
        if (nextNode) {
            nextNode.last = undefined;
            this.next = undefined;
        }

        while (tempNode) {
            if (tempNode instanceof AudioBufferSourceNode) {
                switch (tempNode.state) {
                    case State.Playing:
                        tempNode.start()
                        break;
                    default:
                        break;
                }
                break;
            }
            tempNode = tempNode.last;
        }
    }

}

/**
 * 文档参考：https://developer.mozilla.org/zh-CN/docs/Web/API/AudioParam
 */
class AudioParam {
    constructor(defaultValue: number, maxValue: number, minValue: number, node: AudioNode) {
        this.defaultValue = defaultValue;
        this.maxValue = maxValue;
        this.minValue = minValue;
        this._value = defaultValue;
        this._currentNode = node;
    }

    private _currentNode: AudioNode;

    private _value: number;

    /**
     * 代表被具体的AudioNode 创建的 AudioParam 的属性的初始的音量。
     */
    public readonly defaultValue: number

    /**
     * 代表参数有效范围的最大可能值。
     */
    public readonly maxValue: number;

    /**
     * 代表参数有效范围的最小可能值。
     */
    public readonly minValue: number;


    /**
     * 代表参数的浮点数音量值；初始化设定为 AudioParam.defaultValue 的值。
     * 虽然它可以被设置，但是任何发生在自动化事件（事件被计划用于 AudioParam ）的修改会被忽略，没有任何例外。
     */
    set value(value: number) {
        if (value > this.maxValue) {
            this._value = this.maxValue;
        } else if (value < this.minValue) {
            this._value = this.maxValue;
        } else {
            this._value = value;
        }

        let tempNode = this._currentNode;
        // while(tempNode){
        while (tempNode) {
            if (tempNode instanceof AudioBufferSourceNode) {
                tempNode.start();
            }
            tempNode = tempNode.last;
        }
        // }
    }

    get value(): number {
        return this._value;
    }


    /**
     * 在一个确切的时间，即时更改 AudioParam 的值，按照AudioContext.currentTime 的时间。
     * 新的值会被传递到 value 参数。
     */
    public setValueAtTime() {

    }

    /**
     * 调整 AudioParam 的值，使其逐渐按线性变化。
     * 这个改变会从上一个事件指定的事件开始，跟随一个线性“斜坡”到参数给的新值，并在 endTime 参数给定的时间到达新值。
     */
    public linearRampToValueAtTime() {

    }

    /**
     * 调整 AudioParam 的值，使其逐渐按指数变化。
     * 这个改变会从上一个事件指定的事件开始，跟随一个指数“斜坡”到参数给的新值，并在 endTime 参数给定的时间到达新值。
     */
    public exponentialRampToValueAtTime() {

    }

    /**
     * 将开始计划改变 AudioParam 的值。
     * 这个改变将从 startTime 指定的时间开始，并且以指定的方式向目标参数给定的值改变。
     * 指数衰减速率由 timeConstant 参数定义，time 参数使以秒作为测量单位的时间。
     */
    public setTargetAtTime() {

    }

    /**
     * 调整 AudioParam 的值以跟随一组定义为 Float32Array 的值，
     * 数值会缩放到适应给定的间隔,，从 startTime 时间开始并具有特定的持续时间（duration）。 
     */
    public setValueCurveAtTime() {

    }

    /**
     * 取消全部在 AudioParam 中的未来计划发生的改变
     */
    public cancelScheduledValues() {

    }

    /**
     * 取消全部计划将来对 AudioParam 的改变，但是保持给定时间的值，
     * 直到将来的使用其他方法产生改变。新的值会被赋予到 value 属性中。
     */
    public cancelAndHoldAtTime() {

    }
}

/**
 * 文档参考：https://developer.mozilla.org/zh-CN/docs/Web/API/BaseAudioContext
 */
class BaseAudioContext {
    constructor() {
        this.destination = new AudioDestinationNode(this);
        this.currentTime = 0;
    }

    // control: AudioBufferSourceNodeBase;
    private _onstatechangeCb = null;

    private _audioBufferSource: AudioBufferSourceNode;

    /**
     * 返回AudioWorklet用于通过JavaScript处理创建自定义AudioNode的对象。  
     */
    public audioWorklet: AudioNode;

    /**
     * 返回一个double值，表示用于调度的硬件时间（以秒为单位）不断增加。开始于0。
     */
    public currentTime: number;

    /**
     * 返回AudioDestinationNode表示上下文中所有音频的最终目的地的。可以将其视为音频渲染设备。
     */
    public destination: AudioNode;

    /**
     * 返回AudioListener用于3D空间化的对象。
     */
    public listener: AudioListener;

    /**
     * 返回一个浮点数，该浮点数表示此上下文中所有节点使用的采样率（以每秒采样数为单位）。AudioContext无法更改的采样率。
     */
    public sampleRate: number;

    /**
     * 返回的当前状态AudioContext。
     * 可能的值如下：
     *     suspended: audio context被阻塞了(用AudioContext.suspend() 方法)
     *     running: audio context正常运行
     *     closed: audio context被关闭了(用AudioContext.close()方法)
     */
    public state: string;

    /**
     * 事件类型statechange触发后运行的事件处理程序。
     * 这发生在所述AudioContext的状态变化，由于该状态改变的方法之一的调用（AudioContext.suspend，AudioContext.resume，或AudioContext.close）
     */
    // 
    set onstatechange(cb: (this: BaseAudioContext) => void) {
        this._onstatechangeCb = cb ? cb.bind(this) : null;
    }

    /**
     * 创建一个新的空AudioBuffer对象，然后可以将其填充数据并通过AudioBufferSourceNode播放。
     */
    public createBuffer() {
        return new AudioBuffer();
    }

    /**
     *  创建一个ConstantSourceNode对象，该对象是一个音频源，它连续输出其样本都具有相同值的单声道（单声道）声音信号。
     */
    public createConstantSource() {

    }

    /**
     * 创建一个AudioBufferSourceNode，可用于播放和操作AudioBuffer对象中包含的音频数据。
     * AudioBuffer使用成功解码音轨时使用AudioContext.createBuffer或返回来创建AudioContext.decodeAudioData。
     */
    public createBufferSource() {
        if (!!this._audioBufferSource) {
            this._audioBufferSource.player.detach();
        }
        this._audioBufferSource = new AudioBufferSourceNode(this);

        return this._audioBufferSource;
    }

    /**
     * 创建一个ScriptProcessorNode，可用于通过JavaScript直接进行音频处理。
     */
    public createScriptProcessor() {

    }

    /**
     * 创建一个StereoPannerNode，可用于将立体声声像应用于音频源。
     */
    public createStereoPanner() {

    }

    /**
     * 创建一个AnalyserNode，可用于显示音频时间和频率数据，例如创建数据可视化。
     */
    public createAnalyser() {

    }

    /**
     * 创建一个BiquadFilterNode，代表一个可配置为几种不同的常见滤波器类型的二阶滤波器：高通，低通，带通等。
     */
    public createBiquadFilter() {

    }

    /**
     * 创建一个ChannelMergerNode，用于将来自多个音频流的通道合并为一个音频流。
     */
    public createChannelMerger() {

    }

    /**
     * 创建一个ChannelSplitterNode，用于访问音频流的各个通道并分别处理它们。
     */
    public createChannelSplitter() {

    }

    /**
     * 创建一个ConvolverNode，可用于将卷积效果应用于音频图，例如混响效果。
     */
    public createConvolver() {

    }

    /**
     *  创建一个DelayNode，用于将传入的音频信号延迟一定量。此节点对于在Web Audio API图形中创建反馈循环也很有用。
     */
    public createDelay() {

    }

    /**
     * 创建一个DynamicsCompressorNode，可用于将音频压缩应用于音频信号。
     */
    public createDynamicsCompressor() {

    }

    /**
     * 创建一个GainNode，可用于控制音频图的整体音量。
     */
    public createGain() {
        return new GainNode(this);
    }

    /**
     * 创建一个IIRFilterNode，该代表可配置为几种不同的常用过滤器类型的二阶过滤器。
     */
    public createIIRFilter() {

    }

    /** 
     * 创建一个OscillatorNode，代表周期波形的源。它基本上会产生音调。 
     */
    public createOscillator() {

    }

    /**
    * 创建一个PannerNode，用于在3D空间中对传入的音频流进行空间化。
    */
    public createPanner() {

    }

    /**
     * 创建一个PeriodicWave，用于定义可用于确定的输出的周期波形OscillatorNode。
     */
    public createPeriodicWave() {

    }

    /**
     * 创建一个WaveShaperNode，用于实现非线性失真效果。
     */
    public createWaveShaper() {

    }

    /**
     * 异步解码中包含的音频文件数据ArrayBuffer。
     * 在这种情况下，ArrayBuffer通常是从一个加载XMLHttpRequest的response属性设定后responseType到arraybuffer。此方法仅适用于完整文件，而不适用于音频文件的片段。 
     * 
     */
    public decodeAudioData(audioData: ArrayBuffer, onLoadFunc?: (buffer: AudioBuffer) => void, onErrorFunc?: (error: any) => void) {
        decode_audio_data(audioData).then((res) => {
            if (res instanceof Error) {
                throw res;
            }
            let buffer = new AudioBuffer(res);
            onLoadFunc(buffer);
        }).catch((err) => {
            // onErrorFunc(err);
        });
    }

    /**
     * 在以前已暂停/暂停的音频上下文中恢复时间的进展。
     * 
     * @param v 状态值
     */
    public resume(v: number) {

    }

}

/**
 * 文档参考：https://developer.mozilla.org/zh-CN/docs/Web/API/AudioContext
 */
class AudioContext extends BaseAudioContext {
    constructor() {
        super();

    }

    /**
     * 关闭一个音频环境, 释放任何正在使用系统资源的音频。
     */
    public close() {

    }

    /**
     * 创建一个MediaElementAudioSourceNode接口来关联HTMLMediaElement. 这可以用来播放和处理来自<video>或<audio> 元素的音频。
     */
    public createMediaElementSource() {

    }

    /**
     * 创建一个MediaStreamAudioSourceNode接口来关联可能来自本地计算机麦克风或其他来源的音频流MediaStream。
     */
    public createMediaStreamSource() { }

    /**
     * 创建一个MediaStreamAudioDestinationNode接口来关联可能储存在本地或已发送至其他计算机的MediaStream音频。
     */
    public createMediaStreamDestination() { }

    /**
     * 创建一个MediaStreamTrackAudioSourceNode，它与一个MediaStream相关联，表示一个媒体流轨迹。
     */
    public createMediaStreamTrackSource() { }

    /**
     * 返回一个新的AudioTimestamp对象，该对象包含两个与当前音频上下文相关的音频时间戳。
     */
    public getOutputTimestamp() { }
    /**
     * 恢复之前被暂停的音频上下文中的时间进程。
     */
    public resume() { }

    /**
     * 暂停音频上下文中的时间进程，暂停音频硬件访问并减少进程中的CPU/电池使用。
     */
    public suspend() { }

}
enum State {
    Init,
    Playing,
    Pause,
}

class AudioBufferSourceNode extends AudioNode {
    constructor(context: BaseAudioContext) {
        super(context);
        this.loopEnd = 0.0;
        this.loopStart = 0.0;
        this.player = AudioBufferSourceNodeBase.new();
    }
    private _onendedCb = null;
    private volume: number = 1.0;

    public state = State.Init;
    public player: AudioBufferSourceNodeBase;
    /**
     * 是一个AudioBuffer它定义了要播放的音频，当设置它的计数0时，它会定义一个静默的单通道。
     */
    set buffer(buffer: AudioBuffer) {
        // if (this._playListNumber === 0) {
        this.player.set_audio_buffer(buffer.base.get_ptr());
        // this._playListNumber++;
        // }
    }

    /**
     * 是一个k速率， AudioParam表示以分度计的振荡失谐。其默认值为0。
     */
    public detune: number;

    /**
     * 是一个布尔型属性，指示在到达末尾时是否必须重播音频资产AudioBuffer。其默认值为false。
     */
    set loop(enable: boolean) {
        if (this.player) {
            this.player.set_loop_player(enable);
        }
    }

    /**
     * 是一个双精度值，以秒为单位表示AudioBuffer必须重新开始播放的位置。其默认值为0。
     */
    public loopStart: number;

    /**
     * 是一个双精度值，以秒为单位，指示AudioBuffer播放的重播必须停止的位置（并最终再次循环播放）。其默认值为0。
     */
    public loopEnd: number;

    /**
     * 是一种速率 AudioParam，用于定义播放音频资产的速度因子。由于没有在输出上应用音高校正，因此可以使用它来更改样本的音高。
     */
    public playbackRate = {
        value: 0,
    }

    set onended(cb: (this: BaseAudioContext) => void) {
        this._onendedCb = cb ? cb.bind(this) : null;
    }

    /**
     * 安排开始播放音频。
     */
    private _isFirst = true;
    public start(duration?: number) {
        let tempNode: AudioNode = this;

        while (tempNode) {
            if (tempNode instanceof AudioDestinationNode) {
                if (this.state === State.Init) {
                    this.player.set_destination(tempNode.ptr);
                }

                if (this._isFirst) {
                    this.player.play_volume(this.volume);
                    let res = this.player.start();

                    if (res instanceof Error) {
                        throw res;
                    }
                    this.state = State.Playing;
                    this._isFirst = false;
                } else if (this.state = State.Pause) {
                    this.player.play();
                    this.state = State.Playing;
                }

                return;
            } else if (tempNode instanceof GainNode) {
                if (this.volume !== tempNode.gain.value) {
                    this.volume = tempNode.gain.value;
                    if (this.state !== State.Init) {
                        this.player.play_volume(this.volume);
                    }
                }

            }
            tempNode = tempNode.next;
        }

        return;

    }

    /**
     * 安排音频资产播放的结束。
     */
    public stop(duration?: number) {
        this.player.pause();
        this.state = State.Pause;
    }
}

/**
 * BiquadFilterNode 接口表示一个简单的低频滤波器。
 * 文档参考：https://developer.mozilla.org/zh-CN/docs/Web/API/BiquadFilterNode
 */
class BiquadFilterNode extends AudioNode {
    /**
     * 是a速率 AudioParam，它代表当前滤波算法中以赫兹（Hz）为单位的频率。
     */
    public frequency: AudioParam;

    /**
     * 是一个速率 AudioParam表示频率的失谐美分。
     */
    public detune: AudioParam;

    /**
     * 是一个比率 AudioParam，代表Q因子或品质因子的双精度。
     */
    public Q: AudioParam;

    /**
     * 是a-rate AudioParam，是表示当前滤波算法中使用的增益的两倍。
     */
    public gain: AudioParam;

    /**
     *  报表实现定义不同过滤算法的一个同轴值。
     */
    public type: string;

    /**
     * 从当前的滤波器参数设置中，此方法将计算提供的频率数组中指定的频率的频率响应。
     */
    public getFrequencyResponse() { }

}

/**
 * ConvolverNode 接口是一个对给定 AudioBuffer 上执行线性卷积的 AudioNode，一般用来做音频混响效果。
 * 每一个 ConvolverNode 都会有一个输入值和输出值。
 * 文档参考：https://developer.mozilla.org/zh-CN/docs/Web/API/ConvolverNode
 */
class ConvolverNode extends AudioNode {
    /**
     * 一个被 ConvolverNode 用来产生混响效果的单声道、立体声或四声道的音频缓冲器，包含了(可能是多声道)脉冲反应(IR)。
     */
    public buffer: AudioBuffer;

    /**
     * 布尔值，在设置缓冲区属性时，可绝定是否对来自 buffer 的脉冲反应按等功率归一化进行缩放。
     */
    public normalize: boolean;
}

/**
 * GainNode 接口用于音量变化。它是一个 AudioNode 类型的音频处理模块，输入后应用增益 效果，然后输出。
 */
class GainNode extends AudioNode {
    constructor(context: BaseAudioContext) {
        super(context);
        this.gain = new AudioParam(1, 3.40, -3.40, this);
    }
    /**
     * 是一个a-rateAudioParam表示应用的增益量。必须设置AudioParam.value或者使用AudioParam的方法改变增益效果。
     */
    public gain: AudioParam;
}

/**
 * 该AudioDestinationNode接口表示给定上下文中的音频图表的最终目标，通常是设备的扬声器。
 * 当与一起使用时，它也可以是将“记录”音频数据的节点OfflineAudioContext。
 * 文档参考：https://developer.mozilla.org/zh-CN/docs/Web/API/ConvolverNode
 */
class AudioDestinationNode extends AudioNode {
    constructor(context: BaseAudioContext) {
        super(context);
        let res = AudioDestinationNodeBase.new();
        if (res instanceof Error) {
            throw res;
        } else if (typeof res === typeof "") {
            this.ptr = res;
        }

    }
    public ptr: string;

    public maxChannelCount: number;
}

let g = globalThis as any;
g.AudioContext = AudioContext;
