declare var NativeObject;
declare var _$cid;


class WindowCanvas {
    /**
     * 本地对象
     */
    private self: object;

    /**
     * 类的私有构造方法
     */
    private constructor(self: object) {
        this.self = self;
        if (NativeObject.registry != undefined) {
            NativeObject.registry.register(self, [self]);
        }
    }

    /**
     * 获取本地对象方法
     */
    public get_self() {
        return this.self;
    }

    /**
     * 释放本地对象的方法
     */
    public destroy() {
        if (this.self == undefined) {
            throw new Error("WindowCanvas already destroy");
        }

        NativeObject.release(_$cid, this.self);
        this.self = undefined;
    }

    /**
     * 从指定本地对象构建当前类方法，此方法是不安全的，使用错误的本地对象将会导致调用时异常
     */
    static from(obj: object): WindowCanvas {
        return new WindowCanvas(obj);
    }

    /**
     * 创建canvas
     */
    static new(pid: number): WindowCanvas {
        let result = NativeObject.static_call(2, pid) as object;
        if (result instanceof Error) {
            throw result;
        } else {
            return new WindowCanvas(result);
        }
    }

    /**
     * 获取canvas webgl的渲染上下文,全局只能有一个，与主窗口绑定
     */
    public get_webgl_context(off_screen: boolean): object | undefined {
        if (this.self == undefined) {
            throw new Error("WindowCanvas object already destroy");
        }

        let result = NativeObject.call(19, this.self, off_screen) as object;
        return result;
    }

    /**
     * 获取一个canvas2d 的上下文，主要用于字体渲染，全局可以有多个，
     */
    public get_2d_context(): object {
        if (this.self == undefined) {
            throw new Error("WindowCanvas object already destroy");
        }

        let result = NativeObject.call(20, this.self) as object;
        return result;
    }

    /**
     * 获取canvas宽度
     */
    public get_width(): number {
        if (this.self == undefined) {
            throw new Error("WindowCanvas object already destroy");
        }

        let result = NativeObject.call(21, this.self) as number;
        return result;
    }

    /**
     * 获取canvas高度
     */
    public get_height(): number {
        if (this.self == undefined) {
            throw new Error("WindowCanvas object already destroy");
        }

        let result = NativeObject.call(22, this.self) as number;
        return result;
    }

    /**
     * 获取canvas位图数据
     */
    public get_buffer(): ArrayBuffer | undefined {
        if (this.self == undefined) {
            throw new Error("WindowCanvas object already destroy");
        }

        let result = NativeObject.call(23, this.self) as ArrayBuffer;
        return result;
    }

    /**
     * 设置canvas的宽
     */
    public set_width(width: number): void {
        if (this.self == undefined) {
            throw new Error("WindowCanvas object already destroy");
        }

        NativeObject.call(24, this.self, width);
    }

    /**
     * 设置canvas的高
     */
    public set_height(height: number): void {
        if (this.self == undefined) {
            throw new Error("WindowCanvas object already destroy");
        }

        NativeObject.call(25, this.self, height);
    }

    /**
 * 设置canvas的大小
 */
    public set_size(width: number, height: number): void {
        if (this.self == undefined) {
            throw new Error("WindowCanvas object already destroy");
        }

        NativeObject.call(26, this.self, width, height);
    }
}





export class Canvas {

    private canvas: WindowCanvas;
    _width: number;
    _height: number;
    id: number;

    contextName: string;
    context: any;

    constructor() {
        let pid: number = globalThis.ESProcess.pid;
        this.canvas = WindowCanvas.new(pid);
        this._width = this.canvas.get_width();
        this._height = this.canvas.get_height();
        this.id = isFirst;
        isFirst++;
    }

    public destroy() {
        this.canvas.destroy();
    }

    public getContext(name: string, _option: any) {
        if (this.context) {
            if (this.contextName !== name) {
                return undefined;
            } else {
                return this.context;
            }
        }

        if (this.id === 0) {
            if (name === "webgl") {
                let obj = this.canvas.get_webgl_context(false);
                let gl = globalThis.WebGLRenderingContext.from(obj, this._width, this._height);
                this.context = gl;
                this.contextName = name;
                globalThis.__setFaqContext(this.context);
                globalThis.__setWinContext(this.context);
                return this.context;
            }
            if (name === "webgpu") {
                let obj = this.canvas.get_webgpu_context();
                let context = new globalThis.GPUCanvasContext(obj, this._width, this._height);
                this.context = context;
                this.contextName = name;
                return this.context;
            }
            return undefined;
        } else {
            if (name === "2d") {
                this.canvas.set_width(this._width);
                this.canvas.set_height(this._height);

                let obj = this.canvas.get_2d_context();
                let ctx = globalThis.CanvasRenderingContext2D.from(obj);
                this.context = ctx;
                this.contextName = name;
                return this.context;
            }
            if (name === "webgl") {
                this.canvas.set_width(this._width);
                this.canvas.set_height(this._height);
                let obj = this.canvas.get_webgl_context(true);
                let gl = globalThis.WebGLRenderingContext.from(obj, this._width, this._height);
                this.context = gl;
                this.contextName = name;
                return this.context;
            }
            return undefined;
        }
    }

    public setSize(width: number, height: number) {
        this.canvas.set_size(width, height);
    }

    set width(width: number) {
        // if (this.contextName === "2d") {
        this._width = width;
        this.canvas.set_width(width);
        // }
    }
    set height(height: number) {
        // if (this.contextName === "2d") {
        this._height = height;
        this.canvas.set_height(height);
        // }
    }
    get width() {
        return this._width;
    }
    get height() {
        return this._height;
    }

    public setAttribute(key: string, value: any) {
        this[key] = value;
    }

    public addEventListener(event: String, cb: (this: Canvas) => void, options?: boolean) {
        switch (event) {
            case "pointerdown":
                globalThis.onMouseDown(cb);
                globalThis.onTouchStart(cb);
                break;
            case "pointermove":
                globalThis.onMouseMove(cb);
                globalThis.onTouchMove(cb);
                break;
            case "pointerup":
                globalThis.onMouseUp(cb);
                globalThis.onTouchEnd(cb);
                break;
            case "pointercancel":
                globalThis.onTouchCancel(cb);
                break;
            case "wheel":
                globalThis.onMouseWheel(cb);
                break;
            default:
                break;
        }
    }

    public removeEventListener(event: String, cb: (this: Canvas) => void, options?: boolean) {
        switch (event) {
            case "pointerdown":
                globalThis.offTouchStart(cb);
                globalThis.offMouseDown(cb);
                break;
            case "pointermove":
                globalThis.offTouchMove(cb);
                globalThis.offMouseMove(cb);
                break;
            case "pointerup":
                globalThis.offTouchEnd(cb);
                globalThis.offMouseUp(cb);
                break;
            case "pointercancel":
                globalThis.offTouchCancel(cb);
                break;
            case "wheel":
                globalThis.offMouseWheel(cb);
                break;
            default:
                break;
        }
    }

    public getBoundingClientRect() {
        return {
            x: 0,
            y: 0,
            width: this.width,
            heugght: this.height,
            top: 0,
            right: this.width,
            bottom: this.height,
            left: 0
        }
    }

    public focus() {
        return undefined;
    }

    get buffer(): ArrayBuffer {
        return this.canvas.get_buffer()
    }

    public get_width() {
        return this._width;
    }

    public get_height() {
        return this._height;
    }
    // public set_width(width:number){
    //     this.canvas.set_width(width);
    // }

    // public set_height(height: number){
    //     this.canvas.set_height(height);
    // }
}

let isFirst = 0;

let g: any = globalThis;
g.Canvas = Canvas;
