


const enum FSType {
    File = 1,
    Dir = 2,
    SymLink = 3,
}

interface FSState {
    size: number;
    type: FSType;
    isReadOnly: boolean;
    lastAccessTime: number;
    lastModifyTime: number;
}

const fsMod = globalThis._$pi_native.FileSystemManager;
delete globalThis._$pi_native.FileSystemManager;

/**
 * 
 * @param path 必填，文件/目录的路径
 */
export const access = (path: string): Promise<boolean> => {
    if (typeof path !== typeof "") {
        return Promise.reject(`access failed, path is not string`);
    }

    try {
        let isExist = fsMod.access(path);
        return Promise.resolve(isExist);
    } catch (e) {
        return Promise.reject(e);
    }
}



/**
 * @param {*} path 必填，文件的路径
 */
const stat = (path: string): Promise<FSState> => {
    if (typeof path !== typeof "") {
        return Promise.reject(`stat failed, path is not string`);
    }
    try {
        let info = fsMod.stat(path);
        return Promise.resolve(info);
    } catch (e) {
        return Promise.reject(e);
    }
}


/**
 * 
 * @param {*} from_path 必填，源文件/目录 的路径
 * @param {*} to_path 必填，修改成 的路径
 */

const rename = (fromPath: string, toPath: string): Promise<boolean> => {
    if (typeof fromPath !== typeof "" || typeof toPath !== typeof "") {
        return Promise.reject(`rename failed, path is not string`);
    }
    try {
        return fsMod.rename(fromPath, toPath);
    } catch (e) {
        return Promise.reject(e);
    }
}


/**
 * 
 * @param {*} path 必填 需要创建的目录
 */
const mkdir = (path: string): Promise<void> => {
    if (typeof path !== typeof "") {
        return Promise.reject(`rename failed, path is not string`);
    }
    try {
        return fsMod.mkdir(path);
    } catch (e) {
        return Promise.reject(e);
    }
}


/**
 * 
 * @param {*} path 必填，要读取的目录
 */
const readdir = (path: string): Promise<string[]> => {
    if (typeof path !== typeof "") {
        return Promise.reject(`readir failed, path is not string`);
    }

    try {
        let info = fsMod.readdir(path);
        return Promise.resolve(info);
    } catch (e) {
        return Promise.reject(e);
    }
}


/**
 * 
 * @param {*} path 必填，要删除的目录路径
 */
const rmdir = (path: string): Promise<void> => {
    if (typeof path !== typeof "") {
        return Promise.reject("readFile error, path is not string");
    }

    try {
        return fsMod.rmdir(path);
    } catch (e) {
        return Promise.reject(e);
    }
}

/**
 * 
 * @param {} path 必填，要读取的路径
 * @param {} encoding 可选，编码，只有三个值：utf8, utf-8, binary(默认值)
 * @param {} offset 可选，要读取的二进制偏移量，单位：字节；如果不填，默认为0，代表从文件开头读取；
 * @param {} size 可选，要读取的二进制大小，单位：字节；如果不填，默认读到文件末尾；
 */
const readFile = (path: string, encoding: 'utf8' | 'utf-8' | 'binary', offset: number, size: number): Promise<string | ArrayBuffer> => {

    if (typeof path !== typeof "") {
        return Promise.reject("readFile error, path is not string");
    }

    if (offset === undefined || offset === null) {
        offset = 0;
    } else if (typeof offset !== typeof 0) {
        return Promise.reject(`readFile error, offset is not number, offset = ${offset}`);
    } else if (offset < 0) {
        return Promise.reject(`readFile error, offset is invalid, offset = ${offset}`);
    }

    if (size === undefined || size === null) {
        size = 0; // 默认值
    } else if (typeof size !== typeof 0) {
        return Promise.reject("readFile error, size is not number");
    } else if (size <= 0) { // 用户填进来的数字，不能是非正数；
        return Promise.reject("readFile error, size must >= 0");
    }

    let is_binary = true;
    if (encoding !== undefined && encoding !== null) {
        if (encoding === "utf8" || encoding === "utf-8") {
            is_binary = false;
        } else if (encoding !== "binary") {
            return Promise.reject("readFile error, encoding is not 'utf8' | 'utf-8' | 'binary'");
        }
    }

    try {
        return fsMod.readFile(path, is_binary, offset, size);
    } catch (e) {
        return Promise.reject(e);
    }
}


/**
 * 
 * @param {*} path 必填，需要写入的路径
 * @param {*} data 必填，需要写入的数据
 */
const writeFile = (path: string, data: string | ArrayBuffer): Promise<void> => {
    if (typeof path !== typeof "") {
        return Promise.reject("writeFile error, path is not string");
    }

    if (typeof data !== typeof "" && !(data instanceof ArrayBuffer)) {
        return Promise.reject("writeFile error, data is not string or arrayBuffer");
    }

    try {
        return fsMod.writeFile(path, data);
    } catch (e) {
        return Promise.reject(e);
    }
}

/**
 * 
 * @param {*} path 必填，需要加入的路径
 * @param {*} data 必填，需要加入的数据
 */
const appendFile = (path: string, data: string | ArrayBuffer): Promise<void> => {
    if (typeof path !== typeof "") {
        return Promise.reject("appendFile error, path is not string");
    }

    if (typeof data !== typeof "" && !(data instanceof ArrayBuffer)) {
        return Promise.reject("appendFile error, data is not string or arrayBuffer");
    }

    try {
        return fsMod.appendFile(path, data);
    } catch (e) {
        return Promise.reject(e);
    }
}

/**
 * 
 * @param {*} srcPath 必填，需要拷贝的源文件路径
 * @param {*} destPath 必填，拷贝副本的文件路径
 */
const copyFile = (srcPath: string, destPath: string): Promise<void> => {
    if (typeof srcPath !== typeof "") {
        return Promise.reject("copyFile error, srcPath is not string");
    }

    if (typeof destPath !== typeof "") {
        return Promise.reject("copyFile error, destPast is not string");
    }

    try {
        return fsMod.copyFile(srcPath, destPath);
    } catch (e) {
        return Promise.reject(e);
    }
}


/**
 * 删除文件
 * @param {*} path 必填，需要删除的文件路径
 */
const unlink = (path: string): Promise<void> => {
    if (typeof path !== typeof "") {
        return Promise.reject("unlink error, path is not string");
    }

    try {
        return fsMod.unlink(path);
    } catch (e) {
        return Promise.reject(e);
    }
}

const FileSystemManager = {
    access,
    rename,
    stat,
    mkdir,
    readdir,
    rmdir,
    readFile,
    writeFile,
    appendFile,
    copyFile,
    unlink,
};

globalThis.getFileSystemManager = () => FileSystemManager;
