
interface ImageMod {
    // 传入图片的原始二进制数据，返回rgba8的二进制数据
    decodeImage(data: ArrayBuffer, type: number, callback: {
        onLoad: (rgba_data: ArrayBuffer, width: number, height: number) => void,
        onError: (error: string) => void
    }): Promise<ArrayBuffer>;
}

interface UtilMod {
    createUUID(): string;
    base64Decode(base64Data: string): ArrayBuffer;
}

const imageMod: ImageMod = globalThis._$pi_native.Image;
delete globalThis._$pi_native.Image;

const utilMod: UtilMod = globalThis._$pi_native.Util;
delete globalThis._$pi_native.Util;

// const BolbMap = new Map<string, Blob>();


class Image {
    public width: number;
    public height: number;

    url: string;
    rgba_data: ArrayBuffer;

    error: string;

    private _loadEventCb = new Set<() => void>();
    private _errorEventCb = new Set<() => void>();


    private _onloadCb = null;
    private _onerrorCb = null;
    constructor(width?: number | null, height?: number | null) {
        if (width) {
            this.width = width;
        }
        if (height) {
            this.height = height;
        }
    }

    get buffer(): ArrayBuffer {
        return this.rgba_data;
    }

    set onload(cb: (this: Image) => void) {
        this._onloadCb = cb ? cb.bind(this) : null;
    }

    set onerror(cb: (this: Image) => void) {
        this._onerrorCb = cb ? cb.bind(this) : null;
    }

    set src(url: string) {
        if (typeof url !== typeof "") {
            this.error = "invalid value of url, not string!!"
            this._onerrorCb && this._onerrorCb(this.error);
            for (let cb of this._errorEventCb) {
                cb();
            }
            return;
        }
        this.url = url;
        if (url.indexOf("blob:") !== -1) {
            this.loadImage(url);
        } else if (url.indexOf("http:") !== -1 || url.indexOf("https:") !== -1) {
            let http = new XMLHttpRequest();
            http.onload = () => {
                let blob = new Blob([http.response]);
                let blodUrl = URL.createObjectURL(blob);
                this.loadImage(blodUrl);
            }

            http.onerror = () => {
                this._onerrorCb && this._onerrorCb(this.error);
            }
            http.open("GET", this.url);

            http.send();
        } else if (url.indexOf("data:") !== -1) {
            this.url = url.substring(url.indexOf("base64") + 7);
            let image = utilMod.base64Decode(this.url);
            let blob = new Blob([image]);
            let blodUrl = URL.createObjectURL(blob);
            this.loadImage(blodUrl);
        }
        else {
            this.error = "invalid value of src!!";
            this._onerrorCb && this._onerrorCb(this.error);
            for (let cb of this._errorEventCb) {
                cb();
            }
        }

    }

    addEventListener(event: string, cb: (this: Image) => void, options?: boolean) {
        if (event.indexOf("load") !== -1) {
            this._loadEventCb.add(cb);
        } else if (event.indexOf("error") !== -1) {
            this._errorEventCb.add(cb);
        }
    }
    removeEventListener(event: string, cb: (this: Image) => void, options?: boolean) {
        if (event.indexOf("load") !== -1) {
            this._loadEventCb.delete(cb);
        } else if (event.indexOf("error") !== -1) {
            this._errorEventCb.delete(cb);
        }
    }

    async loadImage(blobUrl: string) {
        let blob = globalThis.BolbMap.get(blobUrl);
        if (!blob) {
            this.error = `invalid src = ${this.url}`;
            this._onerrorCb && this._onerrorCb(this.error);
            for (let cb of this._errorEventCb) {
                cb();
            }
            return;
        }

        const onLoad = (rgba_buffer: ArrayBuffer, width: number, height: number) => {
            if (!this.width) {
                this.width = width;
            } else {
                if (this.width !== width) {
                    throw new Error(`invalid width = ${this.width}`);
                }
            }

            if (!this.height) {
                this.height = height;
            } else {
                if (this.height !== height) {
                    throw new Error(`invalid height = ${this.height}`);
                }
            }
            this.rgba_data = rgba_buffer;

            this._onloadCb && this._onloadCb();
            for (let cb of this._loadEventCb) {
                cb();
            }
        }

        const onError = (error: string) => {
            this.error = error;
            this._onerrorCb && this._onerrorCb(error);
            for (let cb of this._errorEventCb) {
                cb();
            }
        }

        let imageCB = {
            onLoad,
            onError
        }
        if (blob.buffer instanceof Uint8Array) {
            // if (blob.buffer.length !== blob.buffer.buffer.byteLength) {
            let buffer = blob.buffer.slice();
            await imageMod.decodeImage(buffer.buffer, blob.type, imageCB);
            // }

        } else {
            if (ArrayBuffer.isView(blob.buffer)) {
                await imageMod.decodeImage(blob.buffer.buffer, blob.type, imageCB);
            } else {
                await imageMod.decodeImage(blob.buffer, blob.type, imageCB);
            }
        }

    }

}



class Blob {
    buffer: ArrayBuffer;
    type: ImageFormat;
    endings: string;

    constructor(buffer: ArrayBuffer[] | Uint8Array[], args?: { "type": string, "endings": string }) {
        this.buffer = buffer[0];
        if (args) {
            if (args.endings) {
                this.endings = args.endings;
            }
            // TODO: 添加图片格式预处理
            if (args.type) {
                if (args.type.indexOf("image/jpeg") !== -1) {
                    this.type = ImageFormat.Jpeg;
                } else if (args.type.indexOf("image/png") !== -1) {
                    this.type = ImageFormat.Png;
                } else if (args.type.indexOf("image/webp") !== -1) {
                    this.type = ImageFormat.Webp;
                } else {
                    this.type = ImageFormat.Other;
                }
            }
        }
        this.type = ImageFormat.Other;
    }
}
enum ImageFormat {
    Jpeg = 1,
    Png = 2,
    Webp = 3,
    Other = 0,
}

export class URL {

    static createObjectURL(blob: Blob) {
        let uuid = "blob:" + utilMod.createUUID();
        globalThis.BolbMap.set(uuid, blob);
        return uuid;
    }

    static revokeObjectURL(blobUrl: string) {
        if (globalThis.BolbMap.has(blobUrl)) {
            globalThis.BolbMap.delete(blobUrl);
        }
    }
}




let g = globalThis as any;
g.Image = Image;

g.Blob = Blob;
g.URL = URL;
g.BolbMap = new Map<string, Blob>();
