declare var NativeObject;

let _webglContext = undefined;

let restartApp = globalThis.restartApp;
globalThis.restartApp = () => {
    if (!!_webglContext) {
        _webglContext.destoryContext();
    }
    restartApp();
}

export function get_window_width(pid: number): number | undefined {
    let result = NativeObject.static_call(5, pid) as number;
    return result;
}

function get_window_height(pid: number): number | undefined {
    let result = NativeObject.static_call(6, pid) as number;
    return result;
}

function get_external_files(pid: number): string {
    let result = NativeObject.static_call(7, pid) as string;
    return result;
}

/**
 * exe 协议唤醒 参数
 */
function get_arg(): string | undefined {
    let result = NativeObject.static_call(8) as string;
    return result;
}


// 将所有的窗口事件监听函数统一放到以下的队列中
// 当底层触发事件时，遍历执行整个回调函数队列
// 当队列中的监听函数为零时，自动关闭底层相对应的事件回调，不为零时自动开启；
const keyDownCb = new Set<(object: Object) => void>();
const keyUpCb = new Set<(object: Object) => void>();
const mouseDownCb = new Set<(object: Object) => void>();
const mouseUpCb = new Set<(object: Object) => void>();
const mouseMoveCb = new Set<(object: Object) => void>();
const mouseWheelCb = new Set<(object: Object) => void>();
const showCb = new Set<(timeStamp: number) => void>();
const hideCb = new Set<(timeStamp: number) => void>();
const resizeCb = new Set<(width: number, height: number, timeStamp: number) => void>();
const touchStartCb = new Set<(object: Object) => void>();
const touchMoveCb = new Set<(object: Object) => void>();
const touchEndCb = new Set<(object: Object) => void>();
const touchCancelCb = new Set<(object: Object) => void>();
const keyboardInputCb = new Set<(object: Object) => void>();
const ProgressCb = new Set<(object: Object) => void>();
const completeCb = new Set<(object: Object) => void>();
const ImeInputCb = new Set<(object: Object) => void>();

// 和rust层约定的接口
interface MainWindowMod {
    setMainFocus(focus: boolean);
    createMainWindowAndRun(callback: {
        onResize: (width: number, height: number, timeStamp: number) => void,
        onShow: (timeStamp: number) => void,
        onHide: (timeStamp: number) => void,
        onKeyDown: (virtualCode: number, timeStamp: number, onCapsLock: number, onNumLock: number, isShift: boolean, repeat: boolean) => void,
        onKeyUp: (virtualCode: number, timeStamp: number, onCapsLock: number, onNumLock: number, isShift: boolean, repeat: boolean) => void,
        onMouseMove: (x: number, y: number, delta_x: number, delta_y: number, timeStamp: number) => void,
        onMouseWheel: (delta: string, x: number, y: number, phase: string, timeStamp: number) => void,
        onMouseDown: (button: number, timeStamp: number) => void,
        onMouseUp: (button: number, timeStamp: number) => void,
        onTouchStart: (x: number, y: number, timeStamp: number) => void,
        onTouchMove: (x: number, y: number, timeStamp: number) => void,
        onTouchEnd: (x: number, y: number, timeStamp: number) => void,
        onTouchCancel: (x: number, y: number, timeStamp: number) => void,
        onMainEventsCleared: (timeStamp: number) => void,
        onRedrawRequested: (timeStamp: number) => void,
        onRedrawEventsCleared: (timeStamp: number) => void,
    });
    setMainTitle(title: string);
    mainResize(width: number, height: number);
    mainMaximize();
    mainMinimize();
    mainShow(isShow: boolean);
    mainFullscreen(isFullscreen: boolean);
    mainAlwaysOnTop(isOnTop: boolean);
    mainResizable(isResizable: boolean);
    mainRequestRedraw();
    destroyMainWindow();
    showSoftInput(text: string, inputType: string);
    hideSoftInput(flag: boolean);
    setSelection(index: number);
    setCursor(path: string, xHotspot?: number, yHotspot?: number);
    lockCursor(isLock: boolean);
    setImeAllowed(isAllowed: boolean);
    setImePosition(x: number, y: number);
    setCursorVisible(visible: boolean);
    // 关闭事件回调
    offCb(enable: boolean, cbName: string);
};


const mainWindowMod: MainWindowMod = globalThis._$pi_native.mainWindow;
delete globalThis._$pi_native.mainWindow;

class CursorPos {
    x: number;
    y: number;
}

class PointerEvent {
    type: string;
    timeStamp: number;
    button: number;
    movementX: number;
    movementY: number;
    pageX: number;
    pagey: number;
}

enum Platform {
    WINDOW,
    MOBILE,
}

// 窗口上的事件
class MainWindow {
    private _title: string;
    private _cursorPos: CursorPos;
    private _isShow: boolean;
    private _isOnTop: boolean;
    private _isResizable: boolean;
    private _isFullscreen: boolean;
    // private _platform: string;
    private _isMouseLeftPressed: boolean;

    innerWidth: number;
    innerHeight: number;
    _mainEventsClearedCb = null;
    _redrawRequestedCb = null;
    _redrawEventsClearedCb = null;

    constructor() {
        let width = get_window_width(globalThis.ESProcess.pid);
        let height = get_window_height(globalThis.ESProcess.pid);
        this.innerWidth = width;
        this.innerHeight = height;

        globalThis.innerWidth = width;
        globalThis.innerHeight = height;

        const onStart = (url: string) => {
            globalThis._$auth_test = url;
        };

        const onShow = (timeStamp: number) => {
            this._isShow = true;

            for (let cb of showCb) {
                cb(timeStamp);
            }
        }

        const onHide = (timeStamp: number) => {
            this._isShow = false;

            for (let cb of hideCb) {
                cb(timeStamp);
            }
        }

        const onResize = (width: number, height: number, timeStamp: number) => {
            this.innerWidth = width;
            this.innerHeight = height;

            globalThis.innerWidth = width;
            globalThis.innerHeight = height;

            for (let cb of resizeCb) {
                cb(width, height, timeStamp);
            }
        }

        const onKeyDown = (virtualCode: number, timeStamp: number, onCapsLock: number, onNumLock: number, isShift: boolean, repeat: boolean) => {
            let obj = this.getKey(virtualCode, onCapsLock, onNumLock, isShift);
            obj["timeStamp"] = timeStamp;
            obj["repeat"] = repeat;

            for (let cb of keyDownCb) {
                cb(obj);
            }
        }

        const onKeyUp = (virtualCode: number, timeStamp: number, onCapsLock: number, onNumLock: number, isShift: boolean, repeat: boolean) => {
            let obj = this.getKey(virtualCode, onCapsLock, onNumLock, isShift);
            obj["timeStamp"] = timeStamp;
            obj["repeat"] = repeat;

            for (let cb of keyUpCb) {
                cb(obj);
            }
        }

        const onMouseMove = (x: number, y: number, deltaX: number, deltaY: number, timeStamp: number) => {
            this._cursorPos = { x, y };

            let mouseObj = {
                button: 0,
                screenX: this._cursorPos.x,
                screenY: this._cursorPos.y,
                pageX: this._cursorPos.x,
                pageY: this._cursorPos.y,
                clientX: this._cursorPos.x,
                clientY: this._cursorPos.y,
                tiltX: this._cursorPos.x,
                tiltY: this._cursorPos.x,
                movementX: deltaX,
                movementY: deltaY,
                timeStamp: timeStamp,
                type: 'pointermove',
                x: this._cursorPos.x,
                y: this._cursorPos.y,
            };


            // 适配pc与移动平台的鼠标移动事件
            if (globalThis.platform === 'pc') {
                for (let cb of mouseMoveCb) {
                    cb(mouseObj);
                }
            } else if (globalThis.platform === 'mobile') {
                if (this._isMouseLeftPressed) {
                    for (let cb of touchMoveCb) {
                        cb(mouseObj);
                    }
                }
            } else {
                throw new Error("platform not support!!!!")
            }
        }

        const onMouseWheel = (delta: string, x: number, y: number, phase: string, timeStamp: number) => {
            let wheelObj = {
                deltaX: x,
                deltaY: y * -100,
                deltaZ: 0,
                pageX: this._cursorPos.x,
                pageY: this._cursorPos.y,
                clientX: this._cursorPos.x,
                clientY: this._cursorPos.y,
                tiltX: this._cursorPos.x,
                tiltY: this._cursorPos.x,
                timeStamp,
                type: "wheel",
                x: this._cursorPos.x,
                y: this._cursorPos.y,
            };
            if (delta = 'LineDelta') {
                wheelObj['deltaMode'] = 0x00;
            } else {
                wheelObj['deltaMode'] = 0x01;
            }

            for (let cb of mouseWheelCb) {
                cb(wheelObj);
            }
        }

        const onMouseDown = (button: number, timeStamp: number) => {
            let mouseObj = {
                id: 0,
                button,
                screenX: this._cursorPos.x,
                screenY: this._cursorPos.y,
                pageX: this._cursorPos.x,
                pageY: this._cursorPos.y,
                clientX: this._cursorPos.x,
                clientY: this._cursorPos.y,
                tiltX: this._cursorPos.x,
                tiltY: this._cursorPos.x,
                timeStamp: timeStamp,
                type: 'pointerdown',
                x: this._cursorPos.x,
                y: this._cursorPos.y,
            };

            if (globalThis.platform === 'pc') {
                for (let cb of mouseDownCb) {
                    cb(mouseObj);
                }
            } else if (globalThis.platform === 'mobile') {
                if (button === 0) {
                    this._isMouseLeftPressed = true;
                }
                for (let cb of touchStartCb) {
                    cb(mouseObj);
                }
            } else {
                throw new Error('platform not support!!!')
            }

        }

        const onMouseUp = (button: number, timeStamp: number) => {
            // 适配微信小游戏在Window平台下的表现
            // TODO: 补充identifier属性
            let mouseObj = {
                id: 0,
                button,
                screenX: this._cursorPos.x,
                screenY: this._cursorPos.y,
                pageX: this._cursorPos.x,
                pageY: this._cursorPos.y,
                clientX: this._cursorPos.x,
                clientY: this._cursorPos.y,
                tiltX: this._cursorPos.x,
                tiltY: this._cursorPos.x,
                x: this._cursorPos.x,
                y: this._cursorPos.y,
                timeStamp: timeStamp,
                type: 'pointerup',
            };

            if (globalThis.platform === 'pc') {
                for (let cb of mouseUpCb) {
                    cb(mouseObj);
                }
            } else if (globalThis.platform === 'mobile') {
                if (button === 0) {
                    this._isMouseLeftPressed = false;
                }
                for (let cb of touchEndCb) {
                    cb(mouseObj);
                }
            } else {
                throw new Error('platform not support!!!')
            }
        }

        const onTouchStart = (x: number, y: number, id: number, timeStamp: number) => {
            // 兼容旧包
            if (timeStamp === undefined) {
                timeStamp = id;
                id = 1;
            }
            let obj = {
                id,
                button: 0,
                screenX: x,
                screenY: y,
                pageX: x,
                pageY: y,
                clientX: x,
                clientY: y,
                tiltX: x,
                tiltY: x,
                x: x,
                y: y,
                timeStamp: timeStamp,
                type: 'pointerdown',
            };
            for (let cb of touchStartCb) {
                cb(obj);
            }
        }

        const onTouchMove = (x: number, y: number, id: number, timeStamp: number) => {
            // 兼容旧包
            if (timeStamp === undefined) {
                timeStamp = id;
                id = 1;
            }
            // TODO: 补充identifier
            let obj = {
                id,
                button: 0,
                screenX: x,
                screenY: y,
                pageX: x,
                pageY: y,
                clientX: x,
                clientY: y,
                tiltX: x,
                tiltY: x,
                x: x,
                y: y,
                timeStamp: timeStamp,
                type: 'pointermove',
            };

            for (let cb of touchMoveCb) {
                cb(obj);
            }
        }

        const onTouchEnd = (x: number, y: number, id: number, timeStamp: number) => {
            // 兼容旧包
            if (timeStamp === undefined) {
                timeStamp = id;
                id = 1;
            }
            let obj = {
                id,
                button: 0,
                screenX: x,
                screenY: y,
                pageX: x,
                pageY: y,
                clientX: x,
                clientY: y,
                tiltX: x,
                tiltY: x,
                x: x,
                y: y,
                timeStamp: timeStamp,
                type: 'pointerup',
            };

            for (let cb of touchEndCb) {
                cb(obj);
            }
        }

        const onTouchCancel = (x: number, y: number, id: number, timeStamp: number) => {
            // 兼容旧包
            if (timeStamp === undefined) {
                timeStamp = id;
                id = 1;
            }
            let obj = {
                id,
                button: 0,
                screenX: x,
                screenY: y,
                pageX: x,
                pageY: y,
                clientX: x,
                clientY: y,
                tiltX: x,
                tiltY: x,
                x: x,
                y: y,
                timeStamp: timeStamp,
                type: 'pointerup',
            };

            for (let cb of touchCancelCb) {
                cb(obj);
            }
        }

        const onMainEventsCleared = (timeStamp: number) => {
            this._mainEventsClearedCb && this._mainEventsClearedCb(timeStamp);
        }

        const onRedrawRequested = (timeStamp: number) => {
            this._redrawRequestedCb && this._redrawRequestedCb(timeStamp);
        }

        const onRedrawEventsCleared = (timeStamp: number) => {
            this._redrawEventsClearedCb && this._redrawEventsClearedCb(timeStamp);
        }

        const onKeyboardInput = (value: string) => {
            for (let cb of keyboardInputCb) {
                cb({ value });
            }
        }

        const onProgress = (name: string, total: number, loaded: number) => {
            for (let cb of ProgressCb) {
                cb({ name, total, loaded });
            }
        }

        const onSuspended = (_timeStamp: number) => {
            let version = "";
            if (globalThis.systemInfo && !!globalThis.systemInfo.system) {
                console.log("android system version: ", globalThis.systemInfo.system);
                version = globalThis.systemInfo.system;
            }

            if (_webglContext) {
                _webglContext.unbindSurface(version);
            }
        }

        const onResumed = (_timeStamp: number) => {
            if (_webglContext) {
                _webglContext.bindSurface();
            }
        }

        const loadScript = (code: string) => {
            globalThis.loadScript("postMessage", code);
        }

        const onComplete = (text: string, x: number, y: number) => {
            for (let cb of completeCb) {
                cb({ value: text, x, y });
            }
        }

        const onUpdateAudioDevice = (id: number) => {
            globalThis.onUpdateAudioDevice(id);
        }

        const onImeInput = (type: number, posX: number, posY: number, str: string) => {
            let obj = {
                type,
                posX,
                posY,
                str,
            };

            for (let cb of ImeInputCb) {
                cb(obj);
            }
        }

        let windowCB = {
            onResize,
            onShow,
            onHide,
            onKeyDown,
            onKeyUp,
            onMouseMove,
            onMouseWheel,
            onMouseDown,
            onMouseUp,
            onTouchStart,
            onTouchMove,
            onTouchEnd,
            onMainEventsCleared,
            onRedrawRequested,
            onRedrawEventsCleared,
            onTouchCancel,
            onKeyboardInput,
            onProgress,
            loadScript,
            onSuspended,
            onResumed,
            onComplete,
            onUpdateAudioDevice,
            onImeInput
        };
        mainWindowMod.createMainWindowAndRun(windowCB);
    }

    destroy() {
        mainWindowMod.destroyMainWindow();
    }

    setTitle(title: string) {
        if (typeof title !== typeof "") {
            throw new Error(`Window::setTitle failed, title is not string`);
        }

        if (this._title !== title) {
            mainWindowMod.setMainTitle(title);
            this._title = title;
        }

    }

    resize(width: number, height: number) {
        if (typeof width !== typeof 0 || typeof height !== typeof 0) {
            throw new Error(`Window::resize failed, width or height is not number`);
        }

        mainWindowMod.mainResize(width, height);
    }

    minimize() {
        return mainWindowMod.mainMinimize();
    }

    maximize() {
        return mainWindowMod.mainMaximize();
    }

    fullscreen(isFullscreen: boolean) {
        if (isFullscreen !== this._isFullscreen) {
            this._isFullscreen = isFullscreen;
            mainWindowMod.mainFullscreen(isFullscreen);
        }
    }

    show(isShow: boolean) {
        if (this._isShow !== isShow) {
            mainWindowMod.mainShow(isShow);
        }
    }

    resizable(isResizable: boolean) {
        if (this._isResizable !== isResizable) {
            mainWindowMod.mainResizable(isResizable);
            this._isResizable = isResizable;
        }
    }

    alwaysOnTop(isOnTop: boolean) {


        if (this._isOnTop !== isOnTop) {
            mainWindowMod.mainAlwaysOnTop(isOnTop);
            this._isOnTop = isOnTop;
        }
    }

    redraw() {
        mainWindowMod.mainRequestRedraw();
    }

    setFocus(focus: boolean) {
        mainWindowMod.setMainFocus(focus);
    }

    showSoftInput(text: string, inputType: string) {
        mainWindowMod.showSoftInput(text, inputType);
        if (navigator.userAgent.indexOf("PI_WINDOWS") != -1) {
            mainWindowMod.setImeAllowed(true);
        }
    }

    hideSoftInput(flag: boolean) {
        mainWindowMod.hideSoftInput(flag);
        if (navigator.userAgent.indexOf("PI_WINDOWS") != -1) {
            mainWindowMod.setImeAllowed(false);
        }
    }

    setSelection(index: number) {
        mainWindowMod.setSelection(index);
    }

    setCursor(path: string, xHotspot?: number, yHotspot?: number) {
        if (navigator.userAgent.indexOf("PI_WINDOWS") != -1) {
            mainWindowMod.setCursor(path, xHotspot, yHotspot);
        }

    }

    setCursorVisible(visible: boolean) {
        mainWindowMod.setCursorVisible(visible);
    }

    requestPointerLock() {
        if (navigator.userAgent.indexOf("PI_WINDOWS") != -1) {
            mainWindowMod.lockCursor(true);
        }
    }

    exitPointerLock() {
        if (navigator.userAgent.indexOf("PI_WINDOWS") != -1) {
            mainWindowMod.lockCursor(false);
        }
    }

    setImeAllowed(isAllow: boolean) {
        mainWindowMod.setImeAllowed(isAllow);
    }

    setImePosition(x: number, y: number) {
        if (navigator.userAgent.indexOf("PI_WINDOWS") != -1) {
            mainWindowMod.setImePosition(x, y);
        }
    }

    offResize(cb: (this: MainWindow) => void) {
        resizeCb.delete(cb);

        if (resizeCb.size === 0) {
            mainWindowMod.offCb(true, "onResize");
        }
    }

    offMouseDown(cb: (this: MainWindow) => void) {
        mouseDownCb.delete(cb);

        if (mouseDownCb.size === 0) {
            mainWindowMod.offCb(true, "onMouseDown");
        }
    }

    offMouseUp(cb: (this: MainWindow) => void) {
        mouseUpCb.delete(cb);

        if (mouseUpCb.size === 0) {
            mainWindowMod.offCb(true, "onMouseUp");
        }
    }

    offKeyDown(cb: (this: MainWindow) => void) {
        keyDownCb.delete(cb);

        if (keyDownCb.size === 0) {
            mainWindowMod.offCb(true, "onKeyDown");
        }
    }

    offKeyUp(cb: (this: MainWindow) => void) {
        keyUpCb.delete(cb);

        if (keyUpCb.size === 0) {
            mainWindowMod.offCb(true, "onKeyUp");
        }
    }

    offMouseMove(cb: (this: MainWindow) => void) {
        mouseMoveCb.delete(cb);

        if (mouseMoveCb.size === 0) {
            mainWindowMod.offCb(true, "onMouseMove");
        }
    };
    offMouseWheel(cb: (this: MainWindow) => void) {
        mouseWheelCb.delete(cb);

        if (mouseWheelCb.size === 0) {
            mainWindowMod.offCb(true, "onMouseWheel");
        }
    };

    offShow(cb: (this: MainWindow) => void) {
        showCb.delete(cb);

        if (showCb.size === 0) {
            mainWindowMod.offCb(true, "onShow");
        }
    };
    offHide(cb: (this: MainWindow) => void) {
        hideCb.delete(cb);

        if (hideCb.size === 0) {
            mainWindowMod.offCb(true, "onHide");
        }
    }
    offTouchStart(cb: (this: MainWindow) => void) {
        touchStartCb.delete(cb);

        if (touchStartCb.size === 0) {
            mainWindowMod.offCb(true, "onTouchStart");
        }
    }
    offTouchMove(cb: (this: MainWindow) => void) {
        touchMoveCb.delete(cb);

        if (touchMoveCb.size === 0) {
            mainWindowMod.offCb(true, "onTouchMove");
        }
    }
    offTouchEnd(cb: (this: MainWindow) => void) {
        touchEndCb.delete(cb);

        if (touchEndCb.size === 0) {
            mainWindowMod.offCb(true, "onTouchEnd");
        }
    }
    offTouchCancel(cb: (this: MainWindow) => void) {
        touchCancelCb.delete(cb);

        if (touchCancelCb.size === 0) {
            mainWindowMod.offCb(true, "onTouchCancel");
        }
    }

    offKeyboardInput(cb: (this: MainWindow) => void) {
        keyboardInputCb.delete(cb);

        if (keyboardInputCb.size === 0) {
            mainWindowMod.offCb(true, "onKeyboardInput");
        }
    }

    offComplete(cb: (this: MainWindow) => void) {
        completeCb.delete(cb);

        if (completeCb.size === 0) {
            mainWindowMod.offCb(true, "onComplete");
        }
    }
    // offMainEventsCleared(enable: boolean) {
    //     if (this._cbState.onMainEventsCleared !== enable) {
    //         this._cbState.onMainEventsCleared = enable;
    //         mainWindowMod.offCb(enable, "onMainEventsCleared");
    //     }
    // }
    // offRedrawRequested(enable: boolean) {
    //     if (this._cbState.onRedrawRequested !== enable) {
    //         this._cbState.onRedrawRequested = enable;
    //         mainWindowMod.offCb(enable, "onRedrawRequested");
    //     }
    // }

    // offRedrawEventsCleared(enable: boolean) {
    //     if (this._cbState.onRedrawEventsCleared !== enable) {
    //         this._cbState.onRedrawEventsCleared = enable;
    //         mainWindowMod.offCb(enable, "onRedrawEventsCleared");
    //     }
    // }

    onKeyDown(cb: (this: MainWindow) => void) {
        keyDownCb.add(cb);

        if (keyDownCb.size === 1) {
            mainWindowMod.offCb(false, "onKeyDown");
        }
    }

    onKeyUp(cb: (this: MainWindow) => void) {
        keyUpCb.add(cb);

        if (keyUpCb.size === 1) {
            mainWindowMod.offCb(false, "onKeyUp");
        }
    }

    onMouseDown(cb: (this: MainWindow) => void) {
        mouseDownCb.add(cb);

        if (mouseDownCb.size === 1) {
            mainWindowMod.offCb(false, "onMouseDown");
        }
    }

    onMouseUp(cb: (this: MainWindow) => void) {
        mouseUpCb.add(cb);

        if (mouseUpCb.size === 1) {
            mainWindowMod.offCb(false, "onMouseUp");
        }
    }

    onMouseMove(cb: (this: MainWindow) => void) {
        mouseMoveCb.add(cb);

        if (mouseMoveCb.size === 1) {
            mainWindowMod.offCb(false, "onMouseMove");
        }
    }

    onMouseWheel(cb: (this: MainWindow) => void) {
        mouseWheelCb.add(cb);

        if (mouseWheelCb.size === 1) {
            mainWindowMod.offCb(false, "onMouseWheel");
        }
    }

    onShow(cb: (this: MainWindow) => void) {
        showCb.add(cb);

        if (showCb.size === 1) {
            mainWindowMod.offCb(false, "onShow");
        }
    }

    onHide(cb: (this: MainWindow) => void) {
        hideCb.add(cb);

        if (hideCb.size === 1) {
            mainWindowMod.offCb(false, "onHide");
        }
    }

    onResize(cb: (this: MainWindow) => void) {
        resizeCb.add(cb);

        if (resizeCb.size === 1) {
            mainWindowMod.offCb(false, "onResize");
        }
    }

    onTouchStart(cb: (this: MainWindow) => void) {
        touchStartCb.add(cb);

        if (touchStartCb.size === 1) {
            mainWindowMod.offCb(false, "onTouchStart");
        }
    }

    onTouchMove(cb: (this: MainWindow) => void) {
        touchMoveCb.add(cb);

        if (touchMoveCb.size === 1) {
            mainWindowMod.offCb(false, "onTouchMove");
        }
    }

    onTouchEnd(cb: (this: MainWindow) => void) {
        touchEndCb.add(cb);

        if (touchEndCb.size === 1) {
            mainWindowMod.offCb(false, "onTouchEnd");
        }
    }

    onTouchCancel(cb: (this: MainWindow) => void) {
        touchCancelCb.add(cb);

        if (touchCancelCb.size === 1) {
            mainWindowMod.offCb(false, "onTouchCancel");
        }
    }

    onKeyboardInput(cb: (this: MainWindow) => void) {
        keyboardInputCb.add(cb);

        if (keyboardInputCb.size === 1) {
            mainWindowMod.offCb(false, "onKeyboardInput");
        }
    }

    onProgress(cb: (this: MainWindow) => void) {
        ProgressCb.add(cb);
    }

    set onMainEventsCleared(cb: (this: MainWindow) => void) {
        this._mainEventsClearedCb = cb ? cb.bind(this) : null;
    }

    set onRedrawRequested(cb: (this: MainWindow) => void) {
        this._redrawRequestedCb = cb ? cb.bind(this) : null;
    }

    set onRedrawEventsCleared(cb: (this: MainWindow) => void) {
        this._redrawEventsClearedCb = cb ? cb.bind(this) : null;
    }

    onComplete(cb: (this: MainWindow) => void) {
        completeCb.add(cb);

        if (completeCb.size === 1) {
            mainWindowMod.offCb(false, "onComplete");
        }
    }
    private getKey(virtualCode: number, onCapsLock: number, onNumLock: number, isShift: boolean) {
        let key = "unknow";
        let code = "unknow";
        switch (virtualCode) {
            case -1: {
                break;
            }
            case 0: {
                if (isShift) key = "!";
                else key = "1";
                code = "Digit1";
                break;
            }
            case 1: {
                if (isShift) key = "@";
                else key = "2";
                code = "Digit2";
                break;
            }
            case 2: {
                if (isShift) key = "#";
                else key = "3";
                code = "Digit3";
                break;
            }
            case 3: {
                if (isShift) key = "$";
                else key = "4";
                code = "Digit4";
                break;
            }
            case 4: {
                if (isShift) key = "%";
                else key = "5";
                code = "Digit5";
                break;
            }
            case 5: {
                if (isShift) key = "^";
                else key = "6";
                code = "Digit6";
                break;
            }
            case 6: {
                if (isShift) key = "&";
                else key = "7";
                code = "Digit7";
                break;
            }
            case 7: {
                if (isShift) key = "*";
                else key = "8";
                code = "Digit8";
                break;
            }
            case 8: {
                if (isShift) key = "(";
                else key = "9";
                code = "Digit9";
                break;
            }
            case 9: {
                if (isShift) key = ")";
                else key = "0";
                code = "Digit0";
                break;
            }
            case 10: {
                if (onCapsLock) { key = "A"; if (isShift) key = 'a'; }
                else { key = "a"; if (isShift) key = "A" }
                code = "keyA";
                break;
            }
            case 11: {
                if (onCapsLock) { key = "B"; if (isShift) key = 'b'; }
                else { key = "b"; if (isShift) key = "B" }
                code = "keyB";
                break;
            }
            case 12: {
                if (onCapsLock) { key = "C"; if (isShift) key = 'c'; }
                else { key = "c"; if (isShift) key = "C" }
                code = "keyC";
                break;
            }
            case 13: {
                if (onCapsLock) { key = "D"; if (isShift) key = 'd'; }
                else { key = "d"; if (isShift) key = "D" }
                code = "keyD";
                break;
            }
            case 14: {
                if (onCapsLock) { key = "E"; if (isShift) key = 'e'; }
                else { key = "e"; if (isShift) key = "E" }
                code = "keyE";
                break;
            }
            case 15: {
                if (onCapsLock) { key = "F"; if (isShift) key = 'f'; }
                else { key = "f"; if (isShift) key = "F" }
                code = "keyF";
                break;
            }
            case 16: {
                if (onCapsLock) { key = "G"; if (isShift) key = 'g'; }
                else { key = "g"; if (isShift) key = "G" }
                code = "keyG";
                break;
            }
            case 17: {
                if (onCapsLock) { key = "H"; if (isShift) key = 'h'; }
                else { key = "h"; if (isShift) key = "H" }
                code = "keyH";
                break;
            }
            case 18: {
                if (onCapsLock) { key = "I"; if (isShift) key = 'i'; }
                else { key = "i"; if (isShift) key = "I" }
                code = "keyI";
                break;
            }
            case 19: {
                if (onCapsLock) { key = "J"; if (isShift) key = 'j'; }
                else { key = "j"; if (isShift) key = "J" }
                code = "keyJ";
                break;
            }
            case 20: {
                if (onCapsLock) { key = "K"; if (isShift) key = 'k'; }
                else { key = "k"; if (isShift) key = "K" }
                code = "keyK";
                break;
            }
            case 21: {
                if (onCapsLock) { key = "L"; if (isShift) key = 'l'; }
                else { key = "l"; if (isShift) key = "L" }
                code = "keyL";
                break;
            }
            case 22: {
                if (onCapsLock) { key = "M"; if (isShift) key = 'm'; }
                else { key = "m"; if (isShift) key = "M" }
                code = "keyM";
                break;
            }
            case 23: {
                if (onCapsLock) { key = "N"; if (isShift) key = 'n'; }
                else { key = "n"; if (isShift) key = "N" }
                code = "keyN";
                break;
            }
            case 24: {
                if (onCapsLock) { key = "O"; if (isShift) key = 'o'; }
                else { key = "o"; if (isShift) key = "O" }
                code = "keyO";
                break;
            }
            case 25: {
                if (onCapsLock) { key = "P"; if (isShift) key = 'p'; }
                else { key = "p"; if (isShift) key = "P" }
                code = "keyP";
                break;
            }
            case 26: {
                if (onCapsLock) { key = "Q"; if (isShift) key = 'q'; }
                else { key = "q"; if (isShift) key = "Q" }
                code = "keyQ";
                break;
            }
            case 27: {
                if (onCapsLock) { key = "R"; if (isShift) key = 'r'; }
                else { key = "r"; if (isShift) key = "R" }
                code = "keyR";
                break;
            }
            case 28: {
                if (onCapsLock) { key = "S"; if (isShift) key = 's'; }
                else { key = "s"; if (isShift) key = "S" }
                code = "keyS";
                break;
            }
            case 29: {
                if (onCapsLock) { key = "T"; if (isShift) key = 't'; }
                else { key = "t"; if (isShift) key = "T" }
                code = "keyT";
                break;
            }
            case 30: {
                if (onCapsLock) { key = "U"; if (isShift) key = 'u'; }
                else { key = "u"; if (isShift) key = "U" }
                code = "keyU";
                break;
            }
            case 31: {
                if (onCapsLock) { key = "V"; if (isShift) key = 'v'; }
                else { key = "v"; if (isShift) key = "V" }
                code = "keyV";
                break;
            }
            case 32: {
                if (onCapsLock) { key = "W"; if (isShift) key = 'w'; }
                else { key = "w"; if (isShift) key = "W" }
                code = "keyW";
                break;
            }
            case 33: {
                if (onCapsLock) { key = "X"; if (isShift) key = 'x'; }
                else { key = "x"; if (isShift) key = "X" }
                code = "keyX";
                break;
            }
            case 34: {
                if (onCapsLock) { key = "Y"; if (isShift) key = 'y'; }
                else { key = "y"; if (isShift) key = "Y" }
                code = "keyY";
                break;
            }
            case 35: {
                if (onCapsLock) { key = "Z"; if (isShift) key = 'z'; }
                else { key = "z"; if (isShift) key = "Z" }
                code = "keyZ";
                break;
            }
            case 36: {
                key = "Escape";
                code = "Escape";
                break;
            }
            case 37: {
                key = "F1";
                code = "F1";
                break;
            }
            case 38: {
                key = "F2";
                code = "F2";
                break;
            }
            case 39: {
                key = "F3";
                code = "F3";
                break;
            }
            case 40: {
                key = "F4";
                code = "F4";
                break;
            }
            case 41: {
                key = "F5";
                code = "F5"; break;
            }
            case 42: {
                key = "F6";
                code = "F6"; break;
            }
            case 43: {
                key = "F7";
                code = "F7"; break;
            }
            case 44: {
                key = "F8";
                code = "F8"; break;
            }
            case 45: {
                key = "F9";
                code = "F9"; break;
            }
            case 46: {
                key = "F10";
                code = "F10"; break;
            }
            case 47: {
                key = "F11";
                code = "F11"; break;
            }
            case 48: {
                key = "F12";
                code = "F12"; break;
            }
            case 49: { break; }
            case 50: { break; }
            case 51: { break; }
            case 52: { break; }
            case 53: { break; }
            case 54: { break; }
            case 55: { break; }
            case 56: { break; }
            case 57: { break; }
            case 58: { break; }
            case 59: { break; }
            case 60: { break; }
            case 61: { break; }
            case 62: {
                key = "ScrollLock";
                code = "ScrollLock";
                break;
            }
            case 63: {
                key = "NumLock";
                code = "NumLock";
                break;
            }
            case 64: {
                key = "Insert";
                code = "Insert";
                break;
            }
            case 65: {
                key = "Home";
                code = "Home";
                break;
            }
            case 66: {
                key = "Delete";
                code = "Delete"; break;
            }
            case 67: {
                key = "End";
                code = "End"; break;
            }
            case 68: {
                key = "PageDown";
                code = "PageDown"; break;
            }
            case 69: {
                key = "PageUp";
                code = "PageUp"; break;
            }
            case 70: {
                key = "ArrowLeft";
                code = "ArrowLeft";
                break;
            }
            case 71: {
                key = "ArrowUp";
                code = "ArrowUp"; break;
            }
            case 72: {
                key = "ArrowRight";
                code = "ArrowRight"; break;
            }
            case 73: {
                key = "ArrowDown";
                code = "ArrowDown"; break;
            }
            case 74: {
                key = "Backspace";
                code = "Backspace";
                break;
            }
            case 75: {
                key = "Enter";
                code = "Enter";
                break;
            }
            case 76: {
                key = " ";
                code = "Space";
                break;
            }
            case 77: { break; }
            case 78: { break; }
            case 79: { break; }
            case 80: {
                if (isShift) key = "Insert";
                else key = "0";

                if (!onNumLock) key = "Insert";

                code = "Numpad0";
                break;
            }
            case 81: {
                if (isShift) key = "End";
                else key = "1";
                if (!onNumLock) key = "End";
                code = "Numpad1";
                break;
            }
            case 82: {
                if (isShift) key = "ArrowDown";
                else key = "2";
                if (!onNumLock) key = "ArrowDown";
                code = "Numpad2";
                break;
            }
            case 83: {
                if (isShift) key = "PageDown";
                else key = "3";
                if (!onNumLock) key = "ArrowDown";
                code = "Numpad3";
                break;
            }
            case 84: {
                if (isShift) key = "ArrowLeft";
                else key = "4";
                if (!onNumLock) key = "ArrowLeft";
                code = "Numpad4";
                break;
            }
            case 85: {
                if (isShift) key = "Clear";
                else key = "5";
                if (!onNumLock) key = "Clear";
                code = "Numpad5";
                break;
            }
            case 86: {
                if (isShift) key = "ArrowRight";
                else key = "6";
                if (!onNumLock) key = "ArrowRight";
                code = "Numpad6";
                break;
            }
            case 87: {
                if (isShift) key = "Home";
                else key = "7";
                if (!onNumLock) key = "Home";
                code = "Numpad7";
                break;
            }
            case 88: {
                if (isShift) key = "ArrowUp";
                else key = "8";
                code = "Numpad8";
                if (!onNumLock) key = "ArrowUp";
                break;
            }
            case 89: {
                if (isShift) key = "PageUp";
                else key = "9";
                if (!onNumLock) key = "PageUp";
                code = "Numpad9";
                break;
            }
            case 90: {
                key = "+";
                code = "NumpadAdd";
                break;
            }
            case 91: {
                key = "/";
                code = "NumpadDivide";
                break;
            }
            case 92: {
                if (isShift) key = "Delete";
                else key = ".";
                if (!onNumLock) key = "Delete";
                code = "NumpadDecimal";
                break;
            }
            case 95: { break; }
            case 96: {
                key = "*";
                code = "NumpadMultiply";
                break;
            }
            case 97: {
                key = "-";
                code = "NumpadSubtract";
                break;
            }
            case 100: {
                if (isShift) key = "\"";
                else key = "'";
                code = "Quote";
                break;
            }
            case 101: {
                key = "ContextMenu";
                code = "ContextMenu";
                break;
            }
            case 102: { break; }
            case 105: {
                if (isShift) key = "|";
                else key = "\\";
                code = "Backslash";
                break;
            }
            case 107: {
                key = "CapsLock";
                code = "CapsLock";
                break;
            }
            case 108: { break; }
            case 109: {
                if (isShift) key = "<";
                else key = ",";
                code = "Comma";
                break;
            }
            case 111: {
                if (isShift) key = "+";
                else key = "=";
                code = "Equal";
                break;
            }
            case 112: {
                key = "`";
                code = "Backquote";
                if (isShift) key = "~";
                break;
            }
            case 114: { break; }
            case 115: {
                key = "Alt";
                code = "AltLeft"
                break;
            }
            case 116: {
                if (isShift) key = "{";
                else key = "[";
                code = "BracketLeft";
                break;
            }
            case 117: {
                key = "Control";
                code = "ControlLeft";
                break;
            }
            case 118: {
                key = "Shift";
                code = "ShiftLeft"
                break;
            }
            case 119: {
                key = "Meta";
                code = "MetaLeft";
                break;
            }
            case 120: { break; }
            case 121: { break; }
            case 123: {
                if (isShift) key = "_";
                else key = "-";
                code = "Minus";
                break;
            }
            case 131: {
                if (isShift) key = ">";
                else key = ".";
                code = "Period";
                break;
            }
            case 136: {
                key = "Alt";
                code = "AltRight";
                break;
            }
            case 137: {
                if (isShift) key = "}";
                else key = "]";
                code = "BracketRight";
                break;
            }
            case 138: {
                key = "Control";
                code = "ControlRight";
                break;
            }
            case 136: {
                key = "Shift";
                code = "ShiftRight";
                break;
            }
            case 140: {
                key = "Meta";
                code = "MetaRight";
                break;
            }
            case 141: {
                if (isShift) key = ":";
                else key = ";";
                code = "Semicolon";
                break;
            }
            case 142: {
                if (isShift) key = "?";
                else key = "/";
                code = "Slash";
                break;
            }
            case 146: {
                key = "Tab";
                code = "Tab";
                break;
            }
        }
        let obj = {
            key,
            code
        };
        return obj;
    }
}
// 将窗口上的事件绑定到globalThis上

let g: any = globalThis;
g.offResize = (cb: (this: Object) => void) => {
    resizeCb.delete(cb);

    if (resizeCb.size === 0) {
        mainWindowMod.offCb(true, "onResize");
    }
}

g.offMouseDown = (cb: (this: Object) => void) => {
    mouseDownCb.delete(cb);

    if (mouseDownCb.size === 0) {
        mainWindowMod.offCb(true, "onMouseDown");
    }
}

g.offMouseUp = (cb: (this: Object) => void) => {
    mouseUpCb.delete(cb);

    if (mouseUpCb.size === 0) {
        mainWindowMod.offCb(true, "onMouseUp");
    }

}

g.offKeyDown = (cb: (this: Object) => void) => {
    keyDownCb.delete(cb);

    if (keyDownCb.size === 0) {
        mainWindowMod.offCb(true, "onKeyDown");
    }
}

g.offKeyUp = (cb: (this: Object) => void) => {
    keyUpCb.delete(cb);

    if (keyUpCb.size === 0) {
        mainWindowMod.offCb(true, "onKeyUp");
    }
}

g.offMouseMove = (cb: (this: Object) => void) => {
    mouseMoveCb.delete(cb);

    if (mouseMoveCb.size === 0) {
        mainWindowMod.offCb(true, "onMouseMove");
    }
};
g.offMouseWheel = (cb: (this: Object) => void) => {
    mouseWheelCb.delete(cb);

    if (mouseWheelCb.size === 0) {
        mainWindowMod.offCb(true, "onMouseWheel");
    }
};

g.offShow = (cb: (this: Object) => void) => {
    showCb.delete(cb);

    if (showCb.size === 0) {
        mainWindowMod.offCb(true, "onShow");
    }
};
g.offHide = (cb: (this: Object) => void) => {
    hideCb.delete(cb);

    if (hideCb.size === 0) {
        mainWindowMod.offCb(true, "onHide");
    }
}
g.offTouchStart = (cb: (this: Object) => void) => {
    touchStartCb.delete(cb);

    if (touchStartCb.size === 0) {
        mainWindowMod.offCb(true, "onTouchStart");
    }
}
g.offTouchMove = (cb: (this: Object) => void) => {
    touchMoveCb.delete(cb);

    if (touchMoveCb.size === 0) {
        mainWindowMod.offCb(true, "onTouchMove");
    }
}
g.offTouchEnd = (cb: (this: Object) => void) => {
    touchEndCb.delete(cb);

    if (touchEndCb.size === 0) {
        mainWindowMod.offCb(true, "onTouchEnd");
    }

}
g.offTouchCancel = (cb: (this: Object) => void) => {
    touchCancelCb.delete(cb);

    if (touchEndCb.size === 0) {
        mainWindowMod.offCb(true, "onTouchCancel");
    }
}

g.offKeyboardInput = (cb: (this: Object) => void) => {
    if (mainWindow) {
        mainWindow.offKeyboardInput(cb);
    }
}

g.offComplete = (cb: (this: Object) => void) => {
    if (mainWindow) {
        mainWindow.offComplete(cb);
    }
}

g.onKeyDown = (cb: (this: Object) => void) => {
    keyDownCb.add(cb);

    if (keyDownCb.size === 1) {
        mainWindowMod.offCb(false, "onKeyDown");
    }
}

g.onKeyUp = (cb: (this: Object) => void) => {
    keyUpCb.add(cb);

    if (keyUpCb.size === 1) {
        mainWindowMod.offCb(false, "onKeyUp");
    }
}

g.onMouseDown = (cb: (this: Object) => void) => {
    mouseDownCb.add(cb);

    if (mouseDownCb.size === 1) {
        mainWindowMod.offCb(false, "onMouseDown");
    }
}

g.onMouseUp = (cb: (this: Object) => void) => {
    mouseUpCb.add(cb);

    if (mouseUpCb.size === 1) {
        mainWindowMod.offCb(false, "onMouseUp");
    }
}

g.onMouseMove = (cb: (this: Object) => void) => {
    mouseMoveCb.add(cb);

    if (mouseMoveCb.size === 1) {
        mainWindowMod.offCb(false, "onMouseMove");
    }
}

g.onMouseWheel = (cb: (this: Object) => void) => {
    mouseWheelCb.add(cb);

    if (mouseWheelCb.size === 1) {
        mainWindowMod.offCb(false, "onMouseWheel");
    }
}

g.onShow = (cb: (this: Object) => void) => {
    showCb.add(cb);

    if (showCb.size === 1) {
        mainWindowMod.offCb(false, "onShow");
    }
}

g.onHide = (cb: (this: Object) => void) => {
    hideCb.add(cb);

    if (hideCb.size === 1) {
        mainWindowMod.offCb(false, "onHide");
    }
}

g.onResize = (cb: (this: Object) => void) => {
    resizeCb.add(cb);

    if (resizeCb.size === 1) {
        mainWindowMod.offCb(false, "onResize");
    }
}

g.onTouchStart = (cb: (this: Object) => void) => {
    touchStartCb.add(cb);

    if (touchStartCb.size === 1) {
        mainWindowMod.offCb(false, "onTouchStart");
        // 适配微信小游戏在Window平台下的触摸事件
        mainWindowMod.offCb(false, "onMouseDown");
    }
}

g.onTouchMove = (cb: (this: Object) => void) => {
    touchMoveCb.add(cb);

    if (touchMoveCb.size === 1) {
        mainWindowMod.offCb(false, "onTouchMove");
        // 适配微信小游戏在Window平台下的触摸事件
        mainWindowMod.offCb(false, "onMouseMove");
    }
}

g.onTouchEnd = (cb: (this: Object) => void) => {
    touchEndCb.add(cb);

    if (touchEndCb.size === 1) {
        mainWindowMod.offCb(false, "onTouchEnd");
        // 适配微信小游戏在Window平台下的触摸事件
        mainWindowMod.offCb(false, "onMouseUp");
    }
}

g.onTouchCancel = (cb: (this: Object) => void) => {
    touchCancelCb.add(cb);

    if (touchCancelCb.size === 1) {
        mainWindowMod.offCb(false, "onTouchCancel");
    }
}

g.onImeInput = (cb: (this: Object) => void) => {
    ImeInputCb.add(cb);
}

g.offImeInput = (cb: (this: Object) => void) => {
    ImeInputCb.delete(cb);
}

g.onKeyboardInput = (cb: (this: Object) => void) => {
    if (mainWindow) {
        mainWindow.onKeyboardInput(cb);
    }
}

g.onComplete = (cb: (this: Object) => void) => {
    if (mainWindow) {
        mainWindow.onComplete(cb);
    }
}


g.addEventListener = (event: String, listener: (this: Object) => void, options?: boolean) => {

}
g.removeEventListener = (event: String, listener: (this: Object) => void, options?: boolean) => {

}

g.getExternalFilesDir = () => {
    let pid: number = globalThis.ESProcess.pid;
    return get_external_files(pid);
};



let mainWindow;
g.getMainWindow = () => {
    if (!mainWindow) {
        mainWindow = new MainWindow();
        g.Window = mainWindow;
        // exe协议唤醒参数
        g._$auth_test = get_arg();
    }
    return mainWindow;
}

globalThis.__setWinContext = (context: any) => {
    _webglContext = context;
    globalThis.__setWinContext = undefined;
};

globalThis.__setD = (context: any) => {
    _webglContext = context;
    globalThis.__setWinContext = undefined;
};

// g.innerWidth = 1820;
// g.innerHeight = 1024;

g.platform = "pc";
