

let deviceId = null;

interface InnerAudioMod {
    initAudioEngine(): string;
    setAudioBuffer(ptr: string, data: ArrayBuffer, cb: object): void;
    startTime(ptr: string, time: number): void;
    setLoop(ptr: string, loop: boolean): void;
    setVolume(ptr: string, volume: number): void;
    setSpeed(ptr: string, speed: number): void;
    play(ptr: string,): void;
    pause(ptr: string,): void;
    stop(ptr: string,): void;
    seek(ptr: string,): void;
    destroy(ptr: string,): void;
    getDuration(ptr: string,): number;
    getCurrTime(ptr: string,): number;
    setDeviceId(ptr: string, id: number): void;
    restart(ptr: string): void;
}

const innerAudioMod: InnerAudioMod = globalThis._$pi_native.InnerAudio;
delete globalThis._$pi_native.InnerAudio;



enum State {
    Pause,
    Stop,
    Init,
    Play,
}

const canPlayCb = new Set<(object: Object) => void>();
const playCb = new Set<(object: Object) => void>();
const pauseCb = new Set<(object: Object) => void>();
const stopCb = new Set<(object: Object) => void>();
const endedCb = new Set<(object: Object) => void>();
const timeUpdateCb = new Set<(object: Object) => void>();
const errorCb = new Set<(object: Object) => void>();
const waitingCb = new Set<(object: Object) => void>();
const seekingCb = new Set<(object: Object) => void>();
const seekedCb = new Set<(object: Object) => void>();

const audioEngineSet = new Set<string>();

globalThis.onUpdateAudioDevice = (id: number) => {

    if (id !== deviceId) {
        deviceId = id;
        for (let ptr of audioEngineSet) {
            innerAudioMod.setDeviceId(ptr, id);
            innerAudioMod.restart(ptr);
        }
    }
};

export class InnerAudioContext {
    private audioEnginePtr: string = undefined;

    private _onPlayCb;
    private _onErrorCb;
    private _error;
    private _duration;
    private _currentTime;
    private _state: State;
    private _autoplay: boolean = false;
    private _loop: boolean = false;
    private _url: string = "";
    private _bufferOnLoad: boolean = false;

    constructor() {
        let ptr = innerAudioMod.initAudioEngine();

        if (!!deviceId) {
            innerAudioMod.setDeviceId(ptr, deviceId);
        }
        audioEngineSet.add(ptr);
        this.audioEnginePtr = ptr;

    }
    // 音频资源的地址，用于直接播放。2.2.3 开始支持云文件ID
    set src(url: string) {
        if (!this.audioEnginePtr) {
            return;
        }

        if (typeof url !== typeof "") {
            throw new Error("url is not string!!!!");
        }

        if (this._url === url) {
            return;
        }
        this._bufferOnLoad = false;
        this._url = url;
        const onPlay = () => {
            for (let cb of playCb) {
                cb({});
            }
        }
        const onPause = () => {
            for (let cb of pauseCb) {
                cb({});
            }
        }
        const onStop = () => {
            for (let cb of stopCb) {
                cb({});
            }
        }
        const onEnded = () => {
            for (let cb of endedCb) {
                cb({});
            }
        }
        const onError = (code: number, msg: string) => {
            for (let cb of errorCb) {
                cb({ code, msg });
            }
        }
        let audioCB = {
            onPlay,
            onPause,
            onStop,
            onEnded,
            onError,
        };
        this._url = url;
        if (url.indexOf("blob:") !== -1) {
            this.setBlobAudioBuffer(url, audioCB);
        }
        else if (url.indexOf("http:") !== -1 || url.indexOf("https:") !== -1) {
            this.setHttpAudioBuffer(url, audioCB);
        }
        else {
            this.setAssetAudioBuffer(url, audioCB);
        }
    }

    // 开始播放的位置（单位：s），默认为 0
    set startTime(time: number) {
        if (!this.audioEnginePtr) {
            return;
        }

        innerAudioMod.startTime(this.audioEnginePtr, time);
    }

    // 是否自动开始播放，默认为 false
    set autoplay(v: boolean) {
        if (!this.audioEnginePtr) {
            return;
        }
        this._autoplay = v;
    }

    // 是否循环播放，默认为 false
    set loop(v: boolean) {
        if (!this.audioEnginePtr) {
            return;
        }
        innerAudioMod.setLoop(this.audioEnginePtr, v);
    }

    // 是否遵循系统静音开关，默认为 true。当此参数为 false 时，即使用户打开了静音开关，也能继续发出声音。
    // 从 2.3.0 版本开始此参数不生效，使用 wx.setInnerAudioOption 接口统一设置。
    set obeyMuteSwitch(v: boolean) {
        if (!this.audioEnginePtr) {
            return;
        }

    }
    // 基础库 1.9.90 开始支持，低版本需做兼容处理。音量。范围 0~1。默认为 1
    set volume(v: number) {
        if (!this.audioEnginePtr) {
            return;
        }

        innerAudioMod.setVolume(this.audioEnginePtr, v);
    }

    // 基础库 2.11.0 开始支持，低版本需做兼容处理。
    // 播放速度。范围 0.5 - 2.0，默认为 1。（Android 需要 6 及以上版本）
    set playbackRate(v: number) {
        if (!this.audioEnginePtr) {
            return;
        }

        innerAudioMod.setSpeed(this.audioEnginePtr, v);
    }
    // 当前音频的长度（单位 s）。只有在当前有合法的 src 时返回（只读）
    get duration() {
        return innerAudioMod.getDuration(this.audioEnginePtr);
    }

    // 当前音频的播放位置（单位 s）。只有在当前有合法的 src 时返回，时间保留小数点后 6 位（只读）
    get currentTime() {
        return innerAudioMod.getCurrTime(this.audioEnginePtr,);
    }

    // 当前是是否暂停或停止状态（只读）
    get paused() {
        return this._state === State.Pause;
    }

    // 音频缓冲的时间点，仅保证当前播放时间点到此时间点内容已缓冲（只读）
    get buffered() {
        return 0;
    }

    // string referrerPolicy
    // 基础库 2.13.0 开始支持，低版本需做兼容处理。
    // origin: 发送完整的referrer; no - referrer: 不发送。格式固定为 https://servicewechat.com/{appid}/{version}/page-frame.html，其中 {appid} 为小程序的 appid，{version} 为小程序的版本号，版本号为 0 表示为开发版、体验版以及审核版本，版本号为 devtools 表示为开发者工具，其余为正式版本；

    // 播放
    public play() {
        if (!this.audioEnginePtr) {
            return;
        }
        if (this._bufferOnLoad) {
            innerAudioMod.play(this.audioEnginePtr);
            this._state = State.Play;
        } else {
            this._autoplay = true;
        }


    }

    // 暂停。暂停后的音频再播放会从暂停处开始播放
    public pause() {
        if (!this.audioEnginePtr) {
            return;
        }
        innerAudioMod.pause(this.audioEnginePtr);
        this._state = State.Pause;
    }

    // 停止。停止后的音频再播放会从头开始播放。
    public stop() {
        if (!this.audioEnginePtr) {
            return;
        }
        innerAudioMod.stop(this.audioEnginePtr);
        this._state = State.Stop;
    }

    // 跳转到指定位置
    public seek(v: number) {
        if (!this.audioEnginePtr) {
            return;
        }
        innerAudioMod.seek(this.audioEnginePtr);
    }

    // 销毁当前实例
    public destroy() {
        if (!this.audioEnginePtr) {
            return;
        }
        audioEngineSet.delete(this.audioEnginePtr);
        innerAudioMod.destroy(this.audioEnginePtr);
        this.audioEnginePtr = undefined;
    }

    // 监听音频进入可以播放状态的事件。但不保证后面可以流畅播放
    public onCanplay(cb: (this: InnerAudioContext) => void) {
        canPlayCb.add(cb);
    }

    // 取消监听音频进入可以播放状态的事件
    public offCanplay(cb: (this: InnerAudioContext) => void) {
        canPlayCb.delete(cb);
    }
    //监听音频播放事件
    public onPlay(cb: (this: InnerAudioContext) => void) {
        playCb.add(cb);
    }

    // 取消监听音频播放事件
    public offPlay(cb: (this: InnerAudioContext) => void) {
        playCb.delete(cb);
    }

    // 监听音频暂停事件
    public onPause(cb: (this: InnerAudioContext) => void) {
        pauseCb.add(cb);
    }

    // 取消监听音频暂停事件
    public offPause(cb: (this: InnerAudioContext) => void) {
        pauseCb.delete(cb);
    }

    // 监听音频停止事件
    public onStop(cb: (this: InnerAudioContext) => void) {
        stopCb.add(cb);
    }

    // 取消监听音频停止事件
    public offStop(cb: (this: InnerAudioContext) => void) {
        stopCb.delete(cb);
    }

    // 监听音频自然播放至结束的事件
    public onEnded(cb: (this: InnerAudioContext) => void) {
        endedCb.add(cb);
    }

    // 取消监听音频自然播放至结束的事件
    public offEnded(cb: (this: InnerAudioContext) => void) {
        endedCb.delete(cb);
    }

    // 监听音频播放进度更新事件
    public onTimeUpdate(cb: (this: InnerAudioContext) => void) {
        timeUpdateCb.add(cb);
    }

    // 取消监听音频播放进度更新事件
    public offTimeUpdate(cb: (this: InnerAudioContext) => void) {
        timeUpdateCb.delete(cb);
    }

    // 监听音频播放错误事件
    public onError(cb: (this: InnerAudioContext) => void) {
        errorCb.add(cb);
    }

    // 取消监听音频播放错误事件
    public offError(cb: (this: InnerAudioContext) => void) {
        errorCb.delete(cb);
    }

    // 监听音频加载中事件。当音频因为数据不足，需要停下来加载时会触发
    public onWaiting(cb: (this: InnerAudioContext) => void) {
        waitingCb.add(cb);
    }

    // 取消监听音频加载中事件
    public offWaiting(cb: (this: InnerAudioContext) => void) {
        waitingCb.delete(cb);
    }

    // 监听音频进行跳转操作的事件
    public onSeeking(cb: (this: InnerAudioContext) => void) {
        seekingCb.add(cb);
    }

    // 取消监听音频进行跳转操作的事件
    public offSeeking(cb: (this: InnerAudioContext) => void) {
        seekingCb.delete(cb);
    }

    // 监听音频完成跳转操作的事件
    public onSeeked(cb: (this: InnerAudioContext) => void) {
        seekedCb.add(cb);
    }

    // 取消监听音频完成跳转操作的事件
    public offSeeked(cb: (this: InnerAudioContext) => void) {
        seekedCb.delete(cb);
    }

    private setBlobAudioBuffer(url: string, cb: object) {
        let blob = globalThis.BolbMap.get(url);
        if (!blob) {
            throw new Error("Blod url: " + url + " is invalid!!!!");
        }
        try {
            if (blob.buffer instanceof Uint8Array) {
                let buffer = blob.buffer.slice();
                innerAudioMod.setAudioBuffer(this.audioEnginePtr, buffer, cb);
            } else {
                if (ArrayBuffer.isView(blob.buffer)) {
                    innerAudioMod.setAudioBuffer(this.audioEnginePtr, blob.buffer.buffer, cb);
                } else {
                    innerAudioMod.setAudioBuffer(this.audioEnginePtr, blob.buffer, cb);
                }
            }
        } catch {

        }

        this._bufferOnLoad = true;
        for (let cb of canPlayCb) {
            cb({});
        }

        if (this._autoplay) {
            innerAudioMod.play(this.audioEnginePtr);
            this._state = State.Play;
        }

    }

    private setHttpAudioBuffer(url: string, cb: object) {
        let http = new XMLHttpRequest();
        http.onload = () => {
            try {
                innerAudioMod.setAudioBuffer(this.audioEnginePtr, http.response, cb);
                for (let cb of canPlayCb) {
                    cb({});
                }

                this._bufferOnLoad = true;

                if (this._autoplay) {
                    innerAudioMod.play(this.audioEnginePtr,);
                    this._state = State.Play;
                }

            } catch (error) {

            }
        }
        http.onerror = () => {
            throw new Error("Http url: " + url + " is invalid!!!!");
        }
        http.open("GET", url);
        http.send();
    }

    private setAssetAudioBuffer(url: string, cb: object) {
        globalThis.read_asset(url).then((buffer) => {
            try {
                innerAudioMod.setAudioBuffer(this.audioEnginePtr, buffer, cb);
                for (let cb of canPlayCb) {
                    cb({});
                }
                this._bufferOnLoad = true;
                if (this._autoplay) {
                    innerAudioMod.play(this.audioEnginePtr,);
                    this._state = State.Play;
                }

            } catch (error) {

            }
        }).catch((err) => {
            // onErrorFunc(err);
            throw new Error("Assets path: " + url + " is invalid!!!!");
        });
    }

}

let g = globalThis as any;
g.InnerAudioContext = InnerAudioContext;
