

interface StoreInfo {
    keys: string[],      // 当前 storage 中所有的 key
    currentSize: number, // 当前占用的空间大小, 单位 KB
}


const storeMod = globalThis._$pi_native.Store;
delete globalThis._$pi_native.Store;

/**
 * 创建表格
 */
export const createStoreTable = (
    tablePath: string,
    bufLen = 64 * 1024,
    fileLen = 1024 * 1024,
): Promise<StoreTable> => {
    if (typeof tablePath !== typeof "") {
        return Promise.reject(`createTable failed, invalid tablePath`);
    }

    if (typeof bufLen !== typeof 0) {
        return Promise.reject(`Store::setItem failed, invalid bufLen`);
    }

    if (typeof fileLen !== typeof 0) {
        return Promise.reject(`Store::setItem failed, fileLen bufLen`);
    }

    try {
        return storeMod.createTable(tablePath, bufLen, fileLen).then((tablePtr) => {
            try {
                return new StoreTable(tablePtr);
            } catch (e) {
                return Promise.reject(e);
            }
        });
    } catch (e) {
        return Promise.reject(e);
    }
}

class StoreTable {

    tablePtr: string;

    constructor(tablePtr: string) {
        if (!tablePtr || typeof tablePtr !== typeof "") {
            throw new Error("StoreTable::constructor, invalid ptr = " + tablePtr);
        }

        this.tablePtr = tablePtr;
    }

    /**
     * 给定键值对，设置到表格
     */
    setItem(
        key: string,
        value: string | ArrayBuffer
    ): Promise<void> {
        if (!this.tablePtr || typeof this.tablePtr !== typeof "") {
            return Promise.reject(`StoreTable::set failed, invalid tablePtr`);
        }

        if (typeof key !== typeof "") {
            return Promise.reject(`StoreTable::set failed, key is not string`);
        }

        if (typeof value !== typeof "" && !(value instanceof ArrayBuffer)) {
            return Promise.reject("StoreTable::set  failed, value is not string or arrayBuffer");
        }

        try {
            return storeMod.setTableItem(this.tablePtr, key, value);
        } catch (e) {
            return Promise.reject(e);
        }
    }

    /**
     * 取键对应的值
     */
    getItem(
        key: string
    ): Promise<ArrayBuffer> {
        if (!this.tablePtr || typeof this.tablePtr !== typeof "") {
            return Promise.reject(`StoreTable::get failed, invalid tablePtr`);
        }

        if (typeof key !== typeof "") {
            return Promise.reject(`StoreTable::get failed, key is not string`);
        }

        try {
            let info = storeMod.getTableValue(this.tablePtr, key);
            return Promise.resolve(info);
        } catch (e) {
            return Promise.reject(e);
        }
    }

    /**
     * 获取信息
     */
    getInfo(): Promise<StoreInfo> {
        if (!this.tablePtr || typeof this.tablePtr !== typeof "") {
            return Promise.reject(`StoreTable::getInfo failed, invalid tablePtr`);
        }

        try {
            let info = storeMod.getTableInfo(this.tablePtr);
            return Promise.resolve(info);
        } catch (e) {
            return Promise.reject(e);
        }
    }

    /**
     * 清除key对应的项
     */
    remove(
        key: string
    ): Promise<void> {
        if (!this.tablePtr || typeof this.tablePtr !== typeof "") {
            return Promise.reject(`StoreTable::remove failed, invalid tablePtr`);
        }

        if (typeof key !== typeof "") {
            return Promise.reject(`StoreTable::remove failed, key is not string`);
        }

        try {
            return storeMod.removeTableItem(this.tablePtr, key);
        } catch (e) {
            return Promise.reject(e);
        }
    }

    /**
     * 清除表格数据
     */
    clear(): Promise<void> {
        if (!this.tablePtr || typeof this.tablePtr !== typeof "") {
            return Promise.reject(`StoreTable::clear failed, invalid tablePtr`);
        }

        try {
            return storeMod.clearTable(this.tablePtr);
        } catch (e) {
            return Promise.reject(e);
        }
    }

    /**
     * 释放表格对应的指针数据
     */
    destroy(): Promise<void> {
        if (!this.tablePtr) {
            return;
        }

        if (typeof this.tablePtr !== typeof "") {
            return Promise.reject(`StoreTable::destroy failed, invalid tablePtr`);
        }

        try {
            storeMod.destroyTable(this.tablePtr);

            // 注：这里一定要将ptr设置为undefined
            this.tablePtr = undefined;

            return Promise.resolve();
        } catch (e) {
            return Promise.reject(e);
        }
    }
}

globalThis.createStoreTable = createStoreTable;
