
interface IOStream {
    parent: IOStream;
    onlyOutput: boolean;
    args(): [object];
    output(): Promise<any>;
}

interface StreamMod {
    streamOutput(args: [object]): Promise<any>;
};

const streamMod: StreamMod = globalThis._$pi_native.streamMod;
delete globalThis._$pi_native.streamMod;

export class AsyncBufferReader implements IOStream {
    private __buffer: ArrayBuffer;

    onlyOutput: boolean;
    parent: IOStream;

    constructor(buffer: ArrayBuffer) {
        if (!(buffer instanceof ArrayBuffer)) {
            throw Error(`build AsyncBufferReader failed!!! reason: buffer is not ArrayBuffer!!!`)
        }
        this.onlyOutput = true;
        this.__buffer = buffer
    }

    args(): [object] {
        return [{ obj: "AsyncBufferReader", args: [this.__buffer] }];
    }

    output(): Promise<any> {
        return new Promise((resolve, _reject) => {
            resolve(this.__buffer)
        });
    }

    pipe(node: IOStream) {
        if (node.onlyOutput) {
            throw Error(`node can output only!!!`);
        }

        node.parent = this;
        return node;
    }
}

class AsyncFileReader implements IOStream {
    private _path: string;
    private _isAsset: boolean;

    parent: IOStream;
    onlyOutput: boolean;

    constructor(path: string, isAsset: boolean) {
        if (typeof path !== typeof "") {
            throw Error(`build AsyncFileReader failed!!! reason: path is not string!!!`)
        }

        if (typeof isAsset !== typeof true) {
            throw Error(`build AsyncFileReader failed!!! reason: len is not number!!!`)
        }

        this.onlyOutput = true
        this._path = path
        this._isAsset = isAsset;
    }

    args(): [object] {
        return [{ obj: "AsyncFileReader", args: [this._path, this._isAsset] }];
    }

    output(): Promise<any> {
        try {
            let args = this.args();
            return streamMod.streamOutput(args);
        } catch (error) {
            return Promise.reject(`read file failed!! reason: ${error}`);
        }
    }

    pipe(node: IOStream) {
        if (node.onlyOutput) {
            throw Error(`node can output only!!!`);
        }
        node.parent = this;
        return node;
    }
}

class AsyncEncoder implements IOStream {
    private _method: string;
    private _key: string | undefined;
    private _format: string | undefined;
    private _level: number | undefined;

    parent: IOStream;
    onlyOutput: boolean;

    constructor(method: string, key?: string, format?: string, level?: number) {
        if (method !== "compress" && method !== "encrypt" && method !== "compressEncrypt") {
            throw Error(`build AsyncEncoder failed!! reason: method ${method} not surpport!!!`)
        }

        if (!!format && format !== "lz4" && format !== "zlib" && format !== "xz" && format !== "lz4-hc") {
            throw Error(`build AsyncDecoder failed!! reason: format ${format} not surpport!!!`)
        }

        this.onlyOutput = false;
        this._method = method;
        this._key = key;
        this._format = format;
        this._level = level;
    }
    args(): [object] {
        if (!this.parent) {
            throw Error(`AsyncEncoder have not input!!!`)
        }
        let args = this.parent.args();

        args.push({ obj: "AsyncEncoder", args: [this._method, this._key, this._format, this._level] });
        return args;
    }

    output(): Promise<any> {
        try {
            let args = this.args();
            return streamMod.streamOutput(args);
        } catch (error) {
            return Promise.reject(`output encoder failed!! reason: ${error}`);
        }
    }

    pipe(node: IOStream) {
        if (node.onlyOutput) {
            throw Error(`node can output only!!!`);
        }
        node.parent = this;
        return node;
    }
}

class AsyncDecoder implements IOStream {
    private _method: string;
    private _key: string | undefined;
    private _format: string | undefined;

    parent: IOStream;
    onlyOutput: boolean;

    constructor(method: string, key?: string, format?: string) {
        if (method !== "decompress" && method !== "decrypt" && method !== "decompressDecrypt") {
            throw Error(`build AsyncDecoder failed!! reason: method ${method} not surpport!!!`)
        }

        if (!!format && format !== "lz4" && format !== "zlib" && format !== "xz" && format !== "lz4-hc") {
            throw Error(`build AsyncDecoder failed!! reason: format ${format} not surpport!!!`)
        }

        this.onlyOutput = false;
        this._method = method;
        this._key = key;
        this._format = format;
    }

    args(): [object] {
        if (!this.parent) {
            throw Error(`AsyncDecoder have not input!!!`)
        }
        let args = this.parent.args();

        args.push({ obj: "AsyncDecoder", args: [this._method, this._key, this._format] });
        return args;
    }

    output(): Promise<any> {
        try {
            let args = this.args();
            return streamMod.streamOutput(args);
        } catch (error) {
            return Promise.reject(`output decoder failed!! reason: ${error}`);
        }
    }

    pipe(node: IOStream) {
        if (node.onlyOutput) {
            throw Error(`node can output only!!!`);
        }
        node.parent = this;
        return node;
    }
}


class AsyncWriteFile implements IOStream {
    private _path: string;

    parent: IOStream;
    onlyOutput: boolean;

    constructor(path: string) {
        if (typeof path !== typeof "") {
            throw Error(`build AsyncWriteFile failed!! reason: path is not string!!!`)
        }

        this.onlyOutput = false;
        this._path = path;
    }

    args(): [object] {
        if (!this.parent) {
            throw Error(`AsyncWriteFile have not input!!!`)
        }
        let args = this.parent.args();

        args.push({ obj: "AsyncWriteFile", args: [this._path,] });
        return args;
    }

    output(): Promise<any> {
        try {
            let args = this.args();
            return streamMod.streamOutput(args);
        } catch (error) {
            return Promise.reject(`write file failed!! reason: ${error}`);
        }
    }

    pipe(node: IOStream) {
        if (node.onlyOutput) {
            throw Error(`node can output only!!!`);
        }

        throw new Error("AsyncWriteFile cannot connect to input stream!!!");
    }
}

const g = globalThis as any;
g.AsyncBufferReader = AsyncBufferReader;
g.AsyncFileReader = AsyncFileReader;
g.AsyncDecoder = AsyncDecoder;
g.AsyncEncoder = AsyncEncoder;
g.AsyncWriteFile = AsyncWriteFile;
