


// 和rust交互的接口
interface WSMod {

    send(senderPointer: string, data: string | ArrayBuffer): void;

    // close code 
    // Normal = 1000, 表示正常关闭，表示建立连接的目的已经实现。
    // Away = 1001, 指示端点正在“离开”，例如服务器关闭或浏览器已离开页面。
    // Protocol = 1002, 指示端点由于协议错误而终止连接。
    // Unsupported = 1003, 表示端点正在终止连接，因为它收到了它不能接受的数据类型（例如，仅理解文本数据的端点可以在接收到二进制消息时发送该数据）。
    // Status = 1006, 服务器会崩溃, 指示在结束框中没有包含状态码。该关闭代码使得可以使用单个方法“ on_close”来处理甚至没有提供关闭代码的情况。
    // Abnormal = 1006, 服务器会崩溃,表示异常关闭。如果异常关闭是由于错误引起的，则不会使用此关闭代码。取而代之的是，处理程序的on_error方法会被调用并显示错误信息。但是，如果只是简单地删除连接而没有错误，则此关闭代码将发送到处理程序。
    // Invalid = 1007, 指示端点正在终止连接，因为它已接收到消息中与消息类型不一致的数据（例如，文本消息中的非UTF-8 [RFC3629]数据）。
    // Policy = 1008, 指示端点正在终止连接，因为它收到了违反其策略的消息。这是一种通用状态代码，当没有其他更合适的状态代码（例如，不受支持或大小）或需要隐藏有关策略的特定详细信息时，可以返回该状态代码。
    // Size = 1009, 指示端点正在终止连接，因为它收到的消息太大而无法处理。
    // Extension = 1010, 指示端点（客户端）正在终止连接，因为它期望服务器协商一个或多个扩展，但是服务器未在WebSocket握手的响应消息中返回它们。所需扩展名列表应作为关闭原因。请注意，服务器不会使用此状态代码，因为它可能会使WebSocket握手失败。
    // Error = 1011, 表示服务器正在终止连接，因为它遇到了意外情况，阻止了该请求的执行。
    // Restart = 1012, 表示服务器正在重新启动。客户端可以选择重新连接，如果这样做，则应在两次尝试之间使用5-30秒的随机延迟。
    // Again = 1013, 指示服务器超载，并且客户端应连接到其他IP（存在多个目标时），或者在用户执行操作后重新连接到同一IP。
    // Tls = 1006,
    // Empty = 1005
    // Other = 
    close(senderPointer: string, code?: number, reason?: string): void;

    destroySender(senderPointer: string);

    open(
        url: string,
        callback: {
            onopen: (senderPointer: string) => void,
            onmessage: (data: ArrayBuffer | string) => void,
            onclose: (code: number, reason: string) => void,
            onerror: (reason: string) => void,
        },
        protocols?: [string],
    );
}

const wsMod: WSMod = globalThis._$pi_native.WS;
delete globalThis._$pi_native.WS;

export class WebSocket {

    constructor(url: string, protocols?: [string]) {
        if (!url) {
            throw new Error("invalid url");
        }

        this._url = url;
        this.protocols = protocols;
    }

    // ===================== 不可用

    extensions: string; // 服务器选择的扩展。
    bufferedAmount: number; // 未发送至服务器的字节数。

    // ===================== 状态

    static CONNECTING = 0;
    static OPEN = 1;
    static CLOSING = 2;
    static CLOSED = 3;

    // ===================== 只读属性

    get url(): string {
        return this._url;
    }

    protocols: [string] // 服务器选择的下属协议。
    readyState = WebSocket.CONNECTING; // 当前的链接状态。

    // ===================== 读写-属性

    binaryType: string;    // 使用二进制的数据类型连接。

    // ===================== 方法

    // 关闭当前链接。        
    close(code?: number, reason?: string) {
        if (
            this.readyState === WebSocket.CLOSING
            || this.readyState === WebSocket.CLOSED) {
            return;
        }

        this.readyState = WebSocket.CLOSING;
        if (this._senderPointer) {
            wsMod.close(this._senderPointer, code, reason);
        }
    }

    // 对要传输的数据进行排队。
    send(data: string | ArrayBuffer) {
        if (this.readyState !== WebSocket.OPEN) {
            throw new Error("invalid readyState");
        }
        wsMod.send(this._senderPointer, data);
    }

    // ================= 底层属性

    private _isRealOpen = false;

    private _url = undefined;
    private _onopen = undefined;
    private _onclose = undefined;
    private _onerror = undefined;
    private _onmessage = undefined;
    private _senderPointer: string = undefined;

    private _callOpen() {
        const onopen = (senderPointer: string) => {
            if (this.readyState === WebSocket.CLOSING) {
                wsMod.close(senderPointer);
            } else {
                this._senderPointer = senderPointer;
                this.readyState = WebSocket.OPEN;
                this._onopen && this._onopen();
            }
        };

        const onmessage = (data: ArrayBuffer | string) => {
            this._onmessage && this._onmessage({ data });
        };

        const onclose = (code: number, reason: string) => {
            this.readyState = WebSocket.CLOSED;

            if (this._senderPointer) {
                wsMod.destroySender(this._senderPointer);
                this._senderPointer = null;
            }

            this._onclose && this._onclose({ code, reason });
        };

        const onerror = (reason: string) => {
            this.readyState = WebSocket.CLOSED;

            if (this._senderPointer) {
                wsMod.destroySender(this._senderPointer);
                this._senderPointer = null;
            }

            this._onerror && this._onerror({ reason });
        };

        let cb = {
            onopen, onmessage, onclose, onerror
        };

        wsMod.open(this.url, cb, this.protocols);
    }

    // ===================== 事件属性；

    // 回调：接受信息时
    set onmessage(cb: (data: ArrayBuffer | string) => void) {

        this._onmessage = cb;
        if (!this.url) {
            throw new Error("url is null");
        }

        if (cb && !this._isRealOpen) {
            this._isRealOpen = true;
            if (this.readyState !== WebSocket.CONNECTING) {
                this.readyState = WebSocket.CLOSED;
                this._onclose && this._onclose({ code: 0, message: "close" });
            } else {
                this._callOpen();
            }
        }
    }

    // 回调：连接成功后
    set onopen(cb: () => void) {
        this._onopen = cb;
    }

    // 回调：连接失败后
    set onerror(cb: (code: number, reason: string) => void) {
        this._onerror = cb;
    }

    // 回调：连接关闭后
    set onclose(cb: (code: number, reason: string) => void) {
        this._onclose = cb;
    }
}

const g = globalThis as any;
g.WebSocket = WebSocket;

