
// 和rust层约定的接口
interface WindowMod {
    setFocus(_winContextPtr: string, focus: boolean);
    // 创建窗口并运行事件循环
    createWindowAndRun(callback: { onResize: (width: number, height: number) => void });
    // 操作窗口
    setTitle(_winContextPtr: string, title: string);
    resize(_winContextPtr: string, width: number, height: number);
    maximize(_winContextPtr: string);
    minimize(_winContextPtr: string);
    show(_winContextPtr: string, isShow: boolean);
    fullscreen(_winContextPtr: string, isFullscreen: boolean);
    alwaysOnTop(_winContextPtr: string, isOnTop: boolean);
    resizable(_winContextPtr: string, isResizable: boolean);
    requestRedraw(_winContextPtr: string);
    destroyWindow(_winContextPtr: string);

    // 关闭事件回调
    offCb(_winContextPtr: string, enable: boolean, cbName: string);
};

const windowMod: WindowMod = globalThis._$pi_native.Window;
delete globalThis._$pi_native.Window;


export class CursorPos {
    x: number;
    y: number;
}

class Size {
    width: number;
    height: number;
}

class CbState {
    onResize: boolean;
    onShow: boolean;
    onHide: boolean;
    onKeyDown: boolean;
    onKeyUp: boolean;
    onMouseMove: boolean;
    onMouseWheel: boolean;
    onMouseDown: boolean;
    onMouseUp: boolean;
    onMainEventsCleared: boolean;
    onRedrawRequested: boolean;
    onRedrawEventsCleared: boolean;
    constructor() {
        this.onResize = false;
        this.onShow = false;
        this.onHide = false;
        this.onKeyDown = false;
        this.onKeyUp = false;
        this.onMouseMove = false;
        this.onMouseWheel = false;
        this.onMouseDown = false;
        this.onMouseUp = false;
        this.onMainEventsCleared = false;
        this.onRedrawRequested = false;
        this.onRedrawEventsCleared = false;
    }
}

class Window {
    private _winContextPtr: string;

    private _title: string;
    private _size: Size;
    private _cursorPos: CursorPos;
    private _isShow: boolean;
    private _isOnTop: boolean;
    private _isResizable: boolean;
    private _isFullscreen: boolean;

    private _cbState: CbState;

    private _onKeyDownCb = null;
    private _onKeyUpCb = null;
    private _onMouseDownCb = null;
    private _onMouseUpCb = null;
    private _mouseMoveCb = null;
    private _mouseWheelCb = null;
    private _showCb = null;
    private _hideCb = null;
    private _resizeCb = null;
    private _mainEventsClearedCb = null;
    private _redrawRequestedCb = null;
    private _redrawEventsClearedCb = null;

    constructor() {
        this._cbState = new CbState();

        const onOpen = (winPtr: string) => {
            this._winContextPtr = winPtr;
        }

        const onShow = () => {
            this._isShow = true;
            this._showCb && this._showCb();
        }

        const onHide = () => {
            this._isShow = false;
            this._showCb && this._hideCb();
        }

        const onResize = (width: number, height: number) => {
            let obj = {
                "width": width,
                "height": height,
            }
            this._size = obj;
            this._resizeCb && this._resizeCb(width, height);
        }

        const onKeyDown = (virtualCode: number, timeStamp: number, onCapsLock: number, onNumLock: number, isShift: boolean) => {
            let obj = this.getKey(virtualCode, onCapsLock, onNumLock, isShift);
            this._onKeyDownCb && this._onKeyDownCb(obj.key, obj.code, timeStamp);
        }

        const onKeyUp = (virtualCode: number, timeStamp: number, onCapsLock: number, onNumLock: number, isShift: boolean) => {
            let obj = this.getKey(virtualCode, onCapsLock, onNumLock, isShift);
            this._onKeyUpCb && this._onKeyUpCb(obj.key, obj.code, timeStamp);
        }

        const onMouseMove = (x: number, y: number) => {
            let obj = {
                "x": x,
                "y": y,
            }
            this._cursorPos = obj;

            this._mouseMoveCb && this._mouseMoveCb(x, y);
        }

        const onMouseWheel = (delta: string, x: number, y: number, phase: string) => {
            this._mouseWheelCb && this._mouseWheelCb(delta, x, y, phase);
        }

        const onMouseDown = (button: number, timeStamp: number) => {
            this._onMouseDownCb && this._onMouseDownCb(this._cursorPos.x, this._cursorPos.y, button, timeStamp);
        }

        const onMouseUp = (button: number, timeStamp: number) => {
            this._onMouseUpCb && this._onMouseUpCb(this._cursorPos.x, this._cursorPos.y, button, timeStamp);
        }

        const onMainEventsCleared = () => {
            this._mainEventsClearedCb && this._mainEventsClearedCb();
        }

        const onRedrawRequested = () => {
            this._redrawRequestedCb && this._redrawRequestedCb();
        }

        const onRedrawEventsCleared = () => {
            this._redrawEventsClearedCb && this._redrawEventsClearedCb();
        }

        const onClose = () => {
            this._winContextPtr = undefined;
        }

        let windowCB = {
            onResize,
            onOpen,
            onShow,
            onHide,
            onKeyDown,
            onKeyUp,
            onMouseMove,
            onMouseWheel,
            onMouseDown,
            onMouseUp,
            onMainEventsCleared,
            onRedrawRequested,
            onRedrawEventsCleared,
            onClose
        };

        windowMod.createWindowAndRun(windowCB);
    }

    destroy() {
        if (!this._winContextPtr || typeof this._winContextPtr !== typeof "") {
            throw new Error(`Window::minimize failed, invalid windowPtr`);
        }

        windowMod.destroyWindow(this._winContextPtr);
    }

    setTitle(title: string) {
        if (!this._winContextPtr || typeof this._winContextPtr !== typeof "") {
            throw new Error(`Window::setTitle failed, invalid windowPtr`);
        }

        if (typeof title !== typeof "") {
            throw new Error(`Window::setTitle failed, title is not string`);
        }

        if (this._title !== title) {
            windowMod.setTitle(this._winContextPtr, title);
            this._title = title;
        }

    }

    resize(width: number, height: number) {
        if (!this._winContextPtr || typeof this._winContextPtr !== typeof "") {
            throw new Error(`Window::resize failed, invalid windowPtr`);
        }

        if (typeof width !== typeof 0 || typeof height !== typeof 0) {
            throw new Error(`Window::resize failed, width or height is not number`);
        }

        windowMod.resize(this._winContextPtr, width, height);
    }

    minimize() {
        if (!this._winContextPtr || typeof this._winContextPtr !== typeof "") {
            throw new Error(`Window::minimize failed, invalid windowPtr`);
        }

        return windowMod.minimize(this._winContextPtr);
    }

    maximize() {
        if (!this._winContextPtr || typeof this._winContextPtr !== typeof "") {
            throw new Error(`Window::minimize failed, invalid windowPtr`);
        }
        return windowMod.maximize(this._winContextPtr);
    }

    fullscreen(isFullscreen: boolean) {
        if (!this._winContextPtr || typeof this._winContextPtr !== typeof "") {
            throw new Error(`Window::minimize failed, invalid windowPtr`);
        }

        if (isFullscreen !== this._isFullscreen) {
            this._isFullscreen = isFullscreen;
            windowMod.fullscreen(this._winContextPtr, isFullscreen);
        }
    }

    show(isShow: boolean) {
        if (!this._winContextPtr || typeof this._winContextPtr !== typeof "") {
            throw new Error(`Window::minimize failed, invalid windowPtr`);
        }

        if (this._isShow !== isShow) {
            windowMod.show(this._winContextPtr, isShow);
        }
    }

    resizable(isResizable: boolean) {
        if (!this._winContextPtr || typeof this._winContextPtr !== typeof "") {
            throw new Error(`Window::resizable failed, invalid windowPtr`);
        }

        if (this._isResizable !== isResizable) {
            windowMod.resizable(this._winContextPtr, isResizable);
            this._isResizable = isResizable;
        }
    }

    alwaysOnTop(isOnTop: boolean) {
        if (!this._winContextPtr || typeof this._winContextPtr !== typeof "") {
            throw new Error(`Window::alwaysOnTop failed, invalid windowPtr`);
        }

        if (this._isOnTop !== isOnTop) {
            windowMod.alwaysOnTop(this._winContextPtr, isOnTop);
            this._isOnTop = isOnTop;
        }
    }

    redraw() {
        if (!this._winContextPtr || typeof this._winContextPtr !== typeof "") {
            throw new Error(`Window::alwaysOnTop failed, invalid windowPtr`);
        }
        windowMod.requestRedraw(this._winContextPtr);
    }

    setFocus(focus: boolean) {
        if (!this._winContextPtr || typeof this._winContextPtr !== typeof "") {
            throw new Error(`Window::alwaysOnTop failed, invalid windowPtr`);
        }

        windowMod.setFocus(this._winContextPtr, focus);
    }

    offResize(enable: boolean) {
        if (!this._winContextPtr || typeof this._winContextPtr !== typeof "") {
            throw new Error(`Window::alwaysOnTop failed, invalid windowPtr`);
        }
        if (this._cbState.onResize !== enable) {
            this._cbState.onResize = enable;
            windowMod.offCb(this._winContextPtr, enable, "onResize");
        }

    };

    offMouseDown(enable: boolean) {
        if (!this._winContextPtr || typeof this._winContextPtr !== typeof "") {
            throw new Error(`Window::alwaysOnTop failed, invalid windowPtr`);
        }
        if (this._cbState.onMouseDown !== enable) {
            this._cbState.onMouseDown = enable;
            windowMod.offCb(this._winContextPtr, enable, "onMouseDown");
        }

    }

    offMouseUp(enable: boolean) {
        if (!this._winContextPtr || typeof this._winContextPtr !== typeof "") {
            throw new Error(`Window::alwaysOnTop failed, invalid windowPtr`);
        }
        if (this._cbState.onMouseUp !== enable) {
            this._cbState.onMouseUp = enable;
            windowMod.offCb(this._winContextPtr, enable, "onMouseUp");
        }
    }

    offKeyDown(enable: boolean) {
        if (!this._winContextPtr || typeof this._winContextPtr !== typeof "") {
            throw new Error(`Window::alwaysOnTop failed, invalid windowPtr`);
        }
        if (this._cbState.onKeyDown !== enable) {
            this._cbState.onKeyDown = enable;
            windowMod.offCb(this._winContextPtr, enable, "onKeyDown");
        }
    }

    offKeyUp(enable: boolean) {
        if (!this._winContextPtr || typeof this._winContextPtr !== typeof "") {
            throw new Error(`Window::alwaysOnTop failed, invalid windowPtr`);
        }
        if (this._cbState.onKeyUp !== enable) {
            this._cbState.onKeyUp = enable;
            windowMod.offCb(this._winContextPtr, enable, "onKeyUp");
        }
    }

    offMouseMove(enable: boolean) {
        if (!this._winContextPtr || typeof this._winContextPtr !== typeof "") {
            throw new Error(`Window::alwaysOnTop failed, invalid windowPtr`);
        }
        if (this._cbState.onMouseMove !== enable) {
            this._cbState.onMouseMove = enable;
            windowMod.offCb(this._winContextPtr, enable, "onMouseMove");
        }
    };
    offMouseWheel(enable: boolean) {
        if (!this._winContextPtr || typeof this._winContextPtr !== typeof "") {
            throw new Error(`Window::alwaysOnTop failed, invalid windowPtr`);
        }
        if (this._cbState.onMouseWheel !== enable) {
            this._cbState.onMouseWheel = enable;
            windowMod.offCb(this._winContextPtr, enable, "onMouseWheel");
        }
    };
    offShow(enable: boolean) {
        if (!this._winContextPtr || typeof this._winContextPtr !== typeof "") {
            throw new Error(`Window::alwaysOnTop failed, invalid windowPtr`);
        }
        if (this._cbState.onShow !== enable) {
            this._cbState.onShow = enable;
            windowMod.offCb(this._winContextPtr, enable, "onShow");
        }
    };
    offHide(enable: boolean) {
        if (!this._winContextPtr || typeof this._winContextPtr !== typeof "") {
            throw new Error(`Window::alwaysOnTop failed, invalid windowPtr`);
        }
        if (this._cbState.onHide !== enable) {
            this._cbState.onHide = enable;
            windowMod.offCb(this._winContextPtr, enable, "onHide");
        }
    }
    offMainEventsCleared(enable: boolean) {
        if (!this._winContextPtr || typeof this._winContextPtr !== typeof "") {
            throw new Error(`Window::alwaysOnTop failed, invalid windowPtr`);
        }
        if (this._cbState.onMainEventsCleared !== enable) {
            this._cbState.onMainEventsCleared = enable;
            windowMod.offCb(this._winContextPtr, enable, "onMainEventsCleared");
        }
    }
    offRedrawRequested(enable: boolean) {
        if (!this._winContextPtr || typeof this._winContextPtr !== typeof "") {
            throw new Error(`Window::alwaysOnTop failed, invalid windowPtr`);
        }
        if (this._cbState.onRedrawRequested !== enable) {
            this._cbState.onRedrawRequested = enable;
            windowMod.offCb(this._winContextPtr, enable, "onRedrawRequested");
        }
    }

    offRedrawEventsCleared(enable: boolean) {
        if (!this._winContextPtr || typeof this._winContextPtr !== typeof "") {
            throw new Error(`Window::alwaysOnTop failed, invalid windowPtr`);
        }
        if (this._cbState.onRedrawEventsCleared !== enable) {
            this._cbState.onRedrawEventsCleared = enable;
            windowMod.offCb(this._winContextPtr, enable, "onRedrawEventsCleared");
        }
    }

    set onKeyDown(cb: (this: Window) => void) {
        this._onKeyDownCb = cb ? cb.bind(this) : null;
    }

    set onKeyUp(cb: (this: Window) => void) {
        this._onKeyUpCb = cb ? cb.bind(this) : null;
    }

    set onMouseDown(cb: (this: Window) => void) {
        this._onMouseDownCb = cb ? cb.bind(this) : null;
    }

    set onMouseUp(cb: (this: Window) => void) {
        this._onMouseUpCb = cb ? cb.bind(this) : null;
    }

    set onMouseMove(cb: (this: Window) => void) {
        this._mouseMoveCb = cb ? cb.bind(this) : null;
    }

    set onMouseWheel(cb: (this: Window) => void) {
        this._mouseWheelCb = cb ? cb.bind(this) : null;
    }

    set onShow(cb: (this: Window) => void) {
        this._showCb = cb ? cb.bind(this) : null;
    }

    set onResize(cb: (this: Window) => void) {
        this._resizeCb = cb ? cb.bind(this) : null;
    }

    set onMainEventsCleared(cb: (this: Window) => void) {
        this._mainEventsClearedCb = cb ? cb.bind(this) : null;
    }

    set onRedrawRequested(cb: (this: Window) => void) {
        this._redrawRequestedCb = cb ? cb.bind(this) : null;
    }

    set onRedrawEventsCleared(cb: (this: Window) => void) {
        this._redrawEventsClearedCb = cb ? cb.bind(this) : null;
    }

    private getKey(virtualCode: number, onCapsLock: number, onNumLock: number, isShift: boolean) {
        let key = undefined;
        let code = undefined;
        switch (virtualCode) {
            case -1: {
                break;
            }
            case 0: {
                if (isShift) key = "!";
                else key = "1";
                code = "Digit1";
                break;
            }
            case 1: {
                if (isShift) key = "@";
                else key = "2";
                code = "Digit2";
                break;
            }
            case 2: {
                if (isShift) key = "#";
                else key = "3";
                code = "Digit3";
                break;
            }
            case 3: {
                if (isShift) key = "$";
                else key = "4";
                code = "Digit4";
                break;
            }
            case 4: {
                if (isShift) key = "%";
                else key = "5";
                code = "Digit5";
                break;
            }
            case 5: {
                if (isShift) key = "^";
                else key = "6";
                code = "Digit6";
                break;
            }
            case 6: {
                if (isShift) key = "&";
                else key = "7";
                code = "Digit7";
                break;
            }
            case 7: {
                if (isShift) key = "*";
                else key = "8";
                code = "Digit8";
                break;
            }
            case 8: {
                if (isShift) key = "(";
                else key = "9";
                code = "Digit9";
                break;
            }
            case 9: {
                if (isShift) key = ")";
                else key = "0";
                code = "Digit0";
                break;
            }
            case 10: {
                if (onCapsLock) { key = "A"; if (isShift) key = 'a'; }
                else { key = "a"; if (isShift) key = "A" }
                code = "keyA";
                break;
            }
            case 11: {
                if (onCapsLock) { key = "B"; if (isShift) key = 'b'; }
                else { key = "b"; if (isShift) key = "B" }
                code = "keyB";
                break;
            }
            case 12: {
                if (onCapsLock) { key = "C"; if (isShift) key = 'c'; }
                else { key = "c"; if (isShift) key = "C" }
                code = "keyC";
                break;
            }
            case 13: {
                if (onCapsLock) { key = "D"; if (isShift) key = 'd'; }
                else { key = "d"; if (isShift) key = "D" }
                code = "keyD";
                break;
            }
            case 14: {
                if (onCapsLock) { key = "E"; if (isShift) key = 'e'; }
                else { key = "e"; if (isShift) key = "E" }
                code = "keyE";
                break;
            }
            case 15: {
                if (onCapsLock) { key = "F"; if (isShift) key = 'f'; }
                else { key = "f"; if (isShift) key = "F" }
                code = "keyF";
                break;
            }
            case 16: {
                if (onCapsLock) { key = "G"; if (isShift) key = 'g'; }
                else { key = "g"; if (isShift) key = "G" }
                code = "keyG";
                break;
            }
            case 17: {
                if (onCapsLock) { key = "H"; if (isShift) key = 'h'; }
                else { key = "h"; if (isShift) key = "H" }
                code = "keyH";
                break;
            }
            case 18: {
                if (onCapsLock) { key = "I"; if (isShift) key = 'i'; }
                else { key = "i"; if (isShift) key = "I" }
                code = "keyI";
                break;
            }
            case 19: {
                if (onCapsLock) { key = "J"; if (isShift) key = 'j'; }
                else { key = "j"; if (isShift) key = "J" }
                code = "keyJ";
                break;
            }
            case 20: {
                if (onCapsLock) { key = "K"; if (isShift) key = 'k'; }
                else { key = "k"; if (isShift) key = "K" }
                code = "keyK";
                break;
            }
            case 21: {
                if (onCapsLock) { key = "L"; if (isShift) key = 'l'; }
                else { key = "l"; if (isShift) key = "L" }
                code = "keyL";
                break;
            }
            case 22: {
                if (onCapsLock) { key = "M"; if (isShift) key = 'm'; }
                else { key = "m"; if (isShift) key = "M" }
                code = "keyM";
                break;
            }
            case 23: {
                if (onCapsLock) { key = "N"; if (isShift) key = 'n'; }
                else { key = "n"; if (isShift) key = "N" }
                code = "keyN";
                break;
            }
            case 24: {
                if (onCapsLock) { key = "O"; if (isShift) key = 'o'; }
                else { key = "o"; if (isShift) key = "O" }
                code = "keyO";
                break;
            }
            case 25: {
                if (onCapsLock) { key = "P"; if (isShift) key = 'p'; }
                else { key = "p"; if (isShift) key = "P" }
                code = "keyP";
                break;
            }
            case 26: {
                if (onCapsLock) { key = "Q"; if (isShift) key = 'q'; }
                else { key = "q"; if (isShift) key = "Q" }
                code = "keyQ";
                break;
            }
            case 27: {
                if (onCapsLock) { key = "R"; if (isShift) key = 'r'; }
                else { key = "r"; if (isShift) key = "R" }
                code = "keyR";
                break;
            }
            case 28: {
                if (onCapsLock) { key = "S"; if (isShift) key = 's'; }
                else { key = "s"; if (isShift) key = "S" }
                code = "keyS";
                break;
            }
            case 29: {
                if (onCapsLock) { key = "T"; if (isShift) key = 't'; }
                else { key = "t"; if (isShift) key = "T" }
                code = "keyT";
                break;
            }
            case 30: {
                if (onCapsLock) { key = "U"; if (isShift) key = 'u'; }
                else { key = "u"; if (isShift) key = "U" }
                code = "keyU";
                break;
            }
            case 31: {
                if (onCapsLock) { key = "V"; if (isShift) key = 'v'; }
                else { key = "v"; if (isShift) key = "V" }
                code = "keyV";
                break;
            }
            case 32: {
                if (onCapsLock) { key = "W"; if (isShift) key = 'w'; }
                else { key = "w"; if (isShift) key = "W" }
                code = "keyW";
                break;
            }
            case 33: {
                if (onCapsLock) { key = "X"; if (isShift) key = 'x'; }
                else { key = "x"; if (isShift) key = "X" }
                code = "keyX";
                break;
            }
            case 34: {
                if (onCapsLock) { key = "Y"; if (isShift) key = 'y'; }
                else { key = "y"; if (isShift) key = "Y" }
                code = "keyY";
                break;
            }
            case 35: {
                if (onCapsLock) { key = "Z"; if (isShift) key = 'z'; }
                else { key = "z"; if (isShift) key = "Z" }
                code = "keyZ";
                break;
            }
            case 36: {
                key = "Escape";
                code = "Escape";
                break;
            }
            case 37: {
                key = "F1";
                code = "F1";
                break;
            }
            case 38: {
                key = "F2";
                code = "F2";
                break;
            }
            case 39: {
                key = "F3";
                code = "F3";
                break;
            }
            case 40: {
                key = "F4";
                code = "F4";
                break;
            }
            case 41: {
                key = "F5";
                code = "F5"; break;
            }
            case 42: {
                key = "F6";
                code = "F6"; break;
            }
            case 43: {
                key = "F7";
                code = "F7"; break;
            }
            case 44: {
                key = "F8";
                code = "F8"; break;
            }
            case 45: {
                key = "F9";
                code = "F9"; break;
            }
            case 46: {
                key = "F10";
                code = "F10"; break;
            }
            case 47: {
                key = "F11";
                code = "F11"; break;
            }
            case 48: {
                key = "F12";
                code = "F12"; break;
            }
            case 48: {
                key = "F2";
                code = "F2"; break;
            }
            case 49: { break; }
            case 50: { break; }
            case 51: { break; }
            case 52: { break; }
            case 53: { break; }
            case 54: { break; }
            case 55: { break; }
            case 56: { break; }
            case 57: { break; }
            case 58: { break; }
            case 59: { break; }
            case 60: { break; }
            case 61: { break; }
            case 62: {
                key = "ScrollLock";
                code = "ScrollLock";
                break;
            }
            case 63: {
                key = "NumLock";
                code = "NumLock";
                break;
            }
            case 64: {
                key = "Insert";
                code = "Insert";
                break;
            }
            case 65: {
                key = "Home";
                code = "Home";
                break;
            }
            case 66: {
                key = "Delete";
                code = "Delete"; break;
            }
            case 67: {
                key = "End";
                code = "End"; break;
            }
            case 68: {
                key = "PageDown";
                code = "PageDown"; break;
            }
            case 69: {
                key = "PageUp";
                code = "PageUp"; break;
            }
            case 70: {
                key = "ArrowLeft";
                code = "ArrowLeft";
                break;
            }
            case 71: {
                key = "ArrowUp";
                code = "ArrowUp"; break;
            }
            case 72: {
                key = "ArrowRight";
                code = "ArrowRight"; break;
            }
            case 73: {
                key = "ArrowDown";
                code = "ArrowDown"; break;
            }
            case 74: {
                key = "Backspace";
                code = "Backspace";
                break;
            }
            case 75: {
                key = "Enter";
                code = "Enter";
                break;
            }
            case 76: {
                key = " ";
                code = "Space";
                break;
            }
            case 77: { break; }
            case 78: { break; }
            case 79: { break; }
            case 80: {
                if (isShift) key = "Insert";
                else key = "0";

                if (!onNumLock) key = "Insert";

                code = "Numpad0";
                break;
            }
            case 81: {
                if (isShift) key = "End";
                else key = "1";
                if (!onNumLock) key = "End";
                code = "Numpad1";
                break;
            }
            case 82: {
                if (isShift) key = "ArrowDown";
                else key = "2";
                if (!onNumLock) key = "ArrowDown";
                code = "Numpad2";
                break;
            }
            case 83: {
                if (isShift) key = "PageDown";
                else key = "3";
                if (!onNumLock) key = "ArrowDown";
                code = "Numpad3";
                break;
            }
            case 84: {
                if (isShift) key = "ArrowLeft";
                else key = "4";
                if (!onNumLock) key = "ArrowLeft";
                code = "Numpad4";
                break;
            }
            case 85: {
                if (isShift) key = "Clear";
                else key = "5";
                if (!onNumLock) key = "Clear";
                code = "Numpad5";
                break;
            }
            case 86: {
                if (isShift) key = "ArrowRight";
                else key = "6";
                if (!onNumLock) key = "ArrowRight";
                code = "Numpad6";
                break;
            }
            case 87: {
                if (isShift) key = "Home";
                else key = "7";
                if (!onNumLock) key = "Home";
                code = "Numpad7";
                break;
            }
            case 88: {
                if (isShift) key = "ArrowUp";
                else key = "8";
                code = "Numpad8";
                if (!onNumLock) key = "ArrowUp";
                break;
            }
            case 89: {
                if (isShift) key = "PageUp";
                else key = "9";
                if (!onNumLock) key = "PageUp";
                code = "Numpad9";
                break;
            }
            case 90: { break; }
            case 91: { break; }
            case 92: {
                key = "+";
                code = "NumpadAdd";
                break;
            }

            case 93: {
                if (isShift) key = "\"";
                else key = "'";
                code = "Quote";
                break;
            }
            case 94: {
                key = "ContextMenu";
                code = "ContextMenu";
                break;
            }
            case 95: { break; }
            case 96: { break; }
            case 97: {
                if (isShift) key = "|";
                else key = "\\";
                code = "Backslash";
                break;
            }
            case 98: { break; }
            case 99: {
                key = "CapsLock";
                code = "CapsLock";
                break;
            }
            case 100: { break; }
            case 101: {
                if (isShift) key = "<";
                else key = ",";
                code = "Comma";
                break;
            }
            case 102: { break; }
            case 103: {
                if (isShift) key = "Delete";
                else key = ".";
                if (!onNumLock) key = "Delete";
                code = "NumpadDecimal";
                break;
            }
            case 104: {
                key = "/";
                code = "NumpadDivide";
                break;
            }
            case 105: {
                if (isShift) key = "+";
                else key = "=";
                code = "Equal";
                break;
            }
            case 106: {
                key = "`";
                code = "Backquote";
                if (isShift) key = "~";
                break;
            }
            case 107: { break; }
            case 108: { break; }
            case 109: {
                key = "Alt";
                code = "AltLeft"
                break;
            }
            case 110: {
                if (isShift) key = "{";
                else key = "[";
                code = "BracketLeft";
                break;
            }
            case 111: {
                key = "Control";
                code = "ControlLeft";
                break;
            }
            case 112: {
                key = "Shift";
                code = "ShiftLeft"
                break;
            }
            case 113: {
                key = "Meta";
                code = "MetaLeft";
                break;
            }
            case 114: { break; }
            case 115: { break; }
            case 116: { break; }
            case 117: {
                if (isShift) key = "_";
                else key = "-";
                code = "Minus";
                break;
            }
            case 118: {
                key = "*";
                code = "NumpadMultiply";
                break;
            }
            case 119: { break; }
            case 120: { break; }
            case 121: { break; }
            case 129: {
                if (isShift) key = ">";
                else key = ".";
                code = "Period";
                break;
            }
            case 133: {
                key = "Alt";
                code = "AltRight";
                break;
            }
            case 134: {
                if (isShift) key = "}";
                else key = "]";
                code = "BracketRight";
                break;
            }
            case 135: {
                key = "Control";
                code = "ControlRight";
                break;
            }
            case 136: {
                key = "Shift";
                code = "ShiftRight";
                break;
            }
            case 137: {
                key = "Meta";
                code = "MetaRight";
                break;
            }
            case 138: {
                if (isShift) key = ":;";
                else key = ";";
                code = "Semicolon";
                break;
            }
            case 139: {
                if (isShift) key = "?";
                else key = "/";
                code = "Slash";
                break;
            }
            case 142: {
                key = "-";
                code = "NumpadSubtract";
                break;
            }
            case 144: {
                key = "Tab";
                code = "Tab";
                break;
            }
        }
        let obj = {
            "key": key,
            "code": code
        };
        return obj;
    }
}

console.log("========= new pi_sys 22222222222222222")
    // let g: any = globalThis;
    // g.Window = Window;
