/**
 * 存储， 会全局管理并行写的次数，这样提高性能和读的响应性。
 * 当有写请求时， 如果并行写次数容许，则立刻写入，否则挂起等待。 写入完毕后检查挂起的存储，继续写入。
 */

import { readFromLocal, writeToData } from "pi_sys/load/data_transfer";
import { FileLoadParam } from "pi_sys/load/fileload_base";
import { IPiStoreBase, StoreType } from "pi_sys/store/store_base";
import { CompressAlgorithm, CompressLevel, EncryptType } from "pi_sys/util/constant";

import { Logger } from "pi_sys/util/logger";
import { utf8Encode } from "pi_sys/util/utf8";
declare var module;
let logger = new Logger(module.name, "sys");

const { readFile, readdir, unlink, mkdir, rmdir, stat, access } = globalThis.getFileSystemManager();

let root_path = "store_root/";
// if (!!globalThis.getExternalFilesDir) {
//     root_path = globalThis.getExternalFilesDir() + "/store_root/";
// }

// let globalFileTable = [];

export class Store implements IPiStoreBase {
    public readonly dbName: string;
    public readonly tabName: string;
    private style: StoreType;
    tab: any;
    map: Map<string, any>;
    filePath: string;
    typeTable: any;
    immediateWrite: boolean = true;  // 是否立即写入

    /**
     * @description 判断是否支持IndexedDB
     * @example
     */
    static check = () => {
        return !!iDB;
    }
    /**
     * 删除指定的数据库
     */
    static deleteStore = async (dbName: string): Promise<any> => {
        try {
            let dirPath = dbName.toString();
            const isExist = await access(dirPath);
            if (!isExist) return Promise.resolve();
            let dirs = await readdir(dirPath);
            for (let dir of dirs) {
                let path = dirPath + "/" + dir;
                let file = await stat(path);
                if (file.type === 1) {
                    await unlink(path);
                } else if (file.type === 2) {
                    await Store.deleteStore(dbName + "/" + dir)
                }
            }
            await rmdir(dirPath);
            return Promise.resolve();
        } catch (e) {
            return Promise.reject(e);
        }

    }
    /**
     * @description 创建指定名称的存储
     * @example
     */
    static createStore = async (dbName: string, tabName = "_db", style = StoreType.Data): Promise<Store> => { // 返回值类型1|类型2
        if (!iDB) {
            let s = new Store(dbName, tabName);
            s.map = new Map();
            return Promise.resolve(s);
        }
        let path = root_path + dbName.toString() + "/" + tabName.toString() + '/';
        let typePath = root_path + dbName.toString() + "/" + tabName.toString() + '/type';
        if (style === StoreType.Data) {

            // logger.debug("store");
            try {
                let table = await globalThis.createStoreTable(path);
                let typeTable = await globalThis.createStoreTable(typePath);
                let s = new Store(dbName, tabName, style);
                s.tab = table;
                s.typeTable = typeTable;
                return Promise.resolve(s);
            }
            catch (e) {
                return Promise.reject(e);
            }
        }
        else if (style === StoreType.File) {
            let s = new Store(dbName, tabName, style);
            let typeTable = await globalThis.createStoreTable(typePath);
            s.filePath = path;
            s.typeTable = typeTable;
            try {
                if (!(await access(path))) {
                    await mkdir(path);
                }
                return Promise.resolve(s);
            }
            catch (e) {
                return Promise.reject(e);
            }
        }
    }

    constructor(dbName: string, tabName: string, style = StoreType.Data) {
        this.dbName = dbName;
        this.tabName = tabName;
        this.style = style;
    }

    /**
     * 获取相对于data目录的路径,仅File模式有效
     * @param key 
     */
    public getFullPath(key: string) {
        return this.filePath + key;
    }

    public read(key: string, decryptType: EncryptType = EncryptType.None, encryptKey?: string, compressAlgorithm: CompressAlgorithm = CompressAlgorithm.LZ4): Promise<any> {
        return new Promise((resolve, reject) => {
            FileLoadParam.storeWightQueuePool.read(this as any, key, () => { return this.__read(key, decryptType, encryptKey, compressAlgorithm) }, (data, err) => {
                if (err) {
                    reject(err)
                } else {
                    resolve(data);
                }
            })
        });
    }


    /**
     * @description 读取数据
     * @example
     */
    public async __read(key: string, decryptType: EncryptType = EncryptType.None, encryptKey?: string, compressAlgorithm: CompressAlgorithm = CompressAlgorithm.LZ4): Promise<any> {
        if (this.map) {
            return Promise.resolve(this.map.get(key));
        }
        if (this.style === StoreType.Data) {

            let info = await this.tab.getInfo();
            if (info.keys.indexOf(key.toString()) !== -1) {
                let v = await this.tab.getItem(key.toString());
                let type = await this.typeTable.getItem(key.toString());
                let enc = new TextDecoder("utf-8");
                let vType = enc.decode(type);
                if (vType === "Uint8Array") {
                    return Promise.resolve(new Uint8Array(v));
                }
                else if (vType === "String") {
                    return Promise.resolve(enc.decode(v));
                }
                else if (vType === "Number") {
                    let str = enc.decode(v);
                    let r = Number(str);
                    return Promise.resolve(r);
                }
                else if (vType === "Object") {
                    let str = enc.decode(v);
                    return Promise.resolve(JSON.parse(str));
                }
                else if (vType === "boolean") {
                    let str = enc.decode(v);
                    if (str === "0") {
                        return Promise.resolve(false);
                    } else {
                        return Promise.resolve(true);
                    }
                }
                else if (vType === "Map") {
                    let str = enc.decode(v);
                    return Promise.resolve(new Map(JSON.parse(str)));
                }
                else if (vType === "Set") {
                    let str = enc.decode(v);
                    return Promise.resolve(new Set(JSON.parse(str)));
                }
                else {
                    throw new Error("Store.read failed!!!, format is invalid!!!!");
                }
            }
            else {
                return Promise.resolve();
            }
        }
        else if (this.style === StoreType.File) {
            let file = key.toString();
            if (key === '') {
                file = '_';
            }

            let path = this.filePath + file;
            let is_exist = await access(path);
            if (is_exist) {
                let v = await readFromLocal(path, false, decryptType, encryptKey, compressAlgorithm);
                v = v.buffer as any;
                let type = await this.typeTable.getItem(key.toString());
                let enc = new TextDecoder("utf-8");
                let vType = enc.decode(type);
                if (vType === "Uint8Array") {
                    return Promise.resolve(new Uint8Array(v));
                }
                else if (vType === "String") {
                    return Promise.resolve(enc.decode(v));
                }
                else if (vType === "Number") {
                    let str = enc.decode(v);
                    let r = Number(str);
                    return Promise.resolve(r);
                }
                else if (vType === "Object") {

                    let str = enc.decode(v);
                    let r = JSON.parse(str);
                    return Promise.resolve(r);
                }
                else if (vType === "boolean") {
                    let str = enc.decode(v);
                    if (str === "0") {
                        return Promise.resolve(false);
                    } else {
                        return Promise.resolve(true);
                    }
                }
                else if (vType === "Map") {
                    let str = enc.decode(v);
                    return Promise.resolve(new Map(JSON.parse(str)));
                }
                else if (vType === "Set") {
                    let str = enc.decode(v);
                    return Promise.resolve(new Set(JSON.parse(str)));
                }
                else {
                    throw new Error("Store.read failed!!!, format is invalid!!!!");
                }
            } else {
                return Promise.resolve();
            }


        }
    }

    public write(key: string, data: Uint8Array, encryptType: EncryptType = EncryptType.None, encryptKey?: string, compressAlgorithm: CompressAlgorithm = CompressAlgorithm.LZ4, compressLevel: CompressLevel = CompressLevel.MIDDLE, immediate?: boolean): Promise<any> {
        if (immediate === undefined) {
            immediate = this.immediateWrite;
        }
        return new Promise((resolve, reject) => {
            FileLoadParam.storeWightQueuePool.write(this as any, key, () => { return this.__write(key, data, encryptType, encryptKey, compressAlgorithm, compressLevel) }, (_, err) => {
                if (err) {
                    reject(err)
                } else {
                    resolve(_);
                }
            }, immediate)
        });
    }


    /**
     * @description 写入数据，如果键名存在则替换
     * @example
     */
    public async __write(key: string, data: any, encryptType: EncryptType = EncryptType.None, encryptKey?: string, compressAlgorithm: CompressAlgorithm = CompressAlgorithm.LZ4, compressLevel: CompressLevel = CompressLevel.MIDDLE, immediate?: boolean): Promise<any> {
        if (this.map) {
            this.map.set(key, data);
            return Promise.resolve();
        }
        let vType: any;
        let v: any;
        if (data instanceof Uint8Array) {
            vType = 'Uint8Array';
            v = data.slice().buffer;
        }
        else if (typeof data === typeof "") {
            vType = "String";
            v = data;
        }
        else if (typeof data === typeof 0) {
            vType = "Number";
            v = data.toString();
        }
        else if (data instanceof Object) {
            if (data instanceof Map) {
                vType = "Map";
                v = JSON.stringify([...data]);
            } else if (data instanceof Set) {
                vType = "Set";
                v = JSON.stringify([...data]);
            } else {
                vType = "Object";
                v = JSON.stringify(data);
            }
        }
        else if (typeof data === "boolean") {
            vType = "boolean";
            if (data === true) {
                v = "1";
            } else {
                v = "0";
            }
        } else {
            throw new Error("Store.write failed!!, data format is invalid!!");
        }
        if (this.style === StoreType.Data) {
            try {
                await this.tab.setItem(key.toString(), v);
                await this.typeTable.setItem(key.toString(), vType);
                return Promise.resolve();
            }
            catch (e) {
                return Promise.reject(e);
            }
        }
        else if (this.style === StoreType.File) {

            let file = key.toString();
            if (key === '') {
                file = '_';
            }

            let path = this.filePath + file;
            try {
                if (typeof v === "string") {
                    v = utf8Encode(v);
                }
                await writeToData(path, v, encryptType, encryptKey, compressAlgorithm, compressLevel);
                await this.typeTable.setItem(key.toString(), vType);
                return Promise.resolve();
            }
            catch (e) {
                console.error("Store.write failed!!, reason: ", e);
                return Promise.reject(e);
            }
        }
    }
    /**
     * @description 删除数据
     * @example
     */
    public async delete(key: string): Promise<any> {
        if (this.map) {
            this.map.delete(key);
            return Promise.resolve();
        }
        await this.typeTable.remove(key.toString());
        if (this.style === StoreType.Data) {
            try {
                await this.tab.remove(key.toString());
                return Promise.resolve();
            }
            catch (e) {
                return Promise.reject(e);
            }
        }
        else if (this.style === StoreType.File) {
            let path = this.filePath + key;
            try {
                if (await access(path)) {
                    await unlink(path);
                }
                return Promise.resolve();
            }
            catch (e) {
                return Promise.reject(e);
            }
        }
    }
    /**
     * @description 清除存储
     * @example
     */
    public async clear(): Promise<any> {
        if (this.map) {
            this.map.clear();
            return Promise.resolve();
        }
        await this.typeTable.clear();
        if (this.style === StoreType.Data) {
            try {
                await this.tab.clear();
                return Promise.resolve();
            }
            catch (e) {
                return Promise.reject(e);
            }
        }
        else if (this.style === StoreType.File) {
            try {
                let path = this.filePath;
                let dirs = await readdir(path);
                for (let dir of dirs) {
                    let file = await stat(path + "/" + dir);
                    if (file.type === 1) {
                        await unlink(path + dir);
                    }
                    else if (file.type === 2) {
                        await Store.deleteStore(path + dir);
                    }
                }
                return Promise.resolve();
            }
            catch (e) {
                return Promise.reject(e);
            }
        }
    }
    /**
     * @description 迭代, callback返回false表示停止迭代
     * @example
     */
    public async iterate(callback: (result: { key: any, value: any }) => boolean, errorCallback: (err: Event) => void) {
        if (!iDB) {
            return Promise.resolve().then(() => {
                for (let [key, value] of this.map) {
                    if (callback({ key, value }) === false) {
                        return;
                    }
                }
                callback(null);
            });
        }
        if (this.style === StoreType.Data) {
            try {
                let info = await this.tab.getInfo();
                for (let key of info.keys) {
                    let value = new Uint8Array(await this.tab.getItem(key));
                    if (callback({ key, value }) === false) {
                        return;
                    }
                }
                callback(null);
            }
            catch (e) {
                errorCallback(e);
            }
        }
        else if (this.style === StoreType.File) {
            try {
                let path = root_path + this.dbName.toString() + "/" + this.tabName.toString();
                // let dirs = await readdir(path);
                for (let key of this.filePath) {
                    let value = new Uint8Array(await readFile(path + "/" + key));
                    if (callback({ key, value }) === false) {
                        return;
                    }
                }
                callback(null);
            }
            catch (e) {
                errorCallback(e);
            }
        }
    }
}

class Entry {
    arr: Array<{ key: number | string, data: Uint8Array, resolve: (value?: any) => void, reject: (reason?: any) => void }> = [];
    size = 0;

    push(key: number | string, data: Uint8Array, resolve: (value?: any) => void, reject: (reason?: any) => void) {
        this.arr.push({ key, data, resolve, reject });
        this.size += data.byteLength;
    }
}

// globalThis.PiStore = Store;
// ============================== 立即执行
let iDB = {}; // || self.webkitIndexedDB || self.mozIndexedDB || self.msIndexedDB;

