import { ICfg } from "pi_sys/update/update_mgr";
import { compareVersion, getUpdateLevel, UpdateLevel } from "pi_sys/update/version";
import { Env } from "pi_sys/util/env";
import { IUpdateOption, restartApp, updateApk, updateSnapshot, updateSo, updateSys } from "pi_sys/native/base/update";
import { writeFile } from "pi_sys/store/file_system";
import { showRestartModel } from "pi_sys/setup/main";
import { showModel } from "pi_sys/native/ui/dialog";
import { clearCache, getChannelQuick, getFlavorName, installApk, onShow, openAppstore } from "pi_sys/native/base/app";
import { ProgressBar } from "pi_sys/load/progressbar";
import { BlobType, createURL, revokeURL } from "pi_sys/net/bloburl";
import { createImage, PiImage } from "pi_sys/render/image";
import { flushLog, Logger, logStep, LogStepId } from "pi_sys/util/logger";
import { ErrorCode } from "pi_sys/util/constant";
import { get } from "pi_sys/net/http_impl";
import { RESP_TYPE } from "pi_sys/net/http_base";
import { formatError } from "pi_sys/util/util";
declare let module;
let logger = new Logger(module.name, "sys");

enum UpdateKey {
    APK = "apk",
    SO = "so",
    SNAPSHOT = "snapshot",
    SYS = "sys"
}

interface IUpdateItem {
    key: UpdateKey;                   // 更新对象
    level: UpdateLevel;                  // 更新等级
    oldVersion: string;                  // 旧版本
    newVersion: string;                  // 新版本
    url: string;                           // 下载连接
    size: number;                        // 大小
    loaded: number;                       // 已下载的大小
}

const updateFuncs = {
    [UpdateKey.APK]: updateApk,
    [UpdateKey.SO]: updateSo,
    [UpdateKey.SNAPSHOT]: updateSnapshot,
    [UpdateKey.SYS]: updateSys
};

/**
 * 这里只处理大版本更新,中小版本更新交由项目处理
 */
export namespace UpdateMgr {
    let baseCfg: ICfg;  // 当前最新的cfg,有网的时候等于netCfg,无网等于runCfg
    let runCfg: ICfg;      // 当前运行的cfg
    let lang = "zh";   // 语言
    const updateList: IUpdateItem[] = [];    // 更新列表
    let updateTaskCount = 0;   // 更新任务数量
    let progressBar: ProgressBar;
    let progressBarStartTime: number;   // 进度条开始展示时间,因为sys更新可能很快,需要展示至少两秒
    let progressBarShowMinTime = 2 * 1000;
    let needClearCache = false;     // 是否清除底层数据,仅在真正大版本更新的时候才清除

    export const init = (_cfg: ICfg) => {
        baseCfg = _cfg;
        const assetsCfg = Env.get("assetsCfg");
        const dataCfg = Env.get("dataCfg");

        if (dataCfg) {
            const r = compareVersion(dataCfg.apk.version, assetsCfg.apk.version);
            if (r === -1) {// 说明apk有更新
                runCfg = assetsCfg;
            } else {
                runCfg = dataCfg;
            }
        } else {
            runCfg = assetsCfg;
        }
        const obj = {};
        [UpdateKey.APK, UpdateKey.SO, UpdateKey.SNAPSHOT, UpdateKey.SYS].forEach(key => {
            let oldVersion = runCfg[key].run_version || runCfg[key].version;  // 以run_verison为准
            obj[key] = oldVersion;
        });
        logger.info(`runVersion `, JSON.stringify(obj));
    }


    /**
     * 是否是大版本更新
     * @returns 
     */
    export const isMajorUpdate = () => {
        [UpdateKey.APK, UpdateKey.SO, UpdateKey.SNAPSHOT, UpdateKey.SYS].forEach(key => {
            let newValue = baseCfg[key];
            let oldVersion = runCfg[key].run_version || runCfg[key].version;  // 以run_verison为准
            let newVersion = newValue.version;
            let depend;
            if (key === UpdateKey.APK) {
                depend = baseCfg.app_depend;
            } else if (key === UpdateKey.SO) {
                depend = baseCfg.so_depend;
            } else if (key === UpdateKey.SYS) {
                depend = baseCfg.pi_sys_depend;
            }
            let level: UpdateLevel;
            let level1: UpdateLevel;
            if (key === UpdateKey.SO) {
                // SO有可能远端版本低于本地版本,忽略更新
                // 假设项目so更新到1.2.0,此时项目又打了一个新包,新包自带so为1.3.0
                try {
                    level = getUpdateLevel(oldVersion, newVersion, depend);
                    level1 = getUpdateLevel(oldVersion, newVersion);
                } catch (e) {
                    level = UpdateLevel.None;
                    level1 = UpdateLevel.None;
                }
            } else {
                level = getUpdateLevel(oldVersion, newVersion, depend);
                level1 = getUpdateLevel(oldVersion, newVersion);
            }
            if (level === UpdateLevel.Major) {
                const item: IUpdateItem = {
                    key,
                    level,
                    oldVersion,
                    newVersion,
                    url: newValue.url,
                    size: newValue.size || 1,
                    loaded: 0
                };
                updateList.push(item);
            }
            if (level1 === UpdateLevel.Major) {
                needClearCache = true;
            }
            if (level !== UpdateLevel.None) {
                Env.set("hasUpdate", true);
            }
        })
        return updateList.length > 0;
    }

    /**
     * 清除本地数据
     */
    export const clearLocalData = () => {
        if (needClearCache) {
            console.log("============= real clearCache");
            return clearCache().catch(e => {
                // 旧包没有这个接口,所以无论成功还是失败都应该接着往下走
                // 如果新包真的清理失败,那么也让其接着往下走,当真的有数据兼容问题的时候让用户手动清理(大部分情况下大版本更新也不会有数据兼容问题,这里为了确保才主动清理)
                logger.error("UpdateMgr.clearLocalData failed, e = %o", formatError(e));
            });
        } else {
            return Promise.resolve();
        }


    }

    /**
     * 开始更新
     * 返回一个永不返回的promise,更新完成重启
     */
    export const startUpdate = () => {
        return new Promise((resolve, reject) => {
            const firstItem = updateList[0];
            if (firstItem.key === UpdateKey.APK) {  // 第一个更新不是APK,后续更新就不可能是APK
                getChannel().then(channel => {
                    let channelCfg = baseCfg.channelsCfg && baseCfg.channelsCfg[channel];
                    if (channelCfg) {
                        const tip = channelCfg[lang];
                        showModel("", tip, false).then(() => {
                            openAppstore();
                        })
                    } else {
                        doUpdate(firstItem);
                    }
                });
            } else {
                updateList.forEach(item => {
                    doUpdate(item);
                })
            }

        });
    }

    const getChannel = () => {
        return getFlavorName().then(channel => {
            if (channel.includes("quick")) {  // quick渠道,需要再次调用接口确定真正渠道
                return getChannelQuick();
            } else {
                return channel;
            }
        })
    }

    const doUpdate = async (item: IUpdateItem) => {
        if (!progressBar) {
            progressBar = await showUpdateProgress();
        }
        const option: IUpdateOption = {
            url: item.url,
            version: item.newVersion,
            timeout: 120 * 60 * 1000,
            deltaTime: 200,
            onsucceed: () => {
                console.log(`pisys update ${item.key} success`);
                item.loaded = item.size;
                const sizeObj = getDownLoadSize();
                progressBar.onProcess("", "", sizeObj.totalSize, sizeObj.loadedSize);
                finishUpdateTask(item);
            },
            onerror: (e) => {
                const message = `update ${item.key} error,e = ${formatError(e)}`;
                showRestartModel(ErrorCode.UPDATE_DOAWLOAD_FAILED, message);
            },
            onprogress: (total: number, loaded: number) => {
                // total === loaded这次回调有可能没有,底层是以时间来回调的,需要在onsucceed中处理
                //  有可能一次也不会调用
                console.log(`update ${item.key} progress,total = ${total},loaded = ${loaded}`);
                if (item.size === 1) {
                    item.size = total;
                }
                item.loaded = loaded;
                const sizeObj = getDownLoadSize();
                progressBar.onProcess("", "", sizeObj.totalSize, sizeObj.loadedSize);
            }
        };
        const sizeObj = getDownLoadSize();
        progressBar.show("", sizeObj.totalSize, sizeObj.loadedSize);
        updateTaskCount++;
        const updateFunc = updateFuncs[item.key];
        updateFunc(option);
    }


    const showUpdateProgress = () => {
        const updateBg = Env.get("updateBg");
        let p: Promise<any> = Promise.resolve();
        if (updateBg) {
            p = loadUpdateBgImage(updateBg).catch((e) => {
                console.error("LoadImage error ", e)
            });
        }
        return p.then(image => {
            const loadBar = Env.get("load_bar");
            let bar = new ProgressBar({ image, canvas: loadBar.canvas });
            progressBarStartTime = Date.now();
            return bar;
        });
    }


    const loadUpdateBgImage = (filePath: string) => {
        return new Promise((resolve, reject) => {
            const domains = Env.get("domains");
            const currentVersion = Env.get("currentVersion");
            let dir = "origin";    // 这里只获取原始的,因为此时无法处理加密后的
            let url = `${domains[0]}/${currentVersion}/${dir}/${filePath}`;
            console.log("======= loadUpdateBgImage url ", url);
            get(url, null, null, RESP_TYPE.BIN, 10 * 1000).then(data => {
                let ab: ArrayBuffer = data as ArrayBuffer;

                const blobUrl = createURL(ab, BlobType.jpg);

                let img = createImage();
                if ((<HTMLImageElement>img).decode) {
                    imageDecodeCall(img, blobUrl, resolve, reject);
                }
                else {
                    imageOnloadCall(img, blobUrl, resolve, reject);
                }
            }).catch(e => {
                reject(e);
            });
            // Asset.read(filePath).then((data) => {
            //     let ab: ArrayBuffer = data as ArrayBuffer;

            //     const blobUrl = createURL(ab, BlobType.jpg);

            //     let img = createImage();
            //     if ((<HTMLImageElement>img).decode) {
            //         imageDecodeCall(img, blobUrl, resolve, reject);
            //     }
            //     else {
            //         imageOnloadCall(img, blobUrl, resolve, reject);
            //     }
            // }).catch(e => {
            //     reject(`loadImage - Asset read ${filePath} fail!`);
            // })

        })
    }

    const getDownLoadSize = () => {
        let totalSize = 0;  // 更新总大小
        let loadedSize = 0;  // 已更新大小
        updateList.forEach(item => {
            totalSize += item.size;
            loadedSize += item.loaded;
        });
        return { totalSize, loadedSize };
    }

    const finishUpdateTask = (item: IUpdateItem) => {
        updateTaskCount--;
        // 更新dataCfg  在所有任务完成后写入文件
        if (item.key !== UpdateKey.APK) {  // apk更新在下次安装完成首次进入的时候更新dataCfg
            let keyCfg = runCfg[item.key];
            keyCfg.version = item.newVersion;
            keyCfg.try_version = item.newVersion;
            // cfg.size = item.size;
            keyCfg.url = item.url;
        }
        if (updateTaskCount === 0) {
            logStep(LogStepId.UPDATE_DOAWLOAD_SUCCESS);
            flushLog();
            console.log("all tasks finished!!!");
            const duration = Date.now() - progressBarStartTime;
            let writePromise = writeFile("cfg", JSON.stringify([runCfg]));
            const doNext = () => {
                // progressBar.clear();
                if (item.key === UpdateKey.APK) {
                    onShow(() => {  // 用户有可能点取消,强制安装
                        installApk(item.newVersion);
                    })
                    installApk(item.newVersion);
                } else {
                    let systemInfo = Env.get("system_info");
                    let language: string = (systemInfo && systemInfo.language) || "en"; //  ”zh_CN” | “zh-CN” | “zh”
                    if (language.includes("_")) {
                        language = language.split("_")[0];
                    } else if (language.includes("-")) {
                        language = language.split("-")[0];
                    }
                    let title = "Update Success";
                    let content = "Restart App Now!";
                    let confirmText = "Confirm";
                    if (language === "zh") {
                        title = "游戏版本更新";
                        content = "现在重启游戏!";
                        confirmText = "确定"
                    }
                    writePromise.then(() => {
                        showModel(title, content, false, "Cancel", "#000000", confirmText, "#576B95").then(() => {
                            restartApp();
                        });
                    }).catch(e => {
                        const message = `wirte cfg failed, e = ${formatError(e)}`;
                        showRestartModel(ErrorCode.UPDATE_WRITE_CFG_FAILED, message)
                    })
                }

            }
            if (duration < progressBarShowMinTime) {
                setTimeout(() => {
                    doNext();
                }, progressBarShowMinTime - duration);
            } else {
                doNext();
            }
        }
    }
}


export function imageDecodeCall(img: PiImage, blobUrl: string, resolve: Function, reject: Function) {
    img.src = blobUrl;
    (<HTMLImageElement>img).decode()
        .then(() => {
            resolve(img);
            revokeURL(blobUrl);
        })
        .catch((err) => {
            let image = createImage();
            imageOnloadCall(image, blobUrl, resolve, reject);
        });
}

export function imageOnloadCall(img: PiImage, blobUrl: string, resolve: Function, reject: Function) {
    img.onload = () => {
        resolve(img);
        revokeURL(blobUrl);
    }
    img.onerror = (err) => {
        reject(err);
        revokeURL(blobUrl);
    }
    img.src = blobUrl;
}